@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -12,6 +12,7 @@ open Declarations
 open Environ
 open Names
 open Univ
+open UVars
 open Util
 
 [@@@ocaml.warning "+9+27"]
@@ -20,63 +21,64 @@ exception InductiveMismatch of MutInd.t * string
 
 let check mind field b = if not b then raise (InductiveMismatch (mind,field))
 
-let to_entry (mb:mutual_inductive_body) : Entries.mutual_inductive_entry =
+let to_entry mind (mb:mutual_inductive_body) : Entries.mutual_inductive_entry =
   let open Entries in
   let nparams = List.length mb.mind_params_ctxt in (* include letins *)
   let mind_entry_record = match mb.mind_record with
     | NotRecord -> None | FakeRecord -> Some None
     | PrimRecord data -> Some (Some (Array.map (fun (x,_,_,_) -> x) data))
   in
-  let mind_entry_universes = match mb.mind_universes with
-    | Monomorphic _ ->
-      (* We only need to rebuild the set of constraints for template polymorphic
-        inductive types. The set of monomorphic constraints is already part of
-        the graph at that point, but we need to emulate a broken bound variable
-        mechanism for template inductive types. *)
-      let univs = match mb.mind_template with
-      | None -> ContextSet.empty
-      | Some ctx -> ctx.template_context
-      in
-      Monomorphic_entry univs
-    | Polymorphic auctx -> Polymorphic_entry (AUContext.names auctx, AUContext.repr auctx)
-  in
-  let mind_entry_inds = Array.map_to_list (fun ind ->
-      let mind_entry_arity = match ind.mind_arity with
-        | RegularArity ar ->
-          let ctx, arity = Term.decompose_prod_n_assum nparams ar.mind_user_arity in
-          ignore ctx; (* we will check that the produced user_arity is equal to the input *)
-          arity
-        | TemplateArity ar ->
-          let ctx = ind.mind_arity_ctxt in
-          let ctx = List.firstn (List.length ctx - nparams) ctx in
-          Term.mkArity (ctx, Sorts.sort_of_univ ar.template_level)
-      in
-      {
-        mind_entry_typename = ind.mind_typename;
-        mind_entry_arity;
-        mind_entry_consnames = Array.to_list ind.mind_consnames;
-        mind_entry_lc = Array.map_to_list (fun c ->
-            let ctx, c = Term.decompose_prod_n_assum nparams c in
-            ignore ctx; (* we will check that the produced user_lc is equal to the input *)
-            c
-          ) ind.mind_user_lc;
-      })
-      mb.mind_packets
-  in
   let check_template ind = match ind.mind_arity with
   | RegularArity _ -> false
   | TemplateArity _ -> true
   in
   let mind_entry_template = Array.exists check_template mb.mind_packets in
   let () = if mind_entry_template then assert (Array.for_all check_template mb.mind_packets) in
+  let mind_entry_universes = match mb.mind_universes with
+    | Monomorphic ->
+      (* We only need to rebuild the set of constraints for template polymorphic
+        inductive types. The set of monomorphic constraints is already part of
+        the graph at that point, but we need to emulate a broken bound variable
+        mechanism for template inductive types. *)
+      begin match mb.mind_template with
+      | None -> Monomorphic_ind_entry
+      | Some ctx -> Template_ind_entry ctx.template_context
+      end
+    | Polymorphic auctx -> Polymorphic_ind_entry (AbstractContext.repr auctx)
+  in
+  let ntyps = Array.length mb.mind_packets in
+  let mind_entry_inds = Array.map_to_list (fun ind ->
+      let mind_entry_arity = match ind.mind_arity with
+        | RegularArity ar ->
+          let ctx, arity = Term.decompose_prod_n_decls nparams ar.mind_user_arity in
+          ignore ctx; (* we will check that the produced user_arity is equal to the input *)
+          arity
+        | TemplateArity ar ->
+          let ctx = ind.mind_arity_ctxt in
+          let ctx = List.firstn (List.length ctx - nparams) ctx in
+          Term.mkArity (ctx, ar.template_level)
+      in
+      {
+        mind_entry_typename = ind.mind_typename;
+        mind_entry_arity;
+        mind_entry_consnames = Array.to_list ind.mind_consnames;
+        mind_entry_lc = Array.map_to_list (fun c ->
+            let c = Inductive.abstract_constructor_type_relatively_to_inductive_types_context ntyps mind c in
+            let ctx, c = Term.decompose_prod_n_decls nparams c in
+            ignore ctx; (* we will check that the produced user_lc is equal to the input *)
+            c
+          ) ind.mind_user_lc;
+      })
+      mb.mind_packets
+  in
+  let mind_entry_variance = Option.map (Array.map (fun v -> Some v)) mb.mind_variance in
   {
     mind_entry_record;
     mind_entry_finite = mb.mind_finite;
     mind_entry_params = mb.mind_params_ctxt;
     mind_entry_inds;
     mind_entry_universes;
-    mind_entry_template;
-    mind_entry_cumulative= Option.has_some mb.mind_variance;
+    mind_entry_variance;
     mind_entry_private = mb.mind_private;
   }
 
@@ -85,48 +87,71 @@ let check_arity env ar1 ar2 = match ar1, ar2 with
     Constr.equal ar.mind_user_arity mind_user_arity &&
     Sorts.equal ar.mind_sort mind_sort
   | TemplateArity ar, TemplateArity {template_level} ->
-    UGraph.check_leq (universes env) template_level ar.template_level
+    UGraph.check_leq_sort (universes env) template_level ar.template_level
     (* template_level is inferred by indtypes, so functor application can produce a smaller one *)
   | (RegularArity _ | TemplateArity _), _ -> assert false
 
 let check_template ar1 ar2 = match ar1, ar2 with
 | None, None -> true
-| Some ar, Some {template_context; template_param_levels} ->
-  List.equal (Option.equal Univ.Level.equal) ar.template_param_levels template_param_levels &&
+| Some ar, Some {template_context; template_param_arguments} ->
+  List.equal Bool.equal ar.template_param_arguments template_param_arguments &&
   ContextSet.equal template_context ar.template_context
 | None, Some _ | Some _, None -> false
 
-let check_kelim k1 k2 = Sorts.family_leq k1 k2
+(* if the generated inductive is squashed the original one must be squashed *)
+let check_squashed orig generated = match orig, generated with
+  | None, None -> true
+  | Some _, None ->
+    (* the inductive is from functor instantiation which removed the need for squash *)
+    true
+  | None, Some _ ->
+    (* missing squash *)
+    false
+  | Some s1, Some s2 ->
+    (* functor instantiation can change sort qualities
+       (from Type -> Prop)
+       Condition: every quality which can make the generated inductive
+       squashed must also make the original inductive squashed *)
+    match s1, s2 with
+    | AlwaysSquashed, AlwaysSquashed -> true
+    | AlwaysSquashed, SometimesSquashed _ -> true
+    | SometimesSquashed _, AlwaysSquashed -> false
+    | SometimesSquashed s1, SometimesSquashed s2 ->
+      Sorts.Quality.Set.subset s2 s1
 
 (* Use [eq_ind_chk] because when we rebuild the recargs we have lost
    the knowledge of who is the canonical version.
    Try with to see test-suite/coqchk/include.v *)
-let eq_recarg a1 a2 = match a1, a2 with
+let eq_recarg_type ty1 ty2 = match ty1, ty2 with
+  | RecArgInd ind1, RecArgInd ind2 -> eq_ind_chk ind1 ind2
+  | RecArgPrim c1, RecArgPrim c2 -> Names.Constant.CanOrd.equal c1 c2
+  | (RecArgInd _ | RecArgPrim _), _ -> false
+
+let eq_recarg r1 r2 = match r1, r2 with
   | Norec, Norec -> true
-  | Mrec i1, Mrec i2 -> eq_ind_chk i1 i2
-  | Imbr i1, Imbr i2 -> eq_ind_chk i1 i2
-  | (Norec | Mrec _ | Imbr _), _ -> false
+  | Mrec ty1, Mrec ty2 -> eq_recarg_type ty1 ty2
+  | (Norec | Mrec _), _ -> false
 
 let eq_reloc_tbl = Array.equal (fun x y -> Int.equal (fst x) (fst y) && Int.equal (snd x) (snd y))
 
 let eq_in_context (ctx1, t1) (ctx2, t2) =
-  Context.Rel.equal Constr.equal ctx1 ctx2 && Constr.equal t1 t2
+  Context.Rel.equal Sorts.relevance_equal Constr.equal ctx1 ctx2 && Constr.equal t1 t2
 
 let check_packet env mind ind
     { mind_typename; mind_arity_ctxt; mind_arity; mind_consnames; mind_user_lc;
-      mind_nrealargs; mind_nrealdecls; mind_kelim; mind_nf_lc;
+      mind_nrealargs; mind_nrealdecls; mind_squashed; mind_nf_lc;
       mind_consnrealargs; mind_consnrealdecls; mind_recargs; mind_relevance;
       mind_nb_constant; mind_nb_args; mind_reloc_tbl } =
   let check = check mind in
 
   ignore mind_typename; (* passed through *)
-  check "mind_arity_ctxt" (Context.Rel.equal Constr.equal ind.mind_arity_ctxt mind_arity_ctxt);
+  check "mind_arity_ctxt" (Context.Rel.equal Sorts.relevance_equal Constr.equal ind.mind_arity_ctxt mind_arity_ctxt);
   check "mind_arity" (check_arity env ind.mind_arity mind_arity);
   ignore mind_consnames; (* passed through *)
   check "mind_user_lc" (Array.equal Constr.equal ind.mind_user_lc mind_user_lc);
   check "mind_nrealargs" Int.(equal ind.mind_nrealargs mind_nrealargs);
   check "mind_nrealdecls" Int.(equal ind.mind_nrealdecls mind_nrealdecls);
-  check "mind_kelim" (check_kelim ind.mind_kelim mind_kelim);
+  check "mind_squashed" (check_squashed ind.mind_squashed mind_squashed);
 
   check "mind_nf_lc" (Array.equal eq_in_context ind.mind_nf_lc mind_nf_lc);
   (* NB: here syntactic equality is not just an optimisation, we also
@@ -157,15 +182,17 @@ let check_same_record r1 r2 = match r1, r2 with
   | (NotRecord | FakeRecord | PrimRecord _), _ -> false
 
 let check_inductive env mind mb =
-  let entry = to_entry mb in
-  let { mind_packets; mind_record; mind_finite; mind_ntypes; mind_hyps;
+  let entry = to_entry mind mb in
+  let { mind_packets; mind_record; mind_finite; mind_ntypes; mind_hyps; mind_univ_hyps;
         mind_nparams; mind_nparams_rec; mind_params_ctxt;
         mind_universes; mind_template; mind_variance; mind_sec_variance;
         mind_private; mind_typing_flags; }
     =
     (* Locally set typing flags for further typechecking *)
     let env = CheckFlags.set_local_flags mb.mind_typing_flags env in
-    Indtypes.check_inductive env ~sec_univs:None mind entry
+    let mib, not_prim_record = Indtypes.check_inductive env ~sec_univs:None mind entry in
+    assert (Option.is_empty not_prim_record);
+    mib
   in
   let check = check mind in
 
@@ -173,17 +200,18 @@ let check_inductive env mind mb =
   check "mind_record" (check_same_record mb.mind_record mind_record);
   check "mind_finite" (mb.mind_finite == mind_finite);
   check "mind_ntypes" Int.(equal mb.mind_ntypes mind_ntypes);
-  check "mind_hyps" (Context.Named.equal Constr.equal mb.mind_hyps mind_hyps);
+  check "mind_hyps" (List.is_empty mind_hyps);
+  check "mind_univ_hyps" (UVars.Instance.is_empty mind_univ_hyps);
   check "mind_nparams" Int.(equal mb.mind_nparams mind_nparams);
 
   check "mind_nparams_rec" (mb.mind_nparams_rec <= mind_nparams_rec);
   (* module substitution can increase the real number of recursively
      uniform parameters, so be tolerant and use [<=]. *)
 
-  check "mind_params_ctxt" (Context.Rel.equal Constr.equal mb.mind_params_ctxt mind_params_ctxt);
+  check "mind_params_ctxt" (Context.Rel.equal Sorts.relevance_equal Constr.equal mb.mind_params_ctxt mind_params_ctxt);
   ignore mind_universes; (* Indtypes did the necessary checking *)
   check "mind_template" (check_template mb.mind_template mind_template);
-  check "mind_variance" (Option.equal (Array.equal Univ.Variance.equal)
+  check "mind_variance" (Option.equal (Array.equal UVars.Variance.equal)
                            mb.mind_variance mind_variance);
   check "mind_sec_variance" (Option.is_empty mind_sec_variance);
   ignore mind_private; (* passed through Indtypes *)
@@ -192,3 +220,9 @@ let check_inductive env mind mb =
   (* TODO non oracle flags *)
 
   add_mind mind mb env
+
+let check_inductive env mind mb : Environ.env =
+  NewProfile.profile "check_inductive"
+    ~args:(fun () -> [("name", `String (MutInd.to_string mind))])
+    (fun () -> check_inductive env mind mb)
+    ()
