@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -23,17 +23,15 @@ end
 
 module type S = Map.S
 
-module type ExtS =
+module type UExtS =
 sig
-  include CSig.MapS
-  module Set : CSig.SetS with type elt = key
+  include CSig.UMapS
+  module Set : CSig.USetS with type elt = key
   val get : key -> 'a t -> 'a
   val set : key -> 'a -> 'a t -> 'a t
   val modify : key -> (key -> 'a -> 'a) -> 'a t -> 'a t
   val domain : 'a t -> Set.t
   val bind : (key -> 'a) -> Set.t -> 'a t
-  val fold_left : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
-  val fold_right : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
   val height : 'a t -> int
   val filter_range : (key -> int) -> 'a t -> 'a t
   val filter_map: (key -> 'a -> 'b option) -> 'a t -> 'b t (* in OCaml 4.11 *)
@@ -49,11 +47,27 @@ sig
   module Monad(M : MonadS) :
   sig
     val fold : (key -> 'a -> 'b -> 'b M.t) -> 'a t -> 'b -> 'b M.t
-    val fold_left : (key -> 'a -> 'b -> 'b M.t) -> 'a t -> 'b -> 'b M.t
-    val fold_right : (key -> 'a -> 'b -> 'b M.t) -> 'a t -> 'b -> 'b M.t
     val mapi : (key -> 'a -> 'b M.t) -> 'a t -> 'b t M.t
   end
 end
+module type ExtS = sig
+  include CSig.MapS
+
+  module Set : CSig.SetS with type elt = key
+
+  include UExtS with type key := key and type 'a t := 'a t and module Set := Set
+
+  module Monad(M:MonadS) : sig
+    include module type of Monad(M)
+    val fold_left : (key -> 'a -> 'b -> 'b M.t) -> 'a t -> 'b -> 'b M.t
+    val fold_right : (key -> 'a -> 'b -> 'b M.t) -> 'a t -> 'b -> 'b M.t
+  end
+
+  val fold_left : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
+  val fold_right : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
+  val fold_left_map : (key -> 'a -> 'b -> 'b * 'c) -> 'a t -> 'b -> 'b * 'c t
+  val fold_right_map : (key -> 'a -> 'b -> 'b * 'c) -> 'a t -> 'b -> 'b * 'c t
+end
 
 module MapExt (M : Map.OrderedType) :
 sig
@@ -65,6 +79,8 @@ sig
   val bind : (M.t -> 'a) -> Set.Make(M).t -> 'a map
   val fold_left : (M.t -> 'a -> 'b -> 'b) -> 'a map -> 'b -> 'b
   val fold_right : (M.t -> 'a -> 'b -> 'b) -> 'a map -> 'b -> 'b
+  val fold_left_map : (M.t -> 'a -> 'b -> 'b * 'c) -> 'a map -> 'b -> 'b * 'c map
+  val fold_right_map : (M.t -> 'a -> 'b -> 'b * 'c) -> 'a map -> 'b -> 'b * 'c map
   val height : 'a map -> int
   val filter_range : (M.t -> int) -> 'a map -> 'a map
   val filter_map: (M.t -> 'a -> 'b option) -> 'a map -> 'b map (* in OCaml 4.11 *)
@@ -178,6 +194,22 @@ struct
     let accu = f k v (fold_right f r accu) in
     fold_right f l accu
 
+  let rec fold_left_map f (s : 'a map) accu = match map_prj s with
+  | MEmpty -> accu, map_inj MEmpty
+  | MNode {l; v=k; d=v; r; h} ->
+    let accu, l = fold_left_map f l accu in
+    let accu, v = f k v accu in
+    let accu, r = fold_left_map f r accu in
+    accu, map_inj (MNode {l; v=k; d=v; r; h})
+
+  let rec fold_right_map f (s : 'a map) accu = match map_prj s with
+  | MEmpty -> accu, map_inj MEmpty
+  | MNode {l; v=k; d=v; r; h} ->
+    let accu, r = fold_right_map f r accu in
+    let accu, v = f k v accu in
+    let accu, l = fold_right_map f l accu in
+    accu, map_inj (MNode {l; v=k; d=v; r; h})
+
   let height s = match map_prj s with
   | MEmpty -> 0
   | MNode {h;_} -> h
