@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -30,7 +30,7 @@ sig
   val is_sub : string -> string -> int -> bool
   val is_prefix : string -> string -> bool
   val is_suffix : string -> string -> bool
-  module Set : Set.S with type elt = t
+  module Set : CSet.ExtS with type elt = t
   module Map : CMap.ExtS with type key = t and module Set := Set
   module Pred : Predicate.S with type elt = t
   module List : CList.MonoS with type elt = t
@@ -178,7 +178,7 @@ struct
   let compare = compare
 end
 
-module Set = Set.Make(Self)
+module Set = CSet.Make(Self)
 module Map = CMap.Make(Self)
 module Pred = Predicate.Make(Self)
 
