@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -205,16 +205,10 @@ struct
     let fold _ m accu = Set.fold (fun x accu -> x :: accu) m accu in
     Int.Map.fold fold s []
 
-  let min_elt _ = assert false (** Cannot be implemented efficiently *)
-
-  let max_elt _ = assert false (** Cannot be implemented efficiently *)
-
   let choose s =
     let (_, m) = Int.Map.choose s in
     Set.choose m
 
-  let split s x = assert false (** Cannot be implemented efficiently *)
-
 end
 
 module Make(M : HashedType) =
@@ -341,14 +335,6 @@ struct
     let fold _ m accu = Map.fold (fun k x accu -> (k, x) :: accu) m accu in
     Int.Map.fold fold s []
 
-  let min_binding _ = assert false (** Cannot be implemented efficiently *)
-
-  let max_binding _ = assert false (** Cannot be implemented efficiently *)
-
-  let fold_left _ _ _ = assert false (** Cannot be implemented efficiently *)
-
-  let fold_right _ _ _ = assert false (** Cannot be implemented efficiently *)
-
   let choose s =
     let (_, m) = Int.Map.choose s in
     Map.choose m
@@ -377,8 +363,6 @@ struct
     let m = Int.Map.get h s in
     Map.get k m
 
-  let split k s = assert false (** Cannot be implemented efficiently *)
-
   let map f s =
     let fs m = Map.map f m in
     Int.Map.map fs s
@@ -455,9 +439,6 @@ struct
       let ff _ m accu = ExtM.fold f m accu in
       IntM.fold ff s accu
 
-    let fold_left _ _ _ = assert false
-    let fold_right _ _ _ = assert false
-
     let mapi f s =
       IntM.mapi (fun _ m -> ExtM.mapi f m) s
   end
