@@ -39,289 +39,315 @@ function project {
 
 }
 
+# subproject <name> <parent project> <submodulefolder> <submodule giturl> <submodule branch>
+# In the case of nested submodules, each subproject should be declared
+# a subproject of its immediate parent, to ensure overlays are applied
+# in the right order
+function subproject {
+  local var_parent_project=${1}_CI_PARENT_PROJECT
+  local var_submodule_folder=${1}_CI_SUBMODULE_FOLDER
+  local var_submodule_giturl=${1}_CI_SUBMODULE_GITURL
+  local var_submodule_branch=${1}_CI_SUBMODULE_BRANCH
+  local parent_project=$2
+  local submodule_folder=$3
+  local submodule_giturl=$4
+  local submodule_branch=$5
+
+  # register the project in the list of projects
+  projects[${#projects[*]}]=$1
+
+  : "${!var_parent_project:=$parent_project}"
+  : "${!var_submodule_folder:=$submodule_folder}"
+  : "${!var_submodule_giturl:=$submodule_giturl}"
+  : "${!var_submodule_branch:=$submodule_branch}"
+}
+
 ########################################################################
 # MathComp
 ########################################################################
-project mathcomp 'https://github.com/math-comp/math-comp' '28837b16a2358f97d31424fc332498a4780c7afe'
+project mathcomp 'https://github.com/math-comp/math-comp' '21781acfe13bce43aa4b53126c3463ce4a282ca9'
 # Contact @CohenCyril, @proux01 on github
 
-project mathcomp_1 'https://github.com/math-comp/math-comp' '26c1db2a83deffba627719c3e9b754b2bb83e83b'
-# Contact @CohenCyril, @proux01 on github
-
-project fourcolor 'https://github.com/math-comp/fourcolor' '43719c0fb5fb6cb0c8fc1c2db09efc632c23df90'
+project fourcolor 'https://github.com/math-comp/fourcolor' '1739214a760d7844e971cb8170707b0395f55c6d'
 # Contact @ybertot, @proux01 on github
 
-project oddorder 'https://github.com/math-comp/odd-order' '8dbbae0e53a6d1fcf3471c8fae4dd14c8f18bd93'
+project oddorder 'https://github.com/math-comp/odd-order' '2bee03dd95d4282f0c13d2eec748a2dcf92fc473'
 # Contact @gares, @proux01 on github
 
-project mczify 'https://github.com/math-comp/mczify' '371b29c7eb30ce408fa29feaf7b94146eed6660b'
+project mczify 'https://github.com/math-comp/mczify' '0d2b9fd068f6477b3d8ef53b302e6f2719f71d78'
 # Contact @pi8027 on github
 
-project finmap "https://github.com/math-comp/finmap" "cea9f088c9cddea1173bc2f7c4c7ebda35081b60"
-# put back master when Analysis master moves to MathComp 2
-# project finmap "https://github.com/math-comp/finmap" "master"
+project finmap 'https://github.com/math-comp/finmap' 'a2f06d1f0bb05afce1f5a9bb73b6398e521bf3a9'
 # Contact @CohenCyril on github
 
-project bigenough 'https://github.com/math-comp/bigenough' 'ff71b25f31658d80fdae9657e8cf34e5e1052647'
+project bigenough 'https://github.com/math-comp/bigenough' '5a5ff278089847d31a28067eb7ca4fad6a49f11c'
 # Contact @CohenCyril on github
 
-project analysis 'https://github.com/math-comp/analysis' 'b0d400f667f98cedde05414cff99853eb86642b8'
+project analysis 'https://github.com/math-comp/analysis' '3cff9bb331cb0d72bce95bed8e279f93b6d5e23a'
 # Contact @affeldt-aist, @CohenCyril on github
 
 ########################################################################
 # UniMath
 ########################################################################
-project unimath 'https://github.com/UniMath/UniMath' 'ee7af93f47daf41c752e6bd212a7aba78b6a3aaa'
+project unimath 'https://github.com/UniMath/UniMath' '0f7bb63aae9fb6179d653d6c39cb1e8e72f2c789'
 # Contact @benediktahrens, @m-lindgren, @nmvdw, @rmatthes on github
 
 ########################################################################
 # Unicoq + Mtac2
 ########################################################################
-project unicoq 'https://github.com/unicoq/unicoq' 'e0486d27f687c8c34f530158fee743b3e7a3c259'
+project unicoq 'https://github.com/unicoq/unicoq' 'a9b72f755539c0b3280e38e778a09e2b7519a51a'
 # Contact @beta-ziliani, @Janno, @mattam82 on github
 
-project mtac2 'https://github.com/Mtac2/Mtac2' '71640c37a3685db5c7aab069cede807faadacc25'
+project mtac2 'https://github.com/Mtac2/Mtac2' '1cdb2cb628444ffe9abc6535f6d2e11004de7fc1'
 # Contact @beta-ziliani, @Janno, @mattam82 on github
 
 ########################################################################
 # Mathclasses + Corn
 ########################################################################
-project math_classes 'https://github.com/coq-community/math-classes' 'f9712ba67b59c90446522e102c87200c7da756ec'
-# Contact @spitters on github
+project math_classes 'https://github.com/coq-community/math-classes' '9db314291e4db1fd30bb0ef21292c78920298a6e'
+# Contact @Lysxia and @spitters on github
 
-project corn 'https://github.com/coq-community/corn' '79e7ee7cc878cd98077d16cfbac12b8133a13805'
-# Contact @spitters on github
+project corn 'https://github.com/coq-community/corn' 'a75b14fad50313d615881713d4703e6c9cd6e111'
+# Contact @Lysxia and @spitters on github
 
 ########################################################################
 # Iris
 ########################################################################
 
-# NB: stdpp and Iris refs are gotten from the opam files in the Iris
-# and iris_examples repos respectively.
+# NB: stdpp and Iris refs are gotten from the opam files in the Iris and
+# iris_examples repos respectively. So just getting a fix landed in stdpp or
+# Iris is not enough. Ping @RalfJung and @robbertkrebbers if you need the
+# versions of stdpp or Iris to be bumped. Perennial also has its own pinned
+# versions of stdpp and Iris; ping @tchajed and @zeldovich to get that bumped.
 project stdpp "https://gitlab.mpi-sws.org/iris/stdpp" ""
 # Contact @RalfJung, @robbertkrebbers on github
 
 project iris "https://gitlab.mpi-sws.org/iris/iris" ""
 # Contact @RalfJung, @robbertkrebbers on github
 
-project autosubst 'https://github.com/coq-community/autosubst' '6ba0acccef68c75f6cca8928706c726754d69791'
+project autosubst 'https://github.com/coq-community/autosubst' '50dfe574c0bd415925eea47c1f5b1a533aa85269'
 # Contact @RalfJung, @co-dan on github
 
-project iris_examples 'https://gitlab.mpi-sws.org/iris/examples' '7c5ce0d556f4e03f27ddc6463462bf4040f5103d'
+project iris_examples 'https://gitlab.mpi-sws.org/iris/examples' '4f54a3254f47e2d65f92e20a56b97cc2098ac00b'
 # Contact @RalfJung, @robbertkrebbers on github
 
 ########################################################################
 # HoTT
 ########################################################################
-project hott 'https://github.com/HoTT/HoTT' '191c10ed39de2fa9821e63b6578f4fba9175ab31'
+project hott 'https://github.com/HoTT/HoTT' 'bb95699e33b1a496d7d31e418c0cb837b2acbaa3'
 # Contact @Alizter, @jdchristensen on github
 
 ########################################################################
 # CoqHammer
 ########################################################################
-project coqhammer 'https://github.com/lukaszcz/coqhammer' '4131a8ccac570336b5ffefd26fd40d879f1e115f'
+project coqhammer 'https://github.com/lukaszcz/coqhammer' '31442e8178a5d85a9f57a323b65bf9f719ded8ec'
 # Contact @lukaszcz on github
 
 ########################################################################
 # Flocq
 ########################################################################
-project flocq 'https://gitlab.inria.fr/flocq/flocq' '3e8b2e09d79ce3c150a732dd446610d63dabc759'
+project flocq 'https://gitlab.inria.fr/flocq/flocq' '0a685a5c8b86f992e74a3e675acf882afacf2de6'
 # Contact @silene on github
 
 ########################################################################
 # coq-performance-tests
 ########################################################################
-project coq_performance_tests 'https://github.com/coq-community/coq-performance-tests' 'c5e2f8d0b0057a003dcaec727ba01f38e519bf6d'
+project coq_performance_tests 'https://github.com/coq-community/coq-performance-tests' '0e14c7b4475d5ea7c3a6aa416f4a761b0e9f8b2f'
 # Contact @JasonGross on github
 
 ########################################################################
 # coq-tools
 ########################################################################
-project coq_tools 'https://github.com/JasonGross/coq-tools' '32dcc6b48607c34442e64789eaaf894dac3dd020'
+project coq_tools 'https://github.com/JasonGross/coq-tools' 'f513668494d3fb9541cee98625ed8488678048a7'
 # Contact @JasonGross on github
 
 ########################################################################
 # Coquelicot
 ########################################################################
-project coquelicot 'https://gitlab.inria.fr/coquelicot/coquelicot' '218b5cca041712775999f3c5ab0cac2f99defb7f'
+project coquelicot 'https://gitlab.inria.fr/coquelicot/coquelicot' '88672c1eac54f78ee733c42c0bdb2ac9905aad1e'
 # Contact @silene on github
 
 ########################################################################
 # CompCert
 ########################################################################
-project compcert 'https://github.com/AbsInt/CompCert' 'c7b27e30a9057c14d5a02a9f9aa3d8f230715eb0'
+project compcert 'https://github.com/AbsInt/CompCert' '54341fe9728ca74fb446145c327220af33f8a701'
 # Contact @xavierleroy on github
 
 ########################################################################
 # VST
 ########################################################################
-project vst 'https://github.com/PrincetonUniversity/VST' '6d2eb96dd42a4d3ee398eef8224e6a8ccca5c750'
+project vst 'https://github.com/PrincetonUniversity/VST' '21f3ac51a1e53ec74d5926a55c5ef9d4d62fda51'
 # Contact @andrew-appel on github
 
 ########################################################################
 # cross-crypto
 ########################################################################
-project cross_crypto 'https://github.com/mit-plv/cross-crypto' 'c46bbe27ef7a05a82f744230c6f2d5f1fdefdaaa'
+project cross_crypto 'https://github.com/mit-plv/cross-crypto' '433e91211a333e73566b0eaa4da1a7d12ed972aa'
 # Contact @andres-erbsen on github
 
 ########################################################################
 # rewriter
 ########################################################################
-project rewriter 'https://github.com/mit-plv/rewriter' '026f87bfbc5d235c3f7f31b24199305d616daff3'
+project rewriter 'https://github.com/mit-plv/rewriter' '30c8507c1e30626b2aa1e15c0aa7c23913da033c'
 # Contact @JasonGross on github
 
 ########################################################################
 # fiat_parsers
 ########################################################################
-project fiat_parsers 'https://github.com/mit-plv/fiat' '157f13408ef9731ccf0829fbcac3c11b9856fb81'
+project fiat_parsers 'https://github.com/mit-plv/fiat' '41a0b8dd92a252fcf7491ee99e16eb002288e87e'
+# Contact @JasonGross on github
+
+########################################################################
+# fiat_crypto_legacy
+########################################################################
+project fiat_crypto_legacy 'https://github.com/mit-plv/fiat-crypto' '9e79fa6d3e8427c4f75321dbd426c86c483891c1'
 # Contact @JasonGross on github
 
 ########################################################################
 # fiat_crypto
 ########################################################################
-project fiat_crypto 'https://github.com/mit-plv/fiat-crypto' '4d9fadc163b8a09d75a6836e08f52f051fa489d4'
+project fiat_crypto 'https://github.com/mit-plv/fiat-crypto' 'eedc9c86c55487cfcf7b40234e8780b8bd8b24db'
 # Contact @andres-erbsen, @JasonGross on github
 
-########################################################################
-# fiat_crypto_legacy
-########################################################################
-project fiat_crypto_legacy 'https://github.com/mit-plv/fiat-crypto' '47058dc312a5cad924972056cdfc4de06c432a82'
-# Contact @JasonGross on github
+# bedrock2, coqutil, rupicola, kami, riscv_coq
+# fiat-crypto is not guaranteed to build with the latest version of
+# bedrock2, so we use the pinned version of bedrock2 for fiat-crypto
+# overlays do not have to follow suite
+subproject rupicola fiat_crypto "rupicola" "https://github.com/mit-plv/rupicola" "master"
+subproject bedrock2 rupicola "bedrock2" "https://github.com/mit-plv/bedrock2" "master"
+subproject coqutil bedrock2 "deps/coqutil" "https://github.com/mit-plv/coqutil" "master"
+subproject kami bedrock2 "deps/kami" "https://github.com/mit-plv/kami" "rv32i"
+subproject riscv_coq bedrock2 "deps/riscv-coq" "https://github.com/mit-plv/riscv-coq" "master"
+# Contact @samuelgruetter, @andres-erbsen on github
 
 ########################################################################
 # coq_dpdgraph
 ########################################################################
-project coq_dpdgraph 'https://github.com/coq-community/coq-dpdgraph' '8452ebdebb34f66c0b87b39b757784090772fd49'
+project coq_dpdgraph 'https://github.com/coq-community/coq-dpdgraph' '74ced1b11a8df8d4c04c3829fcf273aa63d2c493'
 # Contact @Karmaki, @ybertot on github
 
 ########################################################################
 # CoLoR
 ########################################################################
-project color 'https://github.com/fblanqui/color' '42f4b271d41b81d59508936a375ac2f9fafe6b9c'
+project color 'https://github.com/fblanqui/color' 'c867546cea21f74aad86e0a34ceaf93ed9259f96'
 # Contact @fblanqui on github
 
 ########################################################################
 # TLC
 ########################################################################
-project tlc 'https://github.com/charguer/tlc' '1f1df8cfbdfb1838a446fa1614c8572486c0478f'
+project tlc 'https://github.com/charguer/tlc' '51be762eedec72788490f1fd222eb8b0d82b8bea'
 # Contact @charguer on github
 
 ########################################################################
 # Bignums
 ########################################################################
-project bignums 'https://github.com/coq/bignums' 'cb65c878e764793081ead5fdcaffbe018dd713c3'
+project bignums 'https://github.com/coq/bignums' 'cc2d9ee990e4cfebe5f107d8357198baa526eded'
 # Contact @erikmd, @proux01 on github
 
 ########################################################################
 # coqprime
 ########################################################################
-project coqprime 'https://github.com/thery/coqprime' 'd5935ca3b7f3d2f738b0810a49858d17480d6a58'
+project coqprime 'https://github.com/thery/coqprime' '845c00cadc521aa70c3a8ae3bef9e2fabcbabf19'
 # Contact @thery on github
 
 ########################################################################
 # bbv
 ########################################################################
-project bbv 'https://github.com/mit-plv/bbv' 'f4caa0504125460afa41fb92c9071dd628154d8b'
+project bbv 'https://github.com/mit-plv/bbv' 'c53d5b95c70839ae8e947010d85503bd1ada1120'
 # Contact @JasonGross, @samuelgruetter on github
 
-########################################################################
-# bedrock2
-########################################################################
-project bedrock2 'https://github.com/mit-plv/bedrock2' '65988d2662e522b0a7dcd04ce66c47acb5c66df1'
-# Contact @samuelgruetter, @andres-erbsen on github
-
 ########################################################################
 # Coinduction
 ########################################################################
-project coinduction 'https://github.com/damien-pous/coinduction' '65d020794e817fcf0295566abe1be454ad3284e4'
+project coinduction 'https://github.com/damien-pous/coinduction' '09caaf1f809e3e91ebba05bc38cef6de83ede3b3'
 # Contact @damien-pous on github
 
 ########################################################################
 # coq-lsp
 ########################################################################
-project coq_lsp 'https://github.com/ejgallego/coq-lsp' '75a92e0abeb29cbc3f75bbffa60db04857789070'
+project coq_lsp 'https://github.com/ejgallego/coq-lsp' '25d8b8672beaa39bafa1f91d62910cad49c381f9'
 # Contact @ejgallego on github
 
 ########################################################################
 # Equations
 ########################################################################
-project equations 'https://github.com/mattam82/Coq-Equations' 'a517a3e128719b078397a415dc9ce6b856019245'
+project equations 'https://github.com/mattam82/Coq-Equations' '3431c88a7575a3972191c25809c563aafdc6414e'
 # Contact @mattam82 on github
 
 ########################################################################
 # Elpi + Hierarchy Builder
 ########################################################################
-project elpi 'https://github.com/LPCIC/coq-elpi' 'b753aa43a643ad7c0b2722d9850253b9d737fec7'
+project elpi 'https://github.com/LPCIC/coq-elpi' '185416f6e157e1a2b0502dc8a3880fd02a6a2358'
 # Contact @gares on github
 
-project hierarchy_builder 'https://github.com/math-comp/hierarchy-builder' 'b4e8cbce83cff0f817494adc625a127bd4114b2e'
+project hierarchy_builder 'https://github.com/math-comp/hierarchy-builder' 'be3bc5943053d5186e8933b510e663ac40a25f6b'
 # Contact @CohenCyril, @gares on github
 
 ########################################################################
 # Engine-Bench
 ########################################################################
-project engine_bench 'https://github.com/mit-plv/engine-bench' '8c0b15abc38b1d3f7dc606f4eeef9fba1a986b05'
+project engine_bench 'https://github.com/mit-plv/engine-bench' '08ecd3ae6e73ff6e62b47fd62f5c57e4ec4fb42d'
 # Contact @JasonGross on github
 
 ########################################################################
 # fcsl-pcm
 ########################################################################
-project fcsl_pcm 'https://github.com/imdea-software/fcsl-pcm' '35cb0a2235adb656c2696e33512f0c4555ad753b'
+project fcsl_pcm 'https://github.com/imdea-software/fcsl-pcm' '2da486d949709643ca54026a6a24d5ee36baafc0'
 # Contact @aleksnanevski, @clayrat on github
 
 ########################################################################
 # ext-lib
 ########################################################################
-project ext_lib 'https://github.com/coq-community/coq-ext-lib' '00d3f4e2a260c7c23d2c0b9cbc69516f8be4ac92'
+project ext_lib 'https://github.com/coq-community/coq-ext-lib' 'b27e806daf39a8f1cfc7ced09c1af44d390af4a6'
 # Contact @gmalecha, @liyishuai on github
 
 ########################################################################
 # simple-io
 ########################################################################
-project simple_io 'https://github.com/Lysxia/coq-simple-io' '7b1e6de2c2028a35776f080dc3db117cae117836'
+project simple_io 'https://github.com/Lysxia/coq-simple-io' '59ca9ed99f7e178e40a32e3ea513dfe9be3a12aa'
 # Contact @Lysxia, @liyishuai on github
 
 ########################################################################
 # quickchick
 ########################################################################
-project quickchick 'https://github.com/QuickChick/QuickChick' '7c570ed650f770e29ea6a84d49687b1c0c2b6b72'
+project quickchick 'https://github.com/QuickChick/QuickChick' '7b33d19066aa762629cbbe210d41067f56dce000'
 # Contact @lemonidas, @Lysxia, @liyishuai on github
 
 ########################################################################
 # reduction-effects
 ########################################################################
-project reduction_effects 'https://github.com/coq-community/reduction-effects' 'f0570f498bc8a0d25e878115b4066b140908c4b4'
+project reduction_effects 'https://github.com/coq-community/reduction-effects' '2e6590125885906ec0c5157b114d2afba18ab9c8'
 # Contact @liyishuai, @JasonGross on github
 
 ########################################################################
 # menhirlib
 ########################################################################
 # Note: menhirlib is now in subfolder coq-menhirlib of menhir
-project menhirlib 'https://gitlab.inria.fr/fpottier/menhir' 'bb68b8fdb186266dbed67599cfa69a69a3c01272'
+project menhirlib 'https://gitlab.inria.fr/fpottier/menhir' '809af6503305acaf6d9877e67c40ef651452e115'
 # Contact @fpottier, @jhjourdan on github
 
 ########################################################################
 # coq-neural-net-interp
 ########################################################################
-project neural_net_interp 'https://github.com/JasonGross/neural-net-coq-interp' '06509a99be412d06847de938a19e02be64f6fc4c'
+project neural_net_interp 'https://github.com/JasonGross/neural-net-coq-interp' '1a3d72129fa5e6f05e093e9e19e94f2110c73d0f'
 # Contact @JasonGross on github
 
 ########################################################################
 # aac_tactics
 ########################################################################
-project aac_tactics 'https://github.com/coq-community/aac-tactics' '059dd94f4fbfe23e93d5c3de800cfb71f6a0ed43'
+project aac_tactics 'https://github.com/coq-community/aac-tactics' '109af844f39bf541823271e45e42e40069f3c2c4'
 # Contact @palmskog on github
 
 ########################################################################
 # paco
 ########################################################################
-project paco 'https://github.com/snu-sf/paco' '5c5693f46c8957f36a2349a0d906e911366136de'
+project paco 'https://github.com/snu-sf/paco' '5efd5da088e26bf4351e3abd258f8a3f848fe00c'
 # Contact @minkiminki on github
 
 ########################################################################
 # coq-itree
 ########################################################################
-project itree 'https://github.com/DeepSpec/InteractionTrees' 'dda104937d79e2052d1a26f6cbe89429245ff743'
+project itree 'https://github.com/DeepSpec/InteractionTrees' '1e8bc121ec8e8d6f699f0231bb7f92ebfd54f4b9'
 # Contact @Lysxia on github
 
 ########################################################################
@@ -339,13 +365,13 @@ project ceres 'https://github.com/Lysxia/coq-ceres' 'f61b24d48222db0100de19f88c1
 ########################################################################
 # coq-parsec
 ########################################################################
-project parsec 'https://github.com/liyishuai/coq-parsec' '2efb4437f8451dfbeb174368f860e629135c08ab'
+project parsec 'https://github.com/liyishuai/coq-parsec' '3feabc998705927ca2d2f9249a21a6e15c394162'
 # Contact @liyishuai on github
 
 ########################################################################
 # coq-json
 ########################################################################
-project json 'https://github.com/liyishuai/coq-json' '2d1d11eb1be9ad96614d9cd224c1df7c75ec2869'
+project json 'https://github.com/liyishuai/coq-json' '8069b6b4d544ea3011246dfa04d2c1c550d50455'
 # Contact @liyishuai on github
 
 ########################################################################
@@ -363,13 +389,13 @@ project http 'https://github.com/liyishuai/coq-http' 'cabde79a4a0d978d031475c744
 ########################################################################
 # paramcoq
 ########################################################################
-project paramcoq 'https://github.com/coq-community/paramcoq' 'a61b1b3f8cea3fce6c904a0b8da6236345ed896f'
+project paramcoq 'https://github.com/coq-community/paramcoq' '32609ca4a9bf4a0e456a855ea5118d8c00cda6be'
 # Contact @proux01 on github
 
 ########################################################################
 # relation_algebra
 ########################################################################
-project relation_algebra 'https://github.com/damien-pous/relation-algebra' '38cfe670448c90406fb72ff40526ad4d08b75319'
+project relation_algebra 'https://github.com/damien-pous/relation-algebra' '7966d1a7bb524444120c56c3474717bcc91a5215'
 # Contact @damien-pous on github
 
 ########################################################################
@@ -384,34 +410,34 @@ project inf_seq_ext 'https://github.com/DistributedComponents/InfSeqExt' '601e89
 project cheerios 'https://github.com/uwplse/cheerios' '5c9318c269f9cae1c1c6583a44405969ac0be0dd'
 # Contact @palmskog on github
 
-project verdi 'https://github.com/uwplse/verdi' 'f082f86b2a0cb56cc0703a79da351d8b9a6e7468'
+project verdi 'https://github.com/uwplse/verdi' 'b7f77848819878b1faf0e2e6a730f9bb850130be'
 # Contact @palmskog on github
 
 project verdi_raft 'https://github.com/uwplse/verdi-raft' 'a3375e867326a82225e724cc1a7b4758b029376f'
 # Contact @palmskog on github
 
 ########################################################################
-# stdlib2
+# Stdlib
 ########################################################################
-project stdlib2 'https://github.com/coq/stdlib2' '33212e05c51efa012c9dfccd0b9e735a42f2d924'
-# Contact @maximedenes, @vbgl on github
+project stdlib 'https://github.com/coq/stdlib' '2b50da773d975b742f731fe4d0bed033a8d30844'
+# Contact TODO on github
 
 ########################################################################
 # argosy
 ########################################################################
-project argosy 'https://github.com/mit-pdos/argosy' 'a6a5aa0d3868efd4ada0b40927b5748e5d8967d3'
+project argosy 'https://github.com/mit-pdos/argosy' '9fa42b78b7f9b7b989fb3434dfbfec4abcfcbff8'
 # Contact @tchajed on github
 
 ########################################################################
 # ATBR
 ########################################################################
-project atbr 'https://github.com/coq-community/atbr' '39de8e4c654a530238390704677a672275649d12'
+project atbr 'https://github.com/coq-community/atbr' '47ac8fb6bf244d9a4049e04c01e561191490f543'
 # Contact @palmskog, @tchajed on github
 
 ########################################################################
 # perennial
 ########################################################################
-project perennial 'https://github.com/mit-pdos/perennial' 'd045aca7d0bd3c43ec561179954777f160eed1e4'
+project perennial 'https://github.com/mit-pdos/perennial' '12aec9c9829939c32eb8eca85371d713d73293d9'
 # Contact @upamanyus, @RalfJung, @tchajed on github
 # PRs to fix Perennial failures should be submitted against the Perennial
 # `master` branch. `coq/tested` is automatically updated every night to the
@@ -421,103 +447,104 @@ project perennial 'https://github.com/mit-pdos/perennial' 'd045aca7d0bd3c43ec561
 ########################################################################
 # metacoq
 ########################################################################
-project metacoq 'https://github.com/MetaCoq/metacoq' '194116d4f80dc53253f8730aa822c52225731d71'
+project metacoq 'https://github.com/MetaCoq/metacoq' '17ba45ffc84d37e187ef87a55b840890f1d87f01'
 # Contact @mattam82, @yforster on github
 
 ########################################################################
 # SF suite
 ########################################################################
-project sf 'https://github.com/DeepSpec/sf' 'd7f9ab78cd01f36ce5a2a359dfe4cfe1e160fae9'
+project sf 'https://github.com/DeepSpec/sf' '5e863a9f92e515a0e11641f28077a64919f8482e'
 # Contact @bcpierce00, @liyishuai on github
 
 ########################################################################
 # Coqtail
 ########################################################################
-project coqtail "https://github.com/whonore/Coqtail" "ecca5017b6b27e1bded009f9725186d941d85b53"
+project coqtail 'https://github.com/whonore/Coqtail' '9bd2a1818e08fd6e52d72024a60c51a983770c81'
 # Contact @whonore on github
 
 ########################################################################
 # Deriving
 ########################################################################
-project deriving 'https://github.com/arthuraa/deriving' '9d6aea3df446854aa1aa009e50954e2fd0c02981'
+project deriving 'https://github.com/arthuraa/deriving' '9dcb5de11b26afa8c14a137f164413a61612a53d'
 # Contact @arthuraa on github
 
 ########################################################################
 # VsCoq
 ########################################################################
-project vscoq 'https://github.com/coq-community/vscoq' '84032c7f46e0a21a01da91e5364cfc7d4951e297'
+project vscoq 'https://github.com/coq-community/vscoq' '36f5a395a28cb5d6c556daeae3b287295304d00e'
 # Contact @rtetley, @gares on github
 
 ########################################################################
 # category-theory
 ########################################################################
-project category_theory 'https://github.com/jwiegley/category-theory' '41ab38d0caffaa5176a6f63810ba8d92816514a4'
+project category_theory 'https://github.com/jwiegley/category-theory' '255751fab5650c45095bbc2ac156ca9fcac17c9c'
 # Contact @jwiegley on github
 
 ########################################################################
 # itauto
 ########################################################################
-project itauto 'https://gitlab.inria.fr/fbesson/itauto' 'a527a27d9dd21ac8f69821f35fd677314d5eadbe'
+project itauto 'https://gitlab.inria.fr/fbesson/itauto' 'c13c6b4a0070ecc3ae8ea9755a1d6a163d123127'
 # Contact @fajb on github
 
 ########################################################################
 # Mathcomp-word
 ########################################################################
-project mathcomp_word "https://github.com/jasmin-lang/coqword" "537141ac0dbdd0a0ca2b409826d93a6f8398b0b3"
+project mathcomp_word 'https://github.com/jasmin-lang/coqword' '1abe5ade5240115aed1e3c140e261f1554af2322'
 # Contact @vbgl, @strub on github
-# go back to "main" and change dependency to MC 2 when
-# https://github.com/jasmin-lang/jasmin/pull/560 is merged
 
 ########################################################################
 # Jasmin
 ########################################################################
-project jasmin 'https://github.com/jasmin-lang/jasmin' 'd195c58958771ba26bc479fef9a5f04fd219290e'
+project jasmin 'https://github.com/jasmin-lang/jasmin' '41068bc89855cd0f015e78472e32fda9230e3218'
 # Contact @vbgl, @bgregoir on github
 
 ########################################################################
 # Lean Importer
 ########################################################################
-project lean_importer 'https://github.com/SkySkimmer/coq-lean-import' '10363695d2f9692b4d8a9e7ee29be444ebe96c80'
+project lean_importer 'https://github.com/SkySkimmer/coq-lean-import' 'c513cee4f5edf8e8a06ba553ca58de5142cffde6'
 # Contact @SkySkimmer on github
 
-########################################################################
-# SerAPI
-########################################################################
-project serapi 'https://github.com/ejgallego/coq-serapi' 'ba0bd1d3fb086b7a7c79adc68d1aa53e9d883d42'
-# Contact @ejgallego on github
-
 ########################################################################
 # SMTCoq
 ########################################################################
-project smtcoq 'https://github.com/smtcoq/smtcoq' 'e24b14768d6f61c915bc25734d769bbe9f044775'
+project smtcoq 'https://github.com/smtcoq/smtcoq' '5c6033c906249fcf98a48b4112f6996053124514'
+# Contact @ckeller on github
+
+project smtcoq_trakt 'https://github.com/smtcoq/smtcoq' '9392f7446a174b770110445c155a07b183cdca3d'
 # Contact @ckeller on github
 
 ########################################################################
 # Stalmarck
 ########################################################################
-project stalmarck 'https://github.com/coq-community/stalmarck' '22a05ddcc6a826ffa772282f1d3a0902d3921e7b'
+project stalmarck 'https://github.com/coq-community/stalmarck' 'd32acd3c477c57b48dd92bdd96d53fb8fa628512'
 # Contact @palmskog on github
 
-########################################################################
-# coq-library-undecidability
-########################################################################
-project coq_library_undecidability 'https://github.com/uds-psl/coq-library-undecidability' '91165e2e95d4372d692aca05f5d53ab628e3c8b9'
-# Contact @mrhaandi, @yforster on github
-
 ########################################################################
 # Tactician
 ########################################################################
-project tactician 'https://github.com/coq-tactician/coq-tactician' '938ee07cba9719753f7f975addc611cd115f3504'
+project tactician 'https://github.com/coq-tactician/coq-tactician' '62efb022e34ce141306be9c646f4e2f3fe40c4b0'
 # Contact @LasseBlaauwbroek on github
 
 ########################################################################
 # Ltac2 compiler
 ########################################################################
-project ltac2_compiler 'https://github.com/SkySkimmer/coq-ltac2-compiler' '9fd8b13fd2eb9b3ed49e76e944c5e2f211d7c52d'
+project ltac2_compiler 'https://github.com/SkySkimmer/coq-ltac2-compiler' '30ee5bbf04ab841deb3481465ef4753f2c69338e'
 # Contact @SkySkimmer on github
 
 ########################################################################
 # Waterproof
 ########################################################################
-project waterproof 'https://github.com/impermeable/coq-waterproof' 'aec4fbe2a9b586b246e18b00a52c66ee7a772e79'
+project waterproof 'https://github.com/impermeable/coq-waterproof' '443f794ddc102309d00f1d512ab50b84fdc261aa'
 # Contact @jellooo038, @jim-portegies on github
+
+########################################################################
+# Autosubst (2) OCaml
+########################################################################
+project autosubst_ocaml 'https://github.com/uds-psl/autosubst-ocaml' 'a4e154fb548197572326167809a4c612a2ec71ac'
+# Contact @yforster on github
+
+########################################################################
+# Trakt
+########################################################################
+project trakt 'https://github.com/ecranceMERCE/trakt' '956e78ab3680d186d03d8b58dbd90681174f3157'
+# Contact @ckeller, @louiseddp on github
