@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -16,6 +16,7 @@ open Pp
 open Names
 open Libnames
 open Univ
+open UVars
 open Environ
 open Printer
 open Constr
@@ -23,9 +24,10 @@ open Context
 open Genarg
 open Clenv
 
-let _ = Detyping.print_evar_arguments := true
-let _ = Detyping.print_universes := true
-let _ = Goptions.set_bool_option_value ["Printing";"Matching"] false
+let with_env_evm f x =
+  let env = Global.env() in
+  let sigma = Evd.from_env env in
+  f env sigma x
 
 (* std_ppcmds *)
 let pp   x = Pp.pp_with !Topfmt.std_ft x
@@ -41,22 +43,17 @@ let ppmbid mbid = pp (str (MBId.debug_to_string mbid))
 let ppdir dir = pp (DirPath.print dir)
 let ppmp mp = pp(str (ModPath.debug_to_string mp))
 let ppcon con = pp(Constant.debug_print con)
-let ppproj con = pp(Constant.debug_print (Projection.constant con))
+let ppprojrepr con = pp(Constant.debug_print (Projection.Repr.constant con))
+let ppproj p = pp(Projection.debug_print p)
 let ppkn kn = pp(str (KerName.to_string kn))
 let ppmind kn = pp(MutInd.debug_print kn)
 let ppind (kn,i) = pp(MutInd.debug_print kn ++ str"," ++int i)
 let ppsp sp = pp(pr_path sp)
 let ppqualid qid = pp(pr_qualid qid)
-let ppclindex cl = pp(Coercionops.pr_cl_index cl)
 let ppscheme k = pp (Ind_tables.pr_scheme_kind k)
 
-let prrecarg = function
-  | Declarations.Norec -> str "Norec"
-  | Declarations.Mrec (mind,i) ->
-     str "Mrec[" ++ MutInd.print mind ++ pr_comma () ++ int i ++ str "]"
-  | Declarations.Imbr (mind,i) ->
-     str "Imbr[" ++ MutInd.print mind ++ pr_comma () ++ int i ++ str "]"
-let ppwf_paths x = pp (Rtree.pp_tree prrecarg x)
+let pprecarg r = pp (Declareops.pr_recarg r)
+let ppwf_paths x = pp (Declareops.pr_wf_paths x)
 
 let get_current_context () =
   try Vernacstate.Declare.get_current_context ()
@@ -67,7 +64,6 @@ let get_current_context () =
 
 (* term printers *)
 let envpp pp = let sigma,env = get_current_context () in pp env sigma
-let rawdebug = ref false
 let ppevar evk = pp (Evar.print evk)
 let pr_constr t =
   let sigma, env = get_current_context () in
@@ -78,14 +74,31 @@ let pr_econstr t =
 let ppconstr x = pp (pr_constr x)
 let ppeconstr x = pp (pr_econstr x)
 let ppconstr_expr x = let sigma,env = get_current_context () in pp (Ppconstr.pr_constr_expr env sigma x)
-let ppsconstr x = ppconstr (Mod_subst.force_constr x)
 let ppconstr_univ x = Constrextern.with_universes ppconstr x
-let ppglob_constr = (fun x -> pp(pr_lglob_constr_env (Global.env()) x))
+let ppglob_constr = (fun x -> pp(with_env_evm pr_lglob_constr_env x))
 let pppattern = (fun x -> pp(envpp pr_constr_pattern_env x))
 let pptype = (fun x -> try pp(envpp (fun env evm t -> pr_ltype_env env evm t) x) with e -> pp (str (Printexc.to_string e)))
 let ppfconstr c = ppconstr (CClosure.term_of_fconstr c)
 
-let ppbigint n = pp (str (Bigint.to_string n));;
+let ppuint63 i = pp (str (Uint63.to_string i))
+
+let pp_parray pr a =
+  let a, def = Parray.to_array a in
+  let a = Array.to_list a in
+  pp (str "[|" ++ prlist_with_sep (fun () -> str ";" ++ spc()) pr a ++ spc() ++ str "|" ++ spc() ++ pr def ++ str "|]")
+
+let pp_constr_parray = pp_parray pr_constr
+let pp_fconstr_parray = pp_parray (fun f -> pr_constr (CClosure.term_of_fconstr f))
+
+let ppfsubst s =
+  let (s, k) = Esubst.Internal.repr s in
+  let sep () = str ";" ++ spc () in
+  let pr = function
+  | Esubst.Internal.REL n -> str "<#" ++ int n ++ str ">"
+  | Esubst.Internal.VAL (k, x) -> pr_constr (Vars.lift k (CClosure.term_of_fconstr x))
+  in
+  pp @@ str "[" ++ prlist_with_sep sep pr s ++ str "| " ++ int k ++ str "]"
+
 let ppnumtokunsigned n = pp (NumTok.Unsigned.print n)
 let ppnumtokunsignednat n = pp (NumTok.UnsignedNat.print n)
 
@@ -100,11 +113,35 @@ let pridmap pr l =
   prset' pr (Id.Map.fold (fun a b l -> (a,b)::l) l [])
 let ppidmap pr l = pp (pridmap pr l)
 
-let pridmapgen l =
-  let dom = Id.Set.elements (Id.Map.domain l) in
+let prmapgen pr dom =
   if dom = [] then str "[]" else
-  str "[domain= " ++ hov 0 (prlist_with_sep spc Id.print dom) ++ str "]"
+  str "[domain= " ++ hov 0 (prlist_with_sep spc pr dom) ++ str "]"
+let pridmapgen l = prmapgen Id.print (Id.Set.elements (Id.Map.domain l))
 let ppidmapgen l = pp (pridmapgen l)
+let printmapgen l = prmapgen int (Int.Set.elements (Int.Map.domain l))
+let ppintmapgen l = pp (printmapgen l)
+let prmodidmapgen l = prmapgen Id.print (ModIdset.elements (ModIdmap.domain l))
+let ppmodidmapgen l = pp (prmodidmapgen l)
+
+let ppmpmapgen l =
+  pp (prmapgen
+        (fun mp -> str (ModPath.debug_to_string mp))
+        (MPset.elements (MPmap.domain l)))
+
+let ppdpmapgen l =
+  pp (prmapgen
+        (fun mp -> str (DirPath.to_string mp))
+        (DPset.elements (DPmap.domain l)))
+
+let ppconmapenvgen l =
+  pp (prmapgen
+        (fun mp -> str (Constant.debug_to_string mp))
+        (Cset_env.elements (Cmap_env.domain l)))
+
+let ppmindmapenvgen l =
+  pp (prmapgen
+        (fun mp -> str (MutInd.debug_to_string mp))
+        (Mindmap_env.Set.elements (Mindmap_env.domain l)))
 
 let ppevarsubst = ppidmap (fun id0 -> prset (fun (c,copt,id) ->
   hov 0
@@ -134,7 +171,7 @@ let rec pr_closure {idents=idents;typed=typed;untyped=untyped} =
 and pr_closed_glob_constr_idmap x =
   pridmap (fun _ -> pr_closed_glob_constr) x
 and pr_closed_glob_constr {closure=closure;term=term} =
-  pr_closure closure ++ (pr_lglob_constr_env Global.(env ())) term
+  pr_closure closure ++ with_env_evm pr_lglob_constr_env term
 
 let ppclosure x = pp (pr_closure x)
 let ppclosedglobconstr x = pp (pr_closed_glob_constr x)
@@ -142,6 +179,14 @@ let ppclosedglobconstridmap x = pp (pr_closed_glob_constr_idmap x)
 
 let pP s = pp (hov 0 s)
 
+let pp_as_format s =
+  let fmt, args = pp_as_format s in
+  let pr_escaped s = str ("\"" ^ String.escaped s ^ "\"") in
+  pp
+    (hov 0
+       (str "printf" ++ spc() ++ str "\"" ++ str fmt ++ str "\"" ++
+        pr_non_empty_arg (prlist_with_sep spc pr_escaped) args))
+
 let safe_pr_global = let open GlobRef in function
   | ConstRef kn -> pp (str "CONSTREF(" ++ Constant.debug_print kn ++ str ")")
   | IndRef (kn,i) -> pp (str "INDREF(" ++ MutInd.debug_print kn ++ str "," ++
@@ -168,13 +213,16 @@ let ppdelta s = pp (Mod_subst.debug_pr_delta s)
 let pp_idpred s = pp (pr_idpred s)
 let pp_cpred s = pp (pr_cpred s)
 let pp_transparent_state s = pp (pr_transparent_state s)
-let pp_stack_t n = pp (Reductionops.Stack.pr (EConstr.of_constr %> pr_econstr) n)
-let pp_cst_stack_t n = pp (Reductionops.Cst_stack.pr Global.(env()) Evd.empty n)
+let pp_estack_t n = pp (Reductionops.Stack.pr pr_econstr n)
 let pp_state_t n = pp (Reductionops.pr_state Global.(env()) Evd.empty n)
 
 (* proof printers *)
 let pr_evar ev = Pp.int (Evar.repr ev)
-let ppmetas metas = pp(Termops.pr_metaset metas)
+let ppmetas metas = pp (Unification.Meta.pr_metaset metas)
+let ppmetamap metas =
+  let env = Global.env () in
+  let sigma = Evd.from_env env in
+  pp (Unification.Meta.pr_metamap env sigma metas)
 let ppevm evd = pp(Termops.pr_evar_map ~with_univs:!Detyping.print_universes (Some 2) (Global.env ()) evd)
 let ppevmall evd = pp(Termops.pr_evar_map ~with_univs:!Detyping.print_universes None (Global.env ()) evd)
 let pr_existentialset evars =
@@ -184,14 +232,14 @@ let ppexistentialset evars =
 let ppexistentialfilter filter = match Evd.Filter.repr filter with
 | None -> pp (Pp.str "ø")
 | Some f -> pp (prlist_with_sep spc bool f)
+let pr_goal e = Pp.(str "GOAL:" ++ int (Evar.repr e))
 let ppclenv clenv = pp(pr_clenv clenv)
-let ppgoalgoal gl = pp(Goal.pr_goal gl)
-let ppgoal g = pp(Printer.pr_goal g)
-let ppgoalsigma g = pp(Printer.pr_goal g ++ Termops.pr_evar_map None (Global.env ()) (Refiner.project g))
+let ppgoal g = pp(Printer.Debug.pr_goal g)
+let ppgoal_with_state g = ppevar (Proofview_monad.drop_state g)
 let pphintdb db = pp(envpp Hints.pr_hint_db_env db)
 let ppproofview p =
   let gls,sigma = Proofview.proofview p in
-  pp(pr_enum Goal.pr_goal gls ++ fnl () ++ Termops.pr_evar_map (Some 1) (Global.env ()) sigma)
+  pp(pr_enum pr_goal gls ++ fnl () ++ Termops.pr_evar_map (Some 1) (Global.env ()) sigma)
 
 let ppopenconstr (x : Evd.open_constr) =
   let (evd,c) = x in pp (Termops.pr_evar_map (Some 2) (Global.env ()) evd ++ envpp pr_econstr_env c)
@@ -213,39 +261,51 @@ let pppftreestate p = pp(print_pftreestate p)
 
 let pproof p = pp(Proof.pr_proof p)
 
-let ppuni u = pp(Universe.pr u)
-let ppuni_level u = pp (Level.pr u)
+let ppuni u = pp(Universe.raw_pr u)
+let ppuni_level u = pp (Level.raw_pr u)
+let ppqvar q = pp (QVar.raw_pr q)
+let ppesorts s = pp (Sorts.debug_print (Evd.MiniEConstr.ESorts.unsafe_to_sorts s))
 
-let prlev = UnivNames.pr_with_global_universes
-let ppuniverse_set l = pp (LSet.pr prlev l)
-let ppuniverse_instance l = pp (Instance.pr prlev l)
-let ppuniverse_context l = pp (pr_universe_context prlev l)
+let prlev l = UnivNames.pr_level_with_global_universes l
+let prqvar q = Sorts.QVar.raw_pr q
+let ppqvarset l = pp (hov 1 (str "{" ++ prlist_with_sep spc QVar.raw_pr (QVar.Set.elements l) ++ str "}"))
+let ppuniverse_set l = pp (Level.Set.pr prlev l)
+let ppuniverse_instance l = pp (Instance.pr prqvar prlev l)
+let ppuniverse_context l = pp (pr_universe_context prqvar prlev l)
 let ppuniverse_context_set l = pp (pr_universe_context_set prlev l)
-let ppuniverse_subst l = pp (Univ.pr_universe_subst l)
-let ppuniverse_opt_subst l = pp (UnivSubst.pr_universe_opt_subst l)
-let ppuniverse_level_subst l = pp (Univ.pr_universe_level_subst l)
-let ppevar_universe_context l = pp (Termops.pr_evar_universe_context l)
-let ppconstraints c = pp (pr_constraints Level.pr c)
+let ppuniverse_subst l = pp (UnivSubst.pr_universe_subst Level.raw_pr l)
+let ppuniverse_opt_subst l = pp (UnivFlex.pr Level.raw_pr l)
+let ppqvar_subst l = pp (UVars.pr_quality_level_subst QVar.raw_pr l)
+let ppuniverse_level_subst l = pp (Univ.pr_universe_level_subst Level.raw_pr l)
+let ppustate l = pp (UState.pr l)
+let ppconstraints c = pp (Constraints.pr Level.raw_pr c)
 let ppuniverseconstraints c = pp (UnivProblem.Set.pr c)
 let ppuniverse_context_future c =
   let ctx = Future.force c in
     ppuniverse_context ctx
-let ppuniverses u = pp (UGraph.pr_universes Level.pr u)
+let ppuniverses u = pp (UGraph.pr_universes Level.raw_pr (UGraph.repr u))
 let ppnamedcontextval e =
   let env = Global.env () in
   let sigma = Evd.from_env env in
   pp (pr_named_context env sigma (named_context_of_val e))
 
 let ppaucontext auctx =
-  let nas = AUContext.names auctx in
-  let prlev l = match Level.var_index l with
+  let qnas, unas = AbstractContext.names auctx in
+  let prgen pr var_index nas l = match var_index l with
     | Some n -> (match nas.(n) with
-        | Anonymous -> prlev l
+        | Anonymous -> pr l
         | Name id -> Id.print id)
-    | None -> prlev l
+    | None -> pr l
   in
-  pp (pr_universe_context prlev (AUContext.repr auctx))
+  let prqvar l = prgen prqvar Sorts.QVar.var_index qnas l in
+  let prlev l = prgen prlev Level.var_index unas l in
+  pp (pr_universe_context prqvar prlev (AbstractContext.repr auctx))
 
+let pp_partialfsubst psubst =
+  pp (Partial_subst.pr (fun f -> pr_constr (CClosure.term_of_fconstr f)) (Quality.pr prqvar) (Universe.pr prlev) psubst)
+
+let pp_partialsubst psubst =
+  pp (Partial_subst.pr pr_econstr (Quality.pr prqvar) (Universe.pr prlev) psubst)
 
 let ppenv e = pp
   (str "[" ++ pr_named_context_of e Evd.empty ++ str "]" ++ spc() ++
@@ -270,7 +330,6 @@ let cast_kind_display k =
   match k with
   | VMcast -> "VMcast"
   | DEFAULTcast -> "DEFAULTcast"
-  | REVERTcast -> "REVERTcast"
   | NATIVEcast -> "NATIVEcast"
 
 let constr_display csr =
@@ -289,17 +348,21 @@ let constr_display csr =
       "LetIn("^(name_display na)^","^(term_display b)^","
       ^(term_display t)^","^(term_display c)^")"
   | App (c,l) -> "App("^(term_display c)^","^(array_display l)^")\n"
-  | Evar (e,l) -> "Evar("^(Pp.string_of_ppcmds (Evar.print e))^","^(array_display (Array.of_list l))^")"
+  | Evar (e,l) ->
+    let l = SList.to_list l in
+    let map = function None -> "?" | Some t -> term_display t in
+    let l = List.map map l in
+    "Evar("^(Pp.string_of_ppcmds (Evar.print e))^", [|"^(String.concat "; " l)^"|])"
   | Const (c,u) -> "Const("^(Constant.to_string c)^","^(universes_display u)^")"
   | Ind ((sp,i),u) ->
       "MutInd("^(MutInd.to_string sp)^","^(string_of_int i)^","^(universes_display u)^")"
   | Construct (((sp,i),j),u) ->
       "MutConstruct(("^(MutInd.to_string sp)^","^(string_of_int i)^"),"
       ^","^(universes_display u)^(string_of_int j)^")"
-  | Proj (p, c) -> "Proj("^(Constant.to_string (Projection.constant p))^","^term_display c ^")"
-  | Case (ci,p,c,bl) ->
+  | Proj (p, r, c) -> "Proj("^(Projection.to_string p)^","^term_display c ^")"
+  | Case (ci,u,pms,((_,p),_),iv,c,bl) ->
       "MutCase(<abs>,"^(term_display p)^","^(term_display c)^","
-      ^(array_display bl)^")"
+      ^(array_display (Array.map snd bl))^")"
   | Fix ((t,i),(lna,tl,bl)) ->
       "Fix(([|"^(Array.fold_right (fun x i -> (string_of_int x)^(if not(i="")
         then (";"^i) else "")) t "")^"|],"^(string_of_int i)^"),"
@@ -317,6 +380,9 @@ let constr_display csr =
       "Int("^(Uint63.to_string i)^")"
   | Float f ->
       "Float("^(Float64.to_string f)^")"
+  | String s ->
+      Printf.sprintf "String(%S)" (Pstring.to_string s)
+  | Array (u,t,def,ty) -> "Array("^(array_display t)^","^(term_display def)^","^(term_display ty)^")@{" ^universes_display u^"\n"
 
   and array_display v =
     "[|"^
@@ -325,10 +391,13 @@ let constr_display csr =
        v "")^"|]"
 
   and univ_display u =
-    incr cnt; pp (str "with " ++ int !cnt ++ str" " ++ pr_uni u ++ fnl ())
+    incr cnt; pp (str "with " ++ int !cnt ++ str" " ++ Universe.raw_pr u ++ fnl ())
+
+  and quality_display q =
+    incr cnt; pp (str "with " ++ int !cnt ++ str" " ++ Sorts.Quality.raw_pr q ++ fnl ())
 
   and level_display u =
-    incr cnt; pp (str "with " ++ int !cnt ++ str" " ++ Level.pr u ++ fnl ())
+    incr cnt; pp (str "with " ++ int !cnt ++ str" " ++ Level.raw_pr u ++ fnl ())
 
   and sort_display = function
     | SProp -> "SProp"
@@ -336,10 +405,18 @@ let constr_display csr =
     | Prop -> "Prop"
     | Type u -> univ_display u;
         "Type("^(string_of_int !cnt)^")"
+    | QSort (q, u) -> univ_display u; Printf.sprintf "QSort(%s, %i)" (Sorts.QVar.to_string q) !cnt
 
   and universes_display l =
+    let qs, us = Instance.to_array l in
+    let qs = Array.fold_right (fun x i ->
+        quality_display x;
+        (string_of_int !cnt)^
+        (if not(i="") then (" "^i) else ""))
+        qs ""
+    in
     Array.fold_right (fun x i -> level_display x; (string_of_int !cnt)^(if not(i="")
-        then (" "^i) else "")) (Instance.to_array l) ""
+        then (" "^i) else "")) us (if qs = "" then "" else (qs^" | "))
 
   and name_display x = match x.binder_name with
     | Name id -> "Name("^(Id.to_string id)^")"
@@ -385,14 +462,15 @@ let print_pure_constr csr =
       Array.iter (fun x -> print_space (); box_display x) l;
       print_string ")"
   | Evar (e,l) -> print_string "Evar#"; print_int (Evar.repr e); print_string "{";
-      List.iter (fun x -> print_space (); box_display x) l;
+      let iter = function None -> print_space (); print_string "?" | Some t -> print_space (); box_display t in
+      List.iter iter (SList.to_list l);
       print_string"}"
   | Const (c,u) -> print_string "Cons(";
       sp_con_display c;
       print_string ","; universes_display u;
       print_string ")"
-  | Proj (p,c') -> print_string "Proj(";
-      sp_con_display (Projection.constant p);
+  | Proj (p,_,c') -> print_string "Proj(";
+      sp_prj_display p;
       print_string ",";
       box_display c';
       print_string ")"
@@ -409,13 +487,25 @@ let print_pure_constr csr =
       print_int i; print_string ","; print_int j;
       print_string ","; universes_display u;
       print_string ")"
-  | Case (ci,p,c,bl) ->
+  | Case (ci,u,pms,(p,_),iv,c,bl) ->
+      let pr_ctx (nas, c) =
+        Array.iter (fun na -> print_cut (); name_display na) nas;
+        print_string " |- ";
+        box_display c
+      in
       open_vbox 0;
-      print_string "<"; box_display p; print_string ">";
       print_cut(); print_string "Case";
-      print_space(); box_display c; print_space (); print_string "of";
+      print_space(); box_display c; print_space ();
+      print_cut(); print_string "in";
+      print_cut(); print_string "Ind(";
+      sp_display (fst ci.ci_ind);
+      print_string ","; print_int (snd ci.ci_ind); print_string ")";
+      print_string "@{"; universes_display u; print_string "}";
+      Array.iter (fun x -> print_space (); box_display x) pms;
+      print_cut(); print_string "return <"; pr_ctx p; print_string ">";
+      print_cut(); print_string "with";
       open_vbox 0;
-      Array.iter (fun x ->  print_cut();  box_display x) bl;
+      Array.iter (fun x ->  print_cut();  pr_ctx x) bl;
       close_box();
       print_cut();
       print_string "end";
@@ -451,18 +541,34 @@ let print_pure_constr csr =
      print_string ("Int("^(Uint63.to_string i)^")")
   | Float f ->
       print_string ("Float("^(Float64.to_string f)^")")
+  | String s ->
+      print_string (Printf.sprintf "String(%S)" (Pstring.to_string s))
+  | Array (u,t,def,ty) ->
+      print_string "Array(";
+      Array.iter (fun x -> box_display x; print_space()) t;
+      print_string "|";
+      box_display def;
+      print_string ":";
+      box_display ty;
+      print_string ")@{";
+      universes_display u;
+      print_string "}"
 
   and box_display c = open_hovbox 1; term_display c; close_box()
 
   and universes_display u =
-    Array.iter (fun u -> print_space (); pp (Level.pr u)) (Instance.to_array u)
+    let qs, us = Instance.to_array u in
+    Array.iter (fun u -> print_space (); pp (Sorts.Quality.raw_pr u)) qs;
+    Array.iter (fun u -> print_space (); pp (Level.raw_pr u)) us
 
   and sort_display = function
     | SProp -> print_string "SProp"
     | Set -> print_string "Set"
     | Prop -> print_string "Prop"
     | Type u -> open_hbox();
-        print_string "Type("; pp (pr_uni u); print_string ")"; close_box()
+        print_string "Type("; pp (Universe.raw_pr u); print_string ")"; close_box()
+    | QSort (q, u) -> open_hbox();
+        print_string "QSort("; pp (QVar.raw_pr q); print_string ", "; pp (Universe.raw_pr u); print_string ")"; close_box()
 
   and name_display x = match x.binder_name with
     | Name id -> print_string (Id.to_string id)
@@ -473,7 +579,7 @@ let print_pure_constr csr =
     let ls =
       match List.rev_map Id.to_string (DirPath.repr dir) with
           ("Top"::l)-> l
-        | ("Coq"::_::l) -> l
+        | ("Stdlib"::_::l) -> l
         | l             -> l
     in  List.iter (fun x -> print_string x; print_string ".") ls;*)
       print_string (MutInd.debug_to_string sp)
@@ -482,11 +588,12 @@ let print_pure_constr csr =
     let ls =
       match List.rev_map Id.to_string (DirPath.repr dir) with
           ("Top"::l)-> l
-        | ("Coq"::_::l) -> l
+        | ("Stdlib"::_::l) -> l
         | l             -> l
     in  List.iter (fun x -> print_string x; print_string ".") ls;*)
       print_string (Constant.debug_to_string sp)
-
+  and sp_prj_display sp =
+      print_string (Projection.debug_to_string sp)
   in
     try
      box_display csr; print_flush()
@@ -539,23 +646,25 @@ VERNAC COMMAND EXTEND PrintConstr
 END
 *)
 
-let _ =
+let () =
   let open Vernacextend in
+  let open Vernactypes in
   let ty_constr = Extend.TUentry (get_arg_tag Stdarg.wit_constr) in
   let cmd_sig = TyTerminal("PrintConstr", TyNonTerminal(ty_constr, TyNil)) in
-  let cmd_fn c ~atts = VtDefault (fun () -> in_current_context econstr_display c) in
+  let cmd_fn c ?loc:_ ~atts () = vtdefault (fun () -> in_current_context econstr_display c) in
   let cmd_class _ = VtQuery in
   let cmd : ty_ml = TyML (false, cmd_sig, cmd_fn, Some cmd_class) in
-  vernac_extend ~command:"PrintConstr" [cmd]
+  static_vernac_extend ~plugin:None ~command:"PrintConstr" [cmd]
 
-let _ =
+let () =
   let open Vernacextend in
+  let open Vernactypes in
   let ty_constr = Extend.TUentry (get_arg_tag Stdarg.wit_constr) in
   let cmd_sig = TyTerminal("PrintPureConstr", TyNonTerminal(ty_constr, TyNil)) in
-  let cmd_fn c ~atts = VtDefault (fun () -> in_current_context print_pure_econstr c) in
+  let cmd_fn c ?loc:_ ~atts () = vtdefault (fun () -> in_current_context print_pure_econstr c) in
   let cmd_class _ = VtQuery in
   let cmd : ty_ml = TyML (false, cmd_sig, cmd_fn, Some cmd_class) in
-  vernac_extend ~command:"PrintPureConstr" [cmd]
+  static_vernac_extend ~plugin:None ~command:"PrintPureConstr" [cmd]
 
 (* Setting printer of unbound global reference *)
 open Names
@@ -571,14 +680,14 @@ let encode_path ?loc prefix mpdir suffix id =
 
 let raw_string_of_ref ?loc _ = let open GlobRef in function
   | ConstRef cst ->
-      let (mp,id) = Constant.repr2 cst in
+      let (mp,id) = KerName.repr (Constant.user cst) in
       encode_path ?loc "CST" (Some mp) [] (Label.to_id id)
   | IndRef (kn,i) ->
-      let (mp,id) = MutInd.repr2 kn in
+      let (mp,id) = KerName.repr (MutInd.user kn) in
       encode_path ?loc "IND" (Some mp) [Label.to_id id]
         (Id.of_string ("_"^string_of_int i))
   | ConstructRef ((kn,i),j) ->
-      let (mp,id) = MutInd.repr2 kn in
+      let (mp,id) = KerName.repr (MutInd.user kn) in
       encode_path ?loc "CSTR" (Some mp)
         [Label.to_id id;Id.of_string ("_"^string_of_int i)]
         (Id.of_string ("_"^string_of_int j))
@@ -596,10 +705,3 @@ let short_string_of_ref ?loc _ = let open GlobRef in function
       encode_path ?loc "CSTR" None
         [Label.to_id (MutInd.label kn);Id.of_string ("_"^string_of_int i)]
         (Id.of_string ("_"^string_of_int j))
-
-(* Anticipate that printers can be used from ocamldebug and that
-   pretty-printer should not make calls to the global env since ocamldebug
-   runs in a different process and does not have the proper env at hand *)
-let _ = Flags.in_debugger := true
-let _ = Constrextern.set_extern_reference
-  (if !rawdebug then raw_string_of_ref else short_string_of_ref)
