@@ -1,5 +1,5 @@
 ##########################################################################
-##         #   The Coq Proof Assistant / The Coq Development Team       ##
+##         #      The Rocq Prover / The Rocq Development Team           ##
 ##  v      #         Copyright INRIA, CNRS and contributors             ##
 ## <O___,, # (see version control and CREDITS file for authors & dates) ##
 ##   \VV/  ###############################################################
@@ -20,6 +20,7 @@ could imagine extending it.
 
 import os
 import re
+import shlex
 from itertools import chain
 from collections import defaultdict
 
@@ -29,7 +30,7 @@ from docutils.parsers.rst import Directive, directives
 from docutils.parsers.rst.roles import code_role #, set_classes
 from docutils.parsers.rst.directives.admonitions import BaseAdmonition
 
-from sphinx import addnodes
+from sphinx import addnodes, version_info as sphinx_version
 from sphinx.directives import ObjectDescription
 from sphinx.domains import Domain, ObjType, Index
 from sphinx.errors import ExtensionError
@@ -48,16 +49,33 @@ from .notations.plain import stringify_with_ellipses
 
 # FIXME: Patch this in Sphinx
 # https://github.com/coq/coq/issues/12361
-def visit_desc_signature(self, node):
-    hyper = ''
-    if node.parent['objtype'] != 'describe' and node['ids']:
-        for id in node['ids']:
-            hyper += self.hypertarget(id)
-    self.body.append(hyper)
-    if not node.get('is_multiline'):
-        self._visit_signature_line(node)
-    else:
-        self.body.append('%\n\\pysigstartmultiline\n')
+if sphinx_version >= (4, 5):
+    from sphinx.writers.latex import CR
+
+    def visit_desc_signature(self, node):
+        hyper = ''
+        if node.parent['objtype'] != 'describe' and node['ids']:
+            for id in node['ids']:
+                hyper += self.hypertarget(id)
+        self.body.append(hyper)
+        if not self.in_desc_signature:
+            self.in_desc_signature = True
+            self.body.append(CR + r'\pysigstartsignatures')
+        if not node.get('is_multiline'):
+            self._visit_signature_line(node)
+        else:
+            self.body.append(CR + r'\pysigstartmultiline')
+else:
+    def visit_desc_signature(self, node):
+        hyper = ''
+        if node.parent['objtype'] != 'describe' and node['ids']:
+            for id in node['ids']:
+                hyper += self.hypertarget(id)
+        self.body.append(hyper)
+        if not node.get('is_multiline'):
+            self._visit_signature_line(node)
+        else:
+            self.body.append('%\n\\pysigstartmultiline\n')
 LaTeXTranslator.visit_desc_signature = visit_desc_signature
 
 PARSE_ERROR = """{}:{} Parse error in notation!
@@ -216,12 +234,11 @@ class CoqObject(ObjectDescription):
             signode['first'] = (not self.names)
             self._record_name(name, targetid, signode)
         else:
-            # todo: make the following a real error or warning
-            # todo: then maybe the above "if" is not needed
-            names_in_subdomain = self.subdomain_data()
-            if name in names_in_subdomain:
-                print("Duplicate", self.subdomain, "name: ", name)
-            # self._warn_if_duplicate_name(names_in_subdomain, name, signode)
+            # We don't warn for duplicates in the SSReflect chapter, because
+            # it's the style of this chapter to repeat all the defined
+            # objects at the end.
+            if self.env.docname != 'proof-engine/ssreflect-proof-language':
+                self._warn_if_duplicate_name(self.subdomain_data(), name, signode)
         return targetid
 
     def _add_index_entry(self, name, target):
@@ -668,13 +685,13 @@ class CoqtopDirective(Directive):
 
     Usage::
 
-       .. coqtop:: options…
+       .. rocqtop:: options…
 
           Coq code to send to coqtop
 
     Example::
 
-       .. coqtop:: in reset
+       .. rocqtop:: in reset
 
           Print nat.
           Definition a := 1.
@@ -698,8 +715,12 @@ class CoqtopDirective(Directive):
       - ``warn``: Don't die if a command emits a warning
       - ``restart``: Send a ``Restart`` command before running this block (only works in proof mode)
       - ``abort``: Send an ``Abort All`` command after running this block (leaves all pending proofs if any)
+      - ``extra``: if environment variable 'COQRST_EXTRA' is set (to anything else than '0') this is ignored, otherwise behaves as ``fail``
+        This is typically used to showcase examples of things outside coq-core or rocq-core.
+        Be careful when using it to surround the code block with a sentence explaining what
+        extra requirement is needed to compile it.
 
-    ``coqtop``\ 's state is preserved across consecutive ``.. coqtop::`` blocks
+    ``coqtop``\ 's state is preserved across consecutive ``.. rocqtop::`` blocks
     of the same document (``coqrst`` creates a single ``coqtop`` process per
     reST source file).  Use the ``reset`` option to reset Coq's state.
     """
@@ -708,7 +729,7 @@ class CoqtopDirective(Directive):
     optional_arguments = 0
     final_argument_whitespace = True
     option_spec = { 'name': directives.unchanged }
-    directive_name = "coqtop"
+    directive_name = "rocqtop"
 
     def run(self):
         # Uses a ‘container’ instead of a ‘literal_block’ to disable
@@ -725,13 +746,13 @@ class CoqdocDirective(Directive):
 
     Usage::
 
-       .. coqdoc::
+       .. rocqdoc::
 
           Coq code to highlight
 
     Example::
 
-       .. coqdoc::
+       .. rocqdoc::
 
           Definition test := 1.
     """
@@ -741,7 +762,7 @@ class CoqdocDirective(Directive):
     optional_arguments = 0
     final_argument_whitespace = True
     option_spec = { 'name': directives.unchanged }
-    directive_name = "coqdoc"
+    directive_name = "rocqdoc"
 
     def run(self):
         # Uses a ‘container’ instead of a ‘literal_block’ to disable
@@ -768,7 +789,7 @@ class ExampleDirective(BaseAdmonition):
 
           The following adds ``plus_comm`` to the ``plu`` database:
 
-          .. coqdoc::
+          .. rocqdoc::
 
              Hint Resolve plus_comm : plu.
     """
@@ -936,6 +957,10 @@ class CoqtopBlocksTransform(Transform):
     def is_coqtop_block(node):
         return isinstance(node, nodes.Element) and 'coqtop_options' in node
 
+    @staticmethod
+    def is_coqtop_args_field(node):
+        return isinstance(node, nodes.field) and node.children[0].rawsource == 'COQTOP_ARGS'
+
     @staticmethod
     def split_lines(source):
         r"""Split Coq input into chunks, which may include single- or
@@ -984,22 +1009,27 @@ class CoqtopBlocksTransform(Transform):
         opt_warn = 'warn' in options
         opt_restart = 'restart' in options
         opt_abort = 'abort' in options
-        options = options - {'reset', 'fail', 'warn', 'restart', 'abort'}
+        opt_extra = 'extra' in options
+        options = options - {'reset', 'fail', 'warn', 'restart', 'abort', 'extra'}
 
         unexpected_options = list(options - {'all', 'none', 'in', 'out'})
         if unexpected_options:
             loc = os.path.basename(get_node_location(node))
-            raise ExtensionError("{}: Unexpected options for .. coqtop:: {}".format(loc,unexpected_options))
+            raise ExtensionError("{}: Unexpected options for .. rocqtop:: {}".format(loc,unexpected_options))
 
         # Display options
         if len(options) != 1:
             loc = os.path.basename(get_node_location(node))
-            raise ExtensionError("{}: Exactly one display option must be passed to .. coqtop::".format(loc))
+            raise ExtensionError("{}: Exactly one display option must be passed to .. rocqtop::".format(loc))
 
         opt_all = 'all' in options
         opt_input = 'in' in options
         opt_output = 'out' in options
 
+        # if 'extra' is given and env variable 'COQRST_EXTRA' is not set,
+        # allow errors
+        opt_fail = opt_fail or (opt_extra and os.environ.get('COQRST_EXTRA', '0') == '0')
+
         return {
             'reset': opt_reset,
             'fail': opt_fail,
@@ -1077,7 +1107,9 @@ class CoqtopBlocksTransform(Transform):
         """Add coqtop's responses to a Sphinx AST
 
         Finds nodes to process using is_coqtop_block."""
-        with CoqTop(color=True) as repl:
+        arg_fields = self.document.traverse(CoqtopBlocksTransform.is_coqtop_args_field)
+        additional_args = [arg for field in arg_fields for arg in shlex.split(field.children[1].rawsource)]
+        with CoqTop(color=True, args=additional_args) as repl:
             repl.send_initial_options()
             for node in self.document.traverse(CoqtopBlocksTransform.is_coqtop_block):
                 try:
