@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -41,20 +41,20 @@ type naming_mode =
 
 val new_evar :
   ?src:Evar_kinds.t Loc.located -> ?filter:Filter.t ->
+  ?relevance:ERelevance.t ->
   ?abstract_arguments:Abstraction.t -> ?candidates:constr list ->
   ?naming:intro_pattern_naming_expr ->
   ?typeclass_candidate:bool ->
-  ?principal:bool -> ?hypnaming:naming_mode ->
+  ?hypnaming:naming_mode ->
   env -> evar_map -> types -> evar_map * EConstr.t
 
 (** Alias of {!Evd.new_pure_evar} *)
 val new_pure_evar :
   ?src:Evar_kinds.t Loc.located -> ?filter:Filter.t ->
-  ?relevance:Sorts.relevance ->
+  ?relevance:ERelevance.t ->
   ?abstract_arguments:Abstraction.t -> ?candidates:constr list ->
   ?name:Id.t ->
   ?typeclass_candidate:bool ->
-  ?principal:bool ->
   named_context_val -> evar_map -> types -> evar_map * Evar.t
 
 (** Create a new Type existential variable, as we keep track of
@@ -62,23 +62,20 @@ val new_pure_evar :
 val new_type_evar :
   ?src:Evar_kinds.t Loc.located -> ?filter:Filter.t ->
   ?naming:intro_pattern_naming_expr ->
-  ?principal:bool -> ?hypnaming:naming_mode ->
+  ?hypnaming:naming_mode ->
   env -> evar_map -> rigid ->
-  evar_map * (constr * Sorts.t)
+  evar_map * (constr * ESorts.t)
 
 val new_Type : ?rigid:rigid -> evar_map -> evar_map * constr
 
 (** {6 Unification utils} *)
 
-(** [head_evar c] returns the head evar of [c] if any *)
-exception NoHeadEvar
-val head_evar : evar_map -> constr -> Evar.t (** may raise NoHeadEvar *)
-
 (* Expand head evar if any *)
 val whd_head_evar :  evar_map -> constr -> constr
 
 (* An over-approximation of [has_undefined (nf_evars evd c)] *)
 val has_undefined_evars : evar_map -> constr -> bool
+val has_undefined_evars_or_metas : evar_map -> constr -> bool
 
 val is_ground_term :  evar_map -> constr -> bool
 val is_ground_env  :  evar_map -> env -> bool
@@ -100,13 +97,12 @@ val reachable_from_evars : evar_map -> Evar.Set.t -> Evar.Set.t
 
 val undefined_evars_of_term : evar_map -> constr -> Evar.Set.t
 val undefined_evars_of_named_context : evar_map -> Constr.named_context -> Evar.Set.t
-val undefined_evars_of_evar_info : evar_map -> evar_info -> Evar.Set.t
 
 type undefined_evars_cache
 
 val create_undefined_evars_cache : unit -> undefined_evars_cache
 
-val filtered_undefined_evars_of_evar_info : ?cache:undefined_evars_cache -> evar_map -> evar_info -> Evar.Set.t
+val filtered_undefined_evars_of_evar_info : ?cache:undefined_evars_cache -> evar_map -> 'a evar_info -> Evar.Set.t
 
 (** [occur_evar_upto sigma k c] returns [true] if [k] appears in
     [c]. It looks up recursively in [sigma] for the value of existential
@@ -119,10 +115,7 @@ val judge_of_new_Type : evar_map -> evar_map * unsafe_judgment
 
 (***********************************************************)
 
-val create_clos_infos : env -> evar_map -> CClosure.RedFlags.reds -> CClosure.clos_infos
-
-(** [flush_and_check_evars] raise [Uninstantiated_evar] if an evar remains
-    uninstantiated; [nf_evar] leaves uninstantiated evars as is *)
+val create_clos_infos : env -> evar_map -> RedFlags.reds -> CClosure.clos_infos
 
 val whd_evar :  evar_map -> constr -> constr
 val nf_evar :  evar_map -> constr -> constr
@@ -138,18 +131,15 @@ val nf_named_context_evar : evar_map -> Constr.named_context -> Constr.named_con
 val nf_rel_context_evar : evar_map -> rel_context -> rel_context
 val nf_env_evar : evar_map -> env -> env
 
-val nf_evar_info : evar_map -> evar_info -> evar_info
+val nf_evar_info : evar_map -> 'a evar_info -> 'a evar_info
 val nf_evar_map : evar_map -> evar_map
 val nf_evar_map_undefined : evar_map -> evar_map
+val nf_relevance : evar_map -> Sorts.relevance -> Sorts.relevance
 
 (** Presenting terms without solved evars *)
 
 val nf_evars_universes : evar_map -> Constr.constr -> Constr.constr
 
-(** Replacing all evars, possibly raising [Uninstantiated_evar] *)
-exception Uninstantiated_evar of Evar.t
-val flush_and_check_evars :  evar_map -> constr -> Constr.constr
-
 (** [finalize env sigma f] combines universe minimisation,
    evar-and-universe normalisation and universe restriction.
 
@@ -169,7 +159,7 @@ val finalize : ?abort_on_undefined_evars:bool -> evar_map ->
     as an evar [e] only if [e] is uninstantiated in [sigma]. Otherwise the
     value of [e] in [sigma] is (recursively) used. *)
 val kind_of_term_upto : evar_map -> Constr.constr ->
-  (Constr.constr, Constr.types, Sorts.t, Univ.Instance.t) kind_of_term
+  (Constr.constr, Constr.types, Sorts.t, UVars.Instance.t, Sorts.relevance) kind_of_term
 
 (** [eq_constr_univs_test ~evd ~extended_evd t u] tests equality of
     [t] and [u] up to existential variable instantiation and
@@ -188,14 +178,17 @@ val eq_constr_univs_test :
    constraints such that [u1 cv_pb? u2] according to [variance].
    Additionally flexible universes in irrelevant positions are unified
    if possible. Returns [Inr p] when the former is impossible. *)
-val compare_cumulative_instances : Reduction.conv_pb -> Univ.Variance.t array ->
-  Univ.Instance.t -> Univ.Instance.t -> evar_map ->
+val compare_cumulative_instances : Conversion.conv_pb -> UVars.Variance.t array ->
+  UVars.Instance.t -> UVars.Instance.t -> evar_map ->
   (evar_map, UGraph.univ_inconsistency) Util.union
 
 (** We should only compare constructors at convertible types, so this
-   is only an opportunity to unify universes. *)
+    is only an opportunity to unify universes.
+
+    But what about qualities?
+*)
 val compare_constructor_instances : evar_map ->
-  Univ.Instance.t -> Univ.Instance.t -> evar_map
+  UVars.Instance.t -> UVars.Instance.t -> (evar_map, UGraph.univ_inconsistency) Util.union
 
 (** {6 Unification problems} *)
 type unification_pb = conv_pb * env * constr * constr
@@ -210,7 +203,7 @@ raise OccurHypInSimpleClause if the removal breaks dependencies *)
 
 type clear_dependency_error =
 | OccurHypInSimpleClause of Id.t option
-| EvarTypingBreak of Constr.existential
+| EvarTypingBreak of EConstr.existential
 | NoCandidatesLeft of Evar.t
 
 exception ClearDependencyError of Id.t * clear_dependency_error * GlobRef.t option
@@ -221,7 +214,7 @@ exception ClearDependencyError of Id.t * clear_dependency_error * GlobRef.t opti
     into an empty list. *)
 
 val restrict_evar : evar_map -> Evar.t -> Filter.t ->
-  ?src:Evar_kinds.t Loc.located -> constr list option -> evar_map * Evar.t
+  constr list option -> evar_map * Evar.t
 
 val clear_hyps_in_evi : env -> evar_map -> named_context_val -> types ->
   Id.Set.t -> evar_map * named_context_val * types
