@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -44,15 +44,15 @@ val id_of_name_using_hdchar : env -> evar_map -> types -> Name.t -> Id.t
 val named_hd : env -> evar_map -> types -> Name.t -> Name.t
 val head_name : evar_map -> types -> Id.t option
 
-val mkProd_name : env -> evar_map -> Name.t Context.binder_annot * types * types -> types
-val mkLambda_name : env -> evar_map -> Name.t Context.binder_annot * types * constr -> constr
+val mkProd_name : env -> evar_map -> Name.t EConstr.binder_annot * types * types -> types
+val mkLambda_name : env -> evar_map -> Name.t EConstr.binder_annot * types * constr -> constr
 
 (** Deprecated synonyms of [mkProd_name] and [mkLambda_name] *)
-val prod_name : env -> evar_map -> Name.t Context.binder_annot * types * types -> types
-val lambda_name : env -> evar_map -> Name.t Context.binder_annot * types * constr -> constr
+val prod_name : env -> evar_map -> Name.t EConstr.binder_annot * types * types -> types
+val lambda_name : env -> evar_map -> Name.t EConstr.binder_annot * types * constr -> constr
 
-val prod_create : env -> evar_map -> Sorts.relevance * types * types -> constr
-val lambda_create : env -> evar_map -> Sorts.relevance * types * constr -> constr
+val prod_create : env -> evar_map -> ERelevance.t * types * types -> constr
+val lambda_create : env -> evar_map -> ERelevance.t * types * constr -> constr
 val name_assumption : env -> evar_map -> rel_declaration -> rel_declaration
 val name_context : env -> evar_map -> rel_context -> rel_context
 
@@ -86,7 +86,7 @@ val next_ident_away_from : Id.t -> (Id.t -> bool) -> Id.t
 val next_ident_away : Id.t -> Id.Set.t -> Id.t
 
 (** Avoid clashing with a name already used in current module *)
-val next_ident_away_in_goal : Id.t -> Id.Set.t -> Id.t
+val next_ident_away_in_goal : Environ.env -> Id.t -> Id.Set.t -> Id.t
 
 (** Avoid clashing with a name already used in current module
    but tolerate overwriting section variables, as in goals *)
@@ -105,23 +105,32 @@ val set_reserved_typed_name : (types -> Name.t) -> unit
 (*********************************************************************
    Making name distinct for displaying *)
 
+val make_all_rel_context_name_different : env -> evar_map -> rel_context -> env * rel_context
+val make_all_name_different : env -> evar_map -> env
+
+module Generator :
+sig
+type 'a t
+type 'a input = 'a t * 'a
+val fresh : Nameops.Fresh.t t
+val idset : Id.Set.t t
+
+val next_name_away : 'a t -> Name.t -> 'a -> Id.t * 'a
+
+val max_map : 'a t -> 'a -> Nameops.Subscript.t Id.Map.t
+end
+
 type renaming_flags =
   | RenamingForCasesPattern of (Name.t list * constr) (** avoid only global constructors *)
   | RenamingForGoal (** avoid all globals (as in intro) *)
   | RenamingElsewhereFor of (Name.t list * constr)
 
-val make_all_name_different : env -> evar_map -> env
-
 val compute_displayed_name_in :
-  evar_map -> renaming_flags -> Id.Set.t -> Name.t -> constr -> Name.t * Id.Set.t
-val compute_and_force_displayed_name_in :
-  evar_map -> renaming_flags -> Id.Set.t -> Name.t -> constr -> Name.t * Id.Set.t
+  'a Generator.t -> Environ.env -> evar_map -> renaming_flags -> 'a -> Name.t -> constr -> Name.t * 'a
 val compute_displayed_let_name_in :
-  evar_map -> renaming_flags -> Id.Set.t -> Name.t -> 'a -> Name.t * Id.Set.t
-val rename_bound_vars_as_displayed :
-  evar_map -> Id.Set.t -> Name.t list -> types -> types
+  'a Generator.t -> Environ.env -> evar_map -> renaming_flags -> 'a -> Name.t -> Name.t * 'a
 
 (* Generic function expecting a "not occurn" function *)
 val compute_displayed_name_in_gen :
-  (evar_map -> int -> 'a -> bool) ->
-  evar_map -> Id.Set.t -> Name.t -> 'a -> Name.t * Id.Set.t
+  'a Generator.t -> (evar_map -> int -> 'constr -> bool) ->
+  Environ.env -> evar_map -> 'a -> Name.t -> 'constr -> Name.t * 'a
