@@ -1,7 +1,7 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
-(*  v      *   INRIA, CNRS and contributors - Copyright 1999-2019       *)
-(* <O___,, *       (see CREDITS file for the list of authors)           *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
+(*  v      *         Copyright INRIA, CNRS and contributors             *)
+(* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
 (*    //   *    This file is distributed under the terms of the         *)
 (*         *     GNU Lesser General Public License Version 2.1          *)
@@ -52,6 +52,24 @@ sig
 
 end
 
+module Fresh :
+sig
+  type t
+  val empty : t
+  val add : Id.t -> t -> t
+  val remove : Id.t -> t -> t
+  val mem : Id.t -> t -> bool
+  val next : Id.t -> t -> Id.t
+  val fresh : Id.t -> t -> Id.t * t
+
+  val of_list : Id.t list -> t
+  val of_set : Id.Set.t -> t
+  val of_named_context_val : Environ.named_context_val -> t
+
+  val max_map : t -> Subscript.t Id.Map.t
+end
+
+
 val has_subscript       : Id.t -> bool
 
 val get_subscript : Id.t -> Id.t * Subscript.t
@@ -112,14 +130,15 @@ module Name : sig
       It is [Anonymous,a] otherwise. *)
 
   val get_id : Name.t -> Id.t
-  (** [get_id] associates [id] to [Name id]. @raise IsAnonymous otherwise. *)
+  (** [get_id] associates [id] to [Name id].
+      @raise IsAnonymous otherwise. *)
 
   val pick : Name.t -> Name.t -> Name.t
   (** [pick na na'] returns [Anonymous] if both names are [Anonymous].
       Pick one of [na] or [na'] otherwise. *)
 
-  val pick_annot : Name.t Context.binder_annot -> Name.t Context.binder_annot ->
-    Name.t Context.binder_annot
+  val pick_annot : (Name.t,'r) Context.pbinder_annot -> (Name.t,'r) Context.pbinder_annot ->
+    (Name.t,'r) Context.pbinder_annot
 
   val cons : Name.t -> Id.t list -> Id.t list
   (** [cons na l] returns [id::l] if [na] is [Name id] and [l] otherwise. *)
