@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -32,7 +32,7 @@ open Quality
 
 let sort_inconsistency ?explain cst l r =
   let explain = Option.map (fun p -> UGraph.Other p) explain in
-  raise (UGraph.UniverseInconsistency (cst, l, r, explain))
+  raise (UGraph.UniverseInconsistency (None, (cst, l, r, explain)))
 
 module QState : sig
   type t
@@ -44,8 +44,9 @@ module QState : sig
   val unify_quality : fail:(unit -> t) -> Conversion.conv_pb -> Quality.t -> Quality.t -> t -> t
   val is_above_prop : elt -> t -> bool
   val undefined : t -> QVar.Set.t
+  val collapse_above_prop : to_prop:bool -> t -> t
   val collapse : t -> t
-  val pr : t -> Pp.t
+  val pr : (QVar.t -> Pp.t) -> t -> Pp.t
   val of_set : QVar.Set.t -> t
 end =
 struct
@@ -99,17 +100,20 @@ let set q qv m =
       Some { named = m.named; qmap = QMap.add q (Some qv) m.qmap; above = QSet.remove q m.above }
 
 let set_above_prop q m =
-  (* XXX fail if named? *)
   let q = repr q m in
   let q = match q with QVar q -> q | QConstant _ -> assert false in
-  { named = m.named; qmap = m.qmap; above = QSet.add q m.above }
+  if QSet.mem q m.named then None
+  else Some { named = m.named; qmap = m.qmap; above = QSet.add q m.above }
 
 let unify_quality ~fail c q1 q2 local = match q1, q2 with
 | QConstant QType, QConstant QType
 | QConstant QProp, QConstant QProp
 | QConstant QSProp, QConstant QSProp -> local
 | QConstant QProp, QVar q when c == Conversion.CUMUL ->
-  set_above_prop q local
+  begin match set_above_prop q local with
+  | Some local -> local
+  | None -> fail ()
+  end
 | QVar qv1, QVar qv2 -> begin match set qv1 q2 local with
     | Some local -> local
     | None -> match set qv2 q1 local with
@@ -174,6 +178,16 @@ let undefined m =
   let m = QMap.filter (fun _ v -> Option.is_empty v) m.qmap in
   QMap.domain m
 
+let collapse_above_prop ~to_prop m =
+  let map q v = match v with
+    | None ->
+      if not @@ QSet.mem q m.above then None else
+      if to_prop then Some (QConstant QProp)
+      else Some (QConstant QType)
+  | Some _ -> v
+  in
+  { named = m.named; qmap = QMap.mapi map m.qmap; above = QSet.empty }
+
 let collapse m =
   let map q v = match v with
   | None -> if QSet.mem q m.named then None else Some (QConstant QType)
@@ -181,18 +195,19 @@ let collapse m =
   in
   { named = m.named; qmap = QMap.mapi map m.qmap; above = QSet.empty }
 
-let pr { qmap; above } =
-  (* TODO names *)
+let pr prqvar { qmap; above; named } =
   let open Pp in
   let prbody u = function
   | None ->
     if QSet.mem u above then str " >= Prop"
+    else if QSet.mem u named then
+      str " (internal name " ++ QVar.raw_pr u ++ str ")"
     else mt ()
   | Some q ->
-    let q = Quality.raw_pr q in
+    let q = Quality.pr prqvar q in
     str " := " ++ q
   in
-  h (prlist_with_sep fnl (fun (u, v) -> QVar.raw_pr u ++ prbody u v) (QMap.bindings qmap))
+  h (prlist_with_sep fnl (fun (u, v) -> prqvar u ++ prbody u v) (QMap.bindings qmap))
 
 end
 
@@ -204,38 +219,74 @@ type univ_names = UnivNames.universe_binders * (uinfo QVar.Map.t * uinfo Level.M
 type t =
  { names : univ_names; (** Printing/location information *)
    local : ContextSet.t; (** The local graph of universes (variables and constraints) *)
-   seff_univs : Level.Set.t; (** Local universes used through private constants *)
    univ_variables : UnivFlex.t;
    (** The local universes that are unification variables *)
    sort_variables : QState.t;
    (** Local quality variables. *)
    universes : UGraph.t; (** The current graph extended with the local constraints *)
-   universes_lbound : UGraph.Bound.t; (** The lower bound on universes (e.g. Set or Prop) *)
-   initial_universes : UGraph.t; (** The graph at the creation of the evar_map *)
+   initial_universes : UGraph.t; (** The graph at the creation of the evar_map + local universes
+                                     (but not local constraints) *)
    minim_extra : UnivMinim.extra;
  }
 
 let empty =
   { names = UnivNames.empty_binders, (QVar.Map.empty, Level.Map.empty);
     local = ContextSet.empty;
-    seff_univs = Level.Set.empty;
     univ_variables = UnivFlex.empty;
     sort_variables = QState.empty;
     universes = UGraph.initial_universes;
-    universes_lbound = UGraph.Bound.Set;
     initial_universes = UGraph.initial_universes;
     minim_extra = UnivMinim.empty_extra; }
 
 let make ~lbound univs =
   { empty with
     universes = univs;
-    universes_lbound = lbound;
     initial_universes = univs }
 
 let is_empty uctx =
   ContextSet.is_empty uctx.local &&
   UnivFlex.is_empty uctx.univ_variables
 
+let id_of_level uctx l =
+  try (Level.Map.find l (snd (snd uctx.names))).uname
+  with Not_found -> None
+
+let id_of_qvar uctx l =
+  try (QVar.Map.find l (fst (snd uctx.names))).uname
+  with Not_found -> None
+
+let qualid_of_qvar_names (bind, (qrev,_)) l =
+  try Some (Libnames.qualid_of_ident (Option.get (QVar.Map.find l qrev).uname))
+  with Not_found | Option.IsNone ->
+    None (* no global qvars *)
+
+let qualid_of_level_names (bind, (_,urev)) l =
+  try Some (Libnames.qualid_of_ident (Option.get (Level.Map.find l urev).uname))
+  with Not_found | Option.IsNone ->
+    UnivNames.qualid_of_level bind l
+
+let qualid_of_level uctx l = qualid_of_level_names uctx.names l
+
+let pr_uctx_qvar_names names l =
+  match qualid_of_qvar_names names l with
+  | Some qid -> Libnames.pr_qualid qid
+  | None -> QVar.raw_pr l
+
+let pr_uctx_level_names names l =
+  match qualid_of_level_names names l with
+  | Some qid -> Libnames.pr_qualid qid
+  | None -> Level.raw_pr l
+
+let pr_uctx_level uctx l = pr_uctx_level_names uctx.names l
+
+let pr_uctx_qvar uctx l = pr_uctx_qvar_names uctx.names l
+
+let merge_constraints uctx cstrs g =
+  try UGraph.merge_constraints cstrs g
+  with UGraph.UniverseInconsistency (_, i) ->
+    let printers = (pr_uctx_qvar uctx, pr_uctx_level uctx) in
+    raise (UGraph.UniverseInconsistency (Some printers, i))
+
 let uname_union s t =
   if s == t then s
   else
@@ -256,14 +307,13 @@ let union uctx uctx' =
   else if is_empty uctx' then uctx
   else
     let local = ContextSet.union uctx.local uctx'.local in
-    let seff = Level.Set.union uctx.seff_univs uctx'.seff_univs in
     let names = names_union uctx.names uctx'.names in
     let newus = Level.Set.diff (ContextSet.levels uctx'.local)
                                (ContextSet.levels uctx.local) in
     let newus = Level.Set.diff newus (UnivFlex.domain uctx.univ_variables) in
     let extra = UnivMinim.extra_union uctx.minim_extra uctx'.minim_extra in
     let declarenew g =
-      Level.Set.fold (fun u g -> UGraph.add_universe u ~lbound:uctx.universes_lbound ~strict:false g) newus g
+      Level.Set.fold (fun u g -> UGraph.add_universe u ~lbound:UGraph.Bound.Set ~strict:false g) newus g
     in
     let fail_union s q1 q2 =
       if UGraph.type_in_type uctx.universes then s
@@ -273,7 +323,6 @@ let union uctx uctx' =
     in
       { names;
         local = local;
-        seff_univs = seff;
         univ_variables =
           UnivFlex.biased_union uctx.univ_variables uctx'.univ_variables;
         sort_variables = QState.union ~fail:fail_union uctx.sort_variables uctx'.sort_variables;
@@ -282,8 +331,7 @@ let union uctx uctx' =
           (if local == uctx.local then uctx.universes
            else
              let cstrsr = ContextSet.constraints uctx'.local in
-             UGraph.merge_constraints cstrsr (declarenew uctx.universes));
-        universes_lbound = uctx.universes_lbound;
+             merge_constraints uctx cstrsr (declarenew uctx.universes));
         minim_extra = extra}
 
 let context_set uctx = uctx.local
@@ -310,9 +358,8 @@ let compute_instance_binders (qrev,urev) inst =
   Array.map qmap qinst, Array.map umap uinst
 
 let context uctx =
-  let (_, rbinders) = uctx.names in
   let qvars = QState.undefined uctx.sort_variables in
-  UContext.of_context_set (compute_instance_binders rbinders) qvars uctx.local
+  UContext.of_context_set (compute_instance_binders (snd uctx.names)) qvars uctx.local
 
 type named_universes_entry = universes_entry * UnivNames.universe_binders
 
@@ -335,8 +382,6 @@ let subst uctx = uctx.univ_variables
 
 let ugraph uctx = uctx.universes
 
-let initial_graph uctx = uctx.initial_universes
-
 let is_algebraic l uctx = UnivFlex.is_algebraic l uctx.univ_variables
 
 let of_names (ubind,(revqbind,revubind)) =
@@ -369,15 +414,15 @@ let instantiate_variable l (b : Universe.t) v =
 
 exception UniversesDiffer
 
-let { Goptions.get = drop_weak_constraints } =
+let { Goptions.get = weak_constraints } =
   Goptions.declare_bool_option_and_ref
     ~key:["Cumulativity";"Weak";"Constraints"]
-    ~value:false
+    ~value:true
     ()
 
 let level_inconsistency cst l r =
   let mk u = Sorts.sort_of_univ @@ Universe.make u in
-  raise (UGraph.UniverseInconsistency (cst, mk l, mk r, None))
+  raise (UGraph.UniverseInconsistency (None, (cst, mk l, mk r, None)))
 
 let nf_universe uctx u =
   UnivSubst.(subst_univs_universe (UnivFlex.normalize_univ_variable uctx.univ_variables)) u
@@ -410,7 +455,13 @@ let nf_relevance uctx r = match r with
 let nf_universes uctx c =
   let lsubst = uctx.univ_variables in
   let nf_univ u = UnivFlex.normalize_univ_variable lsubst u in
-  UnivSubst.nf_evars_and_universes_opt_subst (fun _ -> None) (nf_qvar uctx) nf_univ c
+  let rec self () c = match Constr.kind c with
+  | Evar (evk, args) ->
+    let args' = SList.Smart.map (self ()) args in
+    if args == args' then c else Constr.mkEvar (evk, args')
+  | _ -> UnivSubst.map_universes_opt_subst_with_binders ignore self (nf_qvar uctx) nf_univ () c
+  in
+  self ()  c
 
 type small_universe = USet | UProp | USProp
 
@@ -551,22 +602,12 @@ let process_universe_constraints uctx cstrs =
     | QEq (a, b) ->
       (* TODO sort_inconsistency should be able to handle raw
          qualities instead of having to make a dummy sort *)
-      let mk = function
-        | QVar q -> qsort q Universe.type0
-        | QConstant QSProp -> sprop
-        | QConstant QProp -> prop
-        | QConstant QType -> set
-      in
+      let mk q = Sorts.make q Universe.type0 in
       unify_quality univs CONV (mk a) (mk b) local
     | QLeq (a, b) ->
       (* TODO sort_inconsistency should be able to handle raw
          qualities instead of having to make a dummy sort *)
-      let mk = function
-        | QVar q -> qsort q Universe.type0
-        | QConstant QSProp -> sprop
-        | QConstant QProp -> prop
-        | QConstant QType -> set
-      in
+      let mk q = Sorts.make q Universe.type0 in
       unify_quality univs CUMUL (mk a) (mk b) local
     | ULe (l, r) ->
       let local = unify_quality univs CUMUL l r local in
@@ -636,7 +677,7 @@ let process_universe_constraints uctx cstrs =
     | ULub (l, r) ->
       equalize_variables true l r local
     | UWeak (l, r) ->
-      if not (drop_weak_constraints ())
+      if weak_constraints ()
       then { local with local_weak = UPairSet.add (l, r) local.local_weak }
       else local
     | UEq (l, r) ->
@@ -665,7 +706,7 @@ let add_universe_constraints uctx cstrs =
   { uctx with
     local = (univs, Constraints.union local local');
     univ_variables = vars;
-    universes = UGraph.merge_constraints local' uctx.universes;
+    universes = merge_constraints uctx local' uctx.universes;
     sort_variables = sorts;
     minim_extra = extra; }
 
@@ -734,42 +775,9 @@ let constrain_variables diff uctx =
   let local, vars = UnivFlex.constrain_variables diff uctx.univ_variables uctx.local in
   { uctx with local; univ_variables = vars }
 
-let id_of_level uctx l =
-  try (Level.Map.find l (snd (snd uctx.names))).uname
-  with Not_found -> None
-
-let id_of_qvar uctx l =
-  try (QVar.Map.find l (fst (snd uctx.names))).uname
-  with Not_found -> None
-
-let qualid_of_qvar_names (bind, (qrev,_)) l =
-  try Some (Libnames.qualid_of_ident (Option.get (QVar.Map.find l qrev).uname))
-  with Not_found | Option.IsNone ->
-    None (* no global qvars *)
-
-let qualid_of_level_names (bind, (_,urev)) l =
-  try Some (Libnames.qualid_of_ident (Option.get (Level.Map.find l urev).uname))
-  with Not_found | Option.IsNone ->
-    UnivNames.qualid_of_level bind l
-
-let qualid_of_level uctx l = qualid_of_level_names uctx.names l
-
-let pr_uctx_qvar_names names l =
-  match qualid_of_qvar_names names l with
-  | Some qid -> Libnames.pr_qualid qid
-  | None -> QVar.raw_pr l
-
-let pr_uctx_level_names names l =
-  match qualid_of_level_names names l with
-  | Some qid -> Libnames.pr_qualid qid
-  | None -> Level.raw_pr l
-
-let pr_uctx_level uctx l = pr_uctx_level_names uctx.names l
-
-let pr_uctx_qvar uctx l = pr_uctx_qvar_names uctx.names l
-
 type ('a, 'b, 'c) gen_universe_decl = {
   univdecl_qualities : 'a;
+  univdecl_extensible_qualities : bool;
   univdecl_instance : 'b; (* Declared universes *)
   univdecl_extensible_instance : bool; (* Can new universes be added *)
   univdecl_constraints : 'c; (* Declared constraints *)
@@ -780,6 +788,10 @@ type universe_decl =
 
 let default_univ_decl =
   { univdecl_qualities = [];
+    (* in practice non named qualities will get collapsed for toplevel definitions,
+       but side effects see named qualities from the surrounding definitions
+       while using default_univ_decl *)
+    univdecl_extensible_qualities = true;
     univdecl_instance = [];
     univdecl_extensible_instance = true;
     univdecl_constraints = Constraints.empty;
@@ -821,11 +833,31 @@ let pr_error_unbound_universes quals univs names =
 
 exception UnboundUnivs of QVar.Set.t * Level.Set.t * univ_names
 
-(* Deliberately using no location as the location of the univs
-   doesn't correspond to the failing command. *)
-let error_unbound_universes qs us uctx = raise (UnboundUnivs (qs,us,uctx))
+(* XXX when we have multi location errors we won't have to pick an arbitrary error *)
+let error_unbound_universes qs us uctx =
+  let exception Found of Loc.t in
+  let loc =
+    try
+      Level.Set.iter (fun u ->
+          match Level.Map.find_opt u (snd (snd uctx)) with
+          | None -> ()
+          | Some info -> match info.uloc with
+            | None -> ()
+            | Some loc -> raise_notrace (Found loc))
+        us;
+      QVar.Set.iter (fun s ->
+          match QVar.Map.find_opt s (fst (snd uctx)) with
+          | None -> ()
+          | Some info -> match info.uloc with
+            | None -> ()
+            | Some loc -> raise_notrace (Found loc))
+        qs;
+      None
+    with Found loc -> Some loc
+  in
+  Loc.raise ?loc (UnboundUnivs (qs,us,uctx))
 
-let _ = CErrors.register_handler (function
+let () = CErrors.register_handler (function
     | UnboundUnivs (qs,us,uctx) -> Some (pr_error_unbound_universes qs us uctx)
     | _ -> None)
 
@@ -833,9 +865,10 @@ let universe_context_inst decl qvars levels names =
   let leftqs = List.fold_left (fun acc l -> QVar.Set.remove l acc) qvars decl.univdecl_qualities in
   let leftus = List.fold_left (fun acc l -> Level.Set.remove l acc) levels decl.univdecl_instance in
   let () =
+    let unboundqs = if decl.univdecl_extensible_qualities then QVar.Set.empty else leftqs in
     let unboundus = if decl.univdecl_extensible_instance then Level.Set.empty else leftus in
-    if not (QVar.Set.is_empty leftqs && Level.Set.is_empty unboundus)
-    then error_unbound_universes leftqs unboundus names
+    if not (QVar.Set.is_empty unboundqs && Level.Set.is_empty unboundus)
+    then error_unbound_universes unboundqs unboundus names
   in
   let leftqs = UContext.sort_qualities
       (Array.map_of_list (fun q -> Quality.QVar q) (QVar.Set.elements leftqs))
@@ -858,8 +891,8 @@ let check_universe_context_set ~prefix levels names =
   then error_unbound_universes QVar.Set.empty left names
 
 let check_implication uctx cstrs cstrs' =
-  let gr = initial_graph uctx in
-  let grext = UGraph.merge_constraints cstrs gr in
+  let gr = uctx.initial_universes in
+  let grext = merge_constraints uctx cstrs gr in
   let cstrs' = Constraints.filter (fun c -> not (UGraph.check_constraint grext c)) cstrs' in
   if Constraints.is_empty cstrs' then ()
   else CErrors.user_err
@@ -867,6 +900,7 @@ let check_implication uctx cstrs cstrs' =
           Constraints.pr (pr_uctx_level uctx) cstrs')
 
 let check_mono_univ_decl uctx decl =
+  (* Note: if [decl] is [default_univ_decl], behave like [uctx.local] *)
   let () =
     if not (List.is_empty decl.univdecl_qualities)
     || not (QVar.Set.is_empty (QState.undefined uctx.sort_variables))
@@ -887,6 +921,7 @@ let check_mono_univ_decl uctx decl =
   end
 
 let check_poly_univ_decl uctx decl =
+  (* Note: if [decl] is [default_univ_decl], behave like [context uctx] *)
   let levels, csts = uctx.local in
   let qvars = QState.undefined uctx.sort_variables in
   let inst = universe_context_inst decl qvars levels uctx.names in
@@ -903,21 +938,17 @@ let check_poly_univ_decl uctx decl =
   uctx
 
 let check_univ_decl ~poly uctx decl =
+  let (binders, _) = uctx.names in
   let entry =
-    if not poly then
-      let ctx = check_mono_univ_decl uctx decl in
-      Monomorphic_entry ctx
-    else
-      let ctx = check_poly_univ_decl uctx decl in
-      Polymorphic_entry ctx
-  in
-  entry, fst uctx.names
+    if poly then Polymorphic_entry (check_poly_univ_decl uctx decl)
+    else Monomorphic_entry (check_mono_univ_decl uctx decl) in
+  entry, binders
 
 let is_bound l lbound = match lbound with
   | UGraph.Bound.Prop -> false
   | UGraph.Bound.Set -> Level.is_set l
 
-let restrict_universe_context ~lbound (univs, csts) keep =
+let restrict_universe_context ?(lbound = UGraph.Bound.Set) (univs, csts) keep =
   let removed = Level.Set.diff univs keep in
   if Level.Set.is_empty removed then univs, csts
   else
@@ -931,19 +962,22 @@ let restrict_universe_context ~lbound (univs, csts) keep =
   let csts = Constraints.filter (fun (l,d,r) -> not (is_bound l lbound && d == Le)) csts in
   (Level.Set.inter univs keep, csts)
 
-let restrict uctx vars =
-  let vars = Level.Set.union vars uctx.seff_univs in
+let restrict ?lbound uctx vars =
   let vars = Id.Map.fold (fun na l vars -> Level.Set.add l vars)
       (snd (fst uctx.names)) vars
   in
-  let uctx' = restrict_universe_context ~lbound:uctx.universes_lbound uctx.local vars in
+  let uctx' = restrict_universe_context ?lbound uctx.local vars in
   { uctx with local = uctx' }
 
-let restrict_even_binders uctx vars =
-  let vars = Level.Set.union vars uctx.seff_univs in
-  let uctx' = restrict_universe_context ~lbound:uctx.universes_lbound uctx.local vars in
+let restrict_even_binders ?lbound uctx vars =
+  let uctx' = restrict_universe_context ?lbound uctx.local vars in
   { uctx with local = uctx' }
 
+let restrict_constraints uctx csts =
+  let levels, _ = uctx.local in
+  let uctx' = { uctx with local = ContextSet.of_set levels; universes = uctx.initial_universes } in
+  add_constraints uctx' csts
+
 type rigid =
   | UnivRigid
   | UnivFlexible of bool (** Is substitution by an algebraic ok? *)
@@ -960,7 +994,7 @@ let merge ?loc ~sideff rigid uctx uctx' =
   let local = ContextSet.append uctx' uctx.local in
   let declare g =
     Level.Set.fold (fun u g ->
-        try UGraph.add_universe ~lbound:uctx.universes_lbound ~strict:false u g
+        try UGraph.add_universe ~lbound:UGraph.Bound.Set ~strict:false u g
         with UGraph.AlreadyDeclared when sideff -> g)
       levels g
   in
@@ -978,7 +1012,7 @@ let merge ?loc ~sideff rigid uctx uctx' =
   in
   let initial = declare uctx.initial_universes in
   let univs = declare uctx.universes in
-  let universes = UGraph.merge_constraints (ContextSet.constraints uctx') univs in
+  let universes = merge_constraints uctx (ContextSet.constraints uctx') univs in
   let uctx =
     match rigid with
     | UnivRigid -> uctx
@@ -1015,37 +1049,46 @@ let merge_sort_context ?loc ~sideff rigid uctx ((qvars,levels),csts) =
   let uctx = merge_sort_variables ?loc ~sideff uctx qvars in
   merge ?loc ~sideff rigid uctx (levels,csts)
 
-(* Check bug_4363 and bug_6323 when changing this code *)
-let demote_seff_univs univs uctx =
-  let seff = Level.Set.union uctx.seff_univs univs in
-  { uctx with seff_univs = seff }
+let demote_global_univs (lvl_set,csts_set) uctx =
+  let (local_univs, local_constraints) = uctx.local in
+  let local_univs = Level.Set.diff local_univs lvl_set in
+  let univ_variables = Level.Set.fold UnivFlex.remove lvl_set uctx.univ_variables in
+  let update_ugraph g =
+    let g = Level.Set.fold (fun u g ->
+        try UGraph.add_universe u ~lbound:Set ~strict:true g
+        with UGraph.AlreadyDeclared -> g)
+        lvl_set
+        g
+    in
+    UGraph.merge_constraints csts_set g
+  in
+  let initial_universes = update_ugraph uctx.initial_universes in
+  let universes = update_ugraph uctx.universes in
+  { uctx with local = (local_univs, local_constraints); univ_variables; universes; initial_universes }
 
-let demote_global_univs env uctx =
-  let env_ugraph = Environ.universes env in
-  let global_univs = UGraph.domain env_ugraph in
-  let global_constraints, _ = UGraph.constraints_of_universes env_ugraph in
-  let promoted_uctx =
-    ContextSet.(of_set global_univs |> add_constraints global_constraints) in
-  { uctx with local = ContextSet.diff uctx.local promoted_uctx }
+let demote_global_univ_entry entry uctx = match entry with
+  | Monomorphic_entry entry -> demote_global_univs entry uctx
+  | Polymorphic_entry _ -> uctx
+
+(* Check bug_4363 bug_6323 bug_3539 and success/rewrite lemma l1
+   for quick feedback when changing this code *)
+let emit_side_effects eff u =
+  let uctx = Safe_typing.universes_of_private eff in
+  demote_global_univs uctx u
 
 let merge_seff uctx uctx' =
   let levels = ContextSet.levels uctx' in
   let declare g =
     Level.Set.fold (fun u g ->
-        try UGraph.add_universe ~lbound:uctx.universes_lbound ~strict:false u g
+        try UGraph.add_universe ~lbound:UGraph.Bound.Set ~strict:false u g
         with UGraph.AlreadyDeclared -> g)
       levels g
   in
   let initial_universes = declare uctx.initial_universes in
   let univs = declare uctx.universes in
-  let universes = UGraph.merge_constraints (ContextSet.constraints uctx') univs in
+  let universes = merge_constraints uctx (ContextSet.constraints uctx') univs in
   { uctx with universes; initial_universes }
 
-let emit_side_effects eff u =
-  let uctx = Safe_typing.universes_of_private eff in
-  let u = demote_seff_univs (fst uctx) u in
-  merge_seff u uctx
-
 let update_sigma_univs uctx univs =
   let eunivs =
     { uctx with
@@ -1078,7 +1121,8 @@ let add_loc l loc (names, (qnames_rev,unames_rev) as orig) =
   | None -> orig
   | Some _ -> (names, (qnames_rev, Level.Map.add l { uname = None; uloc = loc } unames_rev))
 
-let add_universe ?loc name strict lbound uctx u =
+let add_universe ?loc name strict uctx u =
+  let lbound = UGraph.Bound.Set in
   let initial_universes = UGraph.add_universe ~lbound ~strict u uctx.initial_universes in
   let universes = UGraph.add_universe ~lbound ~strict u uctx.universes in
   let local = ContextSet.add_universe u uctx.local in
@@ -1110,10 +1154,10 @@ let new_univ_variable ?loc rigid name uctx =
       let univ_variables = UnivFlex.add u ~algebraic uctx.univ_variables in
       { uctx with univ_variables }
   in
-  let uctx = add_universe ?loc name false uctx.universes_lbound uctx u in
+  let uctx = add_universe ?loc name false uctx u in
   uctx, u
 
-let add_global_univ uctx u = add_universe None true UGraph.Bound.Set uctx u
+let add_forgotten_univ uctx u = add_universe None true uctx u
 
 let make_with_initial_binders ~lbound univs binders =
   let uctx = make ~lbound univs in
@@ -1123,7 +1167,7 @@ let make_with_initial_binders ~lbound univs binders =
     uctx binders
 
 let from_env ?(binders=[]) env =
-  make_with_initial_binders ~lbound:(Environ.universes_lbound env) (Environ.universes env) binders
+  make_with_initial_binders ~lbound:UGraph.Bound.Set (Environ.universes env) binders
 
 let make_nonalgebraic_variable uctx u =
   { uctx with univ_variables = UnivFlex.make_nonalgebraic_variable uctx.univ_variables u }
@@ -1148,12 +1192,14 @@ let normalize_variables uctx =
 let fix_undefined_variables uctx =
   { uctx with univ_variables = UnivFlex.fix_undefined_variables uctx.univ_variables }
 
+let collapse_above_prop_sort_variables ~to_prop uctx =
+  { uctx with sort_variables = QState.collapse_above_prop ~to_prop uctx.sort_variables }
+
 let collapse_sort_variables uctx =
   { uctx with sort_variables = QState.collapse uctx.sort_variables }
 
-let minimize uctx =
+let minimize ?(lbound = UGraph.Bound.Set) uctx =
   let open UnivMinim in
-  let lbound = uctx.universes_lbound in
   let (vars', us') =
     normalize_context_set ~lbound uctx.universes uctx.local uctx.univ_variables
       uctx.minim_extra
@@ -1163,20 +1209,87 @@ let minimize uctx =
     let universes = UGraph.merge_constraints (snd us') uctx.initial_universes in
       { names = uctx.names;
         local = us';
-        seff_univs = uctx.seff_univs; (* not sure about this *)
         univ_variables = vars';
         sort_variables = uctx.sort_variables;
         universes = universes;
-        universes_lbound = lbound;
         initial_universes = uctx.initial_universes;
         minim_extra = UnivMinim.empty_extra; (* weak constraints are consumed *) }
 
-(* XXX print above_prop too *)
-let pr_weak prl {minim_extra={UnivMinim.weak_constraints=weak}} =
-  let open Pp in
-  prlist_with_sep fnl (fun (u,v) -> prl u ++ str " ~ " ++ prl v) (UPairSet.elements weak)
+let universe_context_inst_decl decl qvars levels names =
+  let leftqs = List.fold_left (fun acc l -> QVar.Set.remove l acc) qvars decl.univdecl_qualities in
+  let leftus = List.fold_left (fun acc l -> Level.Set.remove l acc) levels decl.univdecl_instance in
+  let () =
+    let unboundqs = if decl.univdecl_extensible_qualities then QVar.Set.empty else leftqs in
+    let unboundus = if decl.univdecl_extensible_instance then Level.Set.empty else leftus in
+    if not (QVar.Set.is_empty unboundqs && Level.Set.is_empty unboundus)
+    then error_unbound_universes unboundqs unboundus names
+  in
+  let instq = Array.map_of_list (fun q -> Quality.QVar q) decl.univdecl_qualities in
+  let instu = Array.of_list decl.univdecl_instance in
+  let inst = Instance.of_array (instq,instu) in
+  inst
 
-let pr_sort_opt_subst uctx = QState.pr uctx.sort_variables
+let check_univ_decl_rev uctx decl =
+  let levels, csts = uctx.local in
+  let qvars = QState.undefined uctx.sort_variables in
+  let inst = universe_context_inst_decl decl qvars levels uctx.names in
+  let nas = compute_instance_binders (snd uctx.names) inst in
+  let () =
+    check_implication uctx
+    csts
+    decl.univdecl_constraints
+  in
+  let uctx = fix_undefined_variables uctx in
+  let uctx, csts =
+    if decl.univdecl_extensible_constraints
+    then uctx, csts else restrict_constraints uctx decl.univdecl_constraints, decl.univdecl_constraints
+  in
+  let uctx' = UContext.make nas (inst, csts) in
+  uctx, uctx'
+
+let check_uctx_impl ~fail uctx uctx' =
+  let levels, csts = uctx'.local in
+  let qvars_diff =
+    QVar.Set.diff
+      (QState.undefined uctx'.sort_variables)
+      (QState.undefined uctx.sort_variables)
+  in
+  let levels_diff = Level.Set.diff levels (fst uctx.local) in
+  let () = if not @@ (QVar.Set.is_empty qvars_diff && Level.Set.is_empty levels_diff) then
+    error_unbound_universes qvars_diff levels_diff uctx'.names
+  in
+  let () =
+    let grext = ugraph uctx in
+    let cstrs' = Constraints.filter (fun c -> not (UGraph.check_constraint grext c)) csts in
+    if Constraints.is_empty cstrs' then ()
+    else fail (Constraints.pr (pr_uctx_level uctx) cstrs')
+  in
+  ()
+
+
+(* XXX print above_prop too *)
+let pr_weak prl {minim_extra={UnivMinim.weak_constraints=weak; above_prop}} =
+  let open Pp in
+  v 0 (
+    prlist_with_sep cut (fun (u,v) -> h (prl u ++ str " ~ " ++ prl v)) (UPairSet.elements weak)
+    ++ if UPairSet.is_empty weak || Level.Set.is_empty above_prop then mt() else cut () ++
+    prlist_with_sep cut (fun u -> h (str "Prop <= " ++ prl u)) (Level.Set.elements above_prop))
+
+let pr_sort_opt_subst uctx = QState.pr (pr_uctx_qvar uctx) uctx.sort_variables
+
+let pr ctx =
+  let open Pp in
+  let prl = pr_uctx_level ctx in
+  if is_empty ctx then mt ()
+  else
+    v 0
+      (str"UNIVERSES:"++brk(0,1)++
+       h (Univ.pr_universe_context_set prl (context_set ctx)) ++ fnl () ++
+       UnivFlex.pr prl (subst ctx) ++ fnl() ++
+       str"SORTS:"++brk(0,1)++
+       h (pr_sort_opt_subst ctx) ++ fnl() ++
+       str "WEAK CONSTRAINTS:"++brk(0,1)++
+       h (pr_weak prl ctx) ++ fnl ())
 
 module Internal =
 struct
