@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -8,26 +8,23 @@
 (*         *     (see LICENSE file for the text of the license)         *)
 (************************************************************************)
 
+open Names
 open Univ
-
-val pr_with_global_universes : Level.t -> Pp.t
-val qualid_of_level : Level.t -> Libnames.qualid option
+open Sorts
 
 (** Local universe name <-> level mapping *)
 
-type universe_binders = Univ.Level.t Names.Id.Map.t
+type universe_binders = QVar.t Id.Map.t * Level.t Id.Map.t
+
+type rev_binders = Id.t QVar.Map.t * Id.t Level.Map.t
 
 val empty_binders : universe_binders
 
-val compute_instance_binders : Instance.t -> universe_binders -> Names.Name.t array
+val empty_rev_binders : rev_binders
 
 type univ_name_list = Names.lname list
 
-(** [universe_binders_with_opt_names ref l]
+type full_name_list = lname list * lname list
 
-    If [l] is [Some univs] return the universe binders naming the bound levels
-    of [ref] by [univs] (skipping Anonymous). May error if the lengths mismatch.
-
-    Otherwise return the bound universe names registered for [ref]. *)
-val universe_binders_with_opt_names : AUContext.t ->
-  univ_name_list option -> universe_binders
+val pr_level_with_global_universes : ?binders:universe_binders -> Level.t -> Pp.t
+val qualid_of_level : universe_binders -> Level.t -> Libnames.qualid option
