@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -37,6 +37,7 @@ type definition_object_kind =
   | Instance
   | Method
   | Let
+  | LetContext
 
 type assumption_object_kind = Definitional | Logical | Conjectural | Context
 
@@ -53,6 +54,7 @@ type assumption_object_kind = Definitional | Logical | Conjectural | Context
 
 type logical_kind =
   | IsPrimitive
+  | IsSymbol
   | IsAssumption of assumption_object_kind
   | IsDefinition of definition_object_kind
   | IsProof of theorem_kind
