@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -26,19 +26,40 @@ type glob_output =
   | MultFiles
   | File of string
 
-let glob_output = ref NoGlob
+let glob_output = ref []
 
-let dump () = !glob_output <> NoGlob
+let get_output () = match !glob_output with
+  | [] -> NoGlob
+  | g::_ -> g
 
-let set_glob_output mode =
-  glob_output := mode
+let push_output g = glob_output := g::!glob_output
+
+let pop_output () = glob_output := match !glob_output with
+    | [] -> CErrors.anomaly (Pp.str "No output left to pop")
+    | _::ds -> ds
+
+let pause () = push_output NoGlob
+let continue = pop_output
+
+let with_glob_output g f () =
+  push_output g;
+  try
+    let res = f () in
+    pop_output ();
+    res
+  with reraise ->
+    let reraise = Exninfo.capture reraise in
+    pop_output ();
+    Exninfo.iraise reraise
+
+let dump () = get_output () <> NoGlob
 
 let dump_string s =
-  if dump () && !glob_output != Feedback then
+  if dump () && get_output () != Feedback then
     output_string !glob_file s
 
 let start_dump_glob ~vfile ~vofile =
-  match !glob_output with
+  match get_output () with
   | MultFiles ->
       open_glob_file (Filename.chop_extension vofile ^ ".glob");
       output_string !glob_file "DIGEST ";
@@ -51,21 +72,17 @@ let start_dump_glob ~vfile ~vofile =
       ()
 
 let end_dump_glob () =
-  match !glob_output with
+  match get_output () with
   | MultFiles | File _ -> close_glob_file ()
   | NoGlob | Feedback -> ()
 
-let previous_state = ref MultFiles
-let pause () = previous_state := !glob_output; glob_output := NoGlob
-let continue () = glob_output := !previous_state
-
 open Decls
 open Declarations
 
 let type_of_logical_kind = function
   | IsDefinition def ->
       (match def with
-      | Definition | Let -> "def"
+      | Definition | Let | LetContext -> "def"
       | Coercion -> "coe"
       | SubClass -> "subclass"
       | CanonicalStructure -> "canonstruc"
@@ -93,6 +110,7 @@ let type_of_logical_kind = function
       | Proposition
       | Corollary -> "thm")
   | IsPrimitive -> "prim"
+  | IsSymbol -> "symb"
 
 
 (** Data associated to global parameters and constants *)
@@ -141,7 +159,7 @@ let interval loc =
   loc1, loc2-1
 
 let dump_ref ?loc filepath modpath ident ty =
-  match !glob_output with
+  match get_output () with
   | Feedback ->
     Option.iter (fun loc ->
         Feedback.feedback (Feedback.GlobRef (loc, filepath, modpath, ident, ty))
@@ -242,12 +260,12 @@ let mp_of_kn kn =
 
 let add_glob_kn ?loc kn =
   if dump () then
-    let sp = Nametab.path_of_syndef kn in
-    let lib_dp = Lib.dp_of_mp (mp_of_kn kn) in
-    add_glob_gen ?loc sp lib_dp "syndef"
+    let sp = Nametab.path_of_abbreviation kn in
+    let lib_dp = Names.ModPath.dp (mp_of_kn kn) in
+    add_glob_gen ?loc sp lib_dp "abbrev"
 
 let dump_def ?loc ty secpath id = Option.iter (fun loc ->
-  if !glob_output = Feedback then
+  if get_output () = Feedback then
     Feedback.feedback (Feedback.GlobDef (loc, id, secpath, ty))
   else
     let bl,el = interval loc in
@@ -267,7 +285,7 @@ let dump_moddef ?loc mp ty =
   let mp = Names.DirPath.to_string (Names.DirPath.make l) in
   dump_def ?loc ty "<>" mp
 
-let dump_notation (loc,(df,_)) sc sec = Option.iter (fun loc ->
+let dump_notation {CAst.loc;v=df} sc sec = Option.iter (fun loc ->
   (* We dump the location of the opening '"' *)
   let i = fst (Loc.unloc loc) in
   let location = (Loc.make_loc (i, i+1)) in
