@@ -1,7 +1,7 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
-(*  v      *   INRIA, CNRS and contributors - Copyright 1999-2019       *)
-(* <O___,, *       (see CREDITS file for the list of authors)           *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
+(*  v      *         Copyright INRIA, CNRS and contributors             *)
+(* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
 (*    //   *    This file is distributed under the terms of the         *)
 (*         *     GNU Lesser General Public License Version 2.1          *)
@@ -9,7 +9,6 @@
 (************************************************************************)
 
 open Names
-open Mod_subst
 open Genarg
 
 module Store : Store.S
@@ -17,7 +16,7 @@ module Store : Store.S
 type intern_variable_status = {
   intern_ids : Id.Set.t;
   notation_variable_status :
-    (bool ref * Notation_term.subscopes option ref *
+    (bool ref * Notation_term.subscopes option ref * Notation_term.notation_var_binders option ref *
        Notation_term.notation_var_internalization_type)
       Id.Map.t
 }
@@ -27,9 +26,10 @@ type glob_sign = {
   genv : Environ.env;
   extra : Store.t;
   intern_sign : intern_variable_status;
+  strict_check : bool;
 }
 
-val empty_glob_sign : Environ.env -> glob_sign
+val empty_glob_sign : strict:bool -> Environ.env -> glob_sign
 
 (** In globalize tactics, we need to keep the initial [constr_expr] to recompute
    in the environment by the effective calls to Intro, Inversion, etc
@@ -46,18 +46,17 @@ val intern : ('raw, 'glb, 'top) genarg_type -> ('raw, 'glb) intern_fun
 
 val generic_intern : (raw_generic_argument, glob_generic_argument) intern_fun
 
-(** {5 Substitution functions} *)
+(** {5 Internalization in tactic patterns} *)
 
-type 'glb subst_fun = substitution -> 'glb -> 'glb
-(** The type of functions used for substituting generic arguments. *)
+type ('raw,'glb) intern_pat_fun = ?loc:Loc.t -> ('raw,'glb) intern_fun
 
-val substitute : ('raw, 'glb, 'top) genarg_type -> 'glb subst_fun
+val intern_pat : ('raw, 'glb, 'top) genarg_type -> ('raw, 'glb) intern_pat_fun
 
-val generic_substitute : glob_generic_argument subst_fun
+val generic_intern_pat : (raw_generic_argument, glob_generic_argument) intern_pat_fun
 
 (** {5 Notation functions} *)
 
-type 'glb ntn_subst_fun = glob_constr_and_expr Id.Map.t -> 'glb -> 'glb
+type 'glb ntn_subst_fun = Id.Set.t -> Glob_term.glob_constr Id.Map.t -> 'glb -> 'glb
 
 val substitute_notation : ('raw, 'glb, 'top) genarg_type -> 'glb ntn_subst_fun
 
@@ -68,8 +67,9 @@ val generic_substitute_notation : glob_generic_argument ntn_subst_fun
 val register_intern0 : ('raw, 'glb, 'top) genarg_type ->
   ('raw, 'glb) intern_fun -> unit
 
-val register_subst0 : ('raw, 'glb, 'top) genarg_type ->
-  'glb subst_fun -> unit
+val register_intern_pat : ('raw, 'glb, 'top) genarg_type ->
+  ('raw, 'glb) intern_pat_fun -> unit
+
 
 val register_ntn_subst0 : ('raw, 'glb, 'top) genarg_type ->
   'glb ntn_subst_fun -> unit
