@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -16,13 +16,24 @@ open Constr
 open Globnames
 open Glob_term
 open Declarations
-open Lib
 open Libobject
 open EConstr
 open Reductionops
-open Namegen
 open Constrexpr
 
+let whd_prod env sigma typ =
+  let open CClosure in
+  let infos = Evarutil.create_clos_infos env sigma RedFlags.all in
+  let tab = create_tab () in
+  let typ = inject (EConstr.Unsafe.to_constr typ) in
+  let typ, stk = whd_stack infos tab typ [] in
+  match fterm_of typ with
+  | FProd (na, c1, c2, e) ->
+    let c1 = EConstr.of_constr @@ term_of_fconstr c1 in
+    let c2 = EConstr.of_constr @@ term_of_fconstr (mk_clos (CClosure.usubs_lift e) c2) in
+    Some (EConstr.of_binder_annot na, c1, c2)
+  | _ -> None
+
 module NamedDecl = Context.Named.Declaration
 
 (*s Flags governing the computation of implicit arguments *)
@@ -199,7 +210,7 @@ let is_reversible_pattern sigma bound depth f l =
 (* Precondition: rels in env are for inductive types only *)
 let add_free_rels_until strict strongly_strict revpat bound env sigma m pos acc =
   let rec frec rig (env,depth as ed) c =
-    let hd = if strict then whd_all env sigma c else c in
+    let hd = if strict then Reductionops.clos_whd_flags RedFlags.all env sigma c else c in
     let c = if strongly_strict then hd else c in
     match kind sigma hd with
     | Rel n when (n < bound+depth) && (n >= depth) ->
@@ -210,16 +221,16 @@ let add_free_rels_until strict strongly_strict revpat bound env sigma m pos acc
         acc.(i) <- update pos rig acc.(i)
     | App (f,_) when rig && is_flexible_reference env sigma bound depth f ->
         if strict then () else
-          iter_with_full_binders sigma push_lift (frec false) ed c
-    | Proj (p, _) when rig ->
+          iter_with_full_binders env sigma push_lift (frec false) ed c
+    | Proj (p, _, _) when rig ->
       if strict then () else
-        iter_with_full_binders sigma push_lift (frec false) ed c
+        iter_with_full_binders env sigma push_lift (frec false) ed c
     | Case _ when rig ->
         if strict then () else
-          iter_with_full_binders sigma push_lift (frec false) ed c
+          iter_with_full_binders env sigma push_lift (frec false) ed c
     | Evar _ -> ()
     | _ ->
-        iter_with_full_binders sigma push_lift (frec rig) ed c
+        iter_with_full_binders env sigma push_lift (frec rig) ed c
   in
   let () = if not (Vars.noccur_between sigma 1 bound m) then frec true (env,1) m in
   acc
@@ -229,59 +240,56 @@ let add_free_rels_until strict strongly_strict revpat bound env sigma m pos acc
 let rec is_rigid_head sigma t = match kind sigma t with
   | Rel _ | Evar _ -> false
   | Ind _ | Const _ | Var _ | Sort _ -> true
-  | Case (_,_,f,_) -> is_rigid_head sigma f
-  | Proj (p,c) -> true
+  | Case (_,_,_,_,_,f,_) -> is_rigid_head sigma f
+  | Proj _ -> true
   | App (f,args) ->
       (match kind sigma f with
         | Fix ((fi,i),_) -> is_rigid_head sigma (args.(fi.(i)))
         | _ -> is_rigid_head sigma f)
   | Lambda _ | LetIn _ | Construct _ | CoFix _ | Fix _
-  | Prod _ | Meta _ | Cast _ | Int _ | Float _ -> assert false
+  | Prod _ | Meta _ | Cast _ | Int _ | Float _ | String _ | Array _ -> assert false
 
 let is_rigid env sigma t =
   let open Context.Rel.Declaration in
-  let t = whd_all env sigma t in
-  match kind sigma t with
-  | Prod (na,a,b) ->
-     let (_,t) = splay_prod (push_rel (LocalAssum (na,a)) env) sigma b in
+  match whd_prod env sigma t with
+  | Some (na,a,b) ->
+     let (_,t) = whd_decompose_prod (push_rel (LocalAssum (na,a)) env) sigma b in
      is_rigid_head sigma t
   | _ -> true
 
-let find_displayed_name_in sigma all avoid na (env, b) =
-  let envnames_b = (env, b) in
-  let flag = RenamingElsewhereFor envnames_b in
-  if all then compute_and_force_displayed_name_in sigma flag avoid na b
-  else compute_displayed_name_in sigma flag avoid na b
-
-let compute_implicits_names_gen all env sigma t =
+let compute_implicits_names env sigma t =
   let open Context.Rel.Declaration in
-  let rec aux env avoid names t =
-    let t = whd_all env sigma t in
-    match kind sigma t with
-    | Prod (na,a,b) ->
-       let na',avoid' = find_displayed_name_in sigma all avoid na.Context.binder_name (names,b) in
-       aux (push_rel (LocalAssum (na,a)) env) avoid' (na'::names) b
-    | _ -> List.rev names
-  in aux env Id.Set.empty [] t
-
-let compute_implicits_names = compute_implicits_names_gen true
+  let rec aux env names t = match whd_prod env sigma t with
+  | Some (na, a, b) ->
+    let rels,ids = Termops.free_rels_and_unqualified_refs sigma a in
+    aux (push_rel (LocalAssum (na,a)) env) ((na.Context.binder_name,rels,ids)::names) b
+  | None ->
+    let rels,ids = Termops.free_rels_and_unqualified_refs sigma t in
+    let rec set_names (allrels,ids) = function
+    | [] -> (1,1,[])
+    | (na,rels',ids')::names ->
+      let (absolute_pos,nnondep,names) = set_names (rels'::allrels,Id.Set.union ids ids') names in
+      let isdep = List.exists_i (fun i rels -> Int.Set.mem i rels) 1 allrels in
+      let nnondep',dep_pos = if isdep then nnondep, None else nnondep + 1, Some nnondep in
+      (absolute_pos+1,nnondep',(na,absolute_pos,dep_pos)::names) in
+    let _,_,names = set_names ([rels],ids) names in
+    List.rev names
+  in aux env [] t
 
 let compute_implicits_explanation_gen strict strongly_strict revpat contextual env sigma t =
   let open Context.Rel.Declaration in
-  let rec aux env n t =
-    let t = whd_all env sigma t in
-    match kind sigma t with
-    | Prod (na,a,b) ->
-       add_free_rels_until strict strongly_strict revpat n env sigma a (Hyp (n+1))
-         (aux (push_rel (LocalAssum (na,a)) env) (n+1) b)
-    | _ ->
-       let v = Array.make n None in
-       if contextual then
-         add_free_rels_until strict strongly_strict revpat n env sigma t Conclusion v
-       else v
+  let rec aux env n t = match whd_prod env sigma t with
+  | Some (na, a, b) ->
+    add_free_rels_until strict strongly_strict revpat n env sigma a (Hyp (n+1))
+      (aux (push_rel (LocalAssum (na,a)) env) (n+1) b)
+  | _ ->
+    let v = Array.make n None in
+    if contextual then
+      add_free_rels_until strict strongly_strict revpat n env sigma t Conclusion v
+    else v
   in
-  match kind sigma (whd_all env sigma t) with
-  | Prod (na,a,b) ->
+  match whd_prod env sigma t with
+  | Some (na, a, b) ->
      let v = aux (push_rel (LocalAssum (na,a)) env) 1 b in
      Array.to_list v
   | _ -> []
@@ -291,9 +299,9 @@ let compute_implicits_explanation_flags env sigma f t =
     (f.strict || f.strongly_strict) f.strongly_strict
     f.reversible_pattern f.contextual env sigma t
 
-let compute_implicits_flags env sigma f all t =
+let compute_implicits_flags env sigma f t =
   List.combine
-    (compute_implicits_names_gen all env sigma t)
+    (compute_implicits_names env sigma t)
     (compute_implicits_explanation_flags env sigma f t)
 
 let compute_auto_implicits env sigma flags enriching t =
@@ -307,9 +315,16 @@ let compute_auto_implicits env sigma flags enriching t =
 type maximal_insertion = bool (* true = maximal contextual insertion *)
 type force_inference = bool (* true = always infer, never turn into evar/subgoal *)
 
-type implicit_status =
-    (* None = Not implicit *)
-    (explicitation * implicit_explanation * (maximal_insertion * force_inference)) option
+type implicit_position = Name.t * int * int option
+
+type implicit_status_info = {
+  impl_pos : implicit_position;
+  impl_expl : implicit_explanation;
+  impl_max : maximal_insertion;
+  impl_force : force_inference;
+}
+
+type implicit_status = implicit_status_info option
 
 type implicit_side_condition = DefaultImpArgs | LessArgsThan of int
 
@@ -322,33 +337,52 @@ let is_status_implicit = function
 let name_of_pos k = Id.of_string ("arg_" ^ string_of_int k)
 
 let binding_kind_of_status = function
-  | Some (_, _, (false, _)) -> NonMaxImplicit
-  | Some (_, _, (true, _)) -> MaxImplicit
+  | Some { impl_max = false } -> NonMaxImplicit
+  | Some { impl_max = true } -> MaxImplicit
   | None -> Explicit
 
 let name_of_implicit = function
   | None -> anomaly (Pp.str "Not an implicit argument.")
-  | Some (ExplByName id,_,_) -> id
-  | Some (ExplByPos (k,_),_,_) -> name_of_pos k
+  | Some { impl_pos = (Name id, _, _) } -> id
+  | Some { impl_pos = (Anonymous, k, _) } -> name_of_pos k
+
+let match_implicit imp pos = match imp, pos with
+  | None, _ -> anomaly (Pp.str "Not an implicit argument.")
+  | Some { impl_pos = (Name id, _, _) }, ExplByName id' -> Id.equal id id'
+  | Some { impl_pos = (_, _, Some n) }, ExplByPos n' -> Int.equal n n'
+  | Some { impl_pos = (_, n, _) }, ExplByName id -> Id.equal id (name_of_pos n)
+  | _ -> false
 
 let maximal_insertion_of = function
-  | Some (_,_,(b,_)) -> b
+  | Some imp -> imp.impl_max
   | None -> anomaly (Pp.str "Not an implicit argument.")
 
 let force_inference_of = function
-  | Some (_, _, (_, b)) -> b
+  | Some imp -> imp.impl_force
   | None -> anomaly (Pp.str "Not an implicit argument.")
 
+let is_nondep_implicit p imps =
+  List.exists (function Some { impl_pos = (_,_,Some p') } -> Int.equal p p' | _ -> false) imps
+
 (* [in_ctx] means we know the expected type, [n] is the index of the argument *)
 let is_inferable_implicit in_ctx n = function
   | None -> false
-  | Some (_,DepRigid (Hyp p),_) -> in_ctx || n >= p
-  | Some (_,DepFlex (Hyp p),_) -> false
-  | Some (_,DepFlexAndRigid (_,Hyp q),_) -> in_ctx || n >= q
-  | Some (_,DepRigid Conclusion,_) -> in_ctx
-  | Some (_,DepFlex Conclusion,_) -> false
-  | Some (_,DepFlexAndRigid (_,Conclusion),_) -> in_ctx
-  | Some (_,Manual,_) -> true
+  | Some { impl_expl = DepRigid (Hyp p) } -> in_ctx || n >= p
+  | Some { impl_expl = DepFlex (Hyp p) } -> false
+  | Some { impl_expl = DepFlexAndRigid (_,Hyp q) } -> in_ctx || n >= q
+  | Some { impl_expl = DepRigid Conclusion } -> in_ctx
+  | Some { impl_expl = DepFlex Conclusion } -> false
+  | Some { impl_expl = DepFlexAndRigid (_,Conclusion) } -> in_ctx
+  | Some { impl_expl = Manual } -> true
+
+let explicitation = function
+  | None -> anomaly (Pp.str "not implicit")
+  | Some imp ->
+    let (na, _, p) = imp.impl_pos in
+    match na, p with
+    | Name id, _ -> ExplByName id
+    | _, Some p -> ExplByPos p
+    | _ -> anomaly (Pp.str "Implicit without a position")
 
 let positions_of_implicits (_,impls) =
   let rec aux n = function
@@ -361,10 +395,15 @@ let positions_of_implicits (_,impls) =
 
 let rec prepare_implicits i f = function
   | [] -> []
-  | (Anonymous, Some _)::_ -> anomaly (Pp.str "Unnamed implicit.")
-  | (Name id, Some imp)::imps ->
+  | ((na,_,_ as pos), Some imp)::imps ->
       let imps' = prepare_implicits (i+1) f imps in
-      Some (ExplByName id,imp,(set_maximality Silent (Name id) i imps' f.maximal,true)) :: imps'
+      let data = {
+        impl_pos = pos;
+        impl_expl = imp;
+        impl_max = set_maximality Silent na i imps' f.maximal;
+        impl_force = true
+      } in
+      Some data :: imps'
   | _::imps -> None :: prepare_implicits (i+1) f imps
 
 let set_manual_implicits silent flags enriching autoimps l =
@@ -373,27 +412,39 @@ let set_manual_implicits silent flags enriching autoimps l =
     | autoimp::autoimps, explimp::explimps ->
        let imps' = merge (k+1) autoimps explimps in
        begin match autoimp, explimp.CAst.v with
-       | (Name id,_), Some (_,max) ->
-          Some (ExplByName id, Manual, (set_maximality (if silent then Silent else Error) (Name id) k imps' max, true))
-       | (Name id,Some exp), None when enriching ->
-          Some (ExplByName id, exp, (set_maximality (if silent then Silent else Info) (Name id) k imps' flags.maximal, true))
-       | (Name _,_), None -> None
-       | (Anonymous,_), Some (Name id,max) ->
-          Some (ExplByName id,Manual,(max,true))
-       | (Anonymous,_), Some (Anonymous,max) ->
-          Some (ExplByPos (k,None),Manual,(max,true))
-       | (Anonymous,_), None -> None
+       | ((Name _ as na,_,_ as pos),_), Some (_,max) ->
+          Some {
+            impl_pos = pos;
+            impl_expl = Manual;
+            impl_max = set_maximality (if silent then Silent else Error) na k imps' max;
+            impl_force = true;
+          }
+       | ((Anonymous,n1,n2),_), Some (na,max) ->
+          Some {
+            impl_pos = (na,n1,n2);
+            impl_expl = Manual;
+            impl_max = max;
+            impl_force = true;
+          }
+       | ((na,_,_ as pos),Some exp), None when enriching ->
+          Some {
+            impl_pos = pos;
+            impl_expl = exp;
+            impl_max = set_maximality (if silent then Silent else Info) na k imps' flags.maximal;
+            impl_force = true;
+          }
+       | (pos,_), None -> None
        end :: imps'
     | [], [] -> []
     | [], _ -> assert false
-    (* possibly more automatic than manual implicit arguments n
+    (* possibly more automatic than manual implicit arguments
        when the conclusion is an unfoldable constant *)
     | autoimps, [] -> merge k autoimps [CAst.make None]
   in merge 1 autoimps l
 
 let compute_semi_auto_implicits env sigma f t =
   if not f.auto then [DefaultImpArgs, []]
-  else let l = compute_implicits_flags env sigma f false t in
+  else let l = compute_implicits_flags env sigma f t in
        [DefaultImpArgs, prepare_implicits 1 f l]
 
 (*s Constants. *)
@@ -421,7 +472,7 @@ let compute_mib_implicits flags kn =
         (fun i mip ->
           (* No need to care about constraints here *)
           let ty, _ = Typeops.type_of_global_in_context env (GlobRef.IndRef (kn,i)) in
-          let r = Inductive.relevance_of_inductive env (kn,i) in
+          let r = (snd @@ Inductive.lookup_mind_specif env (kn,i)).mind_relevance in
           Context.Rel.Declaration.LocalAssum (Context.make_annot (Name mip.mind_typename) r, ty))
         mib.mind_packets) in
   let env_ar = Environ.push_rel_context ar env in
@@ -458,13 +509,13 @@ let compute_global_implicits flags = let open GlobRef in function
   | ConstructRef ((kn,i),j) ->
       let (_,cimps) = (compute_mib_implicits flags kn).(i) in snd cimps.(j-1)
 
-(* Merge a manual explicitation with an implicit_status list *)
+(* Merge a manual implicit position with an implicit_status list *)
 
 let merge_impls (cond,oldimpls) (_,newimpls) =
   let oldimpls,usersuffiximpls = List.chop (List.length newimpls) oldimpls in
   cond, (List.map2 (fun orig ni ->
     match orig with
-    | Some (_, Manual, _) -> orig
+    | Some { impl_expl = Manual } -> orig
     | _ -> ni) oldimpls newimpls)@usersuffiximpls
 
 (* Caching implicits *)
@@ -490,18 +541,18 @@ let implicits_of_global ref =
       let rec rename implicits names = match implicits, names with
         | [], _ -> []
         | _, [] -> implicits
-        | Some (_, x,y) :: implicits, Name id :: names ->
-           Some (ExplByName id, x,y) :: rename implicits names
+        | Some ({ impl_pos = (_, n1, n2) } as impl) :: implicits, Name id :: names ->
+           Some { impl with impl_pos = (Name id, n1, n2) } :: rename implicits names
         | imp :: implicits, _ :: names -> imp :: rename implicits names
       in
       List.map (fun (t, il) -> t, rename il rename_l) l
     with Not_found -> l
   with Not_found -> [DefaultImpArgs,[]]
 
-let cache_implicits_decl (ref,imps) =
+let cache_implicits_decl (ref, imps) =
   implicits_table := GlobRef.Map.add ref imps !implicits_table
 
-let load_implicits _ (_,(_,l)) = List.iter cache_implicits_decl l
+let load_implicits _ (_,l) = List.iter cache_implicits_decl l
 
 let cache_implicits o =
   load_implicits 1 o
@@ -517,32 +568,48 @@ let subst_implicits (subst,(req,l)) =
 let sec_implicits =
   Summary.ref Id.Map.empty ~name:"section-implicits"
 
-let impls_of_context ctx =
-  let map decl =
-    let id = NamedDecl.get_id decl in
+let impls_of_context vars =
+  let map n id =
+    let impl_pos = (Name id, n, None) in
     match Id.Map.get id !sec_implicits with
-    | NonMaxImplicit -> Some (ExplByName id, Manual, (false, true))
-    | MaxImplicit -> Some (ExplByName id, Manual, (true, true))
+    | NonMaxImplicit ->
+      Some { impl_pos; impl_expl = Manual; impl_max = false; impl_force = true }
+    | MaxImplicit ->
+      Some { impl_pos; impl_expl = Manual; impl_max = true; impl_force = true }
     | Explicit -> None
   in
-  List.rev_map map (List.filter (NamedDecl.is_local_assum) ctx)
+  List.map_i map 1 vars
 
 let adjust_side_condition p = function
   | LessArgsThan n -> LessArgsThan (n+p)
   | DefaultImpArgs -> DefaultImpArgs
 
+let lift_argument_position p = function
+  | Conclusion -> Conclusion
+  | Hyp q -> Hyp (p+q)
+
+let lift_explanation p = function
+  | DepRigid e -> DepRigid (lift_argument_position p e)
+  | DepFlex e -> DepFlex (lift_argument_position p e)
+  | DepFlexAndRigid (e1,e2) -> DepFlexAndRigid (lift_argument_position p e1,lift_argument_position p e2)
+  | Manual -> Manual
+
+let lift_implicits p imp =
+  let (na, n1, o) = imp.impl_pos in
+  { imp with impl_pos = (na, n1 + p, o); impl_expl = lift_explanation p imp.impl_expl }
+
 let add_section_impls vars extra_impls (cond,impls) =
   let p = List.length vars - List.length extra_impls in
-  adjust_side_condition p cond, extra_impls @ impls
+  adjust_side_condition p cond, extra_impls @ List.map (Option.map (lift_implicits p)) impls
 
-let discharge_implicits (_,(req,l)) =
+let discharge_implicits (req,l) =
   match req with
   | ImplLocal -> None
   | ImplMutualInductive _ | ImplInteractive _ | ImplConstant _ ->
      let l' =
        try
          List.map (fun (gr, l) ->
-             let vars = variable_section_segment_of_reference gr in
+             let vars = Array.map_to_list Constr.destVar (Lib.section_instance gr) in
              let extra_impls = impls_of_context vars in
              let newimpls = List.map (add_section_impls vars extra_impls) l in
              (gr, newimpls)) l
@@ -568,7 +635,7 @@ let rebuild_implicits (req,l) =
 
   | ImplInteractive (flags,o) ->
       let ref,oldimpls = List.hd l in
-      (if isVarRef ref && is_in_section ref then ImplLocal else req),
+      (if isVarRef ref && Lib.is_in_section ref then ImplLocal else req),
       match o with
       | ImplAuto ->
          let newimpls = compute_global_implicits flags ref in
@@ -582,9 +649,9 @@ let rebuild_implicits (req,l) =
          else
            [ref,oldimpls]
 
-let classify_implicits (req,_ as obj) = match req with
+let classify_implicits (req,_) = match req with
 | ImplLocal -> Dispose
-| _ -> Substitute obj
+| _ -> Substitute
 
 type implicits_obj =
     implicit_discharge_request *
@@ -599,11 +666,11 @@ let inImplicits : implicits_obj -> obj =
     discharge_function = discharge_implicits;
     rebuild_function = rebuild_implicits }
 
-let is_local local ref = local || isVarRef ref && is_in_section ref
+let is_local local ref = local || isVarRef ref && Lib.is_in_section ref
 
 let declare_implicits_gen req flags ref =
   let imps = compute_global_implicits flags ref in
-  add_anonymous_leaf (inImplicits (req,[ref,imps]))
+  Lib.add_leaf (inImplicits (req,[ref,imps]))
 
 let declare_implicits local ref =
   let flags = { !implicit_args with auto = true } in
@@ -625,15 +692,15 @@ let declare_mib_implicits kn =
   let imps = Array.map_to_list
     (fun (ind,cstrs) -> ind::(Array.to_list cstrs))
     (compute_mib_implicits flags kn) in
-    add_anonymous_leaf
+    Lib.add_leaf
       (inImplicits (ImplMutualInductive (kn,flags),List.flatten imps))
 
 (* Declare manual implicits *)
 type manual_implicits = (Name.t * bool) option CAst.t list
 
-let compute_implicits_with_manual env sigma typ enriching l =
+let compute_implicits_with_manual env sigma ?(silent=true) typ enriching l =
   let autoimpls = compute_auto_implicits env sigma !implicit_args enriching typ in
-  set_manual_implicits true !implicit_args enriching autoimpls l
+  set_manual_implicits silent !implicit_args enriching autoimpls l
 
 let check_inclusion l =
   (* Check strict inclusion *)
@@ -651,7 +718,7 @@ let check_rigidity isrigid =
 
 let projection_implicits env p impls =
   let npars = Projection.npars p in
-  CList.skipn_at_least npars impls
+  CList.skipn_at_best npars impls
 
 let declare_manual_implicits local ref ?enriching l =
   let flags = !implicit_args in
@@ -665,25 +732,38 @@ let declare_manual_implicits local ref ?enriching l =
   let req =
     if is_local local ref then ImplLocal
     else ImplInteractive(flags,ImplManual (List.length autoimpls))
-  in add_anonymous_leaf (inImplicits (req,[ref,l]))
+  in
+  Lib.add_leaf (inImplicits (req,[ref,l]))
 
 let maybe_declare_manual_implicits local ref ?enriching l =
   if List.exists (fun x -> x.CAst.v <> None) l then
     declare_manual_implicits local ref ?enriching l
 
-let explicit_kind i = function
-  | Name id -> ExplByName id
-  | Anonymous -> ExplByPos (i,None)
+let set_name (na',x,y as pos) = function
+  | Name _ as na -> (na,x,y)
+  | Anonymous -> pos
 
 let compute_implicit_statuses autoimps l =
   let rec aux i = function
-    | _ :: autoimps, Explicit :: manualimps -> None :: aux (i+1) (autoimps, manualimps)
-    | na :: autoimps, MaxImplicit :: manualimps ->
-       Some (explicit_kind i na, Manual, (true, true)) :: aux (i+1) (autoimps, manualimps)
-    | na :: autoimps, NonMaxImplicit :: manualimps ->
-       let imps' = aux (i+1) (autoimps, manualimps) in
-       let max = set_maximality Error na i imps' false in
-       Some (explicit_kind i na, Manual, (max, true)) :: imps'
+    | _ :: autoimps, (_, Explicit) :: manualimps -> None :: aux (i+1) (autoimps, manualimps)
+    | pos :: autoimps, (na, MaxImplicit) :: manualimps ->
+      let imp = {
+        impl_pos = set_name pos na;
+        impl_expl = Manual;
+        impl_max = true;
+        impl_force = true;
+      } in
+      Some imp :: aux (i+1) (autoimps, manualimps)
+    | pos :: autoimps, (na, NonMaxImplicit) :: manualimps ->
+      let imps' = aux (i+1) (autoimps, manualimps) in
+      let max = set_maximality Error na i imps' false in
+      let imp = {
+        impl_pos = set_name pos na;
+        impl_expl = Manual;
+        impl_max = max;
+        impl_force = true;
+      } in
+      Some imp :: imps'
     | autoimps, [] -> List.map (fun _ -> None) autoimps
     | [], _::_ -> assert false
   in aux 1 (autoimps, l)
@@ -703,7 +783,7 @@ let set_implicits local ref l =
        check_rigidity (is_rigid env sigma t);
        (* Sort by number of implicits, decreasing *)
        let is_implicit = function
-         | Explicit -> false
+         | _, Explicit -> false
          | _ -> true in
        let l = List.map (fun imps -> (imps,List.count is_implicit imps)) l in
        let l = List.sort (fun (_,n1) (_,n2) -> n2 - n1) l in
@@ -715,7 +795,8 @@ let set_implicits local ref l =
   let req =
     if is_local local ref then ImplLocal
     else ImplInteractive(flags,ImplManual (List.length autoimpls))
-  in add_anonymous_leaf (inImplicits (req,[ref,l']))
+  in
+  Lib.add_leaf (inImplicits (req,[ref,l']))
 
 let extract_impargs_data impls =
   let rec aux p = function
@@ -732,9 +813,15 @@ let rec drop_first_implicits p l =
   | DefaultImpArgs,imp::impls ->
       drop_first_implicits (p-1) (DefaultImpArgs,impls)
   | LessArgsThan n,imp::impls ->
-      let n = if is_status_implicit imp then n-1 else n in
+      let n = if is_status_implicit imp then n else n-1 in
       drop_first_implicits (p-1) (LessArgsThan n,impls)
 
+let rec chop_and_adjust n l1 l2 =
+  if n = 0 then (List.rev l1, l2) else
+    match l2 with
+    | [] -> chop_and_adjust (n-1) (None :: l1) []
+    | a::l2 -> chop_and_adjust (n-1) (a :: l1) l2
+
 let rec select_impargs_size n = function
   | [] -> [] (* Tolerance for (DefaultImpArgs,[]) *)
   | [_, impls] | (DefaultImpArgs, impls)::_ -> impls
@@ -744,3 +831,34 @@ let rec select_impargs_size n = function
 let select_stronger_impargs = function
   | [] -> [] (* Tolerance for (DefaultImpArgs,[]) *)
   | (_,impls)::_ -> impls
+
+let select_impargs_size_for_proj ~nexpectedparams ~ngivenparams ~nextraargs impls =
+  let split_implicit_params imps =
+    if List.is_empty imps then (nexpectedparams, [], []) else
+    let imps1, imps2 = chop_and_adjust nexpectedparams [] imps in
+    let imp, imps2 = match imps2 with imp :: imps2 -> [imp],imps2 | _ -> [], [] in
+    let nimps1 = nexpectedparams - List.count is_status_implicit imps1 in
+    (* Force the main argument to be explicit *)
+    (nimps1, imps1 @ [None], imps2)
+  in
+  let nallgivenargs = ngivenparams + nextraargs + 1 in
+  let little_enough_all_implicit = function
+  | (DefaultImpArgs, impls) -> Some (split_implicit_params impls)
+  | (LessArgsThan r, impls) when nallgivenargs <= r -> Some (split_implicit_params impls)
+  | _ -> None in
+  (* Compute the implicit signatures with little enough implicit to
+     match the number of arguments *)
+  let impls = List.map_filter little_enough_all_implicit impls in
+  (* Compute those which matches the number of parameters given *)
+  let impls' = List.filter (fun (nimps1,_,_) -> Int.equal nimps1 ngivenparams) impls in
+  match impls' with
+  | (_, imps1, imps2) :: _ -> Inl (imps1, imps2)
+  | [] ->
+    if Int.equal ngivenparams nexpectedparams then Inl ([], [])
+    else Inr (lazy (List.sort_uniq Int.compare (List.rev_map pi1 impls)))
+
+let impargs_for_proj ~nexpectedparams ~nextraargs imps =
+  let imps1, imps2 = try List.chop nexpectedparams imps with Failure _ -> imps, [] in
+  let imp, imps2 = match imps2 with imp :: imps2 -> imp, imps2 | _ -> None, [] in
+  let imps2 = try List.firstn nextraargs imps2 with Failure _ -> [] in
+  imps1@[None], imps2
