@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -15,11 +15,8 @@ open Util
 open Pp
 open Names
 open Nameops
-open Libobject
-open Lib
 open Notation_term
 open Notation_ops
-open Globnames
 
 type key =
   | RefKey of GlobRef.t
@@ -27,8 +24,17 @@ type key =
 
 (** TODO: share code from Notation *)
 
+let canonize_key env k = match k with
+| Oth -> Oth
+| RefKey gr ->
+  let gr' = Environ.QGlobRef.canonize env gr in
+  if gr' == gr then k else RefKey gr'
+
+let mkRefKey env gr =
+  RefKey (Environ.QGlobRef.canonize env gr)
+
 let key_compare k1 k2 = match k1, k2 with
-| RefKey gr1, RefKey gr2 -> GlobRef.Ordered.compare gr1 gr2
+| RefKey gr1, RefKey gr2 -> GlobRef.UserOrd.compare gr1 gr2
 | RefKey _, Oth -> -1
 | Oth, RefKey _ -> 1
 | Oth, Oth -> 0
@@ -63,63 +69,61 @@ struct
 end
 
 
-let keymap_add key data map =
+let keymap_add env key data map =
+  let key = canonize_key env key in
   let old = try KeyMap.find key map with Not_found -> ReservedSet.empty in
   KeyMap.add key (ReservedSet.add data old) map
 
 let reserve_table = Summary.ref Id.Map.empty ~name:"reserved-type"
 let reserve_revtable = Summary.ref KeyMap.empty ~name:"reserved-type-rev"
 
-let notation_constr_key = function (* Rem: NApp(NRef ref,[]) stands for @ref *)
-  | NApp (NRef ref,args) -> RefKey(canonical_gr ref), Some (List.length args)
-  | NList (_,_,NApp (NRef ref,args),_,_)
-  | NBinderList (_,_,NApp (NRef ref,args),_,_) -> RefKey (canonical_gr ref), Some (List.length args)
-  | NRef ref -> RefKey(canonical_gr ref), None
+let notation_constr_key env = function (* Rem: NApp(NRef ref,[]) stands for @ref *)
+  | NApp (NRef (ref,_),args) -> mkRefKey env ref, Some (List.length args)
+  | NList (_,_,NApp (NRef (ref,_),args),_,_)
+  | NBinderList (_,_,NApp (NRef (ref,_),args),_,_) -> mkRefKey env ref, Some (List.length args)
+  | NRef (ref,_) -> mkRefKey env ref, None
   | _ -> Oth, None
 
-let cache_reserved_type (_,(id,t)) =
-  let key = fst (notation_constr_key t) in
+let add_reserved_type env (id,t) =
+  let key = fst (notation_constr_key env t) in
   reserve_table := Id.Map.add id t !reserve_table;
-  reserve_revtable := keymap_add key (id, t) !reserve_revtable
+  reserve_revtable := keymap_add env key (id, t) !reserve_revtable
 
-let in_reserved : Id.t * notation_constr -> obj =
-  declare_object {(default_object "RESERVED-TYPE") with
-    cache_function = cache_reserved_type }
-
-let declare_reserved_type_binding {CAst.loc;v=id} t =
+let declare_reserved_type_binding env {CAst.loc;v=id} t =
   if not (Id.equal id (root_of_id id)) then
-    user_err ?loc ~hdr:"declare_reserved_type"
+    user_err ?loc
       ((Id.print id ++ str
-      " is not reservable: it must have no trailing digits, quote, or _"));
+      " is not reservable: it must have no trailing digits, quote, or _."));
   begin try
     let _ = Id.Map.find id !reserve_table in
-    user_err ?loc ~hdr:"declare_reserved_type"
-    ((Id.print id++str" is already bound to a type"))
+    user_err ?loc
+    ((Id.print id ++ str " is already bound to a type."))
   with Not_found -> () end;
-  add_anonymous_leaf (in_reserved (id,t))
+  add_reserved_type env (id,t)
 
 let declare_reserved_type idl t =
-  List.iter (fun id -> declare_reserved_type_binding id t) (List.rev idl)
+  let env = Global.env () in
+  List.iter (fun id -> declare_reserved_type_binding env id t) (List.rev idl)
 
 let find_reserved_type id = Id.Map.find (root_of_id id) !reserve_table
 
-let constr_key c =
-  try RefKey (canonical_gr (fst @@ Constr.destRef (fst (Constr.decompose_app c))))
+let constr_key env c =
+  try mkRefKey env (fst @@ Constr.destRef (fst (Constr.decompose_app c)))
   with Constr.DestKO -> Oth
 
 let revert_reserved_type t =
   try
-    let t = EConstr.Unsafe.to_constr t in
-    let reserved = KeyMap.find (constr_key t) !reserve_revtable in
-    let t = EConstr.of_constr t in
     let env = Global.env () in
+    let t = EConstr.Unsafe.to_constr t in
+    let reserved = KeyMap.find (constr_key env t) !reserve_revtable in
+    let t = EConstr.of_constr t in
     let evd = Evd.from_env env in
-    let t = Detyping.detype Detyping.Now false Id.Set.empty env evd t in
+    let t = Detyping.detype Detyping.Now env evd t in
     (* pedrot: if [Notation_ops.match_notation_constr] may raise [Failure _]
         then I've introduced a bug... *)
     let filter _ pat =
       try
-        let _ = match_notation_constr false t ([], pat) in
+        let _ = match_notation_constr ~print_univ:false t ~vars:Id.Set.empty ([], pat) in
         true
       with No_match -> false
     in
