@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -44,12 +44,13 @@ type fterm =
   | FCoFix of cofixpoint * usubs
   | FCaseT of case_info * UVars.Instance.t * constr array * case_return * fconstr * case_branch array * usubs (* predicate and branches are closures *)
   | FCaseInvert of case_info * UVars.Instance.t * constr array * case_return * finvert * fconstr * case_branch array * usubs
-  | FLambda of int * (Name.t Context.binder_annot * constr) list * constr * usubs
-  | FProd of Name.t Context.binder_annot * fconstr * constr * usubs
-  | FLetIn of Name.t Context.binder_annot * fconstr * fconstr * constr * usubs
+  | FLambda of int * (Name.t binder_annot * constr) list * constr * usubs
+  | FProd of Name.t binder_annot * fconstr * constr * usubs
+  | FLetIn of Name.t binder_annot * fconstr * fconstr * constr * usubs
   | FEvar of Evar.t * constr list * usubs * evar_repack
   | FInt of Uint63.t
   | FFloat of Float64.t
+  | FString of Pstring.t
   | FArray of UVars.Instance.t * fconstr Parray.t * fconstr
   | FLIFT of int * fconstr
   | FCLOS of constr * usubs
@@ -80,6 +81,8 @@ val check_native_args : CPrimitives.t -> stack -> bool
 val get_native_args1 : CPrimitives.t -> pconstant -> stack ->
   fconstr list * fconstr * fconstr next_native_args * stack
 
+val get_invert : finvert -> fconstr array
+
 val stack_args_size : stack -> int
 
 val inductive_subst : Declarations.mutual_inductive_body
@@ -94,7 +97,7 @@ val usubs_cons : fconstr -> usubs -> usubs
 (** identity if the first instance is empty *)
 val usubst_instance : 'a UVars.puniverses -> UVars.Instance.t -> UVars.Instance.t
 
-val usubst_binder : _ UVars.puniverses -> 'a Context.binder_annot -> 'a Context.binder_annot
+val usubst_binder : _ UVars.puniverses -> 'a binder_annot -> 'a binder_annot
 
 (** To lazy reduce a constr, create a [clos_infos] with
    [create_clos_infos], inject the term to reduce with [inject]; then use
@@ -102,16 +105,16 @@ val usubst_binder : _ UVars.puniverses -> 'a Context.binder_annot -> 'a Context.
 
 val inject : constr -> fconstr
 
-(** mk_atom: prevents a term from being evaluated *)
-val mk_atom : constr -> fconstr
+val mk_clos      : usubs -> constr -> fconstr
+val mk_clos_vect : usubs -> constr array -> fconstr array
 
-(** mk_red: makes a reducible term (used in ring) *)
-val mk_red : fterm -> fconstr
+val zip : fconstr -> stack -> fconstr
 
 val fterm_of : fconstr -> fterm
 val term_of_fconstr : fconstr -> constr
+val term_of_process : fconstr -> stack -> constr
 val destFLambda :
-  (usubs -> constr -> fconstr) -> fconstr -> Name.t Context.binder_annot * fconstr * fconstr
+  (usubs -> constr -> fconstr) -> fconstr -> Name.t binder_annot * fconstr * fconstr
 
 (** Global and local constant cache *)
 type clos_infos
@@ -121,18 +124,18 @@ type 'a evar_expansion =
 | EvarDefined of 'a
 | EvarUndefined of Evar.t * 'a list
 
-type 'constr evar_handler = {
-  evar_expand : 'constr pexistential -> 'constr evar_expansion;
-  evar_repack : Evar.t * 'constr list -> 'constr;
-  evar_irrelevant : 'constr pexistential -> bool;
+type evar_handler = {
+  evar_expand : constr pexistential -> constr evar_expansion;
+  evar_repack : Evar.t * constr list -> constr;
+  evar_irrelevant : constr pexistential -> bool;
   qvar_irrelevant : Sorts.QVar.t -> bool;
 }
 
-val default_evar_handler : env -> 'constr evar_handler
+val default_evar_handler : env -> evar_handler
 val create_conv_infos :
-  ?univs:UGraph.t -> ?evars:constr evar_handler -> reds -> env -> clos_infos
+  ?univs:UGraph.t -> ?evars:evar_handler -> reds -> env -> clos_infos
 val create_clos_infos :
-  ?univs:UGraph.t -> ?evars:constr evar_handler -> reds -> env -> clos_infos
+  ?univs:UGraph.t -> ?evars:evar_handler -> reds -> env -> clos_infos
 val oracle_of_infos : clos_infos -> Conv_oracle.oracle
 
 val create_tab : unit -> clos_tab
@@ -142,8 +145,8 @@ val info_flags: clos_infos -> reds
 val info_univs : clos_infos -> UGraph.t
 val unfold_projection : clos_infos -> Projection.t -> Sorts.relevance -> stack_member option
 
-val push_relevance : clos_infos -> 'b Context.binder_annot -> clos_infos
-val push_relevances : clos_infos -> 'b Context.binder_annot array -> clos_infos
+val push_relevance : clos_infos -> 'b binder_annot -> clos_infos
+val push_relevances : clos_infos -> 'b binder_annot array -> clos_infos
 val set_info_relevances : clos_infos -> Sorts.relevance Range.t -> clos_infos
 
 val info_relevances : clos_infos -> Sorts.relevance Range.t
@@ -170,7 +173,7 @@ val whd_stack :
 
 val skip_irrelevant_stack : clos_infos -> stack -> stack
 
-val eta_expand_stack : clos_infos -> Name.t Context.binder_annot -> stack -> stack
+val eta_expand_stack : clos_infos -> Name.t binder_annot -> stack -> stack
 
 (** [eta_expand_ind_stack env ind c s t] computes stacks corresponding
     to the conversion of the eta expansion of t, considered as an inhabitant
@@ -198,26 +201,7 @@ val unfold_ref_with_args
   -> stack
   -> (fconstr * stack) option
 
+val get_ref_mask : clos_infos -> clos_tab -> table_key -> bool array
+
 (** Hook for Reduction *)
 val set_conv : (clos_infos -> clos_tab -> fconstr -> fconstr -> bool) -> unit
-
-(***********************************************************************
-  i This is for lazy debug *)
-
-val lift_fconstr      : int -> fconstr -> fconstr
-val lift_fconstr_vect : int -> fconstr array -> fconstr array
-
-val mk_clos      : usubs -> constr -> fconstr
-val mk_clos_vect : usubs -> constr array -> fconstr array
-
-val kni: clos_infos -> clos_tab -> fconstr -> stack -> fconstr * stack
-val knr: clos_infos -> clos_tab -> fconstr -> stack -> fconstr * stack
-val kl : clos_infos -> clos_tab -> fconstr -> constr
-
-val zip : fconstr -> stack -> fconstr
-
-val term_of_process : fconstr -> stack -> constr
-
-val to_constr : lift UVars.puniverses -> fconstr -> constr
-
-(** End of cbn debug section i*)
