@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -8,15 +8,15 @@
 (*         *     (see LICENSE file for the text of the license)         *)
 (************************************************************************)
 
-(** This file defines the most important datatype of Coq, namely kernel terms,
+(** This file defines the most important datatype of Rocq, namely kernel terms,
     as well as a handful of generic manipulation functions. *)
 
 open Names
 
 (** {6 Simply type aliases } *)
-type pconstant = Constant.t Univ.puniverses
-type pinductive = inductive Univ.puniverses
-type pconstructor = constructor Univ.puniverses
+type pconstant = Constant.t UVars.puniverses
+type pinductive = inductive UVars.puniverses
+type pconstructor = constructor UVars.puniverses
 
 (** {6 Existential variables } *)
 type metavariable = int
@@ -25,9 +25,7 @@ type metavariable = int
 type case_style = LetStyle | IfStyle | LetPatternStyle | MatchStyle
   | RegularStyle (** infer printing form from number of constructor *)
 type case_printing =
-  { ind_tags : bool list; (** tell whether letin or lambda in the arity of the inductive type *)
-    cstr_tags : bool list array; (** tell whether letin or lambda in the signature of each constructor *)
-    style     : case_style }
+  { style     : case_style }
 
 (* INVARIANT:
  * - Array.length ci_cstr_ndecls = Array.length ci_cstr_nargs
@@ -45,7 +43,6 @@ type case_info =
                                        in addition to the parameters of the related inductive type
                                        NOTE: "lets" are therefore excluded from the count
                                        NOTE: parameters of the inductive type are also excluded from the count *)
-    ci_relevance : Sorts.relevance; (* relevance of the predicate (not of the inductive!) *)
     ci_pp_info    : case_printing   (* not interpreted by the kernel *)
   }
 
@@ -76,14 +73,20 @@ val mkVar : Id.t -> constr
 (** Constructs a machine integer *)
 val mkInt : Uint63.t -> constr
 
+(** Constructs an array *)
+val mkArray : UVars.Instance.t * constr array * constr * types -> constr
+
 (** Constructs a machine float number *)
 val mkFloat : Float64.t -> constr
 
+(** Constructs a machine string. *)
+val mkString : Pstring.t -> constr
+
 (** Constructs an patvar named "?n" *)
 val mkMeta : metavariable -> constr
 
 (** Constructs an existential variable *)
-type existential = Evar.t * constr list
+type existential = Evar.t * constr SList.t
 val mkEvar : existential -> constr
 
 (** Construct a sort *)
@@ -95,60 +98,89 @@ val mkType : Univ.Universe.t -> types
 
 
 (** This defines the strategy to use for verifiying a Cast *)
-type cast_kind = VMcast | NATIVEcast | DEFAULTcast | REVERTcast
+type cast_kind = VMcast | NATIVEcast | DEFAULTcast
 
 (** Constructs the term [t1::t2], i.e. the term t{_ 1} casted with the
    type t{_ 2} (that means t2 is declared as the type of t1). *)
 val mkCast : constr * cast_kind * constr -> constr
 
+type 'a binder_annot = ('a,Sorts.relevance) Context.pbinder_annot
+
 (** Constructs the product [(x:t1)t2] *)
-val mkProd : Name.t Context.binder_annot * types * types -> types
+val mkProd : Name.t binder_annot * types * types -> types
 
 (** Constructs the abstraction \[x:t{_ 1}\]t{_ 2} *)
-val mkLambda : Name.t Context.binder_annot * types * constr -> constr
+val mkLambda : Name.t binder_annot * types * constr -> constr
 
 (** Constructs the product [let x = t1 : t2 in t3] *)
-val mkLetIn : Name.t Context.binder_annot * constr * types * constr -> constr
+val mkLetIn : Name.t binder_annot * constr * types * constr -> constr
 
 (** [mkApp (f, [|t1; ...; tN|]] constructs the application
     {%html:(f t<sub>1</sub> ... t<sub>n</sub>)%}
     {%latex:$(f~t_1\dots f_n)$%}. *)
 val mkApp : constr * constr array -> constr
 
-val map_puniverses : ('a -> 'b) -> 'a Univ.puniverses -> 'b Univ.puniverses
+val map_puniverses : ('a -> 'b) -> 'a UVars.puniverses -> 'b UVars.puniverses
 
 (** Constructs a Constant.t *)
-val mkConst : Constant.t -> constr
 val mkConstU : pconstant -> constr
 
 (** Constructs a projection application *)
-val mkProj : (Projection.t * constr) -> constr
+val mkProj : (Projection.t * Sorts.relevance * constr) -> constr
 
 (** Inductive types *)
 
 (** Constructs the ith (co)inductive type of the block named kn *)
-val mkInd : inductive -> constr
 val mkIndU : pinductive -> constr
 
 (** Constructs the jth constructor of the ith (co)inductive type of the
    block named kn. *)
-val mkConstruct : constructor -> constr
 val mkConstructU : pconstructor -> constr
 val mkConstructUi : pinductive * int -> constr
 
 (** Make a constant, inductive, constructor or variable. *)
-val mkRef : GlobRef.t Univ.puniverses -> constr
+val mkRef : GlobRef.t UVars.puniverses -> constr
+
+module UnsafeMonomorphic : sig
+  val mkConst : Constant.t -> constr
+  val mkInd : inductive -> constr
+  val mkConstruct : constructor -> constr
+end
 
 (** Constructs a destructor of inductive type.
 
-    [mkCase ci p c ac] stand for match [c] as [x] in [I args] return [p] with [ac]
+    [mkCase ci params p c ac] stand for match [c] as [x] in [I args] return [p] with [ac]
     presented as describe in [ci].
 
-    [p] structure is [fun args x -> "return clause"]
+
+    [p] structure is [args x |- "return clause"]
 
     [ac]{^ ith} element is ith constructor case presented as
-    {e lambda construct_args (without params). case_term } *)
-val mkCase : case_info * constr * constr * constr array -> constr
+    {e construct_args |- case_term } *)
+
+type ('constr,'r) pcase_branch = (Name.t,'r) Context.pbinder_annot array * 'constr
+(** Names bound by matching the constructor for this branch. *)
+
+type ('types,'r) pcase_return = ((Name.t,'r) Context.pbinder_annot array * 'types) * 'r
+(** Names of the indices + name of self *)
+
+type 'constr pcase_invert =
+  | NoInvert
+  (** Normal reduction: match when the scrutinee is a constructor. *)
+
+  | CaseInvert of { indices : 'constr array; }
+  (** SProp to non SProp only:
+      No constructors or reduce when the indices match those of the unique constructor. *)
+
+type ('constr, 'types, 'univs, 'r) pcase =
+  case_info * 'univs * 'constr array * ('types,'r) pcase_return * 'constr pcase_invert * 'constr * ('constr,'r) pcase_branch array
+
+type case_invert = constr pcase_invert
+type case_return = (types, Sorts.relevance) pcase_return
+type case_branch = (constr, Sorts.relevance) pcase_branch
+type case = (constr, types, UVars.Instance.t, Sorts.relevance) pcase
+
+val mkCase : case -> constr
 
 (** If [recindxs = [|i1,...in|]]
       [funnames = [|f1,.....fn|]]
@@ -164,26 +196,25 @@ val mkCase : case_info * constr * constr * constr array -> constr
 
    where the length of the {% $ %}j{% $ %}th context is {% $ %}ij{% $ %}.
 *)
-type ('constr, 'types) prec_declaration =
-    Name.t Context.binder_annot array * 'types array * 'constr array
-type ('constr, 'types) pfixpoint =
-    (int array * int) * ('constr, 'types) prec_declaration
-  (* The array of [int]'s tells for each component of the array of
-     mutual fixpoints the number of lambdas to skip before finding the
-     recursive argument (e.g., value is 2 in "fix f (x:A) (y:=t) (z:B)
-     (v:=u) (w:I) {struct w}"), telling to skip x and z and that w is
-     the recursive argument);
-     The second component [int] tells which component of the block is
-     returned *)
+type ('constr, 'types, 'r) prec_declaration =
+    (Name.t, 'r) Context.pbinder_annot array * 'types array * 'constr array
+type ('constr, 'types, 'r) pfixpoint =
+    (int array * int) * ('constr, 'types, 'r) prec_declaration
+(** The array of [int]'s tells for each component of the array of
+   mutual fixpoints the number of lambdas to skip before finding the
+   recursive argument (e.g., value is 2 in "fix f (x:A) (y:=t) (z:B)
+   (v:=u) (w:I) {struct w}"), telling to skip x and z and that w is
+   the recursive argument); The second component [int] tells which
+   component of the block is returned *)
 
-type ('constr, 'types) pcofixpoint =
-    int * ('constr, 'types) prec_declaration
-  (* The component [int] tells which component of the block of
-     cofixpoint is returned *)
+type ('constr, 'types, 'r) pcofixpoint =
+    int * ('constr, 'types, 'r) prec_declaration
+(** The component [int] tells which component of the block of
+   cofixpoint is returned *)
 
-type rec_declaration = (constr, types) prec_declaration
+type rec_declaration = (constr, types, Sorts.relevance) prec_declaration
 
-type fixpoint = (constr, types) pfixpoint
+type fixpoint = (constr, types, Sorts.relevance) pfixpoint
 val mkFix : fixpoint -> constr
 
 (** If [funnames = [|f1,.....fn|]]
@@ -197,7 +228,7 @@ val mkFix : fixpoint -> constr
      ...
      with       fn = bn.]
  *)
-type cofixpoint = (constr, types) pcofixpoint
+type cofixpoint = (constr, types, Sorts.relevance) pcofixpoint
 val mkCoFix : cofixpoint -> constr
 
 
@@ -205,51 +236,73 @@ val mkCoFix : cofixpoint -> constr
 
 (** [constr list] is an instance matching definitional [named_context] in
    the same order (i.e. last argument first) *)
-type 'constr pexistential = Evar.t * 'constr list
-
-type ('constr, 'types, 'sort, 'univs) kind_of_term =
-  | Rel       of int                                  (** Gallina-variable introduced by [forall], [fun], [let-in], [fix], or [cofix]. *)
-
-  | Var       of Id.t                                 (** Gallina-variable that was introduced by Vernacular-command that extends
-                                                          the local context of the currently open section
-                                                          (i.e. [Variable] or [Let]). *)
+type 'constr pexistential = Evar.t * 'constr SList.t
 
+type ('constr, 'types, 'sort, 'univs, 'r) kind_of_term =
+  | Rel       of int
+  (** Gallina-variable introduced by [forall], [fun], [let-in], [fix], or [cofix]. *)
+  | Var       of Id.t
+  (** Gallina-variable that was introduced by Vernacular-command that
+     extends the local context of the currently open section (i.e.
+     [Variable] or [Let]). *)
   | Meta      of metavariable
   | Evar      of 'constr pexistential
   | Sort      of 'sort
   | Cast      of 'constr * cast_kind * 'types
-  | Prod      of Name.t Context.binder_annot * 'types * 'types             (** Concrete syntax ["forall A:B,C"] is represented as [Prod (A,B,C)]. *)
-  | Lambda    of Name.t Context.binder_annot * 'types * 'constr            (** Concrete syntax ["fun A:B => C"] is represented as [Lambda (A,B,C)].  *)
-  | LetIn     of Name.t Context.binder_annot * 'constr * 'types * 'constr  (** Concrete syntax ["let A:C := B in D"] is represented as [LetIn (A,B,C,D)]. *)
-  | App       of 'constr * 'constr array              (** Concrete syntax ["(F P1 P2 ...  Pn)"] is represented as [App (F, [|P1; P2; ...; Pn|])].
+  | Prod      of (Name.t,'r) Context.pbinder_annot * 'types * 'types
+  (** Concrete syntax ["forall A:B,C"] is represented as [Prod (A,B,C)]. *)
+  | Lambda    of (Name.t,'r) Context.pbinder_annot * 'types * 'constr
+  (** Concrete syntax ["fun A:B => C"] is represented as [Lambda (A,B,C)].  *)
+  | LetIn     of (Name.t,'r) Context.pbinder_annot * 'constr * 'types * 'constr
+  (** Concrete syntax ["let A:C := B in D"] is represented as [LetIn (A,B,C,D)]. *)
+  | App       of 'constr * 'constr array
+  (** Concrete syntax ["(F P1 P2 ...  Pn)"] is represented as [App (F, [|P1; P2; ...; Pn|])].
+      The {!mkApp} constructor also enforces the following invariant:
+      - [F] itself is not {!App}
+      - and [[|P1;..;Pn|]] is not empty. *)
+  | Const     of (Constant.t * 'univs)
+  (** Gallina-variable that was introduced by Vernacular-command that
+     extends the global environment (i.e. [Parameter], or [Axiom], or
+     [Definition], or [Theorem], or [Symbol] etc.) *)
+  | Ind       of (inductive * 'univs)
+  (** A name of an inductive type defined by [Variant], [Inductive] or
+     [Record] Vernacular-commands. *)
+  | Construct of (constructor * 'univs)
+  (** A constructor of an inductive type defined by [Variant],
+     [Inductive] or [Record] Vernacular-commands. *)
+  | Case      of case_info * 'univs * 'constr array * ('types,'r) pcase_return * 'constr pcase_invert * 'constr * ('constr,'r) pcase_branch array
+  (** [Case (ci,u,params,p,iv,c,brs)] is a [match c return p with brs]
+     expression. [c] lives in inductive [ci.ci_ind] at universe
+     instance [u] and parameters [params]. If this match has case
+     inversion (ie match on a 1 constructor SProp inductive with
+     proof relevant return type) the indices are in [iv].
 
-                                                          The {!mkApp} constructor also enforces the following invariant:
-                                                          - [F] itself is not {!App}
-                                                          - and [[|P1;..;Pn|]] is not empty. *)
+     The names in [p] are the names of the bound indices and
+     inductive value (ie the [in] and [as] clauses).
 
-  | Const     of (Constant.t * 'univs)                  (** Gallina-variable that was introduced by Vernacular-command that extends the global environment
-                                                          (i.e. [Parameter], or [Axiom], or [Definition], or [Theorem] etc.) *)
-
-  | Ind       of (inductive * 'univs)                 (** A name of an inductive type defined by [Variant], [Inductive] or [Record] Vernacular-commands. *)
-  | Construct of (constructor * 'univs)              (** A constructor of an inductive type defined by [Variant], [Inductive] or [Record] Vernacular-commands. *)
-  | Case      of case_info * 'constr * 'constr * 'constr array
-  | Fix       of ('constr, 'types) pfixpoint
-  | CoFix     of ('constr, 'types) pcofixpoint
-  | Proj      of Projection.t * 'constr
+     The names in the [brs] are the names of the variables bound in the respective branch. *)
+  | Fix       of ('constr, 'types, 'r) pfixpoint
+  | CoFix     of ('constr, 'types, 'r) pcofixpoint
+  | Proj      of Projection.t * 'r * 'constr
+  (** The relevance is the relevance of the whole term *)
   | Int       of Uint63.t
   | Float     of Float64.t
+  | String    of Pstring.t
+  | Array     of 'univs * 'constr array * 'constr * 'types
+  (** [Array (u,vals,def,t)] is an array of [vals] in type [t] with default value [def].
+      [u] is a universe containing [t]. *)
 
 (** User view of [constr]. For [App], it is ensured there is at
    least one argument and the function is not itself an applicative
    term *)
 
-val kind : constr -> (constr, types, Sorts.t, Univ.Instance.t) kind_of_term
-val of_kind : (constr, types, Sorts.t, Univ.Instance.t) kind_of_term -> constr
+val kind : constr -> (constr, types, Sorts.t, UVars.Instance.t, Sorts.relevance) kind_of_term
+val of_kind : (constr, types, Sorts.t, UVars.Instance.t, Sorts.relevance) kind_of_term -> constr
 
-val kind_nocast_gen : ('v -> ('v, 'v, 'sort, 'univs) kind_of_term) ->
-  ('v -> ('v, 'v, 'sort, 'univs) kind_of_term)
+val kind_nocast_gen : ('v -> ('v, 'v, 'sort, 'univs, 'r) kind_of_term) ->
+  ('v -> ('v, 'v, 'sort, 'univs, 'r) kind_of_term)
 
-val kind_nocast : constr -> (constr, types, Sorts.t, Univ.Instance.t) kind_of_term
+val kind_nocast : constr -> (constr, types, Sorts.t, UVars.Instance.t, Sorts.relevance) kind_of_term
 
 (** {6 Simple case analysis} *)
 val isRel  : constr -> bool
@@ -305,44 +358,44 @@ val destSort : constr -> Sorts.t
 val destCast : constr -> constr * cast_kind * constr
 
 (** Destructs the product {% $ %}(x:t_1)t_2{% $ %} *)
-val destProd : types -> Name.t Context.binder_annot * types * types
+val destProd : types -> Name.t binder_annot * types * types
 
 (** Destructs the abstraction {% $ %}[x:t_1]t_2{% $ %} *)
-val destLambda : constr -> Name.t Context.binder_annot * types * constr
+val destLambda : constr -> Name.t binder_annot * types * constr
 
 (** Destructs the let {% $ %}[x:=b:t_1]t_2{% $ %} *)
-val destLetIn : constr -> Name.t Context.binder_annot * constr * types * constr
+val destLetIn : constr -> Name.t binder_annot * constr * types * constr
 
 (** Destructs an application *)
 val destApp : constr -> constr * constr array
 
 (** Decompose any term as an applicative term; the list of args can be empty *)
-val decompose_app : constr -> constr * constr list
+val decompose_app_list : constr -> constr * constr list
 
 (** Same as [decompose_app], but returns an array. *)
-val decompose_appvect : constr -> constr * constr array
+val decompose_app : constr -> constr * constr array
 
 (** Destructs a constant *)
-val destConst : constr -> Constant.t Univ.puniverses
+val destConst : constr -> Constant.t UVars.puniverses
 
 (** Destructs an existential variable *)
 val destEvar : constr -> existential
 
 (** Destructs a (co)inductive type *)
-val destInd : constr -> inductive Univ.puniverses
+val destInd : constr -> inductive UVars.puniverses
 
 (** Destructs a constructor *)
-val destConstruct : constr -> constructor Univ.puniverses
+val destConstruct : constr -> constructor UVars.puniverses
 
 (** Destructs a [match c as x in I args return P with ... |
 Ci(...yij...) => ti | ... end] (or [let (..y1i..) := c as x in I args
 return P in t1], or [if c then t1 else t2])
 @return [(info,c,fun args x => P,[|...|fun yij => ti| ...|])]
 where [info] is pretty-printing information *)
-val destCase : constr -> case_info * constr * constr * constr array
+val destCase : constr -> case
 
 (** Destructs a projection *)
-val destProj : constr -> Projection.t * constr
+val destProj : constr -> Projection.t * Sorts.relevance * constr
 
 (** Destructs the {% $ %}i{% $ %}th function of the block
    [Fixpoint f{_ 1} ctx{_ 1} = b{_ 1}
@@ -355,7 +408,7 @@ val destFix : constr -> fixpoint
 
 val destCoFix : constr -> cofixpoint
 
-val destRef : constr -> GlobRef.t Univ.puniverses
+val destRef : constr -> GlobRef.t UVars.puniverses
 
 (** {6 Equality} *)
 
@@ -371,14 +424,6 @@ val eq_constr_univs : constr UGraph.check_function
     alpha, casts, application grouping and the universe inequalities in [u]. *)
 val leq_constr_univs : constr UGraph.check_function
 
-(** [eq_constr_univs u a b] is [true] if [a] equals [b] modulo alpha, casts,
-   application grouping and the universe equalities in [u]. *)
-val eq_constr_univs_infer : UGraph.t -> constr -> constr -> bool Univ.constrained
-
-(** [leq_constr_univs u a b] is [true] if [a] is convertible to [b] modulo
-    alpha, casts, application grouping and the universe inequalities in [u]. *)
-val leq_constr_univs_infer : UGraph.t -> constr -> constr -> bool Univ.constrained
-
 (** [eq_constr_univs a b] [true, c] if [a] equals [b] modulo alpha, casts,
    application grouping and ignoring universe instances. *)
 val eq_constr_nounivs : constr -> constr -> bool
@@ -388,9 +433,9 @@ val compare : constr -> constr -> int
 
 (** {6 Extension of Context with declarations on constr} *)
 
-type rel_declaration = (constr, types) Context.Rel.Declaration.pt
-type named_declaration = (constr, types) Context.Named.Declaration.pt
-type compacted_declaration = (constr, types) Context.Compacted.Declaration.pt
+type rel_declaration = (constr, types, Sorts.relevance) Context.Rel.Declaration.pt
+type named_declaration = (constr, types, Sorts.relevance) Context.Named.Declaration.pt
+type compacted_declaration = (constr, types, Sorts.relevance) Context.Compacted.Declaration.pt
 type rel_context = rel_declaration list
 type named_context = named_declaration list
 type compacted_context = compacted_declaration list
@@ -409,12 +454,6 @@ val lift : int -> constr -> constr
 (** {6 Functionals working on expressions canonically abstracted over
        a local context (possibly with let-ins)} *)
 
-(** [map_under_context f l c] maps [f] on the immediate subterms of a
-    term abstracted over a context of length [n] (local definitions
-    are counted) *)
-
-val map_under_context : (constr -> constr) -> int -> constr -> constr
-
 (** [map_branches f br] maps [f] on the immediate subterms of an array
    of "match" branches [br] in canonical eta-let-expanded form; it is
    not recursive and the order with which subterms are processed is
@@ -422,7 +461,7 @@ val map_under_context : (constr -> constr) -> int -> constr -> constr
    types and possibly terms occurring in the context of each branch as
    well as the body of each branch *)
 
-val map_branches : (constr -> constr) -> case_info -> constr array -> constr array
+val map_branches : (constr -> constr) -> case_branch array -> case_branch array
 
 (** [map_return_predicate f p] maps [f] on the immediate subterms of a
    return predicate of a "match" in canonical eta-let-expanded form;
@@ -431,16 +470,7 @@ val map_branches : (constr -> constr) -> case_info -> constr array -> constr arr
    the types and possibly terms occurring in the context of each
    branch as well as the body of the predicate *)
 
-val map_return_predicate : (constr -> constr) -> case_info -> constr -> constr
-
-(** [map_under_context_with_binders g f n l c] maps [f] on the
-    immediate subterms of a term abstracted over a context of length
-    [n] (local definitions are counted); it preserves sharing; it
-    carries an extra data [n] (typically a lift index) which is
-    processed by [g] (which typically add 1 to [n]) at each binder
-    traversal *)
-
-val map_under_context_with_binders : ('a -> 'a) -> ('a -> constr -> constr) -> 'a -> int -> constr -> constr
+val map_return_predicate : (constr -> constr) -> case_return -> case_return
 
 (** [map_branches_with_binders f br] maps [f] on the immediate
    subterms of an array of "match" branches [br] in canonical
@@ -452,7 +482,7 @@ val map_under_context_with_binders : ('a -> 'a) -> ('a -> constr -> constr) -> '
    occurring in the context of the branch as well as the body of the
    branch *)
 
-val map_branches_with_binders : ('a -> 'a) -> ('a -> constr -> constr) -> 'a -> case_info -> constr array -> constr array
+val map_branches_with_binders : ('a -> 'a) -> ('a -> constr -> constr) -> 'a -> case_branch array -> case_branch array
 
 (** [map_return_predicate_with_binders f p] maps [f] on the immediate
    subterms of a return predicate of a "match" in canonical
@@ -464,26 +494,7 @@ val map_branches_with_binders : ('a -> 'a) -> ('a -> constr -> constr) -> 'a ->
    occurring in the context of each branch as well as the body of the
    predicate *)
 
-val map_return_predicate_with_binders : ('a -> 'a) -> ('a -> constr -> constr) -> 'a -> case_info -> constr -> constr
-
-(** [map_under_context_with_full_binders g f n l c] is similar to
-    [map_under_context_with_binders] except that [g] takes also a full
-    binder as argument and that only the number of binders (and not
-    their signature) is required *)
-
-val map_under_context_with_full_binders : ((constr, constr) Context.Rel.Declaration.pt -> 'a -> 'a) -> ('a -> constr -> constr) -> 'a -> int -> constr -> constr
-
-(** [map_branches_with_full_binders g f l br] is equivalent to
-   [map_branches_with_binders] but using
-   [map_under_context_with_full_binders] *)
-
-val map_branches_with_full_binders : ((constr, constr) Context.Rel.Declaration.pt -> 'a -> 'a) -> ('a -> constr -> constr) -> 'a -> case_info -> constr array -> constr array
-
-(** [map_return_predicate_with_full_binders g f l p] is equivalent to
-   [map_return_predicate_with_binders] but using
-   [map_under_context_with_full_binders] *)
-
-val map_return_predicate_with_full_binders : ((constr, constr) Context.Rel.Declaration.pt -> 'a -> 'a) -> ('a -> constr -> constr) -> 'a -> case_info -> constr -> constr
+val map_return_predicate_with_binders : ('a -> 'a) -> ('a -> constr -> constr) -> 'a -> case_return -> case_return
 
 (** {6 Functionals working on the immediate subterm of a construction } *)
 
@@ -493,9 +504,7 @@ val map_return_predicate_with_full_binders : ((constr, constr) Context.Rel.Decla
 
 val fold : ('a -> constr -> 'a) -> 'a -> constr -> 'a
 
-val fold_with_full_binders :
-  (rel_declaration -> 'a -> 'a) -> ('a -> 'b -> constr -> 'b) ->
-    'a -> 'b -> constr -> 'b
+val fold_invert : ('a -> 'b -> 'a) -> 'a -> 'b pcase_invert -> 'a
 
 (** [map f c] maps [f] on the immediate subterms of [c]; it is
    not recursive and the order with which subterms are processed is
@@ -503,17 +512,15 @@ val fold_with_full_binders :
 
 val map : (constr -> constr) -> constr -> constr
 
-(** [map_user_view f c] maps [f] on the immediate subterms of [c]; it
-   differs from [map f c] in that the typing context and body of the
-   return predicate and of the branches of a [match] are considered as
-   immediate subterm of a [match] *)
-
-val map_user_view : (constr -> constr) -> constr -> constr
+val map_invert : ('a -> 'a) -> 'a pcase_invert -> 'a pcase_invert
 
 (** Like {!map}, but also has an additional accumulator. *)
 
 val fold_map : ('a -> constr -> 'a * constr) -> 'a -> constr -> 'a * constr
 
+val fold_map_invert : ('a -> 'b -> 'a * 'b) ->
+  'a -> 'b pcase_invert -> 'a * 'b pcase_invert
+
 (** [map_with_binders g f n c] maps [f n] on the immediate
    subterms of [c]; it carries an extra data [n] (typically a lift
    index) which is processed by [g] (which typically add 1 to [n]) at
@@ -529,6 +536,8 @@ val map_with_binders :
 
 val iter : (constr -> unit) -> constr -> unit
 
+val iter_invert : ('a -> unit) -> 'a pcase_invert -> unit
+
 (** [iter_with_binders g f n c] iters [f n] on the immediate
    subterms of [c]; it carries an extra data [n] (typically a lift
    index) which is processed by [g] (which typically add 1 to [n]) at
@@ -553,12 +562,12 @@ type 'constr constr_compare_fn = int -> 'constr -> 'constr -> bool
    the immediate subterms of [c1] of [c2] if needed; Cast's, binders
    name and Cases annotations are not taken into account *)
 
-val compare_head : constr constr_compare_fn -> constr constr_compare_fn
+val compare_head : (existential -> existential -> bool) -> constr constr_compare_fn -> constr constr_compare_fn
 
 (** Convert a global reference applied to 2 instances. The int says
    how many arguments are given (as we can only use cumulativity for
    fully applied inductives/constructors) .*)
-type 'univs instance_compare_fn = GlobRef.t -> int ->
+type 'univs instance_compare_fn = (GlobRef.t * int) option ->
   'univs -> 'univs -> bool
 
 (** [compare_head_gen u s f c1 c2] compare [c1] and [c2] using [f] to
@@ -566,16 +575,18 @@ type 'univs instance_compare_fn = GlobRef.t -> int ->
    compare universe instances, [s] to compare sorts; Cast's, binders
    name and Cases annotations are not taken into account *)
 
-val compare_head_gen : Univ.Instance.t instance_compare_fn ->
+val compare_head_gen : UVars.Instance.t instance_compare_fn ->
   (Sorts.t -> Sorts.t -> bool) ->
+  (existential -> existential -> bool) ->
   constr constr_compare_fn ->
   constr constr_compare_fn
 
 val compare_head_gen_leq_with :
-  ('v -> ('v, 'v, 'sort, 'univs) kind_of_term) ->
-  ('v -> ('v, 'v, 'sort, 'univs) kind_of_term) ->
+  ('v -> ('v, 'v, 'sort, 'univs, 'r) kind_of_term) ->
+  ('v -> ('v, 'v, 'sort, 'univs, 'r) kind_of_term) ->
   'univs instance_compare_fn ->
   ('sort -> 'sort -> bool) ->
+  ('v pexistential -> 'v pexistential -> bool) ->
   'v constr_compare_fn ->
   'v constr_compare_fn ->
   'v constr_compare_fn
@@ -585,10 +596,11 @@ val compare_head_gen_leq_with :
     is used,rather than {!kind}, to expose the immediate subterms of
     [c1] (resp. [c2]). *)
 val compare_head_gen_with :
-  ('v -> ('v, 'v, 'sort, 'univs) kind_of_term) ->
-  ('v -> ('v, 'v, 'sort, 'univs) kind_of_term) ->
+  ('v -> ('v, 'v, 'sort, 'univs, 'r) kind_of_term) ->
+  ('v -> ('v, 'v, 'sort, 'univs, 'r) kind_of_term) ->
   'univs instance_compare_fn ->
   ('sort -> 'sort -> bool) ->
+  ('v pexistential -> 'v pexistential -> bool) ->
   'v constr_compare_fn ->
   'v constr_compare_fn
 
@@ -599,12 +611,16 @@ val compare_head_gen_with :
     [s] to compare sorts for for subtyping; Cast's, binders name and
     Cases annotations are not taken into account *)
 
-val compare_head_gen_leq : Univ.Instance.t instance_compare_fn ->
+val compare_head_gen_leq : UVars.Instance.t instance_compare_fn ->
   (Sorts.t -> Sorts.t -> bool) ->
+  (existential -> existential -> bool) ->
   constr constr_compare_fn ->
   constr constr_compare_fn ->
   constr constr_compare_fn
 
+val eq_invert : ('a -> 'a -> bool)
+  -> 'a pcase_invert -> 'a pcase_invert -> bool
+
 (** {6 Hashconsing} *)
 
 val hash : constr -> int
@@ -612,7 +628,42 @@ val case_info_hash : case_info -> int
 
 (*********************************************************************)
 
+module GenHCons(C:sig
+    type t
+    val kind : t -> (t, t, Sorts.t, UVars.Instance.t, Sorts.relevance) kind_of_term
+    val self : t -> constr
+    val refcount : t -> int
+
+    val via_hconstr : bool
+
+    module Tbl : sig
+      val find_opt : t -> (constr * int) option
+      val add : t -> constr * int -> unit
+    end
+  end) : sig
+  val hcons : C.t -> constr
+end
+
+
 val hcons : constr -> constr
 
+val hasheq_kind : (_ kind_of_term as 'k) -> 'k -> bool
+(** Checks physical equality of every immediate element (goes inside tuples and arrays) *)
+
 val debug_print : constr -> Pp.t
-val debug_print_fix : ('a -> Pp.t) -> ('a, 'a) pfixpoint -> Pp.t
+val debug_print_fix : ('a -> Pp.t) -> ('a, 'a, 'r) pfixpoint -> Pp.t
+
+val mkConst : Constant.t -> constr
+[@@deprecated "(8.18) Use [mkConstU] or if truly needed [UnsafeMonomorphic.mkConst]"]
+
+val mkInd : inductive -> constr
+[@@deprecated "(8.18) Use [mkIndU] or if truly needed [UnsafeMonomorphic.mkInd]"]
+
+val mkConstruct : constructor -> constr
+[@@deprecated "(8.18) Use [mkConstructU] or if truly needed [UnsafeMonomorphic.mkConstruct]"]
+
+val hcons_annot : Name.t binder_annot -> Name.t binder_annot
+
+val hcons_caseinfo : case_info -> case_info
+
+val hash_cast_kind : cast_kind -> int
