@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -28,20 +28,19 @@ let safe_flags oracle = {
   indices_matter = true;
   impredicative_set = false;
   sprop_allowed = true;
-  cumulative_sprop = false;
   allow_uip = false;
 }
 
 (** {6 Arities } *)
 
-let subst_decl_arity f g sub ar =
+let subst_decl_arity f g subst ar =
   match ar with
   | RegularArity x ->
-    let x' = f sub x in
+    let x' = f subst x in
       if x' == x then ar
       else RegularArity x'
   | TemplateArity x ->
-    let x' = g sub x in
+    let x' = g subst x in
       if x' == x then ar
       else TemplateArity x'
 
@@ -50,22 +49,22 @@ let map_decl_arity f g = function
   | TemplateArity a -> TemplateArity (g a)
 
 let hcons_template_arity ar =
-  { template_level = Univ.hcons_univ ar.template_level; }
+  { template_level = Sorts.hcons ar.template_level; }
 
 let hcons_template_universe ar =
-  { template_param_levels = ar.template_param_levels;
+  { template_param_arguments = ar.template_param_arguments;
     template_context = Univ.hcons_universe_context_set ar.template_context }
 
 let universes_context = function
-  | Monomorphic -> Univ.AbstractContext.empty
+  | Monomorphic -> UVars.AbstractContext.empty
   | Polymorphic ctx -> ctx
 
 let abstract_universes = function
   | Entries.Monomorphic_entry ->
-    Univ.empty_level_subst, Monomorphic
+    UVars.empty_sort_subst, Monomorphic
   | Entries.Polymorphic_entry uctx ->
-    let (inst, auctx) = Univ.abstract_universes uctx in
-    let inst = Univ.make_instance_subst inst in
+    let (inst, auctx) = UVars.abstract_universes uctx in
+    let inst = UVars.make_instance_subst inst in
     (inst, Polymorphic auctx)
 
 (** {6 Constants } *)
@@ -77,7 +76,7 @@ let constant_is_polymorphic cb =
 
 
 let constant_has_body cb = match cb.const_body with
-  | Undef _ | Primitive _ -> false
+  | Undef _ | Primitive _ | Symbol _ -> false
   | Def _ | OpaqueDef _ -> true
 
 let constant_polymorphic_context cb =
@@ -85,40 +84,42 @@ let constant_polymorphic_context cb =
 
 let is_opaque cb = match cb.const_body with
   | OpaqueDef _ -> true
-  | Undef _ | Def _ | Primitive _ -> false
+  | Undef _ | Def _ | Primitive _ | Symbol _ -> false
 
 (** {7 Constant substitutions } *)
 
-let subst_rel_declaration sub =
-  RelDecl.map_constr (subst_mps sub)
+let subst_rel_declaration subst =
+  RelDecl.map_constr (subst_mps subst)
 
-let subst_rel_context sub = List.Smart.map (subst_rel_declaration sub)
+let subst_rel_context subst = List.Smart.map (subst_rel_declaration subst)
 
-let subst_const_type sub arity =
-  if is_empty_subst sub then arity
-  else subst_mps sub arity
+let subst_const_type subst arity =
+  if is_empty_subst subst then arity
+  else subst_mps subst arity
 
 (** No need here to check for physical equality after substitution,
     at least for Def due to the delayed substitution [subst_constr_subst]. *)
-let subst_const_def sub def = match def with
-  | Undef _ | Primitive _ -> def
-  | Def c -> Def (subst_mps sub c)
-  | OpaqueDef o -> OpaqueDef (Opaqueproof.subst_opaque sub o)
+let subst_const_def subst def = match def with
+  | Undef _ | Primitive _ | Symbol _ -> def
+  | Def c -> Def (subst_mps subst c)
+  | OpaqueDef o -> OpaqueDef (Opaqueproof.subst_opaque subst o)
 
-let subst_const_body sub cb =
-  assert (List.is_empty cb.const_hyps); (* we're outside sections *)
-  if is_empty_subst sub then cb
+let subst_const_body subst cb =
+  (* we're outside sections *)
+  assert (List.is_empty cb.const_hyps && UVars.Instance.is_empty cb.const_univ_hyps);
+  if is_empty_subst subst then cb
   else
-    let body' = subst_const_def sub cb.const_body in
-    let type' = subst_const_type sub cb.const_type in
+    let body' = subst_const_def subst cb.const_body in
+    let type' = subst_const_type subst cb.const_type in
     if body' == cb.const_body && type' == cb.const_type
     then cb
     else
       { const_hyps = [];
+        const_univ_hyps = UVars.Instance.empty;
         const_body = body';
         const_type = type';
         const_body_code =
-          Option.map (Vmemitcodes.subst_body_code sub) cb.const_body_code;
+          Option.map (Vmemitcodes.subst_body_code subst) cb.const_body_code;
         const_universes = cb.const_universes;
         const_relevance = cb.const_relevance;
         const_inline_code = cb.const_inline_code;
@@ -135,76 +136,73 @@ let hcons_rel_decl =
 
 let hcons_rel_context l = List.Smart.map hcons_rel_decl l
 
-let hcons_const_def = function
+let hcons_const_def ?(hbody=Constr.hcons) = function
   | Undef inl -> Undef inl
   | Primitive p -> Primitive p
+  | Symbol r -> Symbol r
   | Def l_constr ->
-    Def (Constr.hcons l_constr)
+    Def (hbody l_constr)
   | OpaqueDef _ as x -> x (* hashconsed when turned indirect *)
 
 let hcons_universes cbu =
   match cbu with
   | Monomorphic -> Monomorphic
   | Polymorphic ctx ->
-    Polymorphic (Univ.hcons_abstract_universe_context ctx)
+    Polymorphic (UVars.hcons_abstract_universe_context ctx)
 
-let hcons_const_body cb =
+let hcons_const_body ?hbody cb =
   { cb with
-    const_body = hcons_const_def cb.const_body;
+    const_body = hcons_const_def ?hbody cb.const_body;
     const_type = Constr.hcons cb.const_type;
     const_universes = hcons_universes cb.const_universes;
   }
 
 (** {6 Inductive types } *)
-let eq_nested_type t1 t2 = match t1, t2 with
-| NestedInd ind1, NestedInd ind2 -> Names.Ind.CanOrd.equal ind1 ind2
-| NestedInd _, _ -> false
-| NestedPrimitive c1, NestedPrimitive c2 -> Names.Constant.CanOrd.equal c1 c2
-| NestedPrimitive _, _ -> false
+
+let eq_recarg_type t1 t2 = match t1, t2 with
+| RecArgInd ind1, RecArgInd ind2 -> Names.Ind.CanOrd.equal ind1 ind2
+| RecArgPrim c1, RecArgPrim c2 -> Names.Constant.CanOrd.equal c1 c2
+| (RecArgInd _ | RecArgPrim _), _ -> false
 
 let eq_recarg r1 r2 = match r1, r2 with
 | Norec, Norec -> true
 | Norec, _ -> false
-| Mrec i1, Mrec i2 -> Names.Ind.CanOrd.equal i1 i2
+| Mrec t1, Mrec t2 -> eq_recarg_type t1 t2
 | Mrec _, _ -> false
-| Nested ty1, Nested ty2 -> eq_nested_type ty1 ty2
-| Nested _, _ -> false
 
-let pp_recarg = let open Pp in function
-  | Declarations.Norec -> str "Norec"
-  | Declarations.Mrec (mind,i) ->
+let pr_recarg_type = let open Pp in function
+  | RecArgInd (mind,i) ->
      str "Mrec[" ++ Names.MutInd.print mind ++ pr_comma () ++ int i ++ str "]"
-  | Declarations.(Nested (NestedInd (mind,i))) ->
-     str "Nested[" ++ Names.MutInd.print mind ++ pr_comma () ++ int i ++ str "]"
-  | Declarations.(Nested (NestedPrimitive c)) ->
-     str "Nested[" ++ Names.Constant.print c ++ str "]"
+  | RecArgPrim c ->
+     str "Prim[" ++ Names.Constant.print c ++ str "]"
 
-let pp_wf_paths x = Rtree.pp_tree pp_recarg x
+let pr_recarg = let open Pp in function
+  | Declarations.Norec -> str "Norec"
+  | Declarations.Mrec t -> pr_recarg_type t
 
-let subst_nested_type sub ty = match ty with
-| NestedInd (kn,i) ->
-  let kn' = subst_mind sub kn in
-  if kn==kn' then ty else NestedInd (kn',i)
-| NestedPrimitive c ->
-  let c',_ = subst_con sub c in
-  if c==c' then ty else NestedPrimitive c'
+let pr_wf_paths x = Rtree.pr_tree pr_recarg x
 
-let subst_recarg sub r = match r with
+let subst_recarg_type subst ty = match ty with
+| RecArgInd (kn,i) ->
+  let kn' = subst_mind subst kn in
+  if kn==kn' then ty else RecArgInd (kn',i)
+| RecArgPrim c ->
+  let c',_ = subst_con subst c in
+  if c==c' then ty else RecArgPrim c'
+
+let subst_recarg subst r = match r with
   | Norec -> r
-  | Mrec (kn,i) ->
-    let kn' = subst_mind sub kn in
-    if kn==kn' then r else Mrec (kn',i)
-  | Nested ty ->
-    let ty' = subst_nested_type sub ty in
-    if ty==ty' then r else Nested ty'
+  | Mrec ty ->
+    let ty' = subst_recarg_type subst ty in
+    if ty==ty' then r else Mrec ty'
 
 let mk_norec = Rtree.mk_node Norec [||]
 
 let mk_paths r recargs =
   Rtree.mk_node r
-    (Array.map (fun l -> Rtree.mk_node Norec (Array.of_list l)) recargs)
+    (Array.map Array.of_list recargs)
 
-let dest_recarg p = fst (Rtree.dest_node p)
+let dest_recarg p = Rtree.dest_head p
 
 (* dest_subterms returns the sizes of each argument of each constructor of
    an inductive object of size [p]. This should never be done for Norec,
@@ -214,18 +212,18 @@ let dest_recarg p = fst (Rtree.dest_node p)
 let dest_subterms p =
   let (ra,cstrs) = Rtree.dest_node p in
   assert (match ra with Norec -> false | _ -> true);
-  Array.map (fun t -> Array.to_list (snd (Rtree.dest_node t))) cstrs
+  Array.map Array.to_list cstrs
 
 let recarg_length p j =
   let (_,cstrs) = Rtree.dest_node p in
-  Array.length (snd (Rtree.dest_node cstrs.(j-1)))
+  Array.length cstrs.(j-1)
 
-let subst_wf_paths sub p = Rtree.Smart.map (subst_recarg sub) p
+let subst_wf_paths subst p = Rtree.Smart.map (subst_recarg subst) p
 
 (** {7 Substitution of inductive declarations } *)
 
-let subst_regular_ind_arity sub s =
-  let uar' = subst_mps sub s.mind_user_arity in
+let subst_regular_ind_arity subst s =
+  let uar' = subst_mps subst s.mind_user_arity in
     if uar' == s.mind_user_arity then s
     else { mind_user_arity = uar'; mind_sort = s.mind_sort }
 
@@ -235,46 +233,49 @@ let subst_template_ind_arity _sub s = s
 let subst_ind_arity =
   subst_decl_arity subst_regular_ind_arity subst_template_ind_arity
 
-let subst_mind_packet sub mbp =
+let subst_mind_packet subst mbp =
   { mind_consnames = mbp.mind_consnames;
     mind_consnrealdecls = mbp.mind_consnrealdecls;
     mind_consnrealargs = mbp.mind_consnrealargs;
     mind_typename = mbp.mind_typename;
-    mind_nf_lc = Array.Smart.map (fun (ctx, c) -> Context.Rel.map (subst_mps sub) ctx, subst_mps sub c) mbp.mind_nf_lc;
-    mind_arity_ctxt = subst_rel_context sub mbp.mind_arity_ctxt;
-    mind_arity = subst_ind_arity sub mbp.mind_arity;
-    mind_user_lc = Array.Smart.map (subst_mps sub) mbp.mind_user_lc;
+    mind_nf_lc = Array.Smart.map (fun (ctx, c) -> Context.Rel.map (subst_mps subst) ctx, subst_mps subst c) mbp.mind_nf_lc;
+    mind_arity_ctxt = subst_rel_context subst mbp.mind_arity_ctxt;
+    mind_arity = subst_ind_arity subst mbp.mind_arity;
+    mind_user_lc = Array.Smart.map (subst_mps subst) mbp.mind_user_lc;
     mind_nrealargs = mbp.mind_nrealargs;
     mind_nrealdecls = mbp.mind_nrealdecls;
-    mind_kelim = mbp.mind_kelim;
-    mind_recargs = subst_wf_paths sub mbp.mind_recargs (*wf_paths*);
+    mind_squashed = mbp.mind_squashed;
+    mind_recargs = subst_wf_paths subst mbp.mind_recargs (*wf_paths*);
     mind_relevance = mbp.mind_relevance;
     mind_nb_constant = mbp.mind_nb_constant;
     mind_nb_args = mbp.mind_nb_args;
     mind_reloc_tbl = mbp.mind_reloc_tbl }
 
-let subst_mind_record sub r = match r with
+let subst_mind_record subst r = match r with
 | NotRecord -> NotRecord
 | FakeRecord -> FakeRecord
 | PrimRecord infos ->
   let map (id, ps, rs, pb as info) =
-    let pb' = Array.Smart.map (subst_mps sub) pb in
+    let pb' = Array.Smart.map (subst_mps subst) pb in
     if pb' == pb then info
     else (id, ps, rs, pb')
   in
   let infos' = Array.Smart.map map infos in
   if infos' == infos then r else PrimRecord infos'
 
-let subst_mind_body sub mib =
-  { mind_record = subst_mind_record sub mib.mind_record ;
+let subst_mind_body subst mib =
+  (* we're outside sections *)
+  assert (List.is_empty mib.mind_hyps && UVars.Instance.is_empty mib.mind_univ_hyps);
+  { mind_record = subst_mind_record subst mib.mind_record ;
     mind_finite = mib.mind_finite ;
     mind_ntypes = mib.mind_ntypes ;
-    mind_hyps = (match mib.mind_hyps with [] -> [] | _ -> assert false);
+    mind_hyps = [];
+    mind_univ_hyps = UVars.Instance.empty;
     mind_nparams = mib.mind_nparams;
     mind_nparams_rec = mib.mind_nparams_rec;
     mind_params_ctxt =
-      Context.Rel.map (subst_mps sub) mib.mind_params_ctxt;
-    mind_packets = Array.Smart.map (subst_mind_packet sub) mib.mind_packets ;
+      Context.Rel.map (subst_mps subst) mib.mind_params_ctxt;
+    mind_packets = Array.Smart.map (subst_mind_packet subst) mib.mind_packets ;
     mind_universes = mib.mind_universes;
     mind_template = mib.mind_template;
     mind_variance = mib.mind_variance;
@@ -296,34 +297,30 @@ let inductive_is_cumulative mib =
 
 let inductive_make_projection ind mib ~proj_arg =
   match mib.mind_record with
-  | NotRecord | FakeRecord -> None
+  | NotRecord | FakeRecord ->
+    CErrors.anomaly Pp.(str "inductive_make_projection: not a primitive record.")
   | PrimRecord infos ->
-    let _, labs, _, _ = infos.(snd ind) in
-    Some (Names.Projection.Repr.make ind
-            ~proj_npars:mib.mind_nparams
-            ~proj_arg
-            labs.(proj_arg))
+    let _, labs, rs, _ = infos.(snd ind) in
+    if proj_arg < 0 || Array.length labs <= proj_arg
+    then CErrors.anomaly Pp.(str "inductive_make_projection: invalid proj_arg.");
+    let p = Names.Projection.Repr.make ind
+      ~proj_npars:mib.mind_nparams
+      ~proj_arg
+      labs.(proj_arg)
+    in
+    p, rs.(proj_arg)
 
 let inductive_make_projections ind mib =
   match mib.mind_record with
   | NotRecord | FakeRecord -> None
   | PrimRecord infos ->
-    let _, labs, _, _ = infos.(snd ind) in
-    let projs = Array.mapi (fun proj_arg lab ->
-        Names.Projection.Repr.make ind ~proj_npars:mib.mind_nparams ~proj_arg lab)
-        labs
+    let _, labs, relevances, _ = infos.(snd ind) in
+    let projs = Array.map2_i (fun proj_arg lab r ->
+        Names.Projection.Repr.make ind ~proj_npars:mib.mind_nparams ~proj_arg lab, r)
+        labs relevances
     in
     Some projs
 
-let relevance_of_projection_repr mib p =
-  let _mind,i = Names.Projection.Repr.inductive p in
-  match mib.mind_record with
-  | NotRecord | FakeRecord ->
-    CErrors.anomaly ~label:"relevance_of_projection" Pp.(str "not a projection")
-  | PrimRecord infos ->
-    let _,_,rs,_ = infos.(i) in
-    rs.(Names.Projection.Repr.arg p)
-
 (** {6 Hash-consing of inductive declarations } *)
 
 let hcons_regular_ind_arity a =
@@ -356,8 +353,25 @@ let hcons_mind mib =
     mind_template = Option.Smart.map hcons_template_universe mib.mind_template;
     mind_universes = hcons_universes mib.mind_universes }
 
+let subst_rewrite_rules subst ({ rewrules_rules } as rules) =
+  let body' = List.Smart.map (fun (name, ({ rhs; _ } as rule) as orig) ->
+      let rhs' = subst_mps subst rhs in
+      if rhs == rhs' then orig else name, { rule with rhs = rhs' })
+      rewrules_rules
+  in
+  if rewrules_rules == body' then rules else
+    { rewrules_rules = body' }
+
+let mod_expr { mod_expr = ModBodyVal v; _ } = v
+
 (** Hashconsing of modules *)
 
+let hcons_when_mod_body (type a b) (f : b -> b) : (a, b) when_mod_body -> (a, b) when_mod_body = function
+| ModBodyVal v as arg ->
+  let v' = f v in
+  if v == v' then arg else ModBodyVal v'
+| ModTypeNul -> ModTypeNul
+
 let hcons_functorize hty he hself f = match f with
 | NoFunctor e ->
   let e' = he e in
@@ -372,6 +386,14 @@ let hcons_functorize hty he hself f = match f with
 
 let hcons_module_alg_expr me = me
 
+let rec hcons_module_expression me = match me with
+| MENoFunctor malg ->
+  let malg' = hcons_module_alg_expr malg in
+  if malg == malg' then me else MENoFunctor malg'
+| MEMoreFunctor mf ->
+  let mf' = hcons_module_expression mf in
+  if mf' == mf then me else MEMoreFunctor mf'
+
 let rec hcons_structure_field_body sb = match sb with
 | SFBconst cb ->
   let cb' = hcons_const_body cb in
@@ -380,11 +402,12 @@ let rec hcons_structure_field_body sb = match sb with
   let mib' = hcons_mind mib in
   if mib == mib' then sb else SFBmind mib'
 | SFBmodule mb ->
-  let mb' = hcons_module_body mb in
+  let mb' = hcons_generic_module_body mb in
   if mb == mb' then sb else SFBmodule mb'
 | SFBmodtype mb ->
-  let mb' = hcons_module_type mb in
+  let mb' = hcons_generic_module_body mb in
   if mb == mb' then sb else SFBmodtype mb'
+| SFBrules _ -> sb (* TODO? *)
 
 and hcons_structure_body sb =
   (** FIXME *)
@@ -396,10 +419,7 @@ and hcons_structure_body sb =
   List.Smart.map map sb
 
 and hcons_module_signature ms =
-  hcons_functorize hcons_module_type hcons_structure_body hcons_module_signature ms
-
-and hcons_module_expression me =
-  hcons_functorize hcons_module_type hcons_module_alg_expr hcons_module_expression me
+  hcons_functorize hcons_generic_module_body hcons_structure_body hcons_module_signature ms
 
 and hcons_module_implementation mip = match mip with
 | Abstract -> Abstract
@@ -407,15 +427,15 @@ and hcons_module_implementation mip = match mip with
   let me' = hcons_module_expression me in
   if me == me' then mip else Algebraic me'
 | Struct ms ->
-  let ms' = hcons_module_signature ms in
+  let ms' = hcons_structure_body ms in
   if ms == ms' then mip else Struct ms
 | FullStruct -> FullStruct
 
 and hcons_generic_module_body :
-  'a. ('a -> 'a) -> 'a generic_module_body -> 'a generic_module_body =
-  fun hcons_impl mb ->
+  'a. 'a generic_module_body -> 'a generic_module_body =
+  fun mb ->
   let mp' = mb.mod_mp in
-  let expr' = hcons_impl mb.mod_expr in
+  let expr' = hcons_when_mod_body hcons_module_implementation mb.mod_expr in
   let type' = hcons_module_signature mb.mod_type in
   let type_alg' = mb.mod_type_alg in
   let delta' = mb.mod_delta in
@@ -438,8 +458,8 @@ and hcons_generic_module_body :
     mod_retroknowledge = retroknowledge';
   }
 
-and hcons_module_body mb =
-  hcons_generic_module_body hcons_module_implementation mb
+let hcons_module_body =
+  hcons_generic_module_body
 
-and hcons_module_type mb =
-  hcons_generic_module_body (fun () -> ()) mb
+let hcons_module_type =
+  hcons_generic_module_body
