@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -10,14 +10,14 @@
 
 open Declarations
 open Mod_subst
-open Univ
+open UVars
 
 (** Operations concerning types in [Declarations] :
     [constant_body], [mutual_inductive_body], [module_body] ... *)
 
-val universes_context : universes -> AUContext.t
+val universes_context : universes -> AbstractContext.t
 
-val abstract_universes : Entries.universes_entry -> Univ.universe_level_subst * universes
+val abstract_universes : Entries.universes_entry -> UVars.sort_level_subst * universes
 
 (** {6 Arities} *)
 
@@ -26,26 +26,29 @@ val map_decl_arity : ('a -> 'c) -> ('b -> 'd) ->
 
 (** {6 Constants} *)
 
-val subst_const_body : substitution -> Opaqueproof.opaque constant_body -> Opaqueproof.opaque constant_body
+val subst_const_body : substitution -> constant_body -> constant_body
 
 (** Is there a actual body in const_body ? *)
 
-val constant_has_body : 'a constant_body -> bool
+val constant_has_body : ('a, 'b) pconstant_body -> bool
 
-val constant_polymorphic_context : 'a constant_body -> AUContext.t
+val constant_polymorphic_context : ('a, 'b) pconstant_body -> AbstractContext.t
 
 (** Is the constant polymorphic? *)
-val constant_is_polymorphic : 'a constant_body -> bool
+val constant_is_polymorphic : ('a, 'b) pconstant_body -> bool
 
 (** Return the universe context, in case the definition is polymorphic, otherwise
     the context is empty. *)
 
-val is_opaque : 'a constant_body -> bool
+val is_opaque : ('a, 'b) pconstant_body -> bool
 
 (** {6 Inductive types} *)
 
 val eq_recarg : recarg -> recarg -> bool
 
+val pr_recarg : recarg -> Pp.t
+val pr_wf_paths : wf_paths -> Pp.t
+
 val subst_recarg : substitution -> recarg -> recarg
 
 val mk_norec : wf_paths
@@ -58,19 +61,25 @@ val subst_wf_paths : substitution -> wf_paths -> wf_paths
 
 val subst_mind_body : substitution -> mutual_inductive_body -> mutual_inductive_body
 
-val inductive_polymorphic_context : mutual_inductive_body -> AUContext.t
+val subst_rewrite_rules : substitution -> rewrite_rules_body -> rewrite_rules_body
+
+val inductive_polymorphic_context : mutual_inductive_body -> AbstractContext.t
 
 (** Is the inductive polymorphic? *)
 val inductive_is_polymorphic : mutual_inductive_body -> bool
 (** Is the inductive cumulative? *)
 val inductive_is_cumulative : mutual_inductive_body -> bool
 
+(** Anomaly when not a primitive record or invalid proj_arg *)
 val inductive_make_projection : Names.inductive -> mutual_inductive_body -> proj_arg:int ->
-  Names.Projection.Repr.t option
-val inductive_make_projections : Names.inductive -> mutual_inductive_body ->
-  Names.Projection.Repr.t array option
+  Names.Projection.Repr.t * Sorts.relevance
 
-val relevance_of_projection_repr : mutual_inductive_body -> Names.Projection.Repr.t -> Sorts.relevance
+val inductive_make_projections : Names.inductive -> mutual_inductive_body ->
+  (Names.Projection.Repr.t * Sorts.relevance) array option
+
+(** {6 Modules} *)
+
+val mod_expr : module_body -> module_implementation
 
 (** {6 Kernel flags} *)
 
@@ -83,7 +92,9 @@ val safe_flags : Conv_oracle.oracle -> typing_flags
     of the structure, but simply hash-cons all inner constr
     and other known elements *)
 
-val hcons_const_body : 'a constant_body -> 'a constant_body
+val hcons_const_body : ?hbody:(Constr.t -> Constr.t) ->
+  ('a, 'b) pconstant_body -> ('a, 'b) pconstant_body
 val hcons_mind : mutual_inductive_body -> mutual_inductive_body
+val hcons_generic_module_body : 'a generic_module_body -> 'a generic_module_body
 val hcons_module_body : module_body -> module_body
 val hcons_module_type : module_type_body -> module_type_body
