@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -12,6 +12,18 @@ open Environ
 open Entries
 open Declarations
 
+
+module NotPrimRecordReason : sig
+
+  type t =
+    | MustNotBeSquashed
+    | MustHaveRelevantProj
+    | MustHaveProj
+    | MustNotHaveAnonProj
+
+end
+
+
 (** Type checking for some inductive entry.
     Returns:
     - environment with inductives + parameters in rel context
@@ -19,30 +31,23 @@ open Declarations
     - checked variance info
       (variance for section universes is at the beginning of the array)
     - record entry (checked to be OK)
+    - if primitive record was requested and not ok, the reason why it's not ok
     - parameters
     - for each inductive,
       (arity * constructors) (with params)
       * (indices * splayed constructor types) (both without params)
       * top allowed elimination
  *)
-val typecheck_inductive : env -> sec_univs:Univ.Level.t array option
+val typecheck_inductive : env -> sec_univs:UVars.Instance.t option
   -> mutual_inductive_entry
   -> env
   * universes
   * template_universes option
-  * Univ.Variance.t array option
+  * UVars.Variance.t array option
   * Names.Id.t array option option
+  * NotPrimRecordReason.t option
   * Constr.rel_context
   * ((inductive_arity * Constr.types array) *
      (Constr.rel_context * (Constr.rel_context * Constr.types) array) *
-     Sorts.family)
+     squash_info option)
     array
-
-(* Utility function to compute the actual universe parameters
-   of a template polymorphic inductive *)
-val template_polymorphic_univs :
-  ctor_levels:Univ.LSet.t ->
-  Univ.ContextSet.t ->
-  Constr.rel_context ->
-  Univ.Universe.t ->
-  (Univ.LSet.t * Univ.Level.t option list) option
