@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -67,25 +67,25 @@ exception IllFormedInd of ill_formed_ind
    declaration, and checks that they are all present (and all the same)
    for all the given types. *)
 
-let mind_extract_params = decompose_prod_n_assum
+let mind_extract_params = decompose_prod_n_decls
 
 let explain_ind_err id ntyp env nparamsctxt c err =
   let open Type_errors in
   let (_lparams,c') = mind_extract_params nparamsctxt c in
   match err with
     | LocalNonPos kt ->
-        raise (InductiveError (NonPos (env,c',mkRel (kt+nparamsctxt))))
+        raise (InductiveError (env, NonPos (c',mkRel (kt+nparamsctxt))))
     | LocalNotEnoughArgs kt ->
         raise (InductiveError
-                 (NotEnoughArgs (env,c',mkRel (kt+nparamsctxt))))
+                 (env, NotEnoughArgs (c',mkRel (kt+nparamsctxt))))
     | LocalNotConstructor (paramsctxt,nargs)->
         let nparams = Context.Rel.nhyps paramsctxt in
         raise (InductiveError
-                 (NotConstructor (env,id,c',mkRel (ntyp+nparamsctxt),
+                 (env, NotConstructor (id,c',mkRel (ntyp+nparamsctxt),
                                   nparams,nargs)))
     | LocalNonPar (n,i,l) ->
         raise (InductiveError
-                 (NonPar (env,c',n,mkRel i,mkRel (l+nparamsctxt))))
+                 (env, NonPar (c',n,mkRel i,mkRel (l+nparamsctxt))))
 
 let failwith_non_pos n ntypes c =
   for k = n to n + ntypes - 1 do
@@ -160,12 +160,12 @@ let ienv_push_inductive (env, n, ntypes, ra_env) ((mi,u),lrecparams) =
   let specif = (lookup_mind_specif env mi, u) in
   let ty = type_of_inductive specif in
   let env' =
-    let r = (snd (fst specif)).mind_relevance in
+    let r = Inductive.relevance_of_ind_body (snd (fst specif)) u in
     let anon = Context.make_annot Anonymous r in
     let decl = LocalAssum (anon, hnf_prod_applist env ty lrecparams) in
     push_rel decl env in
   let ra_env' =
-    (Imbr mi,(Rtree.mk_rec_calls 1).(0)) ::
+    (Mrec (RecArgInd mi),(Rtree.mk_rec_calls 1).(0)) ::
     List.map (fun (r,t) -> (r,Rtree.lift 1 t)) ra_env in
   (* New index of the inductive types *)
   let newidx = n + auxntyp in
@@ -195,14 +195,14 @@ let array_min nmr a = if Int.equal nmr 0 then 0 else
     If [chkpos] is [false] then positivity is assumed, and
     [check_positivity_one] computes the subterms occurrences in a
     best-effort fashion. *)
-let check_positivity_one ~chkpos recursive (env,_,ntypes,_ as ienv) paramsctxt (_,i as ind) nnonrecargs lcnames indlc =
+let check_positivity_one ~chkpos recursive (env,_,ntypes,_ as ienv) paramsctxt (mind,i as ind) nnonrecargs lcnames indlc =
   let nparamsctxt = Context.Rel.length paramsctxt in
   let nmr = Context.Rel.nhyps paramsctxt in
   (** Positivity of one argument [c] of a constructor (i.e. the
       constructor [cn] has a type of the shape [… -> c … -> P], where,
       more generally, the arrows may be dependent). *)
-  let rec check_pos (env, n, ntypes, ra_env as ienv) nmr c =
-    let x,largs = decompose_app (whd_all env c) in
+  let rec check_strict_positivity (env, n, ntypes, ra_env as ienv) nmr c =
+    let x,largs = decompose_app_list (whd_all env c) in
       match kind x with
         | Prod (na,b,d) ->
             let () = assert (List.is_empty largs) in
@@ -215,32 +215,40 @@ let check_positivity_one ~chkpos recursive (env,_,ntypes,_ as ienv) paramsctxt (
               | None when chkpos ->
                   failwith_non_pos_list n ntypes [b]
               | None ->
-                  check_pos (ienv_push_var ienv (na, b, mk_norec)) nmr d
+                  check_strict_positivity (ienv_push_var ienv (na, b, mk_norec)) nmr d
               | Some b ->
-                  check_pos (ienv_push_var ienv (na, b, mk_norec)) nmr d)
+                  check_strict_positivity (ienv_push_var ienv (na, b, mk_norec)) nmr d)
         | Rel k ->
-            (try let (ra,rarg) = List.nth ra_env (k-1) in
+            (match List.nth_opt ra_env (k-1) with
+            | Some (ra,rarg) ->
             let largs = List.map (whd_all env) largs in
             let nmr1 =
               (match ra with
-                  Mrec _ -> compute_rec_par ienv paramsctxt nmr largs
-                |  _ -> nmr)
+                (* Are we referring to the original block of mutual inductive types? *)
+                | Mrec (RecArgInd (mind',_)) ->
+                  if Names.MutInd.CanOrd.equal mind mind'
+                  then compute_rec_par ienv paramsctxt nmr largs
+                  else nmr
+                | Norec | Mrec (RecArgPrim _) -> nmr)
             in
               (** The case where one of the inductives of the mutually
                   inductive block occurs as an argument of another is not
-                  known to be safe. So Coq rejects it. *)
+                  known to be safe. So Rocq rejects it. *)
               if chkpos &&
                  not (List.for_all (noccur_between n ntypes) largs)
               then failwith_non_pos_list n ntypes largs
               else (nmr1,rarg)
-              with Failure _ | Invalid_argument _ -> (nmr,mk_norec))
+            | None -> (nmr,mk_norec))
         | Ind ind_kn ->
             (** If one of the inductives of the mutually inductive
                 block being defined appears in a parameter, then we
                 have a nested inductive type. The positivity is then
-                discharged to the [check_positive_nested] function. *)
+                discharged to the [check_positivity_nested] function. *)
             if List.for_all (noccur_between n ntypes) largs then (nmr,mk_norec)
-            else check_positive_nested ienv nmr (ind_kn, largs)
+            else check_positivity_nested ienv nmr (ind_kn, largs)
+        | Const (c,_) when is_primitive_positive_container env c ->
+          if List.for_all (noccur_between n ntypes) largs then (nmr,mk_norec)
+          else check_positivity_nested_primitive ienv nmr (c, largs)
         | _err ->
             (** If an inductive of the mutually inductive block
                 appears in any other way, then the positivy check gives
@@ -251,15 +259,15 @@ let check_positivity_one ~chkpos recursive (env,_,ntypes,_ as ienv) paramsctxt (
             then (nmr,mk_norec)
             else failwith_non_pos_list n ntypes (x::largs)
 
-  (** [check_positive_nested] handles the case of nested inductive
+  (** [check_positivity_nested] handles the case of nested inductive
       calls, that is, when an inductive types from the mutually
       inductive block is called as an argument of an inductive types
       (for the moment, this inductive type must be a previously
       defined types, not one of the types of the mutually inductive
       block being defined). *)
   (* accesses to the environment are not factorised, but is it worth? *)
-  and check_positive_nested (env,n,ntypes,_ra_env as ienv) nmr ((mi,u), largs) =
-    let (mib,mip) = lookup_mind_specif env mi in
+  and check_positivity_nested (env,n,ntypes,_ra_env as ienv) nmr (((mind,_ as ind),u), largs) =
+    let (mib,mip) = lookup_mind_specif env ind in
     let auxnrecpar = mib.mind_nparams_rec in
     let auxnnonrecpar = mib.mind_nparams - auxnrecpar in
     let (auxrecparams,auxnonrecargs) =
@@ -275,10 +283,10 @@ let check_positivity_one ~chkpos recursive (env,_,ntypes,_ as ienv) paramsctxt (
       let auxntyp = mib.mind_ntypes in
         if not (Int.equal auxntyp 1) then raise (IllFormedInd (LocalNonPos n));
         (* The nested inductive type with parameters removed *)
-        let auxlcvect = abstract_mind_lc auxntyp auxnrecpar mip.mind_nf_lc in
+        let auxlcvect = abstract_mind_lc auxntyp auxnrecpar mind mip.mind_nf_lc in
           (* Extends the environment with a variable corresponding to
              the inductive def *)
-        let (env',_,_,_ as ienv') = ienv_push_inductive ienv ((mi,u),auxrecparams) in
+        let (env',_,_,_ as ienv') = ienv_push_inductive ienv ((ind,u),auxrecparams) in
           (* Parameters expressed in env' *)
         let auxrecparams' = List.map (lift auxntyp) auxrecparams in
         let irecargs_nmr =
@@ -298,26 +306,35 @@ let check_positivity_one ~chkpos recursive (env,_,ntypes,_ as ienv) paramsctxt (
         let irecargs = Array.map snd irecargs_nmr
         and nmr' = array_min nmr irecargs_nmr
         in
-          (nmr',(Rtree.mk_rec [|mk_paths (Imbr mi) irecargs|]).(0))
+          (nmr',(Rtree.mk_rec [|mk_paths (Mrec (RecArgInd ind)) irecargs|]).(0))
+
+  and check_positivity_nested_primitive (env,n,ntypes,ra_env) nmr (c, largs) =
+    (* We model the primitive type c X1 ... Xn as if it had one constructor
+       C : X1 -> ... -> Xn -> c X1 ... Xn
+       The subterm relation is defined for each primitive in `inductive.ml`. *)
+    let ra_env = List.map (fun (r,t) -> (r,Rtree.lift 1 t)) ra_env in
+    let ienv = (env,n,ntypes,ra_env) in
+    let nmr',recargs = List.fold_left_map (check_strict_positivity ienv) nmr largs in
+    (nmr', (Rtree.mk_rec [| mk_paths (Mrec (RecArgPrim c)) [| recargs |] |]).(0))
 
   (** [check_constructors ienv check_head nmr c] checks the positivity
       condition in the type [c] of a constructor (i.e. that recursive
       calls to the inductives of the mutually inductive definition
       appear strictly positively in each of the arguments of the
-      constructor, see also [check_pos]). If [check_head] is [true],
+      constructor, see also [check_strict_positivity]). If [check_head] is [true],
       then the type of the fully applied constructor (the "head" of
       the type [c]) is checked to be the right (properly applied)
       inductive type. *)
   and check_constructors ienv check_head nmr c =
     let rec check_constr_rec (env,n,ntypes,_ra_env as ienv) nmr lrec c =
-      let x,largs = decompose_app (whd_all env c) in
+      let x,largs = decompose_app_list (whd_all env c) in
         match kind x with
 
           | Prod (na,b,d) ->
               let () = assert (List.is_empty largs) in
               if not recursive && not (noccur_between n ntypes b) then
-                raise (InductiveError Type_errors.BadEntry);
-              let nmr',recarg = check_pos ienv nmr b in
+                raise (InductiveError (env,Type_errors.BadEntry));
+              let nmr',recarg = check_strict_positivity ienv nmr b in
               let ienv' = ienv_push_var ienv (na,b,mk_norec) in
                 check_constr_rec ienv' nmr' (recarg::lrec) d
           | hd ->
@@ -348,7 +365,7 @@ let check_positivity_one ~chkpos recursive (env,_,ntypes,_ as ienv) paramsctxt (
   in
   let irecargs = Array.map snd irecargs_nmr
   and nmr' = array_min nmr irecargs_nmr
-  in (nmr', mk_paths (Mrec ind) irecargs)
+  in (nmr', mk_paths (Mrec (RecArgInd ind)) irecargs)
 
 (** [check_positivity ~chkpos kn env_ar paramsctxt inds] checks that the mutually
     inductive block [inds] is strictly positive.
@@ -359,7 +376,8 @@ let check_positivity_one ~chkpos recursive (env,_,ntypes,_ as ienv) paramsctxt (
 let check_positivity ~chkpos kn names env_ar_par paramsctxt finite inds =
   let ntypes = Array.length inds in
   let recursive = finite != BiFinite in
-  let rc = Array.mapi (fun j t -> (Mrec (kn,j),t)) (Rtree.mk_rec_calls ntypes) in
+  if not recursive && Array.length inds <> 1 then raise (InductiveError (env_ar_par,Type_errors.BadEntry));
+  let rc = Array.mapi (fun j t -> (Mrec (RecArgInd (kn,j)),t)) (Rtree.mk_rec_calls ntypes) in
   let ra_env_ar = Array.rev_to_list rc in
   let nparamsctxt = Context.Rel.length paramsctxt in
   let nmr = Context.Rel.nhyps paramsctxt in
@@ -401,26 +419,21 @@ let used_section_variables env params inds =
   keep_hyps env ids
 
 let rel_vect n m = Array.init m (fun i -> mkRel(n+m-i))
-let rel_list n m = Array.to_list (rel_vect n m)
 
 (** From a rel context describing the constructor arguments,
     build an expansion function.
     The term built is expecting to be substituted first by
     a substitution of the form [params, x : ind params] *)
-let compute_projections (kn, i as ind) mib =
+let compute_projections (_, i as ind) mib =
   let pkt = mib.mind_packets.(i) in
-  let u = Univ.make_abstract_instance (Declareops.inductive_polymorphic_context mib) in
-  let subst = List.init mib.mind_ntypes (fun i -> mkIndU ((kn, mib.mind_ntypes - i - 1), u)) in
-  let (ctx, cty) = pkt.mind_nf_lc.(0) in
-  let cty = it_mkProd_or_LetIn cty ctx in
-  let rctx, _ = decompose_prod_assum (substl subst cty) in
-  let ctx, paramslet = List.chop pkt.mind_consnrealdecls.(0) rctx in
+  let (ctx, _) = pkt.mind_nf_lc.(0) in
+  let ctx, paramslet = List.chop pkt.mind_consnrealdecls.(0) ctx in
   (** We build a substitution smashing the lets in the record parameters so
       that typechecking projections requires just a substitution and not
       matching with a parameter context. *)
   let paramsletsubst =
     (* [Ind inst] is typed in context [params-wo-let] *)
-    let inst' = rel_list 0 mib.mind_nparams in
+    let inst' = rel_vect 0 mib.mind_nparams in
     (* {params-wo-let |- subst:params] *)
     let subst = subst_of_rel_context_instance paramslet inst' in
     (* {params-wo-let, x:Ind inst' |- subst':(params,x:Ind inst)] *)
@@ -455,7 +468,7 @@ let compute_projections (kn, i as ind) mib =
         let projty = substl letsubst t in
         (* from [params, x:I, field1,..,fieldj |- t(field1,..,fieldj)]
            to [params, x:I |- t(proj1 x,..,projj x)] *)
-        let fterm = mkProj (Projection.make kn false, mkRel 1) in
+        let fterm = mkProj (Projection.make kn false, r, mkRel 1) in
         (i + 1, j + 1, lab :: labs, r :: rs, projty :: pbs, fterm :: letsubst)
       | Anonymous -> assert false (* checked by indTyping *)
   in
@@ -472,10 +485,16 @@ let build_inductive env ~sec_univs names prv univs template variance
   (* Compute the set of used section variables *)
   let hyps = used_section_variables env paramsctxt inds in
   let nparamargs = Context.Rel.nhyps paramsctxt in
+  let u = UVars.make_abstract_instance (universes_context univs) in
+  let subst = List.init ntypes (fun i -> mkIndU ((kn, ntypes - i - 1), u)) in
   (* Check one inductive *)
-  let build_one_packet (id,cnames) ((arity,lc),(indices,splayed_lc),kelim) recarg =
+  let build_one_packet (id,cnames) ((arity,lc),(indices,splayed_lc),squashed) recarg =
+    let lc = Array.map (substl subst) lc in
     (* Type of constructors in normal form *)
-    let nf_lc = Array.map (fun (d, b) -> (d@paramsctxt, b)) splayed_lc in
+    let nf_lc =
+      Array.map (fun (d, b) ->
+        decompose_prod_decls (substl subst (it_mkProd_or_LetIn b (d@paramsctxt))))
+        splayed_lc in
     let consnrealdecls =
       Array.map (fun (d,_) -> Context.Rel.length d)
         splayed_lc in
@@ -484,7 +503,7 @@ let build_inductive env ~sec_univs names prv univs template variance
         splayed_lc in
     let mind_relevance = match arity with
       | RegularArity { mind_sort;_ } -> Sorts.relevance_of_sort mind_sort
-      | TemplateArity _ -> Sorts.Relevant
+      | TemplateArity a -> Sorts.relevance_of_sort a.template_level
     in
     (* Assigning VM tags to constructors *)
     let nconst, nblock = ref 0, ref 0 in
@@ -505,7 +524,7 @@ let build_inductive env ~sec_univs names prv univs template variance
         mind_arity_ctxt = indices @ paramsctxt;
         mind_nrealargs = Context.Rel.nhyps indices;
         mind_nrealdecls = Context.Rel.length indices;
-        mind_kelim = kelim;
+        mind_squashed = squashed;
         mind_consnames = Array.of_list cnames;
         mind_consnrealdecls = consnrealdecls;
         mind_consnrealargs = consnrealargs;
@@ -523,9 +542,14 @@ let build_inductive env ~sec_univs names prv univs template variance
     | Some variance -> match sec_univs with
       | None -> Some variance, None
       | Some sec_univs ->
-        let nsec = Array.length sec_univs in
-        Some (Array.sub variance nsec (Array.length variance - nsec)),
-        Some (Array.sub variance 0 nsec)
+        (* no variance for qualities *)
+        let _nsecq, nsecu = UVars.Instance.length sec_univs in
+        Some (Array.sub variance nsecu (Array.length variance - nsecu)),
+        Some (Array.sub variance 0 nsecu)
+  in
+  let univ_hyps = match sec_univs with
+    | None -> UVars.Instance.empty
+    | Some univs -> univs
   in
   let mib =
       (* Build the mutual inductive *)
@@ -533,6 +557,7 @@ let build_inductive env ~sec_univs names prv univs template variance
       mind_ntypes = ntypes;
       mind_finite = isfinite;
       mind_hyps = hyps;
+      mind_univ_hyps = univ_hyps;
       mind_nparams = nparamargs;
       mind_nparams_rec = nmr;
       mind_params_ctxt = paramsctxt;
@@ -563,7 +588,7 @@ let build_inductive env ~sec_univs names prv univs template variance
 
 let check_inductive env ~sec_univs kn mie =
   (* First type-check the inductive definition *)
-  let (env_ar_par, univs, template, variance, record, paramsctxt, inds) =
+  let (env_ar_par, univs, template, variance, record, why_not_prim_record, paramsctxt, inds) =
     IndTyping.typecheck_inductive env ~sec_univs mie
   in
   (* Then check positivity conditions *)
@@ -576,6 +601,9 @@ let check_inductive env ~sec_univs kn mie =
       (Array.map (fun ((_,lc),(indices,_),_) -> Context.Rel.nhyps indices,lc) inds)
   in
   (* Build the inductive packets *)
+  let mib =
     build_inductive env ~sec_univs names mie.mind_entry_private univs template variance
       paramsctxt kn record mie.mind_entry_finite
       inds nmr recargs
+  in
+  mib, why_not_prim_record
