@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -11,8 +11,10 @@
 open Names
 open Constr
 open Univ
+open UVars
 open Declarations
 open Environ
+open CClosure
 
 (** {6 Extracting an inductive type from a construction } *)
 
@@ -22,11 +24,9 @@ open Environ
    only a coinductive type.
    They raise [Not_found] if not convertible to a recursive type. *)
 
-val find_rectype     : env -> types -> pinductive * constr list
-val find_inductive   : env -> types -> pinductive * constr list
-val find_coinductive : env -> types -> pinductive * constr list
-
-type mind_specif = mutual_inductive_body * one_inductive_body
+val find_rectype     : ?evars:evar_handler -> env -> types -> pinductive * constr list
+val find_inductive   : ?evars:evar_handler -> env -> types -> pinductive * constr list
+val find_coinductive : ?evars:evar_handler -> env -> types -> pinductive * constr list
 
 (** {6 ... } *)
 (** Fetching information in the environment about an inductive type.
@@ -46,24 +46,39 @@ val inductive_nonrec_rec_paramdecls : mutual_inductive_body puniverses -> Constr
 val instantiate_inductive_constraints :
   mutual_inductive_body -> Instance.t -> Constraints.t
 
-type param_univs = (unit -> Sorts.t) list
+type template_univ =
+  | TemplateProp
+  | TemplateUniv of Universe.t
 
-val make_param_univs : Environ.env -> constr array -> param_univs
-(** The constr array is the types of the arguments to a template
-    polymorphic inductive. *)
+type param_univs = (expected:Univ.Level.t -> template_univ) list
+
+val instantiate_template_constraints
+  : template_univ Univ.Level.Map.t
+  -> Declarations.template_universes
+  -> Univ.Constraints.t
+
+val instantiate_template_universes : mind_specif -> param_univs ->
+  Constraints.t * rel_context * template_univ Univ.Level.Map.t
 
 val constrained_type_of_inductive : mind_specif puniverses -> types constrained
-val constrained_type_of_inductive_knowing_parameters :
-  mind_specif puniverses -> param_univs -> types constrained
 
-val relevance_of_inductive : env -> inductive -> Sorts.relevance
+val relevance_of_ind_body : one_inductive_body -> UVars.Instance.t -> Sorts.relevance
+
+val relevance_of_inductive : env -> pinductive -> Sorts.relevance
 
 val type_of_inductive : mind_specif puniverses -> types
 
 val type_of_inductive_knowing_parameters :
-  ?polyprop:bool -> mind_specif puniverses -> param_univs -> types
+  mind_specif puniverses -> param_univs -> types constrained
 
-val elim_sort : mind_specif -> Sorts.family
+val quality_leq : Sorts.Quality.t -> Sorts.Quality.t -> bool
+(** For squashing. *)
+
+type squash = SquashToSet | SquashToQuality of Sorts.Quality.t
+
+val is_squashed : mind_specif puniverses -> squash option
+
+val is_allowed_elimination : mind_specif puniverses -> Sorts.t -> bool
 
 val is_private : mind_specif -> bool
 val is_primitive_record : mind_specif -> bool
@@ -72,6 +87,7 @@ val is_primitive_record : mind_specif -> bool
 
 val constrained_type_of_constructor : pconstructor -> mind_specif -> types constrained
 val type_of_constructor : pconstructor -> mind_specif -> types
+val type_of_constructor_knowing_parameters : pconstructor -> mind_specif -> param_univs -> types constrained
 
 (** Return constructor types in normal form *)
 val arities_of_constructors : pinductive -> mind_specif -> types array
@@ -89,35 +105,49 @@ val abstract_constructor_type_relatively_to_inductive_types_context :
 
 val inductive_params : mind_specif -> int
 
+(** Given an inductive type and its parameters, builds the context of the return
+    clause, including the inductive being eliminated. The additional binder
+    array is only used to set the names of the context variables, we use the
+    less general type to make it easy to use this function on Case nodes. *)
+val expand_arity : mind_specif -> pinductive -> constr array ->
+  Name.t binder_annot array -> rel_context
+
+(** Given an inductive type and its parameters, builds the context of the return
+    clause, including the inductive being eliminated. The additional binder
+    array is only used to set the names of the context variables, we use the
+    less general type to make it easy to use this function on Case nodes. *)
+val expand_branch_contexts : mind_specif -> UVars.Instance.t -> constr array ->
+  (Name.t binder_annot array * 'a) array -> rel_context array
+
+type ('constr,'types,'r) pexpanded_case =
+  (case_info * ('constr * 'r) * 'constr pcase_invert * 'constr * 'constr array)
+
+type expanded_case = (constr,types,Sorts.relevance) pexpanded_case
+
 (** Given a pattern-matching represented compactly, expands it so as to produce
     lambda and let abstractions in front of the return clause and the pattern
     branches. *)
-val expand_case : env -> case -> (case_info * constr * case_invert * constr * constr array)
+val expand_case : env -> case -> expanded_case
 
-val expand_case_specif : mutual_inductive_body -> case -> (case_info * constr * case_invert * constr * constr array)
+val expand_case_specif : mutual_inductive_body -> case -> expanded_case
 
 (** Dual operation of the above. Fails if the return clause or branch has not
     the expected form. *)
-val contract_case : env -> (case_info * constr * case_invert * constr * constr array) -> case
+val contract_case : env -> expanded_case -> case
 
 (** [instantiate_context u subst nas ctx] applies both [u] and [subst]
     to [ctx] while replacing names using [nas] (order reversed). In particular,
     assumes that [ctx] and [nas] have the same length. *)
-val instantiate_context : Instance.t -> Vars.substl -> Name.t Context.binder_annot array ->
+val instantiate_context : Instance.t -> Vars.substl -> Name.t binder_annot array ->
   rel_context -> rel_context
 
 val build_branches_type :
   pinductive -> mutual_inductive_body * one_inductive_body ->
     constr list -> constr -> types array
 
-(** Return the arity of an inductive type *)
-val mind_arity : one_inductive_body -> Constr.rel_context * Sorts.family
-
-val inductive_sort_family : one_inductive_body -> Sorts.family
-
 (** Check a [case_info] actually correspond to a Case expression on the
    given inductive type. *)
-val check_case_info : env -> pinductive -> Sorts.relevance -> case_info -> unit
+val check_case_info : env -> pinductive -> case_info -> unit
 
 (** {6 Guard conditions for fix and cofix-points. } *)
 
@@ -130,18 +160,7 @@ val is_primitive_positive_container : env -> Constant.t -> bool
 
 (** When [chk] is false, the guard condition is not actually
     checked. *)
-val check_fix : env -> fixpoint -> unit
-val check_cofix : env -> cofixpoint -> unit
-
-(** {6 Support for sort-polymorphic inductive types } *)
-
-(** The "polyprop" optional argument below controls
-    the "Prop-polymorphism". By default, it is allowed.
-    But when "polyprop=false", the following exception is raised
-    when a polymorphic singleton inductive type becomes Prop due to
-    parameter instantiation. This is used by the Ocaml extraction,
-    which cannot handle (yet?) Prop-polymorphism. *)
-
-exception SingletonInductiveBecomesProp of Id.t
+val check_fix : ?evars:evar_handler -> env -> fixpoint -> unit
+val check_cofix : ?evars:evar_handler -> env -> cofixpoint -> unit
 
 val abstract_mind_lc : int -> int -> MutInd.t -> (rel_context * constr) array -> constr array
