@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -22,9 +22,6 @@ open Environ
 open Modops
 open Mod_subst
 
-type 'alg translation =
-  module_signature * 'alg * delta_resolver * Univ.Constraints.t
-
 let rec mp_from_mexpr = function
   | MEident mp -> mp
   | MEapply (expr,_) -> mp_from_mexpr expr
@@ -32,7 +29,7 @@ let rec mp_from_mexpr = function
 
 let is_modular = function
   | SFBmodule _ | SFBmodtype _ -> true
-  | SFBconst _ | SFBmind _ -> false
+  | SFBconst _ | SFBmind _ | SFBrules _ -> false
 
 (** Split a [structure_body] at some label corresponding to
     a modular definition or not. *)
@@ -55,12 +52,18 @@ let rec rebuild_mp mp l =
   | i::r -> rebuild_mp (MPdot(mp,Label.of_id i)) r
 
 let infer_gen_conv state env c1 c2 =
-  Reduction.generic_conv Reduction.CONV ~l2r:false Constr.default_evar_handler TransparentState.full env state c1 c2
+  Conversion.generic_conv Conversion.CONV ~l2r:false TransparentState.full env state c1 c2
 
 let infer_gen_conv_leq state env c1 c2 =
-  Reduction.generic_conv Reduction.CUMUL ~l2r:false Constr.default_evar_handler TransparentState.full env state c1 c2
+  Conversion.generic_conv Conversion.CUMUL ~l2r:false TransparentState.full env state c1 c2
 
-let rec check_with_def (cst, ustate) env struc (idl,(c,ctx)) mp reso =
+type with_body = {
+  w_def : Constr.t;
+  w_univs : universes;
+  w_bytecode : Vmlibrary.indirect_code option;
+}
+
+let rec check_with_def (cst, ustate) env struc (idl, wth) mp reso =
   let lab,idl = match idl with
     | [] -> assert false
     | id::idl -> Label.of_id id, idl
@@ -78,57 +81,56 @@ let rec check_with_def (cst, ustate) env struc (idl,(c,ctx)) mp reso =
       (* In the spirit of subtyping.check_constant, we accept
          any implementations of parameters and opaque terms,
          as long as they have the right type *)
-      let univs, ctx' =
-        match cb.const_universes, ctx with
-        | Monomorphic, None ->
+      let ctx' =
+        match cb.const_universes, wth.w_univs with
+        | Monomorphic, Monomorphic ->
           let cst = match cb.const_body with
             | Undef _ | OpaqueDef _ ->
-              let j = Typeops.infer env' c in
-              assert (j.uj_val == c); (* relevances should already be correct here *)
+              let j = Typeops.infer env' wth.w_def in
               let typ = cb.const_type in
               let cst = infer_gen_conv_leq (cst, ustate) env' j.uj_type typ in
               cst
             | Def c' ->
-              infer_gen_conv (cst, ustate) env' c c'
-            | Primitive _ ->
+              infer_gen_conv (cst, ustate) env' wth.w_def c'
+            | Primitive _ | Symbol _ ->
               error_incorrect_with_constraint lab
           in
-          Monomorphic, cst
-        | Polymorphic uctx, Some ctx ->
+          begin match cst with
+          | Result.Ok cst -> cst
+          | Result.Error (None | Some _) ->
+            error_incorrect_with_constraint lab
+          end
+        | Polymorphic uctx, Polymorphic ctx ->
           let () =
             if not (UGraph.check_subtype (Environ.universes env) uctx ctx) then
               error_incorrect_with_constraint lab
           in
           (** Terms are compared in a context with De Bruijn universe indices *)
-          let env' = Environ.push_context ~strict:false (Univ.AbstractContext.repr uctx) env in
+          let env' = Environ.push_context ~strict:false (UVars.AbstractContext.repr uctx) env in
           let () = match cb.const_body with
             | Undef _ | OpaqueDef _ ->
-              let j = Typeops.infer env' c in
-              assert (j.uj_val == c); (* relevances should already be correct here *)
+              let j = Typeops.infer env' wth.w_def in
               let typ = cb.const_type in
-              begin
-                try Reduction.conv_leq env' j.uj_type typ
-                with Reduction.NotConvertible -> error_incorrect_with_constraint lab
+              begin match Conversion.conv_leq env' j.uj_type typ with
+              | Result.Ok () -> ()
+              | Result.Error () -> error_incorrect_with_constraint lab
               end
             | Def c' ->
-              begin
-                try Reduction.conv env' c c'
-                with Reduction.NotConvertible -> error_incorrect_with_constraint lab
+              begin match Conversion.conv env' wth.w_def c' with
+              | Result.Ok () -> ()
+              | Result.Error () -> error_incorrect_with_constraint lab
               end
-            | Primitive _ ->
+            | Primitive _ | Symbol _ ->
               error_incorrect_with_constraint lab
           in
-          Polymorphic ctx, cst
+          cst
         | _ -> error_incorrect_with_constraint lab
       in
-      let def = Def c in
-      (*      let ctx' = Univ.UContext.make (newus, cst) in *)
       let cb' =
         { cb with
-          const_body = def;
-          const_universes = univs ;
-          const_body_code =
-              (Vmbytegen.compile_constant_body ~fail_on_error:false env' cb.const_universes def) }
+          const_body = Def wth.w_def;
+          const_universes = wth.w_univs;
+          const_body_code = wth.w_bytecode; }
       in
       before@(lab,SFBconst(cb'))::after, ctx'
     else
@@ -137,11 +139,11 @@ let rec check_with_def (cst, ustate) env struc (idl,(c,ctx)) mp reso =
         | SFBmodule mb -> mb
         | _ -> error_not_a_module_label lab
       in
-      begin match mb.mod_expr with
+      begin match Declareops.mod_expr mb with
         | Abstract ->
           let struc = Modops.destr_nofunctor (MPdot (mp,lab)) mb.mod_type in
           let struc', cst =
-            check_with_def (cst, ustate) env' struc (idl,(c,ctx)) (MPdot(mp,lab)) mb.mod_delta
+            check_with_def (cst, ustate) env' struc (idl, wth) (MPdot(mp,lab)) mb.mod_delta
           in
           let mb' = { mb with
                       mod_type = NoFunctor struc';
@@ -152,7 +154,6 @@ let rec check_with_def (cst, ustate) env struc (idl,(c,ctx)) mp reso =
       end
   with
   | Not_found -> error_no_such_label lab mp
-  | Reduction.NotConvertible -> error_incorrect_with_constraint lab
 
 let rec check_with_mod (cst, ustate) env struc (idl,new_mp) mp reso =
   let lab,idl = match idl with
@@ -170,12 +171,12 @@ let rec check_with_mod (cst, ustate) env struc (idl,new_mp) mp reso =
       (* Toplevel module definition *)
       let new_mb = lookup_module new_mp env in
       let new_mtb = module_type_of_module new_mb in
-      let cst = match old.mod_expr with
+      let cst = match Declareops.mod_expr old with
         | Abstract ->
           let mtb_old = module_type_of_module old in
           let cst = Subtyping.check_subtypes (cst, ustate) env' new_mtb mtb_old in
           cst
-        | Algebraic (NoFunctor (MEident(mp'))) ->
+        | Algebraic (MENoFunctor (MEident(mp'))) ->
           check_modpath_equiv env' new_mp mp';
           cst
         | _ -> error_generative_module_expected lab
@@ -185,7 +186,7 @@ let rec check_with_mod (cst, ustate) env struc (idl,new_mp) mp reso =
       let new_mb' =
         { new_mb with
           mod_mp = mp';
-          mod_expr = Algebraic (NoFunctor (MEident new_mp));
+          mod_expr = ModBodyVal (Algebraic (MENoFunctor (MEident new_mp)));
         }
       in
       let new_reso = add_delta_resolver reso new_mb.mod_delta in
@@ -201,7 +202,7 @@ let rec check_with_mod (cst, ustate) env struc (idl,new_mp) mp reso =
         | SFBmodule msb -> msb
         | _ -> error_not_a_module_label lab
       in
-      begin match old.mod_expr with
+      begin match Declareops.mod_expr old with
       | Abstract ->
         let struc = destr_nofunctor mp' old.mod_type in
         let struc',reso',cst =
@@ -217,7 +218,7 @@ let rec check_with_mod (cst, ustate) env struc (idl,new_mp) mp reso =
         let id_subst = map_mp mp' mp' reso' in
         let new_after = subst_structure id_subst after in
         before@(lab,SFBmodule new_mb)::new_after, new_reso, cst
-      | Algebraic (NoFunctor (MEident mp0)) ->
+      | Algebraic (MENoFunctor (MEident mp0)) ->
         let mpnew = rebuild_mp mp0 idl in
         check_modpath_equiv env' mpnew mp;
         before@(lab,spec)::after, reso, cst
@@ -225,23 +226,28 @@ let rec check_with_mod (cst, ustate) env struc (idl,new_mp) mp reso =
       end
   with
   | Not_found -> error_no_such_label lab mp
-  | Reduction.NotConvertible -> error_incorrect_with_constraint lab
 
-let check_with ustate env mp (sign,reso,cst) = function
+type 'a vm_handler = { vm_handler : env -> universes -> Constr.t -> 'a -> 'a * Vmlibrary.indirect_code option }
+type 'a vm_state = 'a * 'a vm_handler
+
+let check_with ustate vmstate env mp (sign,reso,cst,vm) = function
   | WithDef(idl, (c, ctx)) ->
     let struc = destr_nofunctor mp sign in
-    let struc', cst = check_with_def (cst, ustate) env struc (idl, (c, ctx)) mp reso in
-    NoFunctor struc', reso, cst
+    let univs = match ctx with None -> Monomorphic | Some uctx -> Polymorphic uctx in
+    let vm, bcode = vmstate.vm_handler env univs c vm in
+    let body = { w_def = c; w_univs = univs; w_bytecode = bcode } in
+    let struc', cst = check_with_def (cst, ustate) env struc (idl, body) mp reso in
+    NoFunctor struc', reso, cst, vm
   | WithMod(idl,new_mp) ->
     let struc = destr_nofunctor mp sign in
     let struc',reso',cst = check_with_mod (cst, ustate) env struc (idl,new_mp) mp reso in
-    NoFunctor struc', reso', cst
+    NoFunctor struc', reso', cst, vm
 
-let check_with_alg ustate env mp (sign,alg,reso,cst) wd =
-  let struc, reso, cst = check_with ustate env mp (sign, reso, cst) wd in
-  struc, MEwith (alg, wd), reso, cst
+let check_with_alg ustate vmstate env mp (sign, alg, reso, cst, vm) wd =
+  let struc, reso, cst, vm = check_with ustate vmstate env mp (sign, reso, cst, vm) wd in
+  struc, MEwith (alg, wd), reso, cst, vm
 
-let translate_apply ustate env inl (sign,alg,reso,cst) mp1 mkalg =
+let translate_apply ustate env inl (sign,alg,reso,cst,vm) mp1 mkalg =
   let farg_id, farg_b, fbody_b = destr_functor sign in
   let mtb = module_type_of_module (lookup_module mp1 env) in
   let cst = Subtyping.check_subtypes (cst, ustate) env mtb farg_b in
@@ -251,7 +257,7 @@ let translate_apply ustate env inl (sign,alg,reso,cst) mp1 mkalg =
   let body = subst_signature subst fbody_b in
   let alg' = mkalg alg mp1 in
   let reso' = subst_codom_delta_resolver subst reso in
-  body, alg', reso', cst
+  body, alg', reso', cst, vm
 
 (** Translation of a module struct entry :
     - We translate to a module when a [module_path] is given,
@@ -262,7 +268,7 @@ let translate_apply ustate env inl (sign,alg,reso,cst) mp1 mkalg =
 
 let mk_alg_app alg arg = MEapply (alg,arg)
 
-let rec translate_mse (cst, ustate) env mpo inl = function
+let rec translate_mse (cst, ustate) (vm, vmstate) env mpo inl = function
   | MEident mp1 as me ->
     let mb = match mpo with
       | Some mp -> strengthen_and_subst_module_body (lookup_module mp1 env) mp false
@@ -270,75 +276,91 @@ let rec translate_mse (cst, ustate) env mpo inl = function
         let mt = lookup_modtype mp1 env in
         module_body_of_type mt.mod_mp mt
     in
-    mb.mod_type, me, mb.mod_delta, cst
+    mb.mod_type, me, mb.mod_delta, cst, vm
   | MEapply (fe,mp1) ->
-    translate_apply ustate env inl (translate_mse (cst, ustate) env mpo inl fe) mp1 mk_alg_app
-  |MEwith(me, with_decl) ->
+    translate_apply ustate env inl (translate_mse (cst, ustate) (vm, vmstate) env mpo inl fe) mp1 mk_alg_app
+  | MEwith(me, with_decl) ->
     assert (Option.is_empty mpo); (* No 'with' syntax for modules *)
     let mp = mp_from_mexpr me in
-    check_with_alg ustate env mp (translate_mse (cst, ustate) env None inl me) with_decl
+    check_with_alg ustate vmstate env mp (translate_mse (cst, ustate) (vm, vmstate) env None inl me) with_decl
 
 let mk_mod mp e ty reso =
   { mod_mp = mp;
-    mod_expr = e;
+    mod_expr = ModBodyVal e;
     mod_type = ty;
     mod_type_alg = None;
     mod_delta = reso;
-    mod_retroknowledge = ModBodyRK []; }
+    mod_retroknowledge = ModBodyVal []; }
 
 let mk_modtype mp ty reso =
-  let mb = mk_mod mp Abstract ty reso in
-  { mb with mod_expr = (); mod_retroknowledge = ModTypeRK }
+  { mod_mp = mp;
+    mod_expr = ModTypeNul;
+    mod_type = ty;
+    mod_type_alg = None;
+    mod_delta = reso;
+    mod_retroknowledge = ModTypeNul; }
 
-let rec translate_mse_funct (cst, ustate) env ~is_mod mp inl mse = function
-  |[] ->
-    let sign,alg,reso,cst = translate_mse (cst, ustate) env (if is_mod then Some mp else None) inl mse in
+let rec translate_mse_funct (cst, ustate) (vm, vmstate) env ~is_mod mp inl mse = function
+  | [] ->
+    let sign,alg,reso,cst,vm = translate_mse (cst, ustate) (vm, vmstate) env (if is_mod then Some mp else None) inl mse in
     let sign,reso =
       if is_mod then sign,reso
       else subst_modtype_signature_and_resolver (mp_from_mexpr mse) mp sign reso in
-    sign, NoFunctor alg, reso, cst
-  |(mbid, ty, ty_inl) :: params ->
+    sign, MENoFunctor alg, reso, cst, vm
+  | (mbid, ty, ty_inl) :: params ->
     let mp_id = MPbound mbid in
-    let mtb, cst = translate_modtype (cst, ustate) env mp_id ty_inl ([],ty) in
+    let mtb, cst, vm = translate_modtype (cst, ustate) (vm, vmstate) env mp_id ty_inl ([],ty) in
     let env' = add_module_type mp_id mtb env in
-    let sign,alg,reso,cst = translate_mse_funct (cst, ustate) env' ~is_mod mp inl mse params in
-    let alg' = MoreFunctor (mbid,mtb,alg) in
-    MoreFunctor (mbid, mtb, sign), alg',reso, cst
+    let sign,alg,reso,cst,vm = translate_mse_funct (cst, ustate) (vm, vmstate) env' ~is_mod mp inl mse params in
+    let alg' = MEMoreFunctor alg in
+    MoreFunctor (mbid, mtb, sign), alg',reso, cst, vm
 
-and translate_modtype state env mp inl (params,mte) =
-  let sign,alg,reso,cst = translate_mse_funct state env ~is_mod:false mp inl mte params in
+and translate_modtype state vmstate env mp inl (params,mte) =
+  let sign,alg,reso,cst,vm = translate_mse_funct state vmstate env ~is_mod:false mp inl mte params in
   let mtb = mk_modtype mp sign reso in
-  { mtb with mod_type_alg = Some alg }, cst
+  { mtb with mod_type_alg = Some alg }, cst, vm
 
 (** [finalize_module] :
     from an already-translated (or interactive) implementation and
     an (optional) signature entry, produces a final [module_body] *)
 
-let finalize_module (cst, ustate) env mp (sign,alg,reso) restype = match restype with
+let finalize_module_alg (cst, ustate) (vm, vmstate) env mp (sign,alg,reso) restype = match restype with
   | None ->
     let impl = match alg with Some e -> Algebraic e | None -> FullStruct in
-    mk_mod mp impl sign reso, cst
+    mk_mod mp impl sign reso, cst, vm
   | Some (params_mte,inl) ->
-    let res_mtb, cst = translate_modtype (cst, ustate) env mp inl params_mte in
+    let res_mtb, cst, vm = translate_modtype (cst, ustate) (vm, vmstate) env mp inl params_mte in
     let auto_mtb = mk_modtype mp sign reso in
     let cst = Subtyping.check_subtypes (cst, ustate) env auto_mtb res_mtb in
-    let impl = match alg with Some e -> Algebraic e | None -> Struct sign in
+    let impl = match alg with
+    | Some e -> Algebraic e
+    | None ->
+      let sign = match sign with
+      | NoFunctor s -> s
+      | MoreFunctor _ -> assert false (* All non-algebraic callers enforce this *)
+      in
+      Struct sign
+    in
     { res_mtb with
       mod_mp = mp;
-      mod_expr = impl;
-      mod_retroknowledge = ModBodyRK [];
+      mod_expr = ModBodyVal impl;
+      mod_retroknowledge = ModBodyVal [];
     },
     (** constraints from module body typing + subtyping + module type. *)
-    cst
+    cst,
+    vm
 
-let translate_module (cst, ustate) env mp inl = function
+let finalize_module univs vm env mp (sign, reso) typ =
+  finalize_module_alg univs vm env mp (sign, None, reso) typ
+
+let translate_module (cst, ustate) (vm, vmstate) env mp inl = function
   | MType (params,ty) ->
-    let mtb, cst = translate_modtype (cst, ustate) env mp inl (params,ty) in
-    module_body_of_type mp mtb, cst
-  |MExpr (params,mse,oty) ->
-    let (sg,alg,reso,cst) = translate_mse_funct (cst, ustate) env ~is_mod:true mp inl mse params in
+    let mtb, cst, vm = translate_modtype (cst, ustate) (vm, vmstate) env mp inl (params,ty) in
+    module_body_of_type mp mtb, cst, vm
+  | MExpr (params,mse,oty) ->
+    let (sg,alg,reso,cst,vm) = translate_mse_funct (cst, ustate) (vm, vmstate) env ~is_mod:true mp inl mse params in
     let restype = Option.map (fun ty -> ((params,ty),inl)) oty in
-    finalize_module (cst, ustate) env mp (sg,Some alg,reso) restype
+    finalize_module_alg (cst, ustate) (vm, vmstate) env mp (sg,Some alg,reso) restype
 
 (** We now forbid any Include of functors with restricted signatures.
     Otherwise, we could end with the creation of undesired axioms
@@ -346,38 +368,38 @@ let translate_module (cst, ustate) env mp inl = function
     thanks to strengthening. *)
 
 let rec unfunct = function
-  | NoFunctor me -> me
-  | MoreFunctor(_,_,me) -> unfunct me
+  | MENoFunctor me -> me
+  | MEMoreFunctor me -> unfunct me
 
 let rec forbid_incl_signed_functor env = function
   | MEapply(fe,_) -> forbid_incl_signed_functor env fe
   | MEwith _ -> assert false (* No 'with' syntax for modules *)
   | MEident mp1 ->
     let mb = lookup_module mp1 env in
-    match mb.mod_type, mb.mod_type_alg, mb.mod_expr with
+    match mb.mod_type, mb.mod_type_alg, Declareops.mod_expr mb with
     | MoreFunctor _, Some _, _ ->
       (* functor + restricted signature = error *)
       error_include_restricted_functor mp1
     | MoreFunctor _, None, Algebraic me ->
       (* functor, no signature yet, a definition which may be restricted *)
       forbid_incl_signed_functor env (unfunct me)
-    |_ -> ()
+    | _ -> ()
 
-let rec translate_mse_include_module (cst, ustate) env mp inl = function
+let rec translate_mse_include_module (cst, ustate) (vm, vmstate) env mp inl = function
   | MEident mp1 ->
     let mb = strengthen_and_subst_module_body (lookup_module mp1 env) mp true in
     let sign = clean_bounded_mod_expr mb.mod_type in
-    sign,(),mb.mod_delta,cst
+    sign,(),mb.mod_delta,cst,vm
   | MEapply (fe,arg) ->
-    let ftrans = translate_mse_include_module (cst, ustate) env mp inl fe in
+    let ftrans = translate_mse_include_module (cst, ustate) (vm, vmstate) env mp inl fe in
     translate_apply ustate env inl ftrans arg (fun _ _ -> ())
   | MEwith _ -> assert false (* No 'with' syntax for modules *)
 
-let translate_mse_include is_mod (cst, ustate) env mp inl me =
+let translate_mse_include is_mod (cst, ustate) (vm, vmstate) env mp inl me =
   if is_mod then
     let () = forbid_incl_signed_functor env me in
-    translate_mse_include_module (cst, ustate) env mp inl me
+    translate_mse_include_module (cst, ustate) (vm, vmstate) env mp inl me
   else
-    let mtb, cst = translate_modtype (cst, ustate) env mp inl ([],me) in
+    let mtb, cst, vm = translate_modtype (cst, ustate) (vm, vmstate) env mp inl ([],me) in
     let sign = clean_bounded_mod_expr mtb.mod_type in
-    sign, (), mtb.mod_delta, cst
+    sign, (), mtb.mod_delta, cst, vm
