@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -148,7 +148,7 @@ module ModIdmap = Id.Map
 
 (** Dirpaths are lists of module identifiers.
     The actual representation is reversed to optimise sharing:
-    Coq.A.B is ["B";"A";"Coq"] *)
+    Corelib.A.B is ["B";"A";"Corelib"] *)
 
 let dummy_module_name = "If you see this, it's a bug"
 
@@ -454,6 +454,7 @@ sig
   module CanOrd : EqType with type t = t
   module UserOrd : EqType with type t = t
   module SyntacticOrd : EqType with type t = t
+  val canonize : t -> t
 end
 
 (** For constant and inductive names, we use a kernel name couple (kn1,kn2)
@@ -490,6 +491,10 @@ module KerPair = struct
     | Same kn -> kn
     | Dual (kn,_) -> kn
 
+  let canonize kp = match kp with
+  | Same _ -> kp
+  | Dual (_, kn) -> Same kn
+
   let same kn = Same kn
   let make knu knc = if KerName.equal knu knc then Same knc else Dual (knu,knc)
 
@@ -656,6 +661,10 @@ struct
       Hashset.Combine.combine (MutInd.SyntacticOrd.hash m) (Int.hash i)
   end
 
+  let canonize ((mind, i) as ind) =
+    let mind' = MutInd.canonize mind in
+    if mind' == mind then ind else (mind', i)
+
 end
 
 module Construct =
@@ -702,6 +711,10 @@ struct
       Hashset.Combine.combine (Ind.SyntacticOrd.hash ind) (Int.hash i)
   end
 
+  let canonize ((ind, i) as cstr) =
+    let ind' = Ind.canonize ind in
+    if ind' == ind then cstr else (ind', i)
+
 end
 
 (** Designation of a (particular) inductive type. *)
@@ -876,6 +889,13 @@ struct
     let equal = CanOrd.equal
     let compare = CanOrd.compare
 
+    let canonize p =
+      let { proj_ind; proj_npars; proj_arg; proj_name } = p in
+      let proj_ind' = Ind.canonize proj_ind in
+      let proj_name' = Constant.canonize proj_name in
+      if proj_ind' == proj_ind && proj_name' == proj_name then p
+      else { proj_ind = proj_ind'; proj_name = proj_name'; proj_npars; proj_arg }
+
     module Self_Hashcons = struct
       type nonrec t = t
       type u = (inductive -> inductive) * (Constant.t -> Constant.t)
@@ -931,24 +951,27 @@ struct
 
   module SyntacticOrd = struct
     type nonrec t = t
-    let compare (c, b) (c', b') =
-      if b = b' then Repr.SyntacticOrd.compare c c' else -1
+    let compare (p, b) (p', b') =
+      let c = Bool.compare b b' in
+      if c <> 0 then c else Repr.SyntacticOrd.compare p p'
     let equal (c, b as x) (c', b' as x') =
       x == x' || b = b' && Repr.SyntacticOrd.equal c c'
     let hash (c, b) = (if b then 0 else 1) + Repr.SyntacticOrd.hash c
   end
   module CanOrd = struct
     type nonrec t = t
-    let compare (c, b) (c', b') =
-      if b = b' then Repr.CanOrd.compare c c' else -1
+    let compare (p, b) (p', b') =
+      let c = Bool.compare b b' in
+      if c <> 0 then c else Repr.CanOrd.compare p p'
     let equal (c, b as x) (c', b' as x') =
       x == x' || b = b' && Repr.CanOrd.equal c c'
     let hash (c, b) = (if b then 0 else 1) + Repr.CanOrd.hash c
   end
   module UserOrd = struct
     type nonrec t = t
-    let compare (c, b) (c', b') =
-      if b = b' then Repr.UserOrd.compare c c' else -1
+    let compare (p, b) (p', b') =
+      let c = Bool.compare b b' in
+      if c <> 0 then c else Repr.UserOrd.compare p p'
     let equal (c, b as x) (c', b' as x') =
       x == x' || b = b' && Repr.UserOrd.equal c c'
     let hash (c, b) = (if b then 0 else 1) + Repr.UserOrd.hash c
@@ -964,13 +987,17 @@ struct
       let hash = hash
     end
 
+  let canonize ((r, u) as p) =
+    let r' = Repr.canonize r in
+    if r' == r then p else (r',  u)
+
   module HashProjection = Hashcons.Make(Self_Hashcons)
 
   let hcons = Hashcons.simple_hcons HashProjection.generate HashProjection.hcons Repr.hcons
 
-  let compare (c, b) (c', b') =
-    if b == b' then Repr.compare c c'
-    else if b then 1 else -1
+  let compare (p, b) (p', b') =
+    let c = Bool.compare b b' in
+    if c <> 0 then c else Repr.compare p p'
 
   let map f (c, b as x) =
     let c' = Repr.map f c in
@@ -983,8 +1010,18 @@ struct
   let to_string p = Constant.to_string (constant p)
   let print p = Constant.print (constant p)
 
+  let debug_to_string p =
+    (if unfolded p then "unfolded(" else "") ^
+    Constant.debug_to_string (constant p) ^
+    (if unfolded p then ")" else "")
+  let debug_print p = str (debug_to_string p)
+
 end
 
+module PRmap = HMap.Make(Projection.Repr.CanOrd)
+module PRset = PRmap.Set
+module PRpred = Predicate.Make(Projection.Repr.CanOrd)
+
 module GlobRefInternal = struct
 
   type t =
@@ -1070,6 +1107,18 @@ module GlobRef = struct
     let hash gr = GlobRefInternal.global_hash_gen Constant.SyntacticOrd.hash Ind.SyntacticOrd.hash Construct.SyntacticOrd.hash gr
   end
 
+  let canonize gr = match gr with
+  | VarRef _ -> gr
+  | ConstRef c ->
+    let c' = Constant.canonize c in
+    if c' == c then gr else ConstRef c'
+  | IndRef ind ->
+    let ind' = Ind.canonize ind in
+    if ind' == ind then gr else IndRef ind'
+  | ConstructRef c ->
+    let c' = Construct.canonize c in
+    if c' == c then gr else ConstructRef c'
+
   let is_bound = function
   | VarRef _ -> false
   | ConstRef cst -> ModPath.is_bound (Constant.modpath cst)
