@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -19,7 +19,7 @@
     - DirPath.t represents generic paths as sequences of identifiers.
     - Label.t is an equivalent of Id.t made distinct for semantical purposes.
     - ModPath.t are module paths.
-    - KerName.t are absolute names of objects in Coq.
+    - KerName.t are absolute names of objects in Rocq.
 *)
 
 open Util
@@ -30,7 +30,7 @@ open Util
 module Id :
 sig
   type t
-  (** Values of this type represent (Coq) identifiers. *)
+  (** Values of this type represent (Rocq) identifiers. *)
 
   val equal : t -> t -> bool
   (** Equality over identifiers. *)
@@ -44,6 +44,9 @@ sig
   val is_valid : string -> bool
   (** Check that a string may be converted to an identifier. *)
 
+  val is_valid_ident_part : string -> bool
+  (** Check that a string is a valid part of an identifier *)
+
   val of_bytes : bytes -> t
   val of_string : string -> t
   (** Converts a string into an identifier.
@@ -59,7 +62,7 @@ sig
   val print : t -> Pp.t
   (** Pretty-printer. *)
 
-  module Set : Set.S with type elt = t
+  module Set : Set.ExtS with type elt = t
   (** Finite sets of identifiers. *)
 
   module Map : Map.ExtS with type key = t and module Set := Set
@@ -112,12 +115,12 @@ end
 (** {6 Type aliases} *)
 
 type name = Name.t = Anonymous | Name of Id.t
-[@@ocaml.deprecated "Use Name.t"]
+[@@ocaml.deprecated "(8.8) Use Name.t"]
 
 type variable = Id.t
 type module_ident = Id.t
 
-module ModIdset : Set.S with type elt = module_ident
+module ModIdset : Set.ExtS with type elt = module_ident
 module ModIdmap : Map.ExtS with type key = module_ident and module Set := ModIdset
 
 (** {6 Directory paths = section names paths } *)
@@ -149,19 +152,19 @@ sig
   val is_empty : t -> bool
   (** Test whether a directory path is empty. *)
 
-  val initial : t
-  (** Initial "seed" of the unique identifier generator *)
+  val dummy : t
+  (** Used in [Safe_typing.empty_environment] and similar *)
 
   val hcons : t -> t
   (** Hashconsing of directory paths. *)
 
   val to_string : t -> string
-  (** Print non-empty directory paths as ["coq_root.module.submodule"] *)
+  (** Print non-empty directory paths as ["root.module.submodule"] *)
 
   val print : t -> Pp.t
 end
 
-module DPset : Set.S with type elt = DirPath.t
+module DPset : Set.ExtS with type elt = DirPath.t
 module DPmap : Map.ExtS with type key = DirPath.t and module Set := DPset
 
 (** {6 Names of structure elements } *)
@@ -195,7 +198,7 @@ sig
   val print : t -> Pp.t
   (** Pretty-printer. *)
 
-  module Set : Set.S with type elt = t
+  module Set : Set.ExtS with type elt = t
   module Map : Map.ExtS with type key = t and module Set := Set
 
   val hcons : t -> t
@@ -237,7 +240,7 @@ sig
 
 end
 
-module MBIset : Set.S with type elt = MBId.t
+module MBIset : Set.ExtS with type elt = MBId.t
 module MBImap : Map.ExtS with type key = MBId.t and module Set := MBIset
 
 (** {6 The module part of the kernel name } *)
@@ -255,20 +258,23 @@ sig
 
   val is_bound : t -> bool
 
-  val initial : t
-  (** Name of the toplevel structure ([= MPfile initial_dir]) *)
+  val dummy : t
+  (** ([= MPfile DirPath.dummy]) *)
 
   val dp : t -> DirPath.t
 
   val to_string : t -> string
-  (** Encode as a string (not to be used for user-facing messages). *)
+  (** Converts a identifier into an string. *)
+
+  val print : t -> Pp.t
+  (** Pretty-printer. *)
 
   val debug_to_string : t -> string
   (** Same as [to_string], but outputs extra information related to debug. *)
 
 end
 
-module MPset : Set.S with type elt = ModPath.t
+module MPset : Set.ExtS with type elt = ModPath.t
 module MPmap : Map.ExtS with type key = ModPath.t and module Set := MPset
 
 (** {6 The absolute names of objects seen by kernel } *)
@@ -303,9 +309,66 @@ sig
   val hash : t -> int
 end
 
-module KNset  : CSig.SetS with type elt = KerName.t
+module KNset  : CSig.USetS with type elt = KerName.t
 module KNpred : Predicate.S with type elt = KerName.t
-module KNmap  : Map.ExtS with type key = KerName.t and module Set := KNset
+module KNmap  : Map.UExtS with type key = KerName.t and module Set := KNset
+
+(** {6 Signature for quotiented names} *)
+
+module type EqType =
+sig
+  type t
+  val compare : t -> t -> int
+  val equal : t -> t -> bool
+  val hash : t -> int
+end
+
+module type QNameS =
+sig
+  type t
+  (** A type of reference that implements an implicit quotient by containing
+      two different names. The first one is the user name, i.e. what the user
+      sees when printing. The second one is the canonical name, which is the
+      actual absolute name of the reference.
+
+      This mechanism is fundamentally tied to the module system of Rocq. Functor
+      application and module inclusion are the typical ways to introduce names
+      where the canonical and user components differ. In particular, the two
+      components should be undistinguishable from the point of view of typing,
+      i.e. from a "kernel" ground. This aliasing only makes sense inside an
+      environment, but at this point this notion is not even defined so, this
+      dual name trick is fragile. One has to ensure many invariants when
+      creating such names, but the kernel is quite lenient when it comes to
+      checking that these invariants hold. (Read: there are soundness bugs
+      lurking in the module system.)
+
+      One could enforce the invariants by splitting the names and storing that
+      information in the environment instead, but unfortunately, this wreaks
+      havoc in the upper layers. The latter are infamously not stable by
+      syntactic equality, in particular they might observe the difference
+      between canonical and user names if not packed together.
+
+      For this reason, it is discouraged to use the canonical-accessing API
+      in the upper layers, notably the [CanOrd] module below. Instead, one
+      should use their quotiented versions defined in the [Environ] module.
+      Eventually all uses to [CanOrd] outside of the kernel should be removed.
+
+      CAVEAT: name sets and maps are still exposing a canonical-accessing API
+      surreptitiously. *)
+
+  module CanOrd : EqType with type t = t
+  (** Equality functions over the canonical name. Their use should be
+      restricted to the kernel. *)
+
+  module UserOrd : EqType with type t = t
+  (** Equality functions over the user name. *)
+
+  module SyntacticOrd : EqType with type t = t
+  (** Equality functions using both names, for low-level uses. *)
+
+  val canonize : t -> t
+  (** Returns the canonical version of the name *)
+end
 
 (** {6 Constant Names } *)
 
@@ -322,16 +385,13 @@ sig
   (** Special case of [make] where the user name is canonical.  *)
 
   val make2 : ModPath.t -> Label.t -> t
-  (** Shortcut for [(make1 (KerName.make2 ...))] *)
+  (** Shortcut for [(make1 (KerName.make ...))] *)
 
   (** Projections *)
 
   val user : t -> KerName.t
   val canonical : t -> KerName.t
 
-  val repr2 : t -> ModPath.t * Label.t
-  (** Shortcut for [KerName.repr (user ...)] *)
-
   val modpath : t -> ModPath.t
   (** Shortcut for [KerName.modpath (user ...)] *)
 
@@ -340,28 +400,12 @@ sig
 
   (** Comparisons *)
 
-  module CanOrd : sig
-    val compare : t -> t -> int
-    val equal : t -> t -> bool
-    val hash : t -> int
-  end
+  include QNameS with type t := t
 
-  module UserOrd : sig
-    val compare : t -> t -> int
-    val equal : t -> t -> bool
-    val hash : t -> int
-  end
-
-  module SyntacticOrd : sig
-    val compare : t -> t -> int
-    val equal : t -> t -> bool
-    val hash : t -> int
-  end
-
-  val equal : t -> t -> bool
+  val equal : t -> t -> bool [@@ocaml.deprecated "(8.13) Use QConstant.equal"]
   (** Default comparison, alias for [CanOrd.equal] *)
 
-  val hash : t -> int
+  val hash : t -> int [@@ocaml.deprecated "(8.13) Use QConstant.hash"]
   (** Hashing function *)
 
   val change_label : t -> Label.t -> t
@@ -386,14 +430,14 @@ end
 (** The [*_env] modules consider an order on user part of names
    the others consider an order on canonical part of names*)
 module Cpred : Predicate.S with type elt = Constant.t
-module Cset : CSig.SetS with type elt = Constant.t
-module Cset_env  : CSig.SetS with type elt = Constant.t
+module Cset : CSig.USetS with type elt = Constant.t
+module Cset_env  : CSig.USetS with type elt = Constant.t
 
-module Cmap : Map.ExtS with type key = Constant.t and module Set := Cset
+module Cmap : Map.UExtS with type key = Constant.t and module Set := Cset
 (** A map whose keys are constants (values of the {!Constant.t} type).
     Keys are ordered wrt. "canonical form" of the constant. *)
 
-module Cmap_env : Map.ExtS with type key = Constant.t and module Set := Cset_env
+module Cmap_env : Map.UExtS with type key = Constant.t and module Set := Cset_env
 (** A map whose keys are constants (values of the {!Constant.t} type).
     Keys are ordered wrt. "user form" of the constant. *)
 
@@ -412,16 +456,13 @@ sig
   (** Special case of [make] where the user name is canonical.  *)
 
   val make2 : ModPath.t -> Label.t -> t
-  (** Shortcut for [(make1 (KerName.make2 ...))] *)
+  (** Shortcut for [(make1 (KerName.make ...))] *)
 
   (** Projections *)
 
   val user : t -> KerName.t
   val canonical : t -> KerName.t
 
-  val repr2 : t -> ModPath.t * Label.t
-  (** Shortcut for [KerName.repr (user ...)] *)
-
   val modpath : t -> ModPath.t
   (** Shortcut for [KerName.modpath (user ...)] *)
 
@@ -430,28 +471,12 @@ sig
 
   (** Comparisons *)
 
-  module CanOrd : sig
-    val compare : t -> t -> int
-    val equal : t -> t -> bool
-    val hash : t -> int
-  end
+  include QNameS with type t := t
 
-  module UserOrd : sig
-    val compare : t -> t -> int
-    val equal : t -> t -> bool
-    val hash : t -> int
-  end
+  val equal : t -> t -> bool [@@ocaml.deprecated "(8.13) Use QMutInd.equal"]
+ (** Default comparison, alias for [CanOrd.equal] *)
 
-  module SyntacticOrd : sig
-    val compare : t -> t -> int
-    val equal : t -> t -> bool
-    val hash : t -> int
-  end
-
-  val equal : t -> t -> bool
-  (** Default comparison, alias for [CanOrd.equal] *)
-
-  val hash : t -> int
+  val hash : t -> int [@@ocaml.deprecated "(8.13) Use QMutInd.hash"]
 
   (** Displaying *)
 
@@ -469,55 +494,53 @@ sig
 
 end
 
-module Mindset : CSig.SetS with type elt = MutInd.t
-module Mindmap : Map.ExtS with type key = MutInd.t and module Set := Mindset
-module Mindmap_env : CMap.ExtS with type key = MutInd.t
+module Mindset : CSig.USetS with type elt = MutInd.t
+module Mindmap : Map.UExtS with type key = MutInd.t and module Set := Mindset
+module Mindmap_env : CMap.UExtS with type key = MutInd.t
 
-(** Designation of a (particular) inductive type. *)
-type inductive = MutInd.t      (* the name of the inductive type *)
-               * int           (* the position of this inductive type
-                                  within the block of mutually-recursive inductive types.
-                                  BEWARE: indexing starts from 0. *)
+module Ind :
+sig
+  (** Designation of a (particular) inductive type. *)
+  type t = MutInd.t      (* the name of the inductive type *)
+         * int           (* the position of this inductive type
+                                    within the block of mutually-recursive inductive types.
+                                    BEWARE: indexing starts from 0. *)
+  val modpath : t -> ModPath.t
 
-(** Designation of a (particular) constructor of a (particular) inductive type. *)
-type constructor = inductive   (* designates the inductive type *)
-                 * int         (* the index of the constructor
-                                  BEWARE: indexing starts from 1. *)
+  include QNameS with type t := t
 
-module Indset : CSet.S with type elt = inductive
-module Constrset : CSet.S with type elt = constructor
-module Indset_env : CSet.S with type elt = inductive
-module Constrset_env : CSet.S with type elt = constructor
+end
+
+type inductive = Ind.t
+
+module Construct :
+sig
+  (** Designation of a (particular) constructor of a (particular) inductive type. *)
+  type t = Ind.t   (* designates the inductive type *)
+         * int     (* the index of the constructor
+                                    BEWARE: indexing starts from 1. *)
+
+  val modpath : t -> ModPath.t
+
+  include QNameS with type t := t
+
+end
+
+type constructor = Construct.t
+
+module Indset : CSet.ExtS with type elt = inductive
+module Constrset : CSet.ExtS with type elt = constructor
+module Indset_env : CSet.ExtS with type elt = inductive
+module Constrset_env : CSet.ExtS with type elt = constructor
 module Indmap : CMap.ExtS with type key = inductive and module Set := Indset
 module Constrmap : CMap.ExtS with type key = constructor and module Set := Constrset
 module Indmap_env : CMap.ExtS with type key = inductive and module Set := Indset_env
 module Constrmap_env : CMap.ExtS with type key = constructor and module Set := Constrset_env
 
-val ind_modpath : inductive -> ModPath.t
-val constr_modpath : constructor -> ModPath.t
-
 val ith_mutual_inductive : inductive -> int -> inductive
 val ith_constructor_of_inductive : inductive -> int -> constructor
 val inductive_of_constructor : constructor -> inductive
 val index_of_constructor : constructor -> int
-val eq_ind : inductive -> inductive -> bool
-val eq_user_ind : inductive -> inductive -> bool
-val eq_syntactic_ind : inductive -> inductive -> bool
-val ind_ord : inductive -> inductive -> int
-val ind_hash : inductive -> int
-val ind_user_ord : inductive -> inductive -> int
-val ind_user_hash : inductive -> int
-val ind_syntactic_ord : inductive -> inductive -> int
-val ind_syntactic_hash : inductive -> int
-val eq_constructor : constructor -> constructor -> bool
-val eq_user_constructor : constructor -> constructor -> bool
-val eq_syntactic_constructor : constructor -> constructor -> bool
-val constructor_ord : constructor -> constructor -> int
-val constructor_hash : constructor -> int
-val constructor_user_ord : constructor -> constructor -> int
-val constructor_user_hash : constructor -> int
-val constructor_syntactic_ord : constructor -> constructor -> int
-val constructor_syntactic_hash : constructor -> int
 
 (** {6 Hash-consing } *)
 
@@ -538,6 +561,7 @@ type inv_rel_key = int (** index in the [rel_context] part of environment
                           of de Bruijn indice *)
 
 val eq_table_key : ('a -> 'a -> bool) -> 'a tableKey -> 'a tableKey -> bool
+val hash_table_key : ('a -> int) -> 'a tableKey -> int
 val eq_constant_key : Constant.t -> Constant.t -> bool
 
 (** equalities on constant and inductive names (for the checker) *)
@@ -550,7 +574,7 @@ type module_path = ModPath.t =
   | MPfile of DirPath.t
   | MPbound of MBId.t
   | MPdot of ModPath.t * Label.t
-[@@ocaml.deprecated "Alias type"]
+[@@ocaml.deprecated "(8.8) Alias type"]
 
 module Projection : sig
   module Repr : sig
@@ -558,21 +582,7 @@ module Projection : sig
 
     val make : inductive -> proj_npars:int -> proj_arg:int -> Label.t -> t
 
-    module SyntacticOrd : sig
-      val compare : t -> t -> int
-      val equal : t -> t -> bool
-      val hash : t -> int
-    end
-    module CanOrd : sig
-      val compare : t -> t -> int
-      val equal : t -> t -> bool
-      val hash : t -> int
-    end
-    module UserOrd : sig
-      val compare : t -> t -> int
-      val equal : t -> t -> bool
-      val hash : t -> int
-    end
+    include QNameS with type t := t
 
     val constant : t -> Constant.t
     (** Don't use this if you don't have to. *)
@@ -583,12 +593,12 @@ module Projection : sig
     val arg : t -> int
     val label : t -> Label.t
 
-    val equal : t -> t -> bool
-    val hash : t -> int
-    val compare : t -> t -> int
+    val equal : t -> t -> bool [@@ocaml.deprecated "(8.13) Use QProjection.equal"]
+    val hash : t -> int [@@ocaml.deprecated "(8.13) Use QProjection.hash"]
+    val compare : t -> t -> int [@@ocaml.deprecated "(8.13) Use QProjection.compare"]
 
     val map : (MutInd.t -> MutInd.t) -> t -> t
-    val map_npars : (MutInd.t -> int -> MutInd.t * int) -> t -> t
+    val map_npars : (int -> int) -> t -> t
 
     val to_string : t -> string
     (** Encode as a string (not to be used for user-facing messages). *)
@@ -602,16 +612,7 @@ module Projection : sig
   val make : Repr.t -> bool -> t
   val repr : t -> Repr.t
 
-  module SyntacticOrd : sig
-    val compare : t -> t -> int
-    val equal : t -> t -> bool
-    val hash : t -> int
-  end
-  module CanOrd : sig
-    val compare : t -> t -> int
-    val equal : t -> t -> bool
-    val hash : t -> int
-  end
+  include QNameS with type t := t
 
   val constant : t -> Constant.t
   val mind : t -> MutInd.t
@@ -623,17 +624,21 @@ module Projection : sig
   val unfold : t -> t
 
   val equal : t -> t -> bool
+  [@@ocaml.deprecated "(8.13) Use QProjection.equal"]
   val hash : t -> int
+  [@@ocaml.deprecated "(8.13) Use QProjection.hash"]
   val hcons : t -> t
   (** Hashconsing of projections. *)
 
   val repr_equal : t -> t -> bool
+  [@@ocaml.deprecated "(8.13) Use an explicit projection of Repr"]
   (** Ignoring the unfolding boolean. *)
 
   val compare : t -> t -> int
+  [@@ocaml.deprecated "(8.13) Use QProjection.compare"]
 
   val map : (MutInd.t -> MutInd.t) -> t -> t
-  val map_npars : (MutInd.t -> int -> MutInd.t * int) -> t -> t
+  val map_npars : (int -> int) -> t -> t
 
   val to_string : t -> string
   (** Encode as a string (not to be used for user-facing messages). *)
@@ -641,11 +646,22 @@ module Projection : sig
   val print : t -> Pp.t
   (** Print internal representation (not to be used for user-facing messages). *)
 
+  val debug_to_string : t -> string
+  (** Same as [to_string], but outputs extra information related to debug. *)
+
+  val debug_print : t -> Pp.t
+  (** Same as [print], but outputs extra information related to debug. *)
+
 end
 
+module PRset : CSig.USetS with type elt = Projection.Repr.t
+module PRmap : Map.UExtS with type key = Projection.Repr.t and module Set := PRset
+
+(** Predicate on projection representation (ignoring unfolding state) *)
+module PRpred : Predicate.S with type elt = Projection.Repr.t
+
 (** {6 Global reference is a kernel side type for all references together } *)
 
-(* XXX: Should we define GlobRefCan GlobRefUser? *)
 module GlobRef : sig
 
   type t =
@@ -655,41 +671,29 @@ module GlobRef : sig
     | ConstructRef of constructor  (** A reference to a constructor of an inductive type. *)
 
   val equal : t -> t -> bool
+  [@@ocaml.deprecated "(8.18) Use QGlobRef.equal"]
 
-  module Ordered : sig
-    type nonrec t = t
-    val compare : t -> t -> int
-    val equal : t -> t -> bool
-    val hash : t -> int
-  end
+  val is_bound : t -> bool
 
-  module Ordered_env : sig
-    type nonrec t = t
-    val compare : t -> t -> int
-    val equal : t -> t -> bool
-    val hash : t -> int
-  end
+  include QNameS with type t := t
 
-  module Set_env : CSig.SetS with type elt = t
-  module Map_env : Map.ExtS
+  module Set_env : CSig.USetS with type elt = t
+  module Map_env : Map.UExtS
     with type key = t and module Set := Set_env
 
-  module Set : CSig.SetS with type elt = t
-  module Map : Map.ExtS
+  module Set : CSig.USetS with type elt = t
+  module Map : Map.UExtS
     with type key = t and module Set := Set
 
+  val print : t -> Pp.t
+  (** Print internal representation (not to be used for user-facing messages). *)
+
 end
 
-(** Better to have it here that in Closure, since required in grammar.cma *)
-(* XXX: Move to a module *)
-type evaluable_global_reference =
-  | EvalVarRef of Id.t
-  | EvalConstRef of Constant.t
-
-val eq_egr : evaluable_global_reference ->  evaluable_global_reference -> bool
-
 (** Located identifiers and objects with syntax. *)
 
 type lident = Id.t CAst.t
 type lname = Name.t CAst.t
 type lstring = string CAst.t
+
+val lident_eq : lident -> lident -> bool
