@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -25,7 +25,8 @@ let open_header = ["Nativevalues";
 let open_header = List.map mk_open open_header
 
 (* Directory where compiled files are stored *)
-let output_dir = ref ".coq-native"
+let dft_output_dir = ".coq-native"
+let output_dir = ref dft_output_dir
 
 (* Extension of generated ml files, stored for debugging purposes *)
 let source_ext = ".native"
@@ -37,32 +38,44 @@ let ( / ) = Filename.concat
 let my_temp_dir = lazy (CUnix.mktemp_dir "Coq_native" "")
 
 let () = at_exit (fun () ->
-    if not !Flags.debug && Lazy.is_val my_temp_dir then
+    if not (keep_debug_files ()) && Lazy.is_val my_temp_dir then
       try
         let d = Lazy.force my_temp_dir in
         Array.iter (fun f -> Sys.remove (Filename.concat d f)) (Sys.readdir d);
         Unix.rmdir d
-      with e ->
+      with (Unix.Unix_error _ | Sys_error _) as e ->
         Feedback.msg_warning
           Pp.(str "Native compile: failed to cleanup: " ++
               str(Printexc.to_string e) ++ fnl()))
 
+let delay_cleanup_file =
+  let toclean = ref [] in
+  let () = at_exit (fun () -> List.iter (fun f -> if Sys.file_exists f then Sys.remove f) !toclean) in
+  fun f -> if not (keep_debug_files()) then toclean := f :: !toclean
+
 (* We have to delay evaluation of include_dirs because coqlib cannot
    be guessed until flags have been properly initialized. It also lets
    us avoid forcing [my_temp_dir] if we don't need it (eg stdlib file
    without native compute or native conv uses). *)
 let include_dirs = ref []
+
 let get_include_dirs () =
   let base = match !include_dirs with
   | [] ->
-    [Envars.coqlib () / "kernel"; Envars.coqlib () / "library"]
+    (* EJGA: Should this case go away in favor of always requiring
+       explicit -nI flags once we remove the make-based system? I think
+       so. *)
+    let env = Boot.Env.init () in
+    [ Boot.Env.(Path.to_string (native_cmi env "kernel"))
+    ; Boot.Env.(Path.to_string (native_cmi env "library"))
+    ]
   | _::_ as l -> l
   in
   if Lazy.is_val my_temp_dir
   then (Lazy.force my_temp_dir) :: base
   else base
 
-(* Pointer to the function linking an ML object into coq's toplevel *)
+(* Pointer to the function linking an ML object into Rocq's toplevel *)
 let load_obj = ref (fun _x -> () : string -> unit)
 
 let rt1 = ref (dummy_value ())
@@ -84,7 +97,9 @@ let write_ml_code fn ?(header=[]) code =
 let error_native_compiler_failed e =
   let msg = match e with
   | Inl (Unix.WEXITED 127) -> Pp.(strbrk "The OCaml compiler was not found. Make sure it is installed, together with findlib.")
-  | Inl (Unix.WEXITED n) -> Pp.(strbrk "Native compiler exited with status" ++ str" " ++ int n)
+  | Inl (Unix.WEXITED n) ->
+     Pp.(strbrk "Native compiler exited with status" ++ str" " ++ int n
+         ++ strbrk (if n = 2 then " (in case of stack overflow, increasing stack size (typically with \"ulimit -s\") often helps)" else ""))
   | Inl (Unix.WSIGNALED n) -> Pp.(strbrk "Native compiler killed by signal" ++ str" " ++ int n)
   | Inl (Unix.WSTOPPED n) -> Pp.(strbrk "Native compiler stopped by signal" ++ str" " ++ int n)
   | Inr e -> Pp.(strbrk "Native compiler failed with error: " ++ strbrk (Unix.error_message e))
@@ -92,9 +107,15 @@ let error_native_compiler_failed e =
   CErrors.user_err msg
 
 let call_compiler ?profile:(profile=false) ml_filename =
-  let load_path = !get_load_paths () in
-  let load_path = List.map (fun dn -> dn / !output_dir) load_path in
-  let include_dirs = List.flatten (List.map (fun x -> ["-I"; x]) (get_include_dirs () @ load_path)) in
+  (* The below path is computed from Require statements, by uniquizing
+     the paths, see [Library.get_used_load_paths] This is in general
+     hacky and we should do a bit better once we move loadpath to its
+     own library *)
+  let require_load_path = !get_load_paths () in
+  (* We assume that installed files always go in .coq-native for now *)
+  (* To ease the build we also consider the current dir, but at some point the build system should manage both *)
+  let install_load_path = List.map (fun dn -> dn / dft_output_dir) require_load_path @ require_load_path in
+  let include_dirs = List.flatten (List.map (fun x -> ["-I"; x]) (get_include_dirs () @ install_load_path)) in
   let f = Filename.chop_extension ml_filename in
   let link_filename = f ^ ".cmo" in
   let link_filename = Dynlink.adapt_filename link_filename in
@@ -123,7 +144,7 @@ let call_compiler ?profile:(profile=false) ml_filename =
        ::"-w"::"a"
        ::include_dirs) @
       ["-impl"; ml_filename] in
-  if !Flags.debug then Feedback.msg_debug (Pp.str (Envars.ocamlfind () ^ " " ^ (String.concat " " args)));
+  debug_native_compiler (fun () -> Pp.str (Envars.ocamlfind () ^ " " ^ (String.concat " " args)));
   try
     let res = CUnix.sys_command (Envars.ocamlfind ()) args in
     match res with
@@ -136,15 +157,19 @@ let call_compiler ?profile:(profile=false) ml_filename =
 let compile fn code ~profile:profile =
   write_ml_code fn code;
   let r = call_compiler ~profile fn in
-  if (not !Flags.debug) && Sys.file_exists fn then Sys.remove fn;
+  (* NB: to prevent reusing the same filename we MUST NOT remove the file until exit
+     cf #15263 *)
+  delay_cleanup_file fn;
   r
 
-let compile_library dir code fn =
-  let header = mk_library_header dir in
+type native_library = Nativecode.global list * Nativevalues.symbols
+
+let compile_library (code, symb) fn =
+  let header = mk_library_header symb in
   let fn = fn ^ source_ext in
   let basename = Filename.basename fn in
-  let dirname = Filename.dirname fn in
-  let dirname = dirname / !output_dir in
+  let dirname =
+    if Filename.is_relative !output_dir then Filename.dirname fn / !output_dir else !output_dir in
   let () =
     try Unix.mkdir dirname 0o755
     with Unix.Unix_error (Unix.EEXIST, _, _) -> ()
@@ -152,39 +177,43 @@ let compile_library dir code fn =
   let fn = dirname / basename in
   write_ml_code fn ~header code;
   let _ = call_compiler fn in
-  if (not !Flags.debug) && Sys.file_exists fn then Sys.remove fn
+  delay_cleanup_file fn
 
-let native_symbols = ref Names.DPmap.empty
-
-let get_library_native_symbols dir =
-  try Names.DPmap.find dir !native_symbols
-  with Not_found ->
-    CErrors.user_err ~hdr:"get_library_native_symbols"
-      Pp.((str "Linker error in the native compiler. Are you using Require inside a nested Module declaration?") ++ fnl () ++
-          (str "This use case is not supported, but disabling the native compiler may help."))
-
-(* call_linker links dynamically the code for constants in environment or a  *)
-(* conversion test. *)
-let call_linker ?(fatal=true) env ~prefix f upds =
-  native_symbols := env.Environ.native_symbols;
+let execute_library ~prefix f upds =
   rt1 := dummy_value ();
   rt2 := dummy_value ();
   if not (Sys.file_exists f) then
-    begin
-      let msg = "Cannot find native compiler file " ^ f in
-      if fatal then CErrors.user_err Pp.(str msg)
-      else if !Flags.debug then Feedback.msg_debug (Pp.str msg)
-    end
-  else
-  (try
+    CErrors.user_err Pp.(str "Cannot find native compiler file " ++ str f);
+  if Dynlink.is_native then Dynlink.loadfile f else !load_obj f;
+  register_native_file prefix;
+  update_locations upds;
+  (!rt1, !rt2)
+
+let link_library dirname prefix =
+  let basename = Dynlink.adapt_filename (prefix ^ "cmo") in
+  (* We try both [output_dir] and [.coq-native], unfortunately from
+     [Require] we don't know if we are loading a library in the build
+     dir or in the installed layout *)
+  let install_location = dirname / dft_output_dir / basename in
+  let build_location = dirname / !output_dir / basename in
+  let f = if Sys.file_exists build_location then build_location else install_location in
+  try
     if Dynlink.is_native then Dynlink.loadfile f else !load_obj f;
     register_native_file prefix
-   with Dynlink.Error _ as exn ->
-     let exn = Exninfo.capture exn in
-     if fatal then Exninfo.iraise exn
-     else if !Flags.debug then Feedback.msg_debug CErrors.(iprint exn));
-  match upds with Some upds -> update_locations upds | _ -> ()
+  with
+  | Dynlink.Error _ as exn ->
+      debug_native_compiler (fun () -> CErrors.iprint (Exninfo.capture exn))
 
-let link_library env ~prefix ~dirname ~basename =
-  let f = dirname / !output_dir / basename in
-  call_linker env ~fatal:false ~prefix f None
+let delayed_link = ref []
+
+let link_libraries () =
+  let delayed = List.rev !delayed_link in
+  delayed_link := [];
+  List.iter (fun (dirname, libname) ->
+      let prefix = mod_uid_of_dirpath libname ^ "." in
+      if not (Nativecode.is_loaded_native_file prefix) then
+        link_library dirname prefix
+    ) delayed
+
+let enable_library dirname libname =
+  delayed_link := (dirname, libname) :: !delayed_link
