@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -7,7 +7,6 @@
 (*         *     GNU Lesser General Public License Version 2.1          *)
 (*         *     (see LICENSE file for the text of the license)         *)
 (************************************************************************)
-open Nativecode
 
 (** This file provides facilities to access OCaml compiler and dynamic linker,
 used by the native compiler. *)
@@ -25,29 +24,27 @@ val get_ml_filename : unit -> string * string
 (** [compile file code ~profile] will compile native [code] to [file],
    and return the name of the object file; this name depends on
    whether are in byte mode or not; file is expected to be .ml file *)
-val compile : string -> global list -> profile:bool -> string
+val compile : string -> Nativecode.global list -> profile:bool -> string
 
-(** [compile_library lib code file] is similar to [compile file code]
-   but will perform some extra tweaks to handle [code] as a Coq lib. *)
-val compile_library : Names.DirPath.t -> global list -> string -> unit
+type native_library = Nativecode.global list * Nativevalues.symbols
 
-val call_linker
-  : ?fatal:bool
-  -> Environ.env
-  -> prefix:string
-  -> string
-  -> code_location_updates option
-  -> unit
+(** [compile_library (code, _) file] is similar to [compile file code]
+   but will perform some extra tweaks to handle [code] as a Rocq lib. *)
+val compile_library : native_library -> string -> unit
 
-val link_library
-  : Environ.env
-  -> prefix:string
-  -> dirname:string
-  -> basename:string
-  -> unit
+(** [execute_library file upds] dynamically loads library [file],
+    updates the library locations [upds], and returns the values stored
+    in [rt1] and [rt2] *)
+val execute_library :
+  prefix:string -> string -> Nativecode.code_location_updates ->
+  Nativevalues.t * Nativevalues.t
 
+(** [enable_library] marks the given library for dynamic loading
+    the next time [link_libraries] is called. *)
+val enable_library : string -> Names.DirPath.t -> unit
+
+val link_libraries : unit -> unit
+
+(* used for communication with the loaded libraries *)
 val rt1 : Nativevalues.t ref
 val rt2 : Nativevalues.t ref
-
-val get_library_native_symbols : Names.DirPath.t -> Nativevalues.symbols
-(** Strictly for usage by code produced by native compute. *)
