@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -10,6 +10,8 @@
 
 open Names
 open Univ
+open UVars
+open Cooking
 
 (** Kernel implementation of sections. *)
 
@@ -41,10 +43,10 @@ val close_section : 'a t -> 'a t option * section_entry list * ContextSet.t * 'a
 
 (** {6 Extending sections} *)
 
-val push_local : 'a t -> 'a t
+val push_local : Constr.named_declaration -> 'a t -> 'a t
 (** Extend the current section with a local definition (cf. push_named). *)
 
-val push_context : Name.t array * UContext.t -> 'a t -> 'a t
+val push_local_universe_context : UContext.t -> 'a t -> 'a t
 (** Extend the current section with a local universe context. Assumes that the
     last opened section is polymorphic. *)
 
@@ -52,15 +54,12 @@ val push_constraints : ContextSet.t -> 'a t -> 'a t
 (** Extend the current section with a global universe context.
     Assumes that the last opened section is monomorphic. *)
 
-val push_constant : poly:bool -> Constant.t -> 'a t -> 'a t
-(** Make the constant as having been defined in this section. *)
-
-val push_inductive : poly:bool -> MutInd.t -> 'a t -> 'a t
-(** Make the inductive block as having been defined in this section. *)
+val push_global : Environ.env -> poly:bool -> section_entry -> 'a t -> 'a t
+(** Push a global entry in this section. *)
 
 (** {6 Retrieving section data} *)
 
-val all_poly_univs : 'a t -> Univ.Level.t array
+val all_poly_univs : 'a t -> Instance.t
 (** Returns all polymorphic universes, including those from previous
    sections. Earlier sections are earlier in the array.
 
@@ -68,27 +67,10 @@ val all_poly_univs : 'a t -> Univ.Level.t array
    constraints about monomorphic universes, which prevent declaring
    monomorphic globals. *)
 
-type abstr_info = private {
-  abstr_ctx : Constr.named_context;
-  (** Section variables of this prefix *)
-  abstr_subst : Univ.Instance.t;
-  (** Actual names of the abstracted variables *)
-  abstr_uctx : Univ.AUContext.t;
-  (** Universe quantification, same length as the substitution *)
-}
-(** Data needed to abstract over the section variable and universe hypotheses *)
-
-
-val empty_segment : abstr_info
-(** Nothing to abstract *)
-
-val segment_of_constant : Environ.env -> Constant.t -> 'a t -> abstr_info
+val segment_of_constant : Constant.t -> 'a t -> cooking_info
 (** Section segment at the time of the constant declaration *)
 
-val segment_of_inductive : Environ.env -> MutInd.t -> 'a t -> abstr_info
+val segment_of_inductive : MutInd.t -> 'a t -> cooking_info
 (** Section segment at the time of the inductive declaration *)
 
-val replacement_context : Environ.env -> 'a t -> Opaqueproof.work_list
-(** Section segments of all declarations from this section. *)
-
 val is_in_section : Environ.env -> GlobRef.t -> 'a t -> bool
