@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -55,7 +55,7 @@ sig
       When possible you should use name-aware printing. *)
 
   val pr : t -> Pp.t
-  [@@deprecated "Use [UnivNames.pr_with_global_universes] instead if possible, otherwise [raw_pr]."]
+  [@@deprecated "(8.18) Use [UnivNames.pr_with_global_universes] instead if possible, otherwise [raw_pr]."]
 
   val to_string : t -> string
   (** Debug printing *)
@@ -131,8 +131,9 @@ sig
   (** Try to get a level out of a universe, returns [None] if it
       is an algebraic universe. *)
 
-  val levels : t -> Level.Set.t
-  (** Get the levels inside the universe, forgetting about increments *)
+  val levels : ?init:Level.Set.t -> t -> Level.Set.t
+  (** Get the levels inside the universe, forgetting about increments,
+      and add them to [init] (default empty) *)
 
   val super : t -> t
   (** The universe strictly above *)
@@ -154,7 +155,7 @@ sig
   val repr : t -> (Level.t * int) list
   val unrepr : (Level.t * int) list -> t
 
-  module Set : CSet.S with type elt  = t
+  module Set : CSet.ExtS with type elt  = t
   module Map : CMap.ExtS with type key = t and module Set := Set
 
 end
@@ -174,7 +175,7 @@ type constraint_type = AcyclicGraph.constraint_type = Lt | Le | Eq
 type univ_constraint = Level.t * constraint_type * Level.t
 
 module Constraints : sig
-  include Set.S with type elt = univ_constraint
+  include CSet.ExtS with type elt = univ_constraint
 
   val pr : (Level.t -> Pp.t) -> t -> Pp.t
 end
