@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -12,7 +12,22 @@ open Util
 
 (** {1 Helper functions} *)
 
-let getenv_else s dft = try Sys.getenv s with Not_found -> dft ()
+let warn_deprecated_coq_var = CWarnings.create ~name:"deprecated-coq-env-var" ~category:Deprecation.Version.v9_0
+    Pp.(fun (rocq,coq) ->
+        str "Deprecated environment variable " ++ str coq ++ pr_comma() ++
+        str "use " ++ str rocq ++ str " instead.")
+
+let warn_deprecated_coq_var ?loc ~rocq ~coq () = warn_deprecated_coq_var ?loc (rocq,coq)
+
+let () = Boot.Util.set_warn_deprecated_coq_var (fun ~rocq ~coq ->
+    warn_deprecated_coq_var ~rocq ~coq ())
+
+let getenv_else s dft =
+  match Boot.Util.getenv_opt s with
+  | Some v -> v
+  | None -> dft ()
+
+let getenv_rocq = Boot.Util.getenv_rocq
 
 let safe_getenv warning n =
   getenv_else n (fun () ->
@@ -64,7 +79,7 @@ let expand_path_macros ~warn s =
 
 (** {1 Paths} *)
 
-(** {2 Coq paths} *)
+(** {2 Rocq paths} *)
 
 let coqbin =
   CUnix.canonical_path_name (Filename.dirname Sys.executable_name)
@@ -85,40 +100,6 @@ let _ =
 let use_suffix prefix suffix =
   if String.length suffix > 0 && suffix.[0] = '/' then suffix else prefix / suffix
 
-(** [check_file_else ~dir ~file oth] checks if [file] exists in
-    the installation directory [dir] given relatively to [coqroot],
-    which maybe has been relocated.
-    If the check fails, then [oth ()] is evaluated.
-    Using file system equality seems well enough for this heuristic *)
-let check_file_else ~dir ~file oth =
-  let path = use_suffix coqroot dir in
-  if Sys.file_exists (path / file) then path else oth ()
-
-let guess_coqlib fail =
-  getenv_else "COQLIB" (fun () ->
-  let prelude = "theories/Init/Prelude.vo" in
-  check_file_else ~dir:Coq_config.coqlibsuffix ~file:prelude
-    (fun () ->
-      if not Coq_config.local && Sys.file_exists (Coq_config.coqlib / prelude)
-      then Coq_config.coqlib
-      else
-        fail "cannot guess a path for Coq libraries; please use -coqlib option")
-  )
-
-let coqlib : string option ref = ref None
-let set_user_coqlib path = coqlib := Some path
-
-(** coqlib is now computed once during coqtop initialization *)
-
-let set_coqlib ~fail =
-  match !coqlib with
-  | Some _ -> ()
-  | None ->
-    let lib = guess_coqlib fail in
-    coqlib := Some lib
-
-let coqlib () = Option.default "" !coqlib
-
 let docdir () =
   (* This assumes implicitly that the suffix is non-trivial *)
   let path = use_suffix coqroot Coq_config.docdirsuffix in
@@ -135,17 +116,19 @@ let configdir () =
   if Sys.file_exists path then path else Coq_config.configdir
 
 let coqpath =
-  let coqpath = getenv_else "COQPATH" (fun () -> "") in
   let make_search_path path =
     let paths = path_to_list path in
     let valid_paths = List.filter Sys.file_exists paths in
     List.rev valid_paths
   in
-  make_search_path coqpath
+  let rocqpath = getenv_else "ROCQPATH" (fun () -> "") in
+  let coqpath = getenv_else "COQPATH" (fun () -> "") in
+  let () = if coqpath <> "" then warn_deprecated_coq_var ~rocq:"ROCQPATH" ~coq:"COQPATH" () in
+  make_search_path rocqpath @ make_search_path coqpath
 
 (** {2 Caml paths} *)
 
-let ocamlfind () = Coq_config.ocamlfind
+let ocamlfind () = getenv_else "OCAMLFIND" (fun () -> Coq_config.ocamlfind)
 
 (** {1 XDG utilities} *)
 
@@ -171,15 +154,22 @@ let xdg_dirs ~warn =
 
 (* Print the configuration information *)
 
-let print_config ?(prefix_var_name="") f coq_src_subdirs =
+let print_config ?(prefix_var_name="") f =
+  let env = Boot.Env.init () in
+  let coqlib = Boot.Env.coqlib env |> Boot.Path.to_string in
+  let corelib = Boot.Env.corelib env |> Boot.Path.to_string in
   let open Printf in
-  fprintf f "%sLOCAL=%s\n" prefix_var_name (if Coq_config.local then "1" else "0");
-  fprintf f "%sCOQLIB=%s/\n" prefix_var_name (coqlib ());
+  fprintf f "%sCOQLIB=%s/\n" prefix_var_name coqlib;
+  fprintf f "%sCOQCORELIB=%s/\n" prefix_var_name corelib;
   fprintf f "%sDOCDIR=%s/\n" prefix_var_name (docdir ());
   fprintf f "%sOCAMLFIND=%s\n" prefix_var_name (ocamlfind ());
   fprintf f "%sCAMLFLAGS=%s\n" prefix_var_name Coq_config.caml_flags;
   fprintf f "%sWARN=%s\n" prefix_var_name "-warn-error +a-3";
   fprintf f "%sHASNATDYNLINK=%s\n" prefix_var_name
     (if Coq_config.has_natdynlink then "true" else "false");
-  fprintf f "%sCOQ_SRC_SUBDIRS=%s\n" prefix_var_name (String.concat " " coq_src_subdirs)
-
+  fprintf f "%sCOQ_SRC_SUBDIRS=%s\n" prefix_var_name (String.concat " " Coq_config.all_src_dirs);
+  fprintf f "%sCOQ_NATIVE_COMPILER_DEFAULT=%s\n" prefix_var_name
+    (match Coq_config.native_compiler with
+     | Coq_config.NativeOn {ondemand=false} -> "yes"
+     | Coq_config.NativeOff -> "no"
+     | Coq_config.NativeOn {ondemand=true} -> "ondemand")
