@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -8,32 +8,33 @@
 (*         *     (see LICENSE file for the text of the license)         *)
 (************************************************************************)
 
-(** Coq document type. *)
+(** Rocq document type. *)
+
+(**
+{4 Pretty printing guidelines}
+
+[Pp.t] is the main pretty printing document type
+in the Rocq system. Documents are composed laying out boxes, and
+users can add arbitrary tag metadata that backends are free
+to interpret.
+
+The datatype has a public view to allow serialization or advanced
+uses, however regular users are _strongly_ warned against its use,
+they should instead rely on the available functions below.
+
+Box order and number is indeed an important factor. Try to create
+a proper amount of boxes. The [++] operator provides "efficient"
+concatenation, but using the list constructors is usually preferred.
+
+That is to say, this:
+
+[hov [str "Term"; hov (pr_term t); str "is defined"]]
+
+is preferred to:
+
+[hov (str "Term" ++ hov (pr_term t) ++ str "is defined")]
+*)
 
-(** Pretty printing guidelines ******************************************)
-(*                                                                      *)
-(* `Pp.t` is the main pretty printing document type                     *)
-(* in the Coq system. Documents are composed laying out boxes, and      *)
-(* users can add arbitrary tag metadata that backends are free          *)
-(* to interpret.                                                        *)
-(*                                                                      *)
-(* The datatype has a public view to allow serialization or advanced    *)
-(* uses, however regular users are _strongly_ warned against its use,   *)
-(* they should instead rely on the available functions below.           *)
-(*                                                                      *)
-(* Box order and number is indeed an important factor. Try to create    *)
-(* a proper amount of boxes. The `++` operator provides "efficient"     *)
-(* concatenation, but using the list constructors is usually preferred. *)
-(*                                                                      *)
-(* That is to say, this:                                                *)
-(*                                                                      *)
-(* `hov [str "Term"; hov (pr_term t); str "is defined"]`                *)
-(*                                                                      *)
-(* is preferred to:                                                     *)
-(*                                                                      *)
-(* `hov (str "Term" ++ hov (pr_term t) ++ str "is defined")`            *)
-(*                                                                      *)
-(************************************************************************)
 
 (* XXX: Improve and add attributes *)
 type pp_tag = string
@@ -43,10 +44,11 @@ type pp_tag = string
 type t
 
 type block_type =
-  | Pp_hbox   of int
+  | Pp_hbox
   | Pp_vbox   of int
   | Pp_hvbox  of int
   | Pp_hovbox of int
+(** [Pp_hovbox] produces boxes according to [Format.open_box] not [Format.open_hovbox] *)
 
 type doc_view =
   | Ppcmd_empty
@@ -90,6 +92,7 @@ val spc : unit -> t
 val cut : unit -> t
 val align : unit -> t
 val int : int -> t
+val int64 : Int64.t -> t
 val real : float -> t
 val bool : bool -> t
 val qstring : string -> t
@@ -99,10 +102,11 @@ val strbrk : string -> t
 
 (** {6 Boxing commands} *)
 
-val h : int -> t -> t
+val h : t -> t
 val v : int -> t -> t
 val hv : int -> t -> t
 val hov : int -> t -> t
+(** [hov] produces boxes according to [Format.open_box] not [Format.open_hovbox] *)
 
 (** {6 Tagging} *)
 
@@ -134,6 +138,12 @@ val pr_opt : ('a -> t) -> 'a option -> t
 val pr_opt_no_spc : ('a -> t) -> 'a option -> t
 (** Same as [pr_opt] but without the leading space. *)
 
+val pr_opt_default : (unit -> t) -> ('a -> t) -> 'a option -> t
+(** Prints [pr v] if [ov] is [Some v], else [prdf ()]. *)
+
+val pr_opt_no_spc_default : (unit -> t) -> ('a -> t) -> 'a option -> t
+(** Same as [pr_opt_default] but without the leading space. *)
+
 val pr_nth : int -> t
 (** Ordinal number with the correct suffix (i.e. "st", "nd", "th", etc.). *)
 
@@ -170,7 +180,11 @@ val prvecti_with_sep :
 
 val pr_enum : ('a -> t) -> 'a list -> t
 (** [pr_enum pr [a ; b ; ... ; c]] outputs
-    [pr a ++ str "," ++ pr b ++ str "," ++ ... ++ str "and" ++ pr c]. *)
+    [pr a ++ str "," ++ spc () ++ pr b ++ str "," ++ spc () ++ ... ++ str "and" ++ spc () ++ pr c]. *)
+
+val pr_choice : ('a -> t) -> 'a list -> t
+(** [pr_choice pr [a ; b ; ... ; c]] outputs
+    [pr a ++ str "," ++ spc () ++ pr b ++ str "," ++ spc () ++ ... ++ str "or" ++ spc () ++ pr c]. *)
 
 val pr_sequence : ('a -> t) -> 'a list -> t
 (** Sequence of objects separated by space (unless an element is empty). *)
@@ -187,7 +201,6 @@ val pp_with          : Format.formatter -> t -> unit
 
 val string_of_ppcmds : t -> string
 
-
 (** Tag prefix to start a multi-token diff span *)
 val start_pfx : string
 
@@ -203,5 +216,11 @@ val db_print_pp : Format.formatter -> t -> unit
 (** Print the Pp in tree form for debugging, return as a string *)
 val db_string_of_pp : t -> string
 
+(** Returns [fmt, args] such that [fmt] is a format string which
+    applied to [args] produces the same result as [pp_with].
+
+    [with_tags] default false. *)
+val pp_as_format : ?with_tags:bool -> t -> string * string list
+
 (** Combine nested Ppcmd_glues *)
 val flatten : t -> t
