@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -11,7 +11,7 @@
 open Names
 open Declarations
 
-(** This module defines the global environment of Coq. The functions
+(** This module defines the global environment of Rocq. The functions
    below are exactly the same as the ones in [Safe_typing], operating on
    that global environment. [add_*] functions perform name verification,
    i.e. check that the name given as argument match those provided by
@@ -20,10 +20,7 @@ open Declarations
 val safe_env : unit -> Safe_typing.safe_environment
 val env : unit -> Environ.env
 
-val env_is_initial : unit -> bool
-
 val universes : unit -> UGraph.t
-val universes_lbound : unit -> UGraph.Bound.t
 val named_context_val : unit -> Environ.named_context_val
 val named_context : unit -> Constr.named_context
 
@@ -38,16 +35,16 @@ val set_check_guarded : bool -> unit
 val set_check_positive : bool -> unit
 val set_check_universes : bool -> unit
 val typing_flags : unit -> typing_flags
-val set_cumulative_sprop : bool -> unit
-val is_cumulative_sprop : unit -> bool
 val set_allow_sprop : bool -> unit
 val sprop_allowed : unit -> bool
+val set_rewrite_rules_allowed : bool -> unit
+val rewrite_rules_allowed : unit -> bool
 
 (** Variables, Local definitions, constants, inductive types *)
 
 val push_named_assum : (Id.t * Constr.types) -> unit
 val push_named_def   : (Id.t * Entries.section_def_entry) -> unit
-val push_section_context : Univ.UContext.t -> unit
+val push_section_context : UVars.UContext.t -> unit
 
 val export_private_constants :
   Safe_typing.private_constants ->
@@ -55,18 +52,20 @@ val export_private_constants :
 
 val add_constant :
   ?typing_flags:typing_flags ->
-  Id.t -> Safe_typing.global_declaration -> Constant.t
+  Id.t -> Entries.constant_entry -> Constant.t
 val fill_opaque : Safe_typing.opaque_certificate -> unit
 val add_private_constant :
   Id.t -> Univ.ContextSet.t -> Safe_typing.side_effect_declaration -> Constant.t * Safe_typing.private_constants
+val add_rewrite_rules : Id.t -> rewrite_rules_body -> unit
 val add_mind :
   ?typing_flags:typing_flags ->
-  Id.t -> Entries.mutual_inductive_entry -> MutInd.t
+  Id.t -> Entries.mutual_inductive_entry ->
+  MutInd.t * IndTyping.NotPrimRecordReason.t option
 
 (** Extra universe constraints *)
 val add_constraints : Univ.Constraints.t -> unit
 
-val push_context_set : strict:bool -> Univ.ContextSet.t -> unit
+val push_context_set : Univ.ContextSet.t -> unit
 
 (** Non-interactive modules and module types *)
 
@@ -84,7 +83,7 @@ val add_include :
 val open_section : unit -> unit
 (** [poly] is true when the section should be universe polymorphic *)
 
-val close_section : Summary.frozen -> unit
+val close_section : Summary.Interp.frozen -> unit
 (** Close the section and reset the global state to the one at the time when
     the section what opened. *)
 
@@ -95,11 +94,11 @@ val sections_are_opened : unit -> bool
 val start_module : Id.t -> ModPath.t
 val start_modtype : Id.t -> ModPath.t
 
-val end_module : Summary.frozen -> Id.t ->
+val end_module : Summary.Interp.frozen -> Id.t ->
   (Entries.module_struct_entry * inline) option ->
     ModPath.t * MBId.t list * Mod_subst.delta_resolver
 
-val end_modtype : Summary.frozen -> Id.t -> ModPath.t * MBId.t list
+val end_modtype : Summary.Interp.frozen -> Id.t -> ModPath.t * MBId.t list
 
 val add_module_parameter :
   MBId.t -> Entries.module_struct_entry -> inline ->
@@ -122,13 +121,13 @@ val constant_of_delta_kn : KerName.t -> Constant.t
 val mind_of_delta_kn : KerName.t -> MutInd.t
 
 type indirect_accessor = {
-  access_proof : Opaqueproof.opaque -> (Constr.t * unit Opaqueproof.delayed_universes) option;
+  access_proof : Opaqueproof.opaque -> Opaqueproof.opaque_proofterm option;
 }
 
-val force_proof : indirect_accessor -> Opaqueproof.opaque -> Constr.t * unit Opaqueproof.delayed_universes
+val force_proof : indirect_accessor -> Opaqueproof.opaque -> Opaqueproof.opaque_proofterm
 
 val body_of_constant : indirect_accessor -> Constant.t ->
-  (Constr.constr * unit Opaqueproof.delayed_universes * Univ.AbstractContext.t) option
+  (Constr.constr * unit Opaqueproof.delayed_universes * UVars.AbstractContext.t) option
 (** Returns the body of the constant if it has any, and the polymorphic context
     it lives in. For monomorphic constant, the latter is empty, and for
     polymorphic constants, the term contains De Bruijn universe variables that
@@ -136,16 +135,16 @@ val body_of_constant : indirect_accessor -> Constant.t ->
 
 val body_of_constant_body : indirect_accessor ->
   constant_body ->
-    (Constr.constr * unit Opaqueproof.delayed_universes * Univ.AbstractContext.t) option
+    (Constr.constr * unit Opaqueproof.delayed_universes * UVars.AbstractContext.t) option
 (** Same as {!body_of_constant} but on {!constant_body}. *)
 
 (** {6 Compiled libraries } *)
 
 val start_library : DirPath.t -> ModPath.t
 val export : output_native_objects:bool -> DirPath.t ->
-  ModPath.t * Safe_typing.compiled_library * Nativelib.native_library
+  ModPath.t * Safe_typing.compiled_library * Vmlibrary.compiled_library * Nativelib.native_library
 val import :
-  Safe_typing.compiled_library -> Univ.ContextSet.t -> Safe_typing.vodigest ->
+  Safe_typing.compiled_library -> Vmlibrary.on_disk -> Safe_typing.vodigest ->
   ModPath.t
 
 (** {6 Misc } *)
@@ -156,10 +155,10 @@ val import :
 val env_of_context : Environ.named_context_val -> Environ.env
 
 val is_joined_environment : unit -> bool
+val is_curmod_library : unit -> bool
 
 val is_polymorphic : GlobRef.t -> bool
 val is_template_polymorphic : GlobRef.t -> bool
-val get_template_polymorphic_variables : GlobRef.t -> Univ.Level.t list
 val is_type_in_type : GlobRef.t -> bool
 
 (** {6 Retroknowledge } *)
@@ -169,7 +168,7 @@ val register_inductive : inductive -> 'a CPrimitives.prim_ind -> unit
 
 (** {6 Oracle } *)
 
-val set_strategy : Constant.t Names.tableKey -> Conv_oracle.level -> unit
+val set_strategy : Conv_oracle.evaluable -> Conv_oracle.level -> unit
 
 (** {6 Conversion settings } *)
 
