@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -8,7 +8,7 @@
 (*         *     (see LICENSE file for the text of the license)         *)
 (************************************************************************)
 
-(* This file is the interface between the c-c algorithm and Coq *)
+(* This file is the interface between the c-c algorithm and Rocq *)
 
 open Names
 open Inductiveops
@@ -28,28 +28,30 @@ open Proofview.Notations
 module RelDecl = Context.Rel.Declaration
 module NamedDecl = Context.Named.Declaration
 
-let _f_equal    = lazy (Coqlib.lib_ref "core.eq.congr")
-let _eq_rect    = lazy (Coqlib.lib_ref "core.eq.rect")
-let _refl_equal = lazy (Coqlib.lib_ref "core.eq.refl")
-let _sym_eq     = lazy (Coqlib.lib_ref "core.eq.sym")
-let _trans_eq   = lazy (Coqlib.lib_ref "core.eq.trans")
-let _eq         = lazy (Coqlib.lib_ref "core.eq.type")
-let _False      = lazy (Coqlib.lib_ref "core.False.type")
-let _not        = lazy (Coqlib.lib_ref "core.not.type")
+let _f_equal    = lazy (Rocqlib.lib_ref "core.eq.congr")
+let _eq_rect    = lazy (Rocqlib.lib_ref "core.eq.rect")
+let _refl_equal = lazy (Rocqlib.lib_ref "core.eq.refl")
+let _sym_eq     = lazy (Rocqlib.lib_ref "core.eq.sym")
+let _trans_eq   = lazy (Rocqlib.lib_ref "core.eq.trans")
+let _eq         = lazy (Rocqlib.lib_ref "core.eq.type")
+let _False      = lazy (Rocqlib.lib_ref "core.False.type")
+let _not        = lazy (Rocqlib.lib_ref "core.not.type")
 
 let whd env sigma t =
-  Reductionops.clos_whd_flags CClosure.betaiotazeta env sigma t
+  Reductionops.clos_whd_flags RedFlags.betaiotazeta env sigma t
 
 let whd_delta env sigma t =
-  Reductionops.clos_whd_flags CClosure.all env sigma t
+  Reductionops.clos_whd_flags RedFlags.all env sigma t
+
+let whd_all env sigma c = Reductionops.whd_all env sigma c
 
 let whd_in_concl =
-  reduct_in_concl ~cast:true ~check:false (Reductionops.whd_all, DEFAULTcast)
+  reduct_in_concl ~cast:true ~check:false (whd_all, DEFAULTcast)
 
 (* decompose member of equality in an applicative format *)
 
 (** FIXME: evar leak *)
-let sf_of env sigma c = snd (sort_of env sigma c)
+let sf_of env sigma c = ESorts.kind sigma (snd (sort_of env sigma c))
 
 let rec decompose_term env sigma t =
     match EConstr.kind sigma (whd env sigma t) with
@@ -64,14 +66,11 @@ let rec decompose_term env sigma t =
         ATerm.mkAppli (ATerm.mkAppli (ATerm.mkProduct (sort_a,sort_b),
                     decompose_term env sigma a),
               decompose_term env sigma b)
-    | Construct c ->
-        let (((mind,i_ind),i_con),u)= c in
+    | Construct ((ind, _ as cstr), u) ->
         let u = EInstance.kind sigma u in
-        let canon_mind = MutInd.make1 (MutInd.canonical mind) in
-        let canon_ind = canon_mind,i_ind in
-        let (oib,_)=Global.lookup_inductive (canon_ind) in
-        let nargs=constructor_nallargs env (canon_ind,i_con) in
-          ATerm.mkConstructor {ci_constr= ((canon_ind,i_con),u);
+        let oib = Environ.lookup_mind (fst ind) env in
+        let nargs = constructor_nallargs env cstr in
+        ATerm.mkConstructor env {ci_constr = (cstr, u);
                        ci_arity=nargs;
                        ci_nhyps=nargs-oib.mind_nparams}
     | Ind c ->
@@ -83,7 +82,7 @@ let rec decompose_term env sigma t =
         let u = EInstance.kind sigma u in
         let canon_const = Constant.make1 (Constant.canonical c) in
           ATerm.mkSymb (Constr.mkConstU (canon_const,u))
-    | Proj (p, c) ->
+    | Proj (p, _, c) ->
         let canon_mind kn = MutInd.make1 (MutInd.canonical kn) in
         let p' = Projection.map canon_mind p in
         let c = Retyping.expand_projection env sigma p' c [] in
@@ -99,7 +98,7 @@ let atom_of_constr b env sigma term =
   let kot = EConstr.kind sigma wh in
     match kot with
       App (f,args)->
-        if isRefX sigma (Lazy.force _eq) f && Int.equal (Array.length args) 3
+        if isRefX env sigma (Lazy.force _eq) f && Int.equal (Array.length args) 3
           then `Eq (args.(0),
                    decompose_term env sigma args.(1),
                    decompose_term env sigma args.(2))
@@ -145,7 +144,7 @@ let rec has_open_head = function
 let patterns_of_constr b env sigma nrels term =
   let f,args=
     try destApp sigma ((if b then whd else whd_delta) env sigma term) with DestKO -> raise Not_found in
-        if isRefX sigma (Lazy.force _eq) f && Int.equal (Array.length args) 3
+        if isRefX env sigma (Lazy.force _eq) f && Int.equal (Array.length args) 3
         then
           let patt1,rels1 = pattern_of_constr env sigma args.(1)
           and patt2,rels2 = pattern_of_constr env sigma args.(2) in
@@ -168,12 +167,12 @@ let patterns_of_constr b env sigma nrels term =
 let rec quantified_atom_of_constr b env sigma nrels term =
   match EConstr.kind sigma ((if b then whd else whd_delta) env sigma term) with
       Prod (id,atom,ff) ->
-        if isRefX sigma (Lazy.force _False) ff then
+        if isRefX env sigma (Lazy.force _False) ff then
           let patts=patterns_of_constr b env sigma nrels atom in
               `Nrule patts
         else
           quantified_atom_of_constr b (EConstr.push_rel (RelDecl.LocalAssum (id,atom)) env) sigma (succ nrels) ff
-    | App (f,[|atom|]) when isRefX sigma (Lazy.force _not) f ->
+    | App (f,[|atom|]) when isRefX env sigma (Lazy.force _not) f ->
         let patts=patterns_of_constr b env sigma nrels atom in
               `Nrule patts
     | _ ->
@@ -183,7 +182,7 @@ let rec quantified_atom_of_constr b env sigma nrels term =
 let litteral_of_constr b env sigma term =
   match EConstr.kind sigma ((if b then whd else whd_delta) env sigma term) with
     | Prod (id,atom,ff) ->
-        if isRefX sigma (Lazy.force _False) ff then
+        if isRefX env sigma (Lazy.force _False) ff then
           match (atom_of_constr b env sigma atom) with
               `Eq(t,a,b) -> `Neq(t,a,b)
             | `Other(p) -> `Nother(p)
@@ -194,7 +193,7 @@ let litteral_of_constr b env sigma term =
             with Not_found ->
               `Other (decompose_term env sigma term)
           end
-    | App (f,[|atom|]) when isRefX sigma (Lazy.force _not) f ->
+    | App (f,[|atom|]) when isRefX env sigma (Lazy.force _not) f ->
           begin match (atom_of_constr b env sigma atom) with
               `Eq(t,a,b) -> `Neq(t,a,b)
             | `Other(p) -> `Nother(p)
@@ -250,45 +249,20 @@ let make_prb gls depth additional_terms b =
 
 (* indhyps builds the array of arrays of constructor hyps for (ind largs) *)
 
-let build_projection intype (cstr:pconstructor) special default gls=
-  let open Tacmach in
-  let ci= (snd(fst cstr)) in
-  let sigma = project gls in
-  let body=Equality.build_selector (pf_env gls) sigma ci (mkRel 1) intype special default in
-  let id=pf_get_new_id (Id.of_string "t") gls in
-  sigma, mkLambda(make_annot (Name id) Sorts.Relevant,intype,body)
+let fresh_id env id =
+  Namegen.next_ident_away id (Environ.ids_of_named_context_val @@ Environ.named_context_val env)
+
+let build_projection env sigma intype (cstr : pconstructor) special default =
+  let ci = (snd (fst cstr)) in
+  let body = Combinators.make_selector env sigma ~pos:ci ~special ~default (mkRel 1) intype in
+  let id = fresh_id env (Id.of_string "t") in
+  sigma, mkLambda (make_annot (Name id) ERelevance.relevant, intype, body)
 
 (* generate an adhoc tactic following the proof tree  *)
 
 let app_global f args k =
   Tacticals.pf_constr_of_global (Lazy.force f) >>= fun fc -> k (mkApp (fc, args))
 
-let rec gen_holes env sigma t n accu =
-  if Int.equal n 0 then (sigma, List.rev accu)
-  else match EConstr.kind sigma t with
-  | Prod (_, u, t) ->
-    let (sigma, ev) = Evarutil.new_evar env sigma u in
-    let t = EConstr.Vars.subst1 ev t in
-    gen_holes env sigma t (pred n) (ev :: accu)
-  | _ -> assert false
-
-let app_global_with_holes f args n =
-  Proofview.Goal.enter begin fun gl ->
-    Tacticals.pf_constr_of_global (Lazy.force f) >>= fun fc ->
-    let env = Proofview.Goal.env gl in
-    let concl = Proofview.Goal.concl gl in
-    Refine.refine ~typecheck:false begin fun sigma ->
-      let open Environ in
-      let fj = Retyping.get_judgment_of env sigma fc in
-      let argsj = Array.map (fun c -> Retyping.get_judgment_of env sigma c) args in
-      let sigma, fj = Typing.judge_of_apply env sigma fj argsj in
-      let (sigma, holes) = gen_holes env sigma fj.uj_type n [] in
-      let ans = applist (fj.uj_val, holes) in
-      let sigma = Typing.check env sigma ans concl in
-      (sigma, ans)
-    end
-  end
-
 let assert_before n c =
   Proofview.Goal.enter begin fun gl ->
     let evm, _ = Tacmach.pf_apply type_of gl c in
@@ -300,79 +274,97 @@ let refresh_type env evm ty =
   Evarsolve.refresh_universes ~status:Evd.univ_flexible ~refreshset:true
                               (Some false) env evm ty
 
-let type_and_refresh c k =
-  Proofview.Goal.enter begin fun gl ->
+let type_and_refresh c =
+  Proofview.Goal.enter_one ~__LOC__ begin fun gl ->
     let env = Proofview.Goal.env gl in
     let evm = Tacmach.project gl in
     (* XXX is get_type_of enough? *)
     let evm, ty = Typing.type_of env evm c in
     let evm, ty = refresh_type env evm ty in
-    Proofview.tclTHEN (Proofview.Unsafe.tclEVARS evm) (k ty)
+    Proofview.tclTHEN (Proofview.Unsafe.tclEVARS evm) (Proofview.tclUNIT ty)
   end
 
+let type_and_refresh_ env sigma c =
+  let sigma, ty = Typing.type_of env sigma c in
+  let sigma, ty = refresh_type env sigma ty in
+  sigma, ty
+
 let constr_of_term c = EConstr.of_constr (ATerm.constr c)
 
-let rec proof_tac p : unit Proofview.tactic =
+let app_global_ env sigma ref args =
+  let (sigma, c) = Evd.fresh_global env sigma (Lazy.force ref) in
+  Typing.checked_appvect env sigma c args
+
+(* Assumes ⊢ typ : Sort, ⊢ lhs : typ and ⊢ rhs : typ
+   and p is a reified proof of "@eq typ lhs rhs" *)
+let rec proof_term env sigma (typ, lhs, rhs) p = match p.p_rule with
+| Ax c ->
+  let c = EConstr.of_constr @@ constr_of_axiom c in
+  let sigma, expected = app_global_ env sigma _eq [|typ; lhs; rhs|] in
+  let sigma = Typing.check env sigma c expected in
+  sigma, c
+| SymAx c ->
+  let c = EConstr.of_constr @@ constr_of_axiom c in
+  let sigma, expected = app_global_ env sigma _eq [|typ; rhs; lhs|] in
+  let sigma = Typing.check env sigma c expected in
+  app_global_ env sigma _sym_eq [|typ; rhs; lhs; c|]
+| Refl t ->
+  let t = constr_of_term t in
+  app_global_ env sigma _refl_equal [|typ; t|]
+| Trans (p1, p2) ->
+  let t1 = constr_of_term p1.p_lhs in
+  let t2 = constr_of_term p1.p_rhs in
+  let t3 = constr_of_term p2.p_rhs in
+  let sigma, p1 = proof_term env sigma (typ, t1, t2) p1 in
+  let sigma, p2 = proof_term env sigma (typ, t2, t3) p2 in
+  app_global_ env sigma _trans_eq [|typ; t1; t2; t3; p1; p2|]
+| Congr (p1, p2) ->
+  (* p1 : ⊢ f = g : forall x : A, B *)
+  (* p2 : ⊢ t = u : A *)
+  let f = constr_of_term p1.p_lhs in
+  let g = constr_of_term p1.p_rhs in
+  let t = constr_of_term p2.p_lhs in
+  let u = constr_of_term p2.p_rhs in
+  let sigma, funty = type_and_refresh_ env sigma f in
+  let sigma, argty = type_and_refresh_ env sigma t in
+  let id = fresh_id env (Id.of_string "f") in
+  let appf = mkLambda (make_annot (Name id) ERelevance.relevant, funty, mkApp (mkRel 1, [|t|])) in
+  let sigma, p1 = proof_term env sigma (funty, f, g) p1 in
+  let sigma, p2 = proof_term env sigma (argty, t, u) p2 in
+  (* lemma1 : ⊢ f t = g t : B{t} *)
+  let sigma, lemma1 = app_global_ env sigma _f_equal [|funty; typ; appf; f; g; p1|] in
+  (* lemma2 : ⊢ g t = g u : B{t}, this only type-checks when B{t} ≡ B{u} *)
+  let sigma, lemma2 =
+    try app_global_ env sigma _f_equal [|argty; typ; g; t; u; p2|]
+    with e when CErrors.noncritical e ->
+      (* Fallback if ⊢ g t ≡ g u *)
+      begin match Evarconv.unify_delay env sigma (mkApp (g, [|t|])) (mkApp (g, [|u|])) with
+      | sigma ->
+        app_global_ env sigma _refl_equal [|typ; mkApp (g, [|t|])|]
+      | exception Evarconv.UnableToUnify _ ->
+        CErrors.user_err (Pp.str "I don't know how to handle dependent equality")
+      end
+  in
+  app_global_ env sigma _trans_eq [|typ; mkApp (f, [|t|]); mkApp (g, [|t|]); mkApp (g, [|u|]); lemma1; lemma2|]
+| Inject (prf, cstr, nargs, argind) ->
+  (* prf : ⊢ ci v = ci w : Ind(args) *)
+  let ti = constr_of_term prf.p_lhs in
+  let tj = constr_of_term prf.p_rhs in
+  let default = constr_of_term p.p_lhs in
+  let special = mkRel (1 + nargs - argind) in
+  let sigma, argty = type_and_refresh_ env sigma ti in
+  let sigma, proj = build_projection env sigma argty cstr special default in
+  let sigma, prf = proof_term env sigma (argty, ti, tj) prf in
+  app_global_ env sigma _f_equal [|argty; typ; proj; ti; tj; prf|]
+
+let proof_tac (typ, lhs, rhs) p : unit Proofview.tactic =
   Proofview.Goal.enter begin fun gl ->
-  match p.p_rule with
-      Ax c -> exact_check (EConstr.of_constr (constr_of_axiom c))
-    | SymAx c ->
-        let c = EConstr.of_constr (constr_of_axiom c) in
-        let l=constr_of_term p.p_lhs and
-            r=constr_of_term p.p_rhs in
-        type_and_refresh l (fun typ ->
-        app_global _sym_eq [|typ;r;l;c|] exact_check)
-    | Refl t ->
-        let lr = constr_of_term t in
-        type_and_refresh lr (fun typ ->
-        app_global _refl_equal [|typ;constr_of_term t|] exact_check)
-    | Trans (p1,p2)->
-        let t1 = constr_of_term p1.p_lhs and
-            t2 = constr_of_term p1.p_rhs and
-            t3 = constr_of_term p2.p_rhs in
-        type_and_refresh t2 (fun typ ->
-        let prf = app_global_with_holes _trans_eq [|typ;t1;t2;t3;|] 2 in
-          Tacticals.tclTHENS prf [(proof_tac p1);(proof_tac p2)])
-    | Congr (p1,p2)->
-        let tf1=constr_of_term p1.p_lhs
-        and tx1=constr_of_term p2.p_lhs
-        and tf2=constr_of_term p1.p_rhs
-        and tx2=constr_of_term p2.p_rhs in
-        type_and_refresh tf1 (fun typf ->
-        type_and_refresh tx1 (fun typx ->
-        type_and_refresh (mkApp (tf1,[|tx1|])) (fun typfx ->
-        let id = Tacmach.pf_get_new_id (Id.of_string "f") gl in
-        let appx1 = mkLambda(make_annot (Name id) Sorts.Relevant,typf,mkApp(mkRel 1,[|tx1|])) in
-        let lemma1 = app_global_with_holes _f_equal [|typf;typfx;appx1;tf1;tf2|] 1 in
-        let lemma2 = app_global_with_holes _f_equal [|typx;typfx;tf2;tx1;tx2|] 1 in
-        let prf =
-          app_global_with_holes _trans_eq
-                [|typfx;
-                  mkApp(tf1,[|tx1|]);
-                  mkApp(tf2,[|tx1|]);
-                  mkApp(tf2,[|tx2|])|] 2 in
-          Tacticals.tclTHENS prf
-            [Tacticals.tclTHEN lemma1 (proof_tac p1);
-             Tacticals.tclFIRST
-               [Tacticals.tclTHEN lemma2 (proof_tac p2);
-                reflexivity;
-                Tacticals.tclZEROMSG
-                    (Pp.str
-                       "I don't know how to handle dependent equality")]])))
-  | Inject (prf,cstr,nargs,argind) ->
-         let ti=constr_of_term prf.p_lhs in
-         let tj=constr_of_term prf.p_rhs in
-         let default=constr_of_term p.p_lhs in
-         let special=mkRel (1+nargs-argind) in
-         type_and_refresh ti (fun intype ->
-         type_and_refresh default (fun outtype ->
-         let sigma, proj =
-           build_projection intype cstr special default gl
-         in
-         let injt=
-           app_global_with_holes _f_equal [|intype;outtype;proj;ti;tj|] 1 in
-         Tacticals.tclTHEN (Proofview.Unsafe.tclEVARS sigma)
-                               (Tacticals.tclTHEN injt (proof_tac prf))))
+    let env = Proofview.Goal.env gl in
+    let sigma = Proofview.Goal.sigma gl in
+    let concl = Proofview.Goal.concl gl in
+    let sigma, p = proof_term env sigma (typ, lhs, rhs) p in
+    let sigma = Typing.check env sigma p concl in
+    Proofview.Unsafe.tclEVARS sigma <*> exact_no_check p
   end
 
 let refute_tac c t1 t2 p =
@@ -383,8 +375,8 @@ let refute_tac c t1 t2 p =
   let k intype =
     let neweq= app_global _eq [|intype;tt1;tt2|] in
     Tacticals.tclTHENS (neweq (assert_before (Name hid)))
-      [proof_tac p; simplest_elim false_t]
-  in type_and_refresh tt1 k
+      [proof_tac (intype, tt1, tt2) p; simplest_elim false_t]
+  in type_and_refresh tt1 >>= k
   end
 
 let refine_exact_check c =
@@ -400,11 +392,11 @@ let convert_to_goal_tac c t1 t2 p =
     let neweq= app_global _eq [|sort;tt1;tt2|] in
     let e = Tacmach.pf_get_new_id (Id.of_string "e") gl in
     let x = Tacmach.pf_get_new_id (Id.of_string "X") gl in
-    let identity=mkLambda (make_annot (Name x) Sorts.Relevant,sort,mkRel 1) in
+    let identity=mkLambda (make_annot (Name x) ERelevance.relevant,sort,mkRel 1) in
     let endt = app_global _eq_rect [|sort; tt1; identity; mkVar c; tt2; mkVar e|] in
     Tacticals.tclTHENS (neweq (assert_before (Name e)))
-                           [proof_tac p; endt refine_exact_check]
-  in type_and_refresh tt2 k
+                           [proof_tac (sort, tt1, tt2) p; endt refine_exact_check]
+  in type_and_refresh tt2 >>= k
   end
 
 let convert_to_hyp_tac c1 t1 c2 t2 p =
@@ -429,21 +421,15 @@ let discriminate_tac cstru p =
     let neweq=app_global _eq [|intype;lhs;rhs|] in
     Tacticals.tclTHEN (Proofview.Unsafe.tclEVARS evm)
                           (Tacticals.tclTHENS (neweq (assert_before (Name hid)))
-      [proof_tac p; Equality.discrHyp hid])
+      [proof_tac (intype, lhs, rhs) p; Equality.discrHyp hid])
   end
 
 (* wrap everything *)
 
-let build_term_to_complete uf pac =
-  let cinfo = get_constructor_info uf pac.cnode in
-  let real_args = List.rev_map (fun i -> constr_of_term (aterm uf i)) pac.args in
-  let (kn, u) = cinfo.ci_constr in
-  (applist (mkConstructU (kn, EInstance.make u), real_args), pac.arity)
-
 let cc_tactic depth additional_terms b =
   Proofview.Goal.enter begin fun gl ->
     let sigma = Tacmach.project gl in
-    Coqlib.(check_required_library logic_module_name);
+    Rocqlib.(check_required_library logic_module_name);
     let _ = debug_congruence (fun () -> Pp.str "Reading goal ...") in
     let state = make_prb gl depth additional_terms b in
     let _ = debug_congruence (fun () -> Pp.str "Problem built, solving ...") in
@@ -453,19 +439,18 @@ let cc_tactic depth additional_terms b =
     match sol with
       None -> Tacticals.tclFAIL (str (if b then "simple congruence failed" else "congruence failed"))
     | Some reason ->
-      debug_congruence (fun () -> Pp.str "Goal solved, generating proof ...");
+      Proofview.tclORELSE (debug_congruence (fun () -> Pp.str "Goal solved, generating proof ...");
       match reason with
         Discrimination (i,ipac,j,jpac) ->
         let p=build_proof (Tacmach.pf_env gl) sigma uf (`Discr (i,ipac,j,jpac)) in
         let cstr=(get_constructor_info uf ipac.cnode).ci_constr in
         discriminate_tac cstr p
-      | Incomplete ->
+      | Incomplete terms_to_complete ->
         let open Glob_term in
         let env = Proofview.Goal.env gl in
-        let terms_to_complete = List.map (build_term_to_complete uf) (epsilons uf) in
-        let hole = DAst.make @@ GHole (Evar_kinds.InternalHole, Namegen.IntroAnonymous, None) in
+        let hole = DAst.make @@ GHole (GInternalHole) in
         let pr_missing (c, missing) =
-          let c = Detyping.detype Detyping.Now ~lax:true false Id.Set.empty env sigma c in
+          let c = Detyping.detype Detyping.Now env sigma c in
           let holes = List.init missing (fun _ -> hole) in
           Printer.pr_glob_constr_env env sigma (DAst.make @@ GApp (c, holes))
         in
@@ -488,15 +473,26 @@ let cc_tactic depth additional_terms b =
         let p=build_proof env sigma uf (`Prove (dis.lhs,dis.rhs)) in
         let ta=aterm uf dis.lhs and tb=aterm uf dis.rhs in
         match dis.rule with
-          Goal -> proof_tac p
+        | Goal ->
+          let lhs = constr_of_term ta in
+          let rhs = constr_of_term tb in
+          let sigma, typ = type_and_refresh_ env sigma lhs in
+          Proofview.Unsafe.tclEVARS sigma <*> proof_tac (typ, lhs, rhs) p
         | Hyp id -> refute_tac (EConstr.of_constr id) ta tb p
         | HeqG id ->
           convert_to_goal_tac id ta tb p
         | HeqnH (ida,idb) ->
-          convert_to_hyp_tac ida ta idb tb p
+          convert_to_hyp_tac ida ta idb tb p)
+      begin function (e, info) -> match e with
+        | Tactics.NotConvertible ->
+          Tacticals.tclFAIL
+            (str (if b then "simple congruence failed" else "congruence failed") ++
+             str " (cannot build a well-typed proof)")
+        | e -> Proofview.tclZERO ~info e
+      end
   end
 
-let id t = mkLambda (make_annot Anonymous Sorts.Relevant, t, mkRel 1)
+let id t = mkLambda (make_annot Anonymous ERelevance.relevant, t, mkRel 1)
 
 (* convertible to (not False) -> P -> not P *)
 let mk_neg_ty ff t nt =
@@ -504,8 +500,8 @@ let mk_neg_ty ff t nt =
 
 (* proof of ((not False) -> P -> not P) -> not P *)
 let mk_neg_tm ff t nt =
-  mkLambda (make_annot Anonymous Sorts.Relevant, mk_neg_ty ff t nt,
-    mkLambda (make_annot Anonymous Sorts.Relevant, t,
+  mkLambda (make_annot Anonymous ERelevance.relevant, mk_neg_ty ff t nt,
+    mkLambda (make_annot Anonymous ERelevance.relevant, t,
       mkApp (mkRel 2,[|id ff; mkRel 1; mkRel 1|])))
 
 (* for [simple congruence] process conclusion (not P) *)
@@ -516,8 +512,8 @@ let negative_concl_introf =
     let concl = Proofview.Goal.concl gl in
     let nt = whd env sigma concl in
     match EConstr.kind sigma nt with
-      Prod (_,_,ff) when isRefX sigma (Lazy.force _False) ff -> introf
-    | App (f,[|t|]) when isRefX sigma (Lazy.force _not) f ->
+      Prod (_,_,ff) when isRefX env sigma (Lazy.force _False) ff -> introf
+    | App (f,[|t|]) when isRefX env sigma (Lazy.force _not) f ->
         Tacticals.pf_constr_of_global (Lazy.force _False) >>= fun ff ->
         Refine.refine ~typecheck:true begin fun sigma ->
           let sigma, e = Evarutil.new_evar env sigma (mk_neg_ty ff t nt) in sigma, (mkApp (mk_neg_tm ff t nt, [|e|]))
@@ -526,12 +522,14 @@ let negative_concl_introf =
   end
 
 let congruence_tac depth l =
+  let depth = Option.default 1000 depth in
   Tacticals.tclTHEN
     (Tacticals.tclREPEAT (Tacticals.tclFIRST [intro; Tacticals.tclTHEN whd_in_concl intro]))
     (cc_tactic depth l false)
 
 
 let simple_congruence_tac depth l =
+  let depth = Option.default 1000 depth in
   Tacticals.tclTHENLIST [
     Tacticals.tclREPEAT intro;
     negative_concl_introf;
@@ -562,6 +560,7 @@ let mk_eq f c1 c2 k =
 let f_equal =
   Proofview.Goal.enter begin fun gl ->
     let concl = Proofview.Goal.concl gl in
+    let env = Proofview.Goal.env gl in
     let sigma = Tacmach.project gl in
     let cut_eq c1 c2 =
         Tacticals.tclTHENS
@@ -570,11 +569,11 @@ let f_equal =
     in
     Proofview.tclORELSE
       begin match EConstr.kind sigma concl with
-      | App (r,[|_;t;t'|]) when isRefX sigma (Lazy.force _eq) r ->
+      | App (r,[|_;t;t'|]) when isRefX env sigma (Lazy.force _eq) r ->
           begin match EConstr.kind sigma t, EConstr.kind sigma t' with
           | App (f,v), App (f',v') when Int.equal (Array.length v) (Array.length v') ->
               let rec cuts i =
-                if i < 0 then Tacticals.tclTRY (congruence_tac 1000 [])
+                if i < 0 then Tacticals.tclTRY (congruence_tac None [])
                 else Tacticals.tclTHENFIRST (cut_eq v.(i) v'.(i)) (cuts (i-1))
               in cuts (Array.length v - 1)
           | _ -> Proofview.tclUNIT ()
