@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -8,7 +8,7 @@
 (*         *     (see LICENSE file for the text of the license)         *)
 (************************************************************************)
 
-(*s Extraction from Coq terms to Miniml. *)
+(*s Extraction from Rocq terms to Miniml. *)
 
 open Names
 open Declarations
@@ -16,9 +16,9 @@ open Environ
 open Evd
 open Miniml
 
-val extract_constant : env -> Constant.t -> Opaqueproof.opaque constant_body -> ml_decl
+val extract_constant : Global.indirect_accessor -> env -> Constant.t -> constant_body -> ml_decl
 
-val extract_constant_spec : env -> Constant.t -> 'a constant_body -> ml_spec
+val extract_constant_spec : env -> Constant.t -> ('a, 'b) pconstant_body -> ml_spec
 
 (** For extracting "module ... with ..." declaration *)
 
@@ -26,8 +26,7 @@ val extract_with_type :
   env -> evar_map -> EConstr.t -> ( Id.t list * ml_type ) option
 
 val extract_fixpoint :
-  env -> evar_map -> Constant.t array ->
-    (EConstr.t, EConstr.types) Constr.prec_declaration -> ml_decl
+  env -> evar_map -> Constant.t array -> EConstr.rec_declaration -> ml_decl
 
 val extract_inductive : env -> MutInd.t -> ml_ind
 
