@@ -1,7 +1,7 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
-(*  v      *   INRIA, CNRS and contributors - Copyright 1999-2019       *)
-(* <O___,, *       (see CREDITS file for the list of authors)           *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
+(*  v      *         Copyright INRIA, CNRS and contributors             *)
+(* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
 (*    //   *    This file is distributed under the terms of the         *)
 (*         *     GNU Lesser General Public License Version 2.1          *)
@@ -124,10 +124,12 @@ and ml_ast =
   | MLfix    of int * Id.t array * ml_ast array
   | MLexn    of string
   | MLdummy  of kill_reason
-  | MLaxiom
+  | MLaxiom  of string
   | MLmagic  of ml_ast
   | MLuint   of Uint63.t
   | MLfloat  of Float64.t
+  | MLstring of Pstring.t
+  | MLparray of ml_ast array * ml_ast
 
 and ml_pattern =
   | Pcons   of GlobRef.t * ml_pattern list
