@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -18,9 +18,9 @@ open Declarations
 
 module RelDecl = Context.Rel.Declaration
 
-let qflag=ref true
-
-let red_flags=ref CClosure.betaiotazeta
+type flags = {
+  reds : RedFlags.reds;
+}
 
 let (=?) f g i1 i2 j1 j2=
   let c=f i1 i2 in
@@ -34,7 +34,61 @@ type ('a,'b) sum = Left of 'a | Right of 'b
 
 type counter = bool -> metavariable
 
-exception Is_atom of constr
+module Env :
+sig
+type t
+type uid
+val empty : t
+val add : flags -> Environ.env -> Evd.evar_map -> EConstr.t -> t -> t * uid
+val find : uid -> t -> EConstr.t
+val repr : uid -> int
+val hole : uid
+end =
+struct
+
+module CM = Map.Make(Constr)
+
+type t = {
+  max_uid : int;
+  repr : int CM.t;
+  data : Constr.t Int.Map.t;
+}
+
+type uid = int
+
+let empty = {
+  max_uid = 1; (* uid 0 is reserved for Meta 1 *)
+  repr = CM.singleton (Constr.mkMeta 1) 0;
+  data = Int.Map.singleton 0 (Constr.mkMeta 1);
+}
+
+let hole = 0
+
+(* This is nonsense, but backwards compatibility mandates it *)
+let add flags env sigma c e =
+  let c = Reductionops.clos_norm_flags flags.reds env sigma c in
+  let c = EConstr.to_constr ~abort_on_undefined_evars:false sigma c in
+  match CM.find_opt c e.repr with
+  | Some i -> e, i
+  | None ->
+    let i = e.max_uid in
+    let e = {
+      max_uid = i + 1;
+      repr = CM.add c i e.repr;
+      data = Int.Map.add i c e.data;
+    } in
+    e, i
+
+let find id e = EConstr.of_constr (Int.Map.find id e.data)
+
+let repr id = id
+
+end
+
+type atom = { atom : Env.uid }
+let repr_atom state a = Env.find a.atom state
+let compare_atom a1 a2 =
+  Int.compare (Env.repr a1.atom) (Env.repr a2.atom)
 
 let meta_succ m = m+1
 
@@ -45,27 +99,25 @@ let rec nb_prod_after n c=
     | _ -> 0
 
 let construct_nhyps env ind =
+  let ind = on_snd EInstance.make ind in
   let nparams = (fst (Global.lookup_inductive (fst ind))).mind_nparams in
   let constr_types = Inductiveops.arities_of_constructors env ind in
-  let hyp = nb_prod_after nparams in
-    Array.map hyp constr_types
+  let hyp c = nb_prod_after nparams (EConstr.Unsafe.to_constr c) in
+  Array.map hyp constr_types
 
 (* indhyps builds the array of arrays of constructor hyps for (ind largs)*)
 let ind_hyps env sigma nevar ind largs =
+  let ind = on_snd EInstance.make ind in
   let types= Inductiveops.arities_of_constructors env ind in
   let myhyps t =
-    let t = EConstr.of_constr t in
     let nparam_decls = Context.Rel.length (fst (Global.lookup_inductive (fst ind))).mind_params_ctxt in
-    let t1=Termops.prod_applist_assum sigma nparam_decls t largs in
-    let t2=snd (decompose_prod_n_assum sigma nevar t1) in
-      fst (decompose_prod_assum sigma t2) in
+    let t1=Termops.prod_applist_decls sigma nparam_decls t largs in
+    let t2=snd (decompose_prod_n_decls sigma nevar t1) in
+      fst (decompose_prod_decls sigma t2) in
     Array.map myhyps types
 
-let special_nf env sigma t =
-  Reductionops.clos_norm_flags !red_flags env sigma t
-
-let special_whd env sigma t =
-  Reductionops.clos_whd_flags !red_flags env sigma t
+let special_whd ~flags env sigma t =
+  Reductionops.clos_whd_flags flags.reds env sigma t
 
 type kind_of_formula=
     Arrow of constr*constr
@@ -74,13 +126,19 @@ type kind_of_formula=
   | Or of pinductive*constr list*bool
   | Exists of pinductive*constr list
   | Forall of constr*constr
-  | Atom of constr
+  | Atom of EConstr.t
 
 let pop t = Vars.lift (-1) t
 
-let kind_of_formula env sigma term =
-  let normalize = special_nf env sigma in
-  let cciterm = special_whd env sigma term in
+let fresh_atom ~flags state env sigma atm =
+  let st, uid = Env.add flags env sigma atm !state in
+  let () = state := st in
+  { atom = uid }
+
+let hole_atom = { atom = Env.hole }
+
+let kind_of_formula ~flags env sigma term =
+  let cciterm = special_whd ~flags env sigma term in
     match match_with_imp_term env sigma cciterm with
         Some (a,b)-> Arrow (a, pop b)
       |_->
@@ -115,23 +173,21 @@ let kind_of_formula env sigma term =
                           let (ind, u) = EConstr.destInd sigma i in
                           let u = EConstr.EInstance.kind sigma u in
                           Exists((ind, u), l)
-                      |_-> Atom (normalize cciterm)
+                      |_->
+                        Atom cciterm
 
-type atoms = {positive:constr list;negative:constr list}
+type atoms = {positive:atom list;negative:atom list}
 
-type side = Hyp | Concl | Hint
+type _ side =
+| Hyp : bool -> [ `Hyp ] side (* true if treated as hint *)
+| Concl : [ `Goal ] side
 
-let no_atoms = (false,{positive=[];negative=[]})
-
-let dummy_id=GlobRef.VarRef (Id.of_string "_") (* "_" cannot be parsed *)
-
-let build_atoms env sigma metagen side cciterm =
+let build_atoms (type a) ~flags state env sigma metagen (side : a side) cciterm =
   let trivial =ref false
   and positive=ref []
   and negative=ref [] in
-  let normalize=special_nf env sigma in
   let rec build_rec subst polarity cciterm=
-    match kind_of_formula env sigma cciterm with
+    match kind_of_formula ~flags env sigma cciterm with
         False(_,_)->if not polarity then trivial:=true
       | Arrow (a,b)->
           build_rec subst (not polarity) a;
@@ -139,7 +195,7 @@ let build_atoms env sigma metagen side cciterm =
       | And(i,l,b) | Or(i,l,b)->
           if b then
             begin
-              let unsigned=normalize (substnl subst 0 cciterm) in
+              let unsigned= fresh_atom ~flags state env sigma (substnl subst 0 cciterm) in
                 if polarity then
                   positive:= unsigned :: !positive
                 else
@@ -164,17 +220,17 @@ let build_atoms env sigma metagen side cciterm =
           let var=mkMeta (metagen true) in
             build_rec (var::subst) polarity b
       | Atom t->
-          let unsigned=substnl subst 0 t in
-            if not (isMeta sigma unsigned) then (* discarding wildcard atoms *)
+          let unsigned = fresh_atom ~flags state env sigma (substnl subst 0 t) in
+            if not (isMeta sigma (repr_atom !state unsigned)) then (* discarding wildcard atoms *)
               if polarity then
                 positive:= unsigned :: !positive
               else
                 negative:= unsigned :: !negative in
     begin
       match side with
-          Concl    -> build_rec [] true cciterm
-        | Hyp      -> build_rec [] false cciterm
-        | Hint     ->
+          Concl     -> build_rec [] true cciterm
+        | Hyp false -> build_rec [] false cciterm
+        | Hyp true  ->
             let rels,head=decompose_prod sigma cciterm in
             let subst=List.rev_map (fun _->mkMeta (metagen true)) rels in
               build_rec subst false head;trivial:=false (* special for hints *)
@@ -206,26 +262,40 @@ type left_pattern=
   | Lor of pinductive
   | Lforall of metavariable*constr*bool
   | Lexists of pinductive
-  | LA of constr*left_arrow_pattern
+  | LA of atom*left_arrow_pattern
 
-type t={id:GlobRef.t;
-        constr:constr;
-        pat:(left_pattern,right_pattern) sum;
-        atoms:atoms}
+type _ identifier =
+| GoalId : [ `Goal ] identifier
+| FormulaId : GlobRef.t -> [ `Hyp ] identifier
 
-let build_formula env sigma side nam typ metagen=
-  let normalize = special_nf env sigma in
+let goal_id = GoalId
+let formula_id env gr = FormulaId (Environ.QGlobRef.canonize env gr)
+
+type _ pattern =
+| LeftPattern : left_pattern -> [ `Hyp ] pattern
+| RightPattern : right_pattern -> [ `Goal ] pattern
+
+type 'a t = {
+  id : 'a identifier;
+  constr : atom;
+  pat : 'a pattern;
+  atoms : atoms;
+}
+
+type any_formula = AnyFormula : 'a t -> any_formula
+
+exception Is_atom of EConstr.t
+
+let build_formula (type a) ~flags state env sigma (side : a side) (nam : a identifier) typ metagen : Env.t * (a t, atom) sum =
     try
+      let state = ref state in
       let m=meta_succ(metagen false) in
-      let trivial,atoms=
-        if !qflag then
-          build_atoms env sigma metagen side typ
-        else no_atoms in
-      let pattern=
+      let trivial, atoms = build_atoms ~flags state env sigma metagen side typ in
+      let pattern : a pattern =
         match side with
             Concl ->
               let pat=
-                match kind_of_formula env sigma typ with
+                match kind_of_formula ~flags env sigma typ with
                     False(_,_)        -> Rfalse
                   | Atom a       -> raise (Is_atom a)
                   | And(_,_,_)        -> Rand
@@ -235,27 +305,25 @@ let build_formula env sigma side nam typ metagen=
                         Rexists(m,d,trivial)
                   | Forall (_,a) -> Rforall
                   | Arrow (a,b) -> Rarrow in
-                Right pat
-          | _ ->
+                RightPattern pat
+          | Hyp _ ->
               let pat=
-                match kind_of_formula env sigma typ with
+                match kind_of_formula ~flags env sigma typ with
                     False(i,_)        ->  Lfalse
                   | Atom a       ->  raise (Is_atom a)
                   | And(i,_,b)         ->
-                      if b then
-                        let nftyp=normalize typ in raise (Is_atom nftyp)
+                      if b then raise (Is_atom typ)
                       else Land i
                   | Or(i,_,b)          ->
-                      if b then
-                        let nftyp=normalize typ in raise (Is_atom nftyp)
+                      if b then raise (Is_atom typ)
                       else Lor i
                   | Exists (ind,_) ->  Lexists ind
                   | Forall (d,_) ->
                       Lforall(m,d,trivial)
                   | Arrow (a,b) ->
-                      let nfa=normalize a in
+                    let nfa = fresh_atom ~flags state env sigma a in
                         LA (nfa,
-                            match kind_of_formula env sigma a with
+                            match kind_of_formula ~flags env sigma a with
                                 False(i,l)-> LLfalse(i,l)
                               | Atom t->     LLatom
                               | And(i,l,_)-> LLand(i,l)
@@ -263,11 +331,12 @@ let build_formula env sigma side nam typ metagen=
                               | Arrow(a,c)-> LLarrow(a,c,b)
                               | Exists(i,l)->LLexists(i,l)
                               | Forall(_,_)->LLforall a) in
-                Left pat
+                LeftPattern pat
       in
-        Left {id=nam;
-              constr=normalize typ;
-              pat=pattern;
-              atoms=atoms}
-    with Is_atom a-> Right a (* already in nf *)
+      let typ = fresh_atom ~flags state env sigma typ in
+      !state, Left { id = nam; constr = typ; pat = pattern; atoms = atoms}
+    with Is_atom a ->
+      let state = ref state in
+      let a = fresh_atom ~flags state env sigma a in
+      !state, Right a (* already in nf *)
 
