@@ -1,7 +1,7 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
-(*  v      *   INRIA, CNRS and contributors - Copyright 1999-2019       *)
-(* <O___,, *       (see CREDITS file for the list of authors)           *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
+(*  v      *         Copyright INRIA, CNRS and contributors             *)
+(* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
 (*    //   *    This file is distributed under the terms of the         *)
 (*         *     GNU Lesser General Public License Version 2.1          *)
@@ -12,9 +12,9 @@ open Names
 open Constr
 open EConstr
 
-val qflag : bool ref
-
-val red_flags: CClosure.RedFlags.reds ref
+type flags = {
+  reds : RedFlags.reds;
+}
 
 val (=?) : ('a -> 'a -> int) -> ('b -> 'b -> int) ->
   'a -> 'a -> 'b -> 'b -> int
@@ -31,14 +31,23 @@ val construct_nhyps : Environ.env -> pinductive -> int array
 val ind_hyps : Environ.env -> Evd.evar_map -> int -> pinductive ->
   constr list -> EConstr.rel_context array
 
-type atoms = {positive:constr list;negative:constr list}
+module Env :
+sig
+  type t
+  val empty : t
+end
 
-type side = Hyp | Concl | Hint
+type atom
 
-val dummy_id: GlobRef.t
+val hole_atom : atom
+val repr_atom : Env.t -> atom -> EConstr.t
+val compare_atom : atom -> atom -> int
 
-val build_atoms : Environ.env -> Evd.evar_map -> counter ->
-  side -> constr -> bool * atoms
+type atoms = { positive:atom list; negative:atom list }
+
+type _ side =
+| Hyp : bool -> [ `Hyp ] side (* true if treated as hint *)
+| Concl : [ `Goal ] side
 
 type right_pattern =
     Rarrow
@@ -63,15 +72,26 @@ type left_pattern=
   | Lor of pinductive
   | Lforall of metavariable*constr*bool
   | Lexists of pinductive
-  | LA of constr*left_arrow_pattern
+  | LA of atom*left_arrow_pattern
 
-type t={id: GlobRef.t;
-        constr: constr;
-        pat: (left_pattern,right_pattern) sum;
+type _ identifier = private
+| GoalId : [ `Goal ] identifier
+| FormulaId : GlobRef.t -> [ `Hyp ] identifier
+
+val goal_id : [ `Goal ] identifier
+val formula_id : Environ.env -> GlobRef.t -> [ `Hyp ] identifier
+
+type _ pattern =
+| LeftPattern : left_pattern -> [ `Hyp ] pattern
+| RightPattern : right_pattern -> [ `Goal ] pattern
+
+type 'a t = private {
+        id: 'a identifier;
+        constr: atom;
+        pat: 'a pattern;
         atoms: atoms}
 
-(*exception Is_atom of constr*)
-
-val build_formula : Environ.env -> Evd.evar_map -> side -> GlobRef.t -> types ->
-  counter -> (t,types) sum
+type any_formula = AnyFormula : 'a t -> any_formula
 
+val build_formula : flags:flags -> Env.t -> Environ.env -> Evd.evar_map -> 'a side -> 'a identifier -> types ->
+  counter -> Env.t * ('a t, atom) sum
