@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -47,11 +47,12 @@ let wrap ~flags n b continue seq =
               List.exists (occur_var_in_decl env sigma id) ctx then
                 (aux (i-1) q (nd::ctx))
             else
-              add_formula ~flags env sigma Hyp (GlobRef.VarRef id) (NamedDecl.get_type nd) (aux (i-1) q (nd::ctx)) in
+              add_formula ~flags ~hint:false env sigma (GlobRef.VarRef id) (NamedDecl.get_type nd) (aux (i-1) q (nd::ctx)) in
   let seq1=aux n nc [] in
-  let seq2=if b then
-    add_formula ~flags env sigma Concl dummy_id (pf_concl gls) seq1 else seq1 in
-    continue seq2
+  let seq2 =
+    if b then add_concl ~flags env sigma (pf_concl gls) seq1 else seq1
+  in
+  continue seq2
   end
 
 let clear_global=function
@@ -60,12 +61,11 @@ let clear_global=function
 
 (* connection rules *)
 
-let axiom_tac t seq =
+let axiom_tac seq =
   Proofview.Goal.enter begin fun gl ->
-  try
-    pf_constr_of_global (find_left (project gl) t seq) >>= fun c ->
-    exact_no_check c
-  with Not_found -> tclFAIL (Pp.str "No axiom link")
+  match Sequent.find_goal (project gl) seq with
+  | gr -> pf_constr_of_global gr >>= fun c -> exact_no_check c
+  | exception Not_found -> tclFAIL (Pp.str "No axiom link")
   end
 
 let ll_atom_tac ~flags a backtrack id continue seq =
@@ -80,7 +80,7 @@ let ll_atom_tac ~flags a backtrack id continue seq =
           gr >>= fun gr ->
           pf_constr_of_global gr >>= fun left ->
           pf_constr_of_global id >>= fun id ->
-            generalize [(mkApp(id, [|left|]))]);
+            Generalize.generalize [(mkApp(id, [|left|]))]);
          clear_global id;
          intro])
     (wrap ~flags 1 false continue seq) backtrack
@@ -153,7 +153,7 @@ let ll_ind_tac ~flags (ind,u as indu) largs backtrack id continue seq =
        let newhyps idc =List.init lp (myterm idc) in
          tclIFTHENELSE
            (tclTHENLIST
-              [(pf_constr_of_global id >>= fun idc -> generalize (newhyps idc));
+              [(pf_constr_of_global id >>= fun idc -> Generalize.generalize (newhyps idc));
                clear_global id;
                tclDO lp intro])
            (wrap ~flags lp false continue seq) backtrack
@@ -162,9 +162,9 @@ let ll_ind_tac ~flags (ind,u as indu) largs backtrack id continue seq =
 let ll_arrow_tac ~flags a b c backtrack id continue seq=
   let open EConstr in
   let open Vars in
-  let cc=mkProd(Context.make_annot Anonymous Sorts.Relevant,a,(lift 1 b)) in
-  let d idc = mkLambda (Context.make_annot Anonymous Sorts.Relevant,b,
-                  mkApp (idc, [|mkLambda (Context.make_annot Anonymous Sorts.Relevant,(lift 1 a),(mkRel 2))|])) in
+  let cc=mkProd(Context.make_annot Anonymous ERelevance.relevant,a,(lift 1 b)) in
+  let d idc = mkLambda (Context.make_annot Anonymous ERelevance.relevant,b,
+                  mkApp (idc, [|mkLambda (Context.make_annot Anonymous ERelevance.relevant,(lift 1 a),(mkRel 2))|])) in
     tclORELSE
       (tclTHENS (cut c)
          [tclTHENLIST
@@ -174,7 +174,7 @@ let ll_arrow_tac ~flags a b c backtrack id continue seq=
           tclTHENS (cut cc)
             [(pf_constr_of_global id >>= fun c -> exact_no_check c);
              tclTHENLIST
-               [(pf_constr_of_global id >>= fun idc -> generalize [d idc]);
+               [(pf_constr_of_global id >>= fun idc -> Generalize.generalize [d idc]);
                 clear_global id;
                 introf;
                 introf;
@@ -189,10 +189,7 @@ let forall_tac ~flags backtrack continue seq=
        (tclORELSE
           (tclTHEN introf (tclCOMPLETE (wrap ~flags 0 true continue seq)))
           backtrack))
-    (if flags.qflag then
-       tclFAIL (Pp.str "reversible in 1st order mode")
-     else
-       backtrack)
+    (tclFAIL (Pp.str "reversible in 1st order mode"))
 
 let left_exists_tac ~flags ind backtrack id continue seq =
   Proofview.Goal.enter begin fun gl ->
@@ -215,7 +212,7 @@ let ll_forall_tac ~flags prod backtrack id continue seq=
               let open EConstr in
               let id0 = List.nth (pf_ids_of_hyps gls) 0 in
               let term=mkApp(idc,[|mkVar(id0)|]) in
-              tclTHEN (generalize [term]) (clear [id0])
+              tclTHEN (Generalize.generalize [term]) (clear [id0])
            end);
            clear_global id;
            intro;
