@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -12,11 +12,14 @@ open Util
 open Names
 open Constr
 open EConstr
-open Tacmach.New
+open Tacmach
 open Tactics
-open Tacticals.New
+open Tacticals
 open Indfun_common
 
+(* funind doesn't support univ poly *)
+open UnsafeMonomorphic
+
 (***********************************************)
 
 (* [revert_graph kn post_tac hid] transforme an hypothesis [hid] having type Ind(kn,num) t1 ... tn res
@@ -27,12 +30,13 @@ open Indfun_common
 *)
 let revert_graph kn post_tac hid =
   Proofview.Goal.enter (fun gl ->
+      let env = Proofview.Goal.env gl in
       let sigma = project gl in
       let typ = pf_get_hyp_typ hid gl in
       match EConstr.kind sigma typ with
       | App (i, args) when isInd sigma i ->
         let ((kn', num) as ind'), u = destInd sigma i in
-        if MutInd.equal kn kn' then
+        if Environ.QMutInd.equal env kn kn' then
           (* We have generated a graph hypothesis so that we must change it if we can *)
           let info =
             match find_Function_of_graph ind' with
@@ -50,7 +54,7 @@ let revert_graph kn post_tac hid =
           | Some f_complete ->
             let f_args, res = Array.chop (Array.length args - 1) args in
             tclTHENLIST
-              [ generalize
+              [ Generalize.generalize
                   [ applist
                       ( mkConst f_complete
                       , Array.to_list f_args @ [res.(0); mkVar hid] ) ]
@@ -92,15 +96,15 @@ let functional_inversion kn hid fconst f_correct =
             ((fun hid -> intros_symmetry (Locusops.onHyp hid)), f_args, args.(2))
           | _, App (f, f_args) when EConstr.eq_constr sigma f fconst ->
             ((fun hid -> tclIDTAC), f_args, args.(1))
-          | _ -> ((fun hid -> tclFAIL 1 Pp.(mt ())), [||], args.(2))
+          | _ -> ((fun hid -> tclFAILn 1 Pp.(mt ())), [||], args.(2))
         in
         tclTHENLIST
           [ pre_tac hid
-          ; generalize
+          ; Generalize.generalize
               [applist (f_correct, Array.to_list f_args @ [res; mkVar hid])]
           ; clear [hid]
           ; Simple.intro hid
-          ; Inv.inv Inv.FullInversion None (Tactypes.NamedHyp hid)
+          ; Inv.inv Inv.FullInversion None (Tactypes.NamedHyp (CAst.make hid))
           ; Proofview.Goal.enter (fun gl ->
                 let new_ids =
                   List.filter
@@ -108,7 +112,7 @@ let functional_inversion kn hid fconst f_correct =
                     (pf_ids_of_hyps gl)
                 in
                 tclMAP (revert_graph kn pre_tac) (hid :: new_ids)) ]
-      | _ -> tclFAIL 1 Pp.(mt ()))
+      | _ -> tclFAILn 1 Pp.(mt ()))
 
 let invfun qhyp f =
   let f =
