@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -163,14 +163,14 @@ let string_of_genarg_arg (ArgumentType arg) =
           end
         | _ -> default
 
-  let pr_with_occurrences pr c = Ppred.pr_with_occurrences pr keyword c
+  let pr_with_occurrences prvar pr c = Ppred.pr_with_occurrences prvar pr keyword c
   let pr_red_expr env sigma pr c = Ppred.pr_red_expr_env env sigma pr keyword c
 
-  let pr_may_eval env sigma test prc prlc pr2 pr3 = function
+  let pr_may_eval env sigma test prc prlc pr2 pr3 pr4 = function
     | ConstrEval (r,c) ->
       hov 0
         (keyword "eval" ++ brk (1,1) ++
-           pr_red_expr env sigma (prc,prlc,pr2,pr3) r ++ spc () ++
+           pr_red_expr env sigma (prc,prlc,pr2,pr3,pr4) r ++ spc () ++
            keyword "in" ++ spc() ++ prc env sigma c)
     | ConstrContext ({CAst.v=id},c) ->
       hov 0
@@ -179,7 +179,7 @@ let string_of_genarg_arg (ArgumentType arg) =
     | ConstrTypeOf c ->
       hov 1 (keyword "type of" ++ spc() ++ prc env sigma c)
     | ConstrTerm c when test c ->
-      h 0 (str "(" ++ prc env sigma c ++ str ")")
+      h (str "(" ++ prc env sigma c ++ str ")")
     | ConstrTerm c ->
       prc env sigma c
 
@@ -191,12 +191,13 @@ let string_of_genarg_arg (ArgumentType arg) =
   let pr_and_short_name pr (c,_) = pr c
 
   let pr_evaluable_reference = function
-    | EvalVarRef id -> pr_id id
-    | EvalConstRef sp -> pr_global (GlobRef.ConstRef sp)
+    | Evaluable.EvalVarRef id -> pr_id id
+    | Evaluable.EvalConstRef sp -> pr_global (GlobRef.ConstRef sp)
+    | Evaluable.EvalProjectionRef p -> str "TODO projection" (* TODO *)
 
   let pr_quantified_hypothesis = function
     | AnonHyp n -> int n
-    | NamedHyp id -> pr_id id
+    | NamedHyp id -> pr_id id.CAst.v
 
   let pr_clear_flag clear_flag pp x =
     match clear_flag with
@@ -293,7 +294,7 @@ let string_of_genarg_arg (ArgumentType arg) =
       let pr _ = str "_" in
       KerName.print key ++ spc() ++ pr_sequence pr l ++ str" (* Generic printer *)"
 
-  let pr_farg prtac arg = prtac LevelSome (TacArg (CAst.make arg))
+  let pr_farg prtac arg = prtac LevelSome (CAst.make (TacArg  arg))
 
   let is_genarg tag wit =
     let ArgT.Any tag = tag in
@@ -335,11 +336,11 @@ let string_of_genarg_arg (ArgumentType arg) =
   let pr_targ prtac symb arg = match symb with
   | Extend.Uentry tag when is_genarg tag (ArgumentType wit_tactic) ->
     prtac LevelSome arg
-  | Extend.Uentryl (_, l) -> prtac LevelSome arg
+  | Extend.Uentryl (_, l) -> prtac (LevelLe l) arg
   | _ ->
     match arg with
-    | TacGeneric arg ->
-      let pr l arg = prtac l (TacGeneric arg) in
+    | TacGeneric (isquot,arg) ->
+      let pr l arg = prtac l (TacGeneric (isquot,arg)) in
       pr_any_arg pr symb arg
     | _ -> str "ltac:(" ++ prtac LevelSome arg ++ str ")"
 
@@ -349,9 +350,9 @@ let string_of_genarg_arg (ArgumentType arg) =
     pr_extend_gen (pr_farg prtac)
 
   let pr_raw_alias prtac lev key args =
-    pr_alias_gen (pr_targ (fun l a -> prtac l (TacArg (CAst.make a)))) lev key args
+    pr_alias_gen (pr_targ (fun l a -> prtac l (CAst.make @@ TacArg a))) lev key args
   let pr_glob_alias prtac lev key args =
-    pr_alias_gen (pr_targ (fun l a -> prtac l (TacArg (CAst.make a)))) lev key args
+    pr_alias_gen (pr_targ (fun l a -> prtac l (CAst.make @@ TacArg a))) lev key args
 
   (**********************************************************************)
   (* The tactic printer                                                 *)
@@ -361,7 +362,7 @@ let string_of_genarg_arg (ArgumentType arg) =
       match ty.CAst.v with
           Constrexpr.CProdN(bll,a) ->
             let bll = List.map (function
-            | CLocalAssum (nal,_,t) -> nal,t
+            | CLocalAssum (nal,_,_,t) -> nal,t
             | _ -> user_err Pp.(str "Cannot translate fix tactic: not only products")) bll in
             let nb = List.fold_left (fun i (nal,t) -> i + List.length nal) 0 bll in
             if nb >= n then (List.rev (bll@acc)), a
@@ -381,9 +382,11 @@ let string_of_genarg_arg (ArgumentType arg) =
         str "<" ++ KerName.print kn ++ str ">"
 
   let pr_evaluable_reference_env env = function
-    | EvalVarRef id -> pr_id id
-    | EvalConstRef sp ->
+    | Evaluable.EvalVarRef id -> pr_id id
+    | Evaluable.EvalConstRef sp ->
       Nametab.pr_global_env (Termops.vars_of_env env) (GlobRef.ConstRef sp)
+    | Evaluable.EvalProjectionRef p ->
+      str "TODO projection" (* TODO *)
 
   let pr_as_disjunctive_ipat prc ipatl =
     keyword "as" ++ spc () ++
@@ -441,13 +444,13 @@ let string_of_genarg_arg (ArgumentType arg) =
     | None -> mt()
 
   let pr_hyp_location pr_id = function
-    | occs, InHyp -> pr_with_occurrences pr_id occs
+    | occs, InHyp -> pr_with_occurrences (pr_or_var int) pr_id occs
     | occs, InHypTypeOnly ->
-      pr_with_occurrences (fun id ->
+      pr_with_occurrences (pr_or_var int) (fun id ->
         str "(" ++ keyword "type of" ++ spc () ++ pr_id id ++ str ")"
       ) occs
     | occs, InHypValueOnly ->
-      pr_with_occurrences (fun id ->
+      pr_with_occurrences (pr_or_var int) (fun id ->
         str "(" ++ keyword "value of" ++ spc () ++ pr_id id ++ str ")"
       ) occs
 
@@ -458,41 +461,49 @@ let string_of_genarg_arg (ArgumentType arg) =
     | l -> pr_in (spc () ++ prlist_with_sep spc pr_id l)
 
   let pr_in_hyp_as prc pr_id = function
-    | None -> mt ()
-    | Some (id,ipat) -> pr_in (spc () ++ pr_id id) ++ pr_as_ipat prc ipat
+    | [] -> mt ()
+    | l -> pr_in (spc () ++ prlist_with_sep pr_comma (fun (id,ipat) -> pr_id id ++ pr_as_ipat prc ipat) l)
 
   let pr_in_clause pr_id = function
     | { onhyps=None; concl_occs=NoOccurrences } ->
       (str "* |-")
     | { onhyps=None; concl_occs=occs } ->
-      (pr_with_occurrences (fun () -> str "*") (occs,()))
+      (pr_with_occurrences (pr_or_var int) (fun () -> str "*") (occs,()))
     | { onhyps=Some l; concl_occs=NoOccurrences } ->
       prlist_with_sep (fun () -> str ", ") (pr_hyp_location pr_id) l
     | { onhyps=Some l; concl_occs=occs } ->
-      let pr_occs = pr_with_occurrences (fun () -> str" |- *") (occs,()) in
+      let pr_occs = pr_with_occurrences (pr_or_var int) (fun () -> str" |- *") (occs,()) in
       (prlist_with_sep (fun () -> str", ") (pr_hyp_location pr_id) l ++ pr_occs)
 
   (* Some true = default is concl; Some false = default is all; None = no default *)
   let pr_clauses has_default pr_id = function
     | { onhyps=Some []; concl_occs=occs }
         when (match has_default with Some true -> true | _ -> false) ->
-      pr_with_occurrences mt (occs,())
+      pr_with_occurrences (pr_or_var int) mt (occs,())
     | { onhyps=None; concl_occs=AllOccurrences }
         when (match has_default with Some false -> true | _ -> false) -> mt ()
     | { onhyps=None; concl_occs=NoOccurrences } ->
       pr_in (str " * |-")
     | { onhyps=None; concl_occs=occs } ->
-      pr_in (pr_with_occurrences (fun () -> str " *") (occs,()))
+      pr_in (pr_with_occurrences (pr_or_var int) (fun () -> str " *") (occs,()))
     | { onhyps=Some l; concl_occs=occs } ->
       let pr_occs = match occs with
         | NoOccurrences -> mt ()
-        | _ -> pr_with_occurrences (fun () -> str" |- *") (occs,())
+        | _ -> pr_with_occurrences (pr_or_var int) (fun () -> str" |- *") (occs,())
       in
       pr_in
         (prlist_with_sep (fun () -> str",")
            (fun id -> spc () ++ pr_hyp_location pr_id id) l ++ pr_occs)
 
-  let pr_orient b = if b then mt () else str "<- "
+
+  (* When the [ssreflect.SsrSynax] module is imported, ssreflect operates
+     in reduced compatibility mode. During printing, we try to account for
+     this when this module is imported. See [plugins/ssr/ssrparser.mlg] for
+     the code that enables the reduced compatibility mode. *)
+  let ssr_loaded = ref (fun () -> false)
+  let ssr_loaded_hook f = ssr_loaded := f
+
+  let pr_orient b = if b then if !ssr_loaded () then str "-> " else mt () else str "<- "
 
   let pr_multi = let open Equality in function
     | Precisely 1 -> mt ()
@@ -514,19 +525,8 @@ let string_of_genarg_arg (ArgumentType arg) =
     | FullInversion -> primitive "inversion"
     | FullInversionClear -> primitive "inversion_clear"
 
-  let pr_range_selector (i, j) =
-    if Int.equal i j then int i
-    else int i ++ str "-" ++ int j
-
-let pr_goal_selector toplevel = let open Goal_select in function
-  | SelectAlreadyFocused -> str "!:"
-  | SelectNth i -> int i ++ str ":"
-  | SelectList l -> prlist_with_sep (fun () -> str ", ") pr_range_selector l ++ str ":"
-  | SelectId id -> str "[" ++ Id.print id ++ str "]:"
-  | SelectAll -> assert toplevel; str "all:"
-
 let pr_goal_selector ~toplevel s =
-  (if toplevel then mt () else str "only ") ++ pr_goal_selector toplevel s
+  (if toplevel then mt () else str "only ") ++ Goal_select.pr_goal_selector s ++ str ":"
 
   let pr_lazy = function
     | General -> keyword "multi"
@@ -571,14 +571,14 @@ let pr_goal_selector ~toplevel s =
 
   let pr_let_clause k pr_gen pr_arg (na,(bl,t)) =
     let pr = function
-      | TacGeneric arg ->
+      | TacGeneric (_,arg) ->
          let name = string_of_genarg_arg (genarg_tag arg) in
          if name = "unit" || name = "int" then
            (* Hard-wired parsing rules *)
            pr_gen  arg
          else
            str name ++ str ":" ++ surround (pr_gen arg)
-      | _ -> pr_arg (TacArg (CAst.make t)) in
+      | _ -> pr_arg (CAst.make (TacArg t)) in
     hov 0 (keyword k ++ spc () ++ pr_lname na ++ prlist pr_funvar bl ++
              str " :=" ++ brk (1,1) ++ pr t)
 
@@ -600,7 +600,7 @@ let pr_goal_selector ~toplevel s =
             str " ]")
 
   let pr_opt_tactic pr = function
-    | TacId [] -> mt ()
+    | {CAst.v=(TacId [])} -> mt ()
     | t -> pr t
 
   let pr_tac_extend_gen pr tf tm tl =
@@ -633,16 +633,14 @@ let pr_goal_selector ~toplevel s =
   let lseq = 4
   let ltactical = 3
   let lorelse = 2
-  let llet = 5
+  let llet = 5 (* in principle necessary only on the left-hand side of binary operators *)
   let lfun = 5
-  let lcomplete = 1
   let labstract = 3
   let lmatch = 1
   let latom = 0
   let lcall = 1
   let leval = 1
   let ltatom = 1
-  let linfo = 5
 
   let level_of p = match p with LevelLe n -> n | LevelLt n -> n-1 | LevelSome -> lseq
 
@@ -654,11 +652,13 @@ let pr_goal_selector ~toplevel s =
     pr_constr    : Environ.env -> Evd.evar_map -> 'trm -> Pp.t;
     pr_lconstr   : Environ.env -> Evd.evar_map -> 'trm -> Pp.t;
     pr_dconstr   : Environ.env -> Evd.evar_map -> 'dtrm -> Pp.t;
+    pr_red_pattern   : Environ.env -> Evd.evar_map -> 'rpat -> Pp.t;
     pr_pattern   : Environ.env -> Evd.evar_map -> 'pat -> Pp.t;
     pr_lpattern  : Environ.env -> Evd.evar_map -> 'pat -> Pp.t;
     pr_constant  : 'cst -> Pp.t;
     pr_reference : 'ref -> Pp.t;
     pr_name      : 'nam -> Pp.t;
+    pr_occvar    : 'occvar -> Pp.t;
     pr_generic   : Environ.env -> Evd.evar_map -> 'lev generic_argument -> Pp.t;
     pr_extend    : int -> ml_tactic_entry -> 'a gen_tactic_arg list -> Pp.t;
     pr_alias     : int -> KerName.t -> 'a gen_tactic_arg list -> Pp.t;
@@ -668,9 +668,11 @@ let pr_goal_selector ~toplevel s =
       term      :'trm;
       dterm     :'dtrm;
       pattern   :'pat;
+      red_pattern :'rpat;
       constant  :'cst;
       reference :'ref;
       name      :'nam;
+      occvar    :'occvar;
       tacexpr   :'tacexpr;
       level     :'lev
     >
@@ -679,7 +681,7 @@ let pr_goal_selector ~toplevel s =
       let pr_with_bindings = pr_with_bindings (pr.pr_constr env sigma) (pr.pr_lconstr env sigma) in
       let pr_with_bindings_arg_full = pr_with_bindings_arg in
       let pr_with_bindings_arg = pr_with_bindings_arg (pr.pr_constr env sigma) (pr.pr_lconstr env sigma) in
-      let pr_red_expr = pr_red_expr env sigma (pr.pr_constr,pr.pr_lconstr,pr.pr_constant,pr.pr_pattern) in
+      let pr_red_expr = pr_red_expr env sigma (pr.pr_constr,pr.pr_lconstr,pr.pr_constant,pr.pr_red_pattern,pr.pr_occvar) in
 
       let _pr_constrarg c = spc () ++ pr.pr_constr env sigma c in
       let pr_lconstrarg c = spc () ++ pr.pr_lconstr env sigma c in
@@ -790,7 +792,7 @@ let pr_goal_selector ~toplevel s =
           hov 1 (
             primitive "generalize" ++ spc ()
             ++ prlist_with_sep pr_comma (fun (cl,na) ->
-              pr_with_occurrences (pr.pr_constr env sigma) cl ++ pr_as_name na)
+              pr_with_occurrences pr.pr_occvar (pr.pr_constr env sigma) cl ++ pr_as_name na)
               l
           )
         | TacLetTac (ev,na,c,cl,true,_) when Locusops.is_nowhere cl ->
@@ -802,16 +804,6 @@ let pr_goal_selector ~toplevel s =
                 else pr_pose_as_style (pr.pr_constr env sigma) na c) ++
               pr_opt (fun p -> pr_eqn_ipat p ++ spc ()) e ++
               pr_non_empty_arg (pr_clauses (Some b) pr.pr_name) cl)
-        (*  | TacInstantiate (n,c,ConclLocation ()) ->
-            hov 1 (str "instantiate" ++ spc() ++
-            hov 1 (str"(" ++ pr_arg int n ++ str" :=" ++
-            pr_lconstrarg c ++ str ")" ))
-            | TacInstantiate (n,c,HypLocation (id,hloc)) ->
-            hov 1 (str "instantiate" ++ spc() ++
-            hov 1 (str"(" ++ pr_arg int n ++ str" :=" ++
-            pr_lconstrarg c ++ str ")" )
-            ++ str "in" ++ pr_hyp_location pr.pr_name (id,[],(hloc,ref None)))
-        *)
 
         (* Derived basic tactics *)
         | TacInductionDestruct (isrec,ev,(l,el)) ->
@@ -832,7 +824,7 @@ let pr_goal_selector ~toplevel s =
             ++ pr_non_empty_arg (pr_clauses (Some true) pr.pr_name) h
           )
         | TacChange (check,op,c,h) ->
-          let name = if check then "change_no_check" else "change" in
+          let name = if check then "change" else "change_no_check" in
           hov 1 (
             primitive name ++ brk (1,1)
             ++ (
@@ -840,7 +832,7 @@ let pr_goal_selector ~toplevel s =
                   None ->
                     mt ()
                 | Some p ->
-                  pr.pr_pattern env sigma p ++ spc ()
+                  pr.pr_red_pattern env sigma p ++ spc ()
                   ++ keyword "with" ++ spc ()
             ) ++ pr.pr_dconstr env sigma c ++ pr_non_empty_arg (pr_clauses (Some true) pr.pr_name) h
           )
@@ -884,12 +876,12 @@ let pr_goal_selector ~toplevel s =
       pr_atom1
 
     let make_pr_tac env sigma pr strip_prod_binders tag_atom tag =
-
         let extract_binders = function
-          | Tacexp (TacFun (lvar,body)) -> (lvar,Tacexp body)
+          | Tacexp { CAst.loc; v=(TacFun (lvar,body))} -> (lvar,Tacexp body)
           | body -> ([],body) in
         let rec pr_tac inherited tac =
           let return (doc, l) = (tag tac doc, l) in
+          let (loc, tac) = CAst.(tac.loc, tac.v) in
           let (strm, prec) = return (match tac with
             | TacAbstract (t,None) ->
               keyword "abstract " ++ pr_tac (LevelLt labstract) t, labstract
@@ -983,16 +975,6 @@ let pr_goal_selector ~toplevel s =
                 keyword "progress" ++ spc ()
                 ++ pr_tac (LevelLe ltactical) t),
               ltactical
-            | TacShowHyps t ->
-              hov 1 (
-                keyword "infoH" ++ spc ()
-                ++ pr_tac (LevelLe ltactical) t),
-              ltactical
-            | TacInfo t ->
-              hov 1 (
-                keyword "info" ++ spc ()
-                ++ pr_tac (LevelLe ltactical) t),
-              linfo
             | TacOr (t1,t2) ->
               hov 1 (
                 pr_tac (LevelLt lorelse) t1 ++ spc ()
@@ -1040,46 +1022,45 @@ let pr_goal_selector ~toplevel s =
               keyword "first" ++ spc () ++ pr_seq_body (pr_tac ltop) tl, llet
             | TacSolve tl ->
               keyword "solve" ++ spc () ++ pr_seq_body (pr_tac ltop) tl, llet
-            | TacComplete t ->
-              pr_tac (LevelLe lcomplete) t, lcomplete
-            | TacSelect (s, tac) -> pr_goal_selector ~toplevel:false s ++ spc () ++ pr_tac ltop tac, latom
+            | TacSelect (s, tac) -> pr_goal_selector ~toplevel:false s ++ spc () ++ pr_tac ltop tac, ltactical
             | TacId l ->
               keyword "idtac" ++ prlist (pr_arg (pr_message_token pr.pr_name)) l, latom
-            | TacAtom { CAst.loc; v=t } ->
+            | TacAtom t ->
               pr_with_comments ?loc (hov 1 (pr_atom env sigma pr strip_prod_binders tag_atom t)), ltatom
-            | TacArg { CAst.v=Tacexp e } ->
+            | TacArg (Tacexp e) ->
               pr_tac inherited e, latom
-            | TacArg { CAst.v=ConstrMayEval (ConstrTerm c) } ->
+            | TacArg (ConstrMayEval (ConstrTerm c)) ->
               keyword "constr:" ++ pr.pr_constr env sigma c, latom
-            | TacArg { CAst.v=ConstrMayEval c } ->
-              pr_may_eval env sigma pr.pr_constr pr.pr_lconstr pr.pr_constant pr.pr_pattern c, leval
-            | TacArg { CAst.v=TacFreshId l } ->
+            | TacArg (ConstrMayEval c) ->
+              pr_may_eval env sigma pr.pr_constr pr.pr_lconstr pr.pr_constant pr.pr_red_pattern pr.pr_occvar c, leval
+            | TacArg (TacFreshId l) ->
               primitive "fresh" ++ pr_fresh_ids l, latom
-            | TacArg { CAst.v=TacGeneric arg } ->
-              pr.pr_generic env sigma arg, latom
-            | TacArg { CAst.v=TacCall {CAst.v=(f,[])} } ->
+            | TacArg (TacGeneric (isquot,arg)) ->
+              let p = pr.pr_generic env sigma arg in
+              (match isquot with Some name -> str name ++ str ":(" ++ p ++ str ")" | None -> p), latom
+            | TacArg (TacCall {CAst.v=(f,[])}) ->
               pr.pr_reference f, latom
-            | TacArg { CAst.v=TacCall {CAst.loc; v=(f,l)} } ->
+            | TacArg (TacCall {CAst.loc; v=(f,l)}) ->
               pr_with_comments ?loc (hov 1 (
                 pr.pr_reference f ++ spc ()
                 ++ prlist_with_sep spc pr_tacarg l)),
               lcall
-            | TacArg { CAst.v=a } ->
+            | TacArg a ->
               pr_tacarg a, latom
-            | TacML { CAst.loc; v=(s,l) } ->
+            | TacML (s,l) ->
               pr_with_comments ?loc (pr.pr_extend 1 s l), lcall
-            | TacAlias { CAst.loc; v=(kn,l) } ->
+            | TacAlias (kn,l) ->
               pr_with_comments ?loc (pr.pr_alias (level_of inherited) kn l), latom
           )
           in
-          if prec_less prec inherited then strm
+          if Notation.prec_less prec inherited then strm
           else str"(" ++ strm ++ str")"
 
         and pr_tacarg = function
           | Reference r ->
             pr.pr_reference r
           | ConstrMayEval c ->
-            pr_may_eval env sigma pr.pr_constr pr.pr_lconstr pr.pr_constant pr.pr_pattern c
+            pr_may_eval env sigma pr.pr_constr pr.pr_lconstr pr.pr_constant pr.pr_red_pattern pr.pr_occvar c
           | TacFreshId l ->
             keyword "fresh" ++ pr_fresh_ids l
           | TacPretype c ->
@@ -1087,7 +1068,7 @@ let pr_goal_selector ~toplevel s =
           | TacNumgoals ->
             keyword "numgoals"
           | (TacCall _|Tacexp _ | TacGeneric _) as a ->
-            hov 0 (keyword "ltac:" ++ surround (pr_tac ltop (TacArg (CAst.make a))))
+            hov 0 (keyword "ltac:" ++ surround (pr_tac ltop (CAst.make (TacArg a))))
 
         in pr_tac
 
@@ -1095,7 +1076,7 @@ let pr_goal_selector ~toplevel s =
     let rec strip_ty acc n ty =
       if Int.equal n 0 then (List.rev acc, (ty,None)) else
         match DAst.get ty with
-            Glob_term.GProd(na,Glob_term.Explicit,a,b) ->
+            Glob_term.GProd(na,_,Glob_term.Explicit,a,b) ->
               strip_ty (([CAst.make na],(a,None))::acc) (n-1) b
           | _ -> user_err Pp.(str "Cannot translate fix tactic: not enough products") in
     strip_ty [] n ty
@@ -1109,11 +1090,13 @@ let pr_goal_selector ~toplevel s =
       pr_constr = pr_constr_expr;
       pr_dconstr = pr_constr_expr;
       pr_lconstr = pr_lconstr_expr;
+      pr_red_pattern = pr_constr_expr;
       pr_pattern = pr_constr_pattern_expr;
       pr_lpattern = pr_lconstr_pattern_expr;
       pr_constant = pr_or_by_notation pr_qualid;
       pr_reference = pr_qualid;
       pr_name = pr_lident;
+      pr_occvar = pr_or_var int;
       pr_generic = Pputils.pr_raw_generic;
       pr_extend = pr_raw_extend_rec @@ pr_raw_tactic_level env sigma;
       pr_alias = pr_raw_alias @@ pr_raw_tactic_level env sigma;
@@ -1136,14 +1119,16 @@ let pr_goal_selector ~toplevel s =
     let rec prtac n (t:glob_tactic_expr) =
       let pr = {
         pr_tactic = prtac;
-        pr_constr = (fun env sigma -> pr_and_constr_expr (pr_glob_constr_env env));
-        pr_dconstr = (fun env sigma -> pr_and_constr_expr (pr_glob_constr_env env));
-        pr_lconstr = (fun env sigma -> pr_and_constr_expr (pr_lglob_constr_env env));
-        pr_pattern = (fun env sigma -> pr_pat_and_constr_expr (pr_glob_constr_env env));
-        pr_lpattern = (fun env sigma -> pr_pat_and_constr_expr (pr_lglob_constr_env env));
+        pr_constr = (fun env sigma -> pr_and_constr_expr (pr_glob_constr_env env sigma));
+        pr_dconstr = (fun env sigma -> pr_and_constr_expr (pr_glob_constr_env env sigma));
+        pr_lconstr = (fun env sigma -> pr_and_constr_expr (pr_lglob_constr_env env sigma));
+        pr_red_pattern = (fun env sigma -> pr_and_constr_expr (pr_glob_constr_env env sigma));
+        pr_pattern = (fun env sigma -> pr_pat_and_constr_expr (pr_glob_constr_env env sigma));
         pr_constant = pr_or_var (pr_and_short_name (pr_evaluable_reference_env env));
+        pr_lpattern = (fun env sigma -> pr_pat_and_constr_expr (pr_lglob_constr_env env sigma));
         pr_reference = pr_ltac_or_var (pr_located pr_ltac_constant);
         pr_name = pr_lident;
+        pr_occvar = pr_or_var int;
         pr_generic = Pputils.pr_glb_generic;
         pr_extend = pr_glob_extend_rec prtac;
         pr_alias = pr_glob_alias prtac;
@@ -1172,13 +1157,15 @@ let pr_goal_selector ~toplevel s =
       let pr = {
         pr_tactic = (fun _ _ -> str "<tactic>");
         pr_constr = pr_econstr_env;
-        pr_dconstr = (fun env sigma -> pr_and_constr_expr (pr_glob_constr_env env));
+        pr_dconstr = (fun env sigma -> pr_and_constr_expr (pr_glob_constr_env env sigma));
         pr_lconstr = pr_leconstr_env;
+        pr_red_pattern = pr_constr_pattern_env;
         pr_pattern = pr_constr_pattern_env;
         pr_lpattern = pr_lconstr_pattern_env;
         pr_constant = pr_evaluable_reference_env env;
         pr_reference = pr_located pr_ltac_constant;
         pr_name = pr_id;
+        pr_occvar = int;
         (* Those are not used by the atomic printer *)
         pr_generic = (fun _ -> assert false);
         pr_extend = (fun _ _ _ -> assert false);
@@ -1195,7 +1182,7 @@ let pr_goal_selector ~toplevel s =
 
   let pr_raw_extend env sigma = pr_raw_extend_rec @@ pr_raw_tactic_level env sigma
 
-  let pr_glob_extend env sigma = pr_glob_extend_rec (pr_glob_tactic_level env)
+  let pr_glob_extend env = pr_glob_extend_rec (pr_glob_tactic_level env)
 
   let pr_alias pr lev key args =
     pr_alias_gen (fun _ arg -> pr arg) lev key args
@@ -1205,6 +1192,20 @@ let pr_goal_selector ~toplevel s =
 
   let pr_atomic_tactic env sigma c = pr_atomic_tactic_level env sigma c
 
+let pp_ltac_call_kind = function
+  | LtacNotationCall s -> pr_alias_key s
+  | LtacNameCall cst -> pr_ltac_constant cst
+  (* todo: don't want the KerName instead? *)
+  | LtacVarCall (_, id, t) -> Names.Id.print id
+  | LtacAtomCall te ->
+    pr_glob_tactic (Global.env ())
+      (CAst.make (TacAtom te))
+  | LtacConstrInterp (env, sigma, c, _) ->
+    pr_glob_constr_env env sigma c
+  | LtacMLCall te ->
+    (pr_glob_tactic (Global.env ())
+       te)
+
 let declare_extra_genarg_pprule wit
   (f : 'a raw_extra_genarg_printer)
   (g : 'b glob_extra_genarg_printer)
@@ -1218,8 +1219,8 @@ let declare_extra_genarg_pprule wit
         f env sigma pr_constr_expr pr_lconstr_expr pr_raw_tactic_level x) in
   let g x =
     Genprint.PrinterBasic (fun env sigma ->
-    g env sigma (fun env sigma -> pr_and_constr_expr (pr_glob_constr_env env))
-      (fun env sigma -> pr_and_constr_expr (pr_lglob_constr_env env))
+    g env sigma (fun env sigma -> pr_and_constr_expr (pr_glob_constr_env env sigma))
+      (fun env sigma -> pr_and_constr_expr (pr_lglob_constr_env env sigma))
       (fun env sigma -> pr_glob_tactic_level env) x)
   in
   let h x =
@@ -1248,8 +1249,8 @@ let declare_extra_genarg_pprule_with_level wit
         default_already_surrounded = default_surrounded;
         default_ensure_surrounded = default_non_surrounded;
         printer = (fun env sigma n ->
-          g env sigma (fun env sigma -> pr_and_constr_expr (pr_glob_constr_env env))
-            (fun env sigma -> pr_and_constr_expr (pr_lglob_constr_env env))
+          g env sigma (fun env sigma -> pr_and_constr_expr (pr_glob_constr_env env sigma))
+            (fun env sigma -> pr_and_constr_expr (pr_lglob_constr_env env sigma))
             (fun env sigma -> pr_glob_tactic_level env) n x) }
   in
   let h x =
@@ -1261,10 +1262,6 @@ let declare_extra_genarg_pprule_with_level wit
   in
   Genprint.register_print0 wit f g h
 
-let declare_extra_vernac_genarg_pprule wit f =
-  let f x = Genprint.PrinterBasic (fun env sigma -> f env sigma pr_constr_expr pr_lconstr_expr pr_raw_tactic_level x) in
-  Genprint.register_vernac_print0 wit f
-
 (** Registering *)
 
 let pr_intro_pattern_env p = Genprint.TopPrinterNeedsContext (fun env sigma ->
@@ -1273,7 +1270,8 @@ let pr_intro_pattern_env p = Genprint.TopPrinterNeedsContext (fun env sigma ->
 
 let pr_red_expr_env r = Genprint.TopPrinterNeedsContext (fun env sigma ->
   pr_red_expr env sigma ((fun e -> pr_econstr_env e), (fun e -> pr_leconstr_env e),
-                         pr_evaluable_reference_env env, pr_constr_pattern_env) r)
+                         pr_evaluable_reference_env env, pr_constr_pattern_env,
+                         int) r)
 
 let pr_bindings_env bl = Genprint.TopPrinterNeedsContext (fun env sigma ->
   let sigma, bl = bl env sigma in
@@ -1307,10 +1305,10 @@ let register_basic_print0 wit f g h =
   Genprint.register_print0 wit (lift f) (lift g) (lift_top h)
 
 let pr_glob_constr_pptac env sigma c =
-  pr_glob_constr_env env c
+  pr_glob_constr_env env sigma c
 
 let pr_lglob_constr_pptac env sigma c =
-  pr_lglob_constr_env env c
+  pr_lglob_constr_env env sigma c
 
 let pr_raw_intro_pattern =
   lift_env (fun env sigma -> Miscprint.pr_intro_pattern @@ pr_constr_expr env sigma)
@@ -1323,12 +1321,13 @@ let () =
   let pr_unit _ = str "()" in
   let open Genprint in
   register_basic_print0 wit_int_or_var (pr_or_var int) (pr_or_var int) int;
+  register_basic_print0 wit_nat_or_var (pr_or_var int) (pr_or_var int) int;
   register_basic_print0 wit_ref
     pr_qualid (pr_or_var (pr_located pr_global)) pr_global;
   register_basic_print0 wit_smart_global
     (pr_or_by_notation pr_qualid) (pr_or_var (pr_located pr_global)) pr_global;
   register_basic_print0 wit_ident pr_id pr_id pr_id;
-  register_basic_print0 wit_var pr_lident pr_lident pr_id;
+  register_basic_print0 wit_hyp pr_lident pr_lident pr_id;
   register_print0 wit_intropattern pr_raw_intro_pattern pr_glob_intro_pattern pr_intro_pattern_env [@warning "-3"];
   register_print0 wit_simple_intropattern pr_raw_intro_pattern pr_glob_intro_pattern pr_intro_pattern_env;
   Genprint.register_print0
@@ -1339,8 +1338,8 @@ let () =
   ;
   Genprint.register_print0
     wit_constr
-    (lift_env Ppconstr.pr_lconstr_expr)
-    (lift_env (fun env sigma (c, _) -> pr_lglob_constr_pptac env sigma c))
+    (lift_env Ppconstr.pr_constr_expr)
+    (lift_env (fun env sigma (c, _) -> pr_glob_constr_pptac env sigma c))
     (make_constr_printer Printer.pr_econstr_n_env)
   ;
   Genprint.register_print0
@@ -1356,13 +1355,14 @@ let () =
     (make_constr_printer Printer.pr_econstr_n_env)
   ;
   Genprint.register_print0
-    wit_red_expr
-    (lift_env (fun env sigma -> pr_red_expr env sigma (pr_constr_expr, pr_lconstr_expr, pr_or_by_notation pr_qualid, pr_constr_pattern_expr)))
+    Redexpr.wit_red_expr
+    (lift_env (fun env sigma -> pr_red_expr env sigma (pr_constr_expr, pr_lconstr_expr, pr_or_by_notation pr_qualid, pr_constr_pattern_expr,pr_or_var int)))
     (lift_env (fun env sigma -> pr_red_expr env sigma
                   ((fun env sigma -> pr_and_constr_expr @@ pr_glob_constr_pptac env sigma),
                    (fun env sigma -> pr_and_constr_expr @@ pr_lglob_constr_pptac env sigma),
                    pr_or_var (pr_and_short_name pr_evaluable_reference),
-                   (fun env sigma -> pr_pat_and_constr_expr @@ pr_glob_constr_pptac env sigma))))
+                   (fun env sigma -> pr_and_constr_expr @@ pr_glob_constr_pptac env sigma),
+                   (pr_or_var int))))
     pr_red_expr_env
   ;
   register_basic_print0 wit_quant_hyp pr_quantified_hypothesis pr_quantified_hypothesis pr_quantified_hypothesis;
