@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -15,7 +15,7 @@ open Names
 open Libobject
 open Genarg
 open Extend
-open Pcoq
+open Procq
 open Egramml
 open Vernacexpr
 open Libnames
@@ -31,34 +31,20 @@ type argument = Genarg.ArgT.any Extend.user_symbol
 (**********************************************************************)
 (* Interpret entry names of the form "ne_constr_list" as entry keys   *)
 
-let coincide s pat off =
-  let len = String.length pat in
-  let break = ref true in
-  let i = ref 0 in
-  while !break && !i < len do
-    let c = Char.code s.[off + !i] in
-    let d = Char.code pat.[!i] in
-    break := Int.equal c d;
-    incr i
-  done;
-  !break
-
 let atactic n =
-  if n = 5 then Pcoq.Symbol.nterm Pltac.binder_tactic
-  else Pcoq.Symbol.nterml Pltac.tactic_expr (string_of_int n)
+  Procq.Symbol.nterml Pltac.ltac_expr (string_of_int n)
 
 type entry_name = EntryName :
-  'a raw_abstract_argument_type * (Tacexpr.raw_tactic_expr, _, 'a) Pcoq.Symbol.t -> entry_name
+  'a raw_abstract_argument_type * (Tacexpr.raw_tactic_expr, _, 'a) Procq.Symbol.t -> entry_name
 
 (** Quite ad-hoc *)
 let get_tacentry n m =
   let check_lvl n =
     Int.equal m n
-    && not (Int.equal m 5) (* Because tactic5 is at binder_tactic *)
     && not (Int.equal m 0) (* Because tactic0 is at simple_tactic *)
   in
-  if check_lvl n then EntryName (rawwit Tacarg.wit_tactic, Pcoq.Symbol.self)
-  else if check_lvl (n + 1) then EntryName (rawwit Tacarg.wit_tactic, Pcoq.Symbol.next)
+  if check_lvl n then EntryName (rawwit Tacarg.wit_tactic, Procq.Symbol.self)
+  else if check_lvl (n + 1) then EntryName (rawwit Tacarg.wit_tactic, Procq.Symbol.next)
   else EntryName (rawwit Tacarg.wit_tactic, atactic n)
 
 let get_separator = function
@@ -70,28 +56,37 @@ let check_separator ?loc = function
 | Some _ -> user_err ?loc (str "Separator is only for arguments with suffix _list_sep.")
 
 let rec parse_user_entry ?loc s sep =
-  let l = String.length s in
-  if l > 8 && coincide s "ne_" 0 && coincide s "_list" (l - 5) then
-    let entry = parse_user_entry ?loc (String.sub s 3 (l-8)) None in
+  let open CString in
+  let matches pre suf s =
+    String.length s > (String.length pre + String.length suf) &&
+      is_prefix pre s && is_suffix suf s
+  in
+  let basename pre suf s =
+    let plen = String.length pre in
+    String.sub s plen (String.length s - (plen + String.length suf))
+  in
+  let tactic_len = String.length "tactic" in
+  if matches "ne_" "_list" s then
+    let entry = parse_user_entry ?loc (basename "ne_" "_list" s) None in
     check_separator ?loc sep;
     Ulist1 entry
-  else if l > 12 && coincide s "ne_" 0 &&
-                   coincide s "_list_sep" (l-9) then
-    let entry = parse_user_entry ?loc (String.sub s 3 (l-12)) None in
+  else if matches "ne_" "_list_sep" s then
+    let entry = parse_user_entry ?loc (basename "ne_" "_list_sep" s) None in
     Ulist1sep (entry, get_separator sep)
-  else if l > 5 && coincide s "_list" (l-5) then
-    let entry = parse_user_entry ?loc (String.sub s 0 (l-5)) None in
+  else if matches "" "_list" s then
+    let entry = parse_user_entry ?loc (basename "" "_list" s) None in
     check_separator ?loc sep;
     Ulist0 entry
-  else if l > 9 && coincide s "_list_sep" (l-9) then
-    let entry = parse_user_entry ?loc (String.sub s 0 (l-9)) None in
+  else if matches "" "_list_sep" s then
+    let entry = parse_user_entry ?loc (basename "" "_list_sep" s) None in
     Ulist0sep (entry, get_separator sep)
-  else if l > 4 && coincide s "_opt" (l-4) then
-    let entry = parse_user_entry ?loc (String.sub s 0 (l-4)) None in
+  else if matches "" "_opt" s then
+    let entry = parse_user_entry ?loc (basename "" "_opt" s) None in
     check_separator ?loc sep;
     Uopt entry
-  else if Int.equal l 7 && coincide s "tactic" 0 && '5' >= s.[6] && s.[6] >= '0' then
-    let n = Char.code s.[6] - 48 in
+  else if String.length s = tactic_len + 1 && is_prefix "tactic" s
+      && '5' >= s.[tactic_len] && s.[tactic_len] >= '0' then
+    let n = Char.code s.[tactic_len] - Char.code '0' in
     check_separator ?loc sep;
     Uentryl ("tactic", n)
   else
@@ -111,15 +106,13 @@ let interp_entry_name interp symb =
   eval symb
 
 (**********************************************************************)
-(** Grammar declaration for Tactic Notation (Coq level)               *)
+(** Grammar declaration for Tactic Notation (Rocq level)               *)
 
 let get_tactic_entry n =
   if Int.equal n 0 then
     Pltac.simple_tactic, None
-  else if Int.equal n 5 then
-    Pltac.binder_tactic, None
-  else if 1<=n && n<5 then
-    Pltac.tactic_expr, Some (Gramlib.Gramext.Level (string_of_int n))
+  else if 1<=n && n<=5 then
+    Pltac.ltac_expr, Some (string_of_int n)
   else
     user_err Pp.(str ("Invalid Tactic Notation level: "^(string_of_int n)^"."))
 
@@ -140,26 +133,26 @@ let head_is_ident tg = match tg.tacgram_prods with
 let rec prod_item_of_symbol lev = function
 | Extend.Ulist1 s ->
   let EntryName (Rawwit typ, e) = prod_item_of_symbol lev s in
-  EntryName (Rawwit (ListArg typ), Pcoq.Symbol.list1 e)
+  EntryName (Rawwit (ListArg typ), Procq.Symbol.list1 e)
 | Extend.Ulist0 s ->
   let EntryName (Rawwit typ, e) = prod_item_of_symbol lev s in
-  EntryName (Rawwit (ListArg typ), Pcoq.Symbol.list0 e)
+  EntryName (Rawwit (ListArg typ), Procq.Symbol.list0 e)
 | Extend.Ulist1sep (s, sep) ->
   let EntryName (Rawwit typ, e) = prod_item_of_symbol lev s in
-  EntryName (Rawwit (ListArg typ), Pcoq.Symbol.list1sep e (Pcoq.Symbol.token (CLexer.terminal sep)) false)
+  EntryName (Rawwit (ListArg typ), Procq.Symbol.list1sep e (Procq.Symbol.tokens [Procq.TPattern (Procq.terminal sep)]) false)
 | Extend.Ulist0sep (s, sep) ->
   let EntryName (Rawwit typ, e) = prod_item_of_symbol lev s in
-  EntryName (Rawwit (ListArg typ), Pcoq.Symbol.list0sep e (Pcoq.Symbol.token (CLexer.terminal sep)) false)
+  EntryName (Rawwit (ListArg typ), Procq.Symbol.list0sep e (Procq.Symbol.tokens [Procq.TPattern (Procq.terminal sep)]) false)
 | Extend.Uopt s ->
   let EntryName (Rawwit typ, e) = prod_item_of_symbol lev s in
-  EntryName (Rawwit (OptArg typ), Pcoq.Symbol.opt e)
+  EntryName (Rawwit (OptArg typ), Procq.Symbol.opt e)
 | Extend.Uentry arg ->
   let ArgT.Any tag = arg in
   let wit = ExtraArg tag in
-  EntryName (Rawwit wit, Pcoq.Symbol.nterm (genarg_grammar wit))
+  EntryName (Rawwit wit, Procq.Symbol.nterm (genarg_grammar wit))
 | Extend.Uentryl (s, n) ->
   let ArgT.Any tag = s in
-  assert (coincide (ArgT.repr tag) "tactic" 0);
+  assert (CString.is_suffix "tactic" (ArgT.repr tag));
   get_tacentry n lev
 
 (** Tactic grammar extensions *)
@@ -174,10 +167,10 @@ let add_tactic_entry (kn, ml, tg) state =
       if Genarg.has_type arg wit && not ml then
         Tacexp (Genarg.out_gen wit arg)
       else
-        TacGeneric arg
+        TacGeneric (None, arg)
     in
     let l = List.map map l in
-    (TacAlias (CAst.make ~loc (kn,l)):raw_tactic_expr)
+    (CAst.make ~loc (TacAlias (kn,l)):raw_tactic_expr)
   in
   let () =
     if Int.equal tg.tacgram_level 0 && not (head_is_ident tg) then
@@ -191,11 +184,11 @@ let add_tactic_entry (kn, ml, tg) state =
   in
   let prods = List.map map tg.tacgram_prods in
   let rules = make_rule mkact prods in
-  let r = ExtendRule (entry, { pos; data=[(None, None, [rules])]}) in
+  let r = ExtendRule (entry, Procq.Reuse (pos, [rules])) in
   ([r], state)
 
 let tactic_grammar =
-  create_grammar_command "TacticGrammar" add_tactic_entry
+  create_grammar_command "TacticGrammar" { gext_fun = add_tactic_entry; gext_eq = (==) (* FIXME *) }
 
 let extend_tactic_grammar kn ml ntn = extend_grammar_command tactic_grammar (kn, ml, ntn)
 
@@ -219,7 +212,9 @@ let interp_prod_item = function
     | None ->
       if String.Map.mem s !entry_names then String.Map.find s !entry_names
       else begin match ArgT.name s with
-      | None -> user_err Pp.(str ("Unknown entry "^s^"."))
+      | None ->
+         if s = "var" then user_err Pp.(str ("var is deprecated, use hyp.")) (* to remove in 8.14 *)
+         else user_err Pp.(str ("Unknown entry "^s^"."))
       | Some arg -> arg
       end
     | Some n ->
@@ -233,7 +228,7 @@ let interp_prod_item = function
     TacNonTerm (loc, (symbol, ido))
 
 let make_fresh_key =
-  let id = Summary.ref ~name:"TACTIC-NOTATION-COUNTER" 0 in
+  let id = Summary.ref ~stage:Summary.Stage.Synterp ~name:"TACTIC-NOTATION-COUNTER" 0 in
   fun prods ->
     let cur = incr id; !id in
     let map = function
@@ -252,7 +247,6 @@ type tactic_grammar_obj = {
   tacobj_key : KerName.t;
   tacobj_local : locality_flag;
   tacobj_tacgram : tactic_grammar;
-  tacobj_body : Tacenv.alias_tactic;
   tacobj_forml : bool;
 }
 
@@ -266,69 +260,103 @@ let check_key key =
     user_err Pp.(str "Conflicting tactic notations keys. This can happen when including \
     twice the same module.")
 
-let cache_tactic_notation (_, tobj) =
+let cache_tactic_notation (tobj, body) =
   let key = tobj.tacobj_key in
   let () = check_key key in
-  Tacenv.register_alias key tobj.tacobj_body;
-  extend_tactic_grammar key tobj.tacobj_forml tobj.tacobj_tacgram;
-  Pptactic.declare_notation_tactic_pprule key (pprule tobj.tacobj_tacgram)
+  Tacenv.register_alias key body
 
-let open_tactic_notation i (_, tobj) =
-  let key = tobj.tacobj_key in
-  if Int.equal i 1 && not tobj.tacobj_local then
-    extend_tactic_grammar key tobj.tacobj_forml tobj.tacobj_tacgram
+let open_tactic_notation i _ = ()
 
-let load_tactic_notation i (_, tobj) =
+let load_tactic_notation i (tobj, body) =
   let key = tobj.tacobj_key in
   let () = check_key key in
   (* Only add the printing and interpretation rules. *)
-  Tacenv.register_alias key tobj.tacobj_body;
-  Pptactic.declare_notation_tactic_pprule key (pprule tobj.tacobj_tacgram);
-  if Int.equal i 1 && not tobj.tacobj_local then
-    extend_tactic_grammar key tobj.tacobj_forml tobj.tacobj_tacgram
+  Tacenv.register_alias key body
 
-let subst_tactic_notation (subst, tobj) =
+let subst_tactic_notation (subst, (tobj, body)) =
   let open Tacenv in
-  let alias = tobj.tacobj_body in
   { tobj with
-    tacobj_key = Mod_subst.subst_kn subst tobj.tacobj_key;
-    tacobj_body = { alias with alias_body = Tacsubst.subst_tactic subst alias.alias_body };
-  }
+    tacobj_key = Mod_subst.subst_kn subst tobj.tacobj_key
+  },
+  { body with alias_body = Tacsubst.subst_tactic subst body.alias_body }
 
-let classify_tactic_notation tacobj = Substitute tacobj
+let classify_tactic_notation tacobj = Substitute
 
-let inTacticGrammar : tactic_grammar_obj -> obj =
+let ltac_notation_cat = Libobject.create_category "ltac.notations"
+
+let inTacticGrammar : tactic_grammar_obj * Tacenv.alias_tactic -> obj =
   declare_object {(default_object "TacticGrammar") with
-       open_function = simple_open open_tactic_notation;
+       open_function = simple_open ~cat:ltac_notation_cat open_tactic_notation;
        load_function = load_tactic_notation;
        cache_function = cache_tactic_notation;
        subst_function = subst_tactic_notation;
        classify_function = classify_tactic_notation}
 
+let cache_tactic_syntax tobj =
+  let key = tobj.tacobj_key in
+  extend_tactic_grammar key tobj.tacobj_forml tobj.tacobj_tacgram;
+  Pptactic.declare_notation_tactic_pprule key (pprule tobj.tacobj_tacgram)
+
+let open_tactic_syntax i tobj =
+  let key = tobj.tacobj_key in
+  if Int.equal i 1 && not tobj.tacobj_local then
+    extend_tactic_grammar key tobj.tacobj_forml tobj.tacobj_tacgram
+
+let load_tactic_syntax i tobj =
+  let key = tobj.tacobj_key in
+  (* Only add the printing and interpretation rules. *)
+  Pptactic.declare_notation_tactic_pprule key (pprule tobj.tacobj_tacgram);
+  if Int.equal i 1 && not tobj.tacobj_local then
+    extend_tactic_grammar key tobj.tacobj_forml tobj.tacobj_tacgram
+
+let subst_tactic_syntax (subst, tobj) =
+  { tobj with
+    tacobj_key = Mod_subst.subst_kn subst tobj.tacobj_key
+  }
+
+let classify_tactic_syntax tacobj = Substitute
+
+let inTacticSyntax : tactic_grammar_obj -> obj =
+  declare_object {(default_object "TacticSyntax") with
+       object_stage = Summary.Stage.Synterp;
+       open_function = simple_open ~cat:ltac_notation_cat open_tactic_syntax;
+       load_function = load_tactic_syntax;
+       cache_function = cache_tactic_syntax;
+       subst_function = subst_tactic_syntax;
+       classify_function = classify_tactic_syntax}
+
 let cons_production_parameter = function
 | TacTerm _ -> None
 | TacNonTerm (_, (_, ido)) -> ido
 
-let add_glob_tactic_notation local ~level ?deprecation prods forml ids tac =
+let add_glob_tactic_notation ?deprecation tacobj ids tac =
+  let open Tacenv in
+  let body =
+    { alias_args = ids; alias_body = tac; alias_deprecation = deprecation } in
+  Lib.add_leaf (inTacticGrammar (tacobj, body))
+
+let add_glob_tactic_notation_syntax local ~level ?deprecation prods forml =
   let parule = {
     tacgram_level = level;
     tacgram_prods = prods;
   } in
-  let open Tacenv in
   let tacobj = {
     tacobj_key = make_fresh_key prods;
     tacobj_local = local;
     tacobj_tacgram = parule;
-    tacobj_body = { alias_args = ids; alias_body = tac; alias_deprecation = deprecation };
     tacobj_forml = forml;
   } in
-  Lib.add_anonymous_leaf (inTacticGrammar tacobj)
+  Lib.add_leaf (inTacticSyntax tacobj);
+  tacobj
 
-let add_tactic_notation local n ?deprecation prods e =
-  let ids = List.map_filter cons_production_parameter prods in
-  let prods = List.map interp_prod_item prods in
+let add_tactic_notation ?deprecation tacobj e =
+  let ids = List.map_filter cons_production_parameter tacobj.tacobj_tacgram.tacgram_prods in
   let tac = Tacintern.glob_tactic_env ids (Global.env()) e in
-  add_glob_tactic_notation local ~level:n ?deprecation prods false ids tac
+  add_glob_tactic_notation ?deprecation tacobj ids tac
+
+let add_tactic_notation_syntax local n ?deprecation prods =
+  let prods = List.map interp_prod_item prods in
+  add_glob_tactic_notation_syntax local ~level:n ?deprecation prods false
 
 (**********************************************************************)
 (* ML Tactic entries                                                  *)
@@ -349,11 +377,11 @@ let extend_atomic_tactic name entries =
     | TacNonTerm (_, (symb, _)) ->
       let EntryName (typ, e) = prod_item_of_symbol 0 symb in
       let Genarg.Rawwit wit = typ in
-      let inj x = TacArg (CAst.make @@ TacGeneric (Genarg.in_gen typ x)) in
+      let inj x = CAst.make @@ TacArg ( TacGeneric (None, Genarg.in_gen typ x)) in
       let default = epsilon_value inj e in
       match default with
       | None -> raise NonEmptyArgument
-      | Some def -> Tacintern.intern_tactic_or_tacarg (Genintern.empty_glob_sign Environ.empty_env) def
+      | Some def -> Tacintern.intern_tactic_or_tacarg (Genintern.empty_glob_sign ~strict:true Environ.empty_env) def
     in
     try Some (hd, List.map empty_value rem) with NonEmptyArgument -> None
   in
@@ -363,7 +391,7 @@ let extend_atomic_tactic name entries =
   | Some (id, args) ->
     let args = List.map (fun a -> Tacexp a) args in
     let entry = { mltac_name = name; mltac_index = i } in
-    let body = TacML (CAst.make (entry, args)) in
+    let body = CAst.make (TacML (entry, args)) in
     Tacenv.register_ltac false false (Names.Id.of_string id) body
   in
   List.iteri add_atomic entries
@@ -379,12 +407,13 @@ let add_ml_tactic_notation name ~level ?deprecation prods =
     let ids = List.map_filter get_id prods in
     let entry = { mltac_name = name; mltac_index = len - i - 1 } in
     let map id = Reference (Locus.ArgVar (CAst.make id)) in
-    let tac = TacML (CAst.make (entry, List.map map ids)) in
-    add_glob_tactic_notation false ~level ?deprecation prods true ids tac
+    let tac = CAst.make (TacML (entry, List.map map ids)) in
+    let tacobj = add_glob_tactic_notation_syntax false ~level ?deprecation prods true in
+    add_glob_tactic_notation ?deprecation tacobj ids tac
   in
   List.iteri iter (List.rev prods);
   (* We call [extend_atomic_tactic] only for "basic tactics" (the ones
-     at tactic_expr level 0) *)
+     at ltac_expr level 0) *)
   if Int.equal level 0 then extend_atomic_tactic name prods
 
 (**********************************************************************)
@@ -392,36 +421,34 @@ let add_ml_tactic_notation name ~level ?deprecation prods =
 
 let ltac_quotations = ref String.Set.empty
 
-let create_ltac_quotation name cast (e, l) =
+let create_ltac_quotation ~plugin name cast (e, l) =
   let () =
     if String.Set.mem name !ltac_quotations then
       failwith ("Ltac quotation " ^ name ^ " already registered")
   in
   let () = ltac_quotations := String.Set.add name !ltac_quotations in
   let entry = match l with
-  | None -> Pcoq.Symbol.nterm e
-  | Some l -> Pcoq.Symbol.nterml e (string_of_int l)
+  | None -> Procq.Symbol.nterm e
+  | Some l -> Procq.Symbol.nterml e (string_of_int l)
   in
-(*   let level = Some "1" in *)
-  let level = None in
-  let assoc = None in
   let rule =
-    Pcoq.(
+    Procq.(
       Rule.next
         (Rule.next
            (Rule.next
               (Rule.next
                  (Rule.next
                     Rule.stop
-                    (Symbol.token (CLexer.terminal name)))
-                 (Symbol.token (CLexer.terminal ":")))
-              (Symbol.token (CLexer.terminal "(")))
+                    (Symbol.token (Procq.terminal name)))
+                 (Symbol.token (Procq.terminal ":")))
+              (Symbol.token (Procq.terminal "(")))
            entry)
-        (Symbol.token (CLexer.terminal ")")))
+        (Symbol.token (Procq.terminal ")")))
   in
   let action _ v _ _ _ loc = cast (Some loc, v) in
-  let gram = (level, assoc, [Pcoq.Production.make rule action]) in
-  Pcoq.grammar_extend Pltac.tactic_arg {pos=None; data=[gram]}
+  let gram = [Procq.Production.make rule action] in
+  let plugin_uid = (plugin, "tacquot:"^name) in
+  Egramml.grammar_extend ~plugin_uid Pltac.tactic_value (Procq.Reuse (None, gram))
 
 (** Command *)
 
@@ -434,7 +461,7 @@ let is_defined_tac kn =
   try ignore (Tacenv.interp_ltac kn); true with Not_found -> false
 
 let warn_unusable_identifier =
-  CWarnings.create ~name:"unusable-identifier" ~category:"parsing"
+  CWarnings.create ~name:"unusable-identifier" ~category:CWarnings.CoreCategories.parsing
       (fun id -> strbrk "The Ltac name" ++ spc () ++ Id.print id ++ spc () ++
         strbrk "may be unusable because of a conflict with a notation.")
 
@@ -450,8 +477,8 @@ let register_ltac local ?deprecation tacl =
         in
         let is_shadowed =
           try
-            match Pcoq.parse_string Pltac.tactic (Id.to_string id) with
-            | Tacexpr.TacArg _ -> false
+            match Procq.parse_string Pltac.tactic (Id.to_string id) with
+            | { CAst.v=(Tacexpr.TacArg _) } -> false
             | _ -> true (* most probably TacAtom, i.e. a primitive tactic ident *)
           with e when CErrors.noncritical e -> true (* prim tactics with args, e.g. "apply" *)
         in
@@ -474,9 +501,9 @@ let register_ltac local ?deprecation tacl =
     in
     List.fold_left fold [] rfun
   in
-  let ist = Tacintern.make_empty_glob_sign () in
+  let ist = Tacintern.make_empty_glob_sign ~strict:true in
   let map (name, body) =
-    let body = Flags.with_option Tacintern.strict_check (Tacintern.intern_tactic_or_tacarg ist) body in
+    let body = Tacintern.intern_tactic_or_tacarg ist body in
     (name, body)
   in
   let defs () =
@@ -489,7 +516,7 @@ let register_ltac local ?deprecation tacl =
   in
   (* STATE XXX: Review what is going on here. Why does this needs
      protection? Why is not the STM level protection enough? Fishy *)
-  let defs = States.with_state_protection defs () in
+  let defs = Vernacstate.System.protect defs () in
   let iter (def, tac) = match def with
   | NewTac id ->
     Tacenv.register_ltac false local id tac ?deprecation;
@@ -519,7 +546,7 @@ let print_ltacs () =
   let entries = List.map_filter map entries in
   let pr_entry (qid, body) =
     let (l, t) = match body with
-    | Tacexpr.TacFun (l, t) -> (l, t)
+    | {CAst.v=(Tacexpr.TacFun (l, t))} -> (l, t)
     | _ -> ([], body)
     in
     let pr_ltac_fun_arg n = spc () ++ Name.print n in
@@ -529,16 +556,40 @@ let print_ltacs () =
 
 let locatable_ltac = "Ltac"
 
+let split_ltac_fun = function
+  | {CAst.v=(Tacexpr.TacFun (l, t))} -> (l,t)
+  | t -> ([],t)
+
+let pr_ltac_fun_arg n = spc () ++ Name.print n
+
+let print_ltac_body qid tac =
+  let filter mp =
+    try Some (Nametab.shortest_qualid_of_module mp)
+    with Not_found -> None
+  in
+  let mods = List.map_filter filter tac.Tacenv.tac_redef in
+  let redefined = match mods with
+  | [] -> mt ()
+  | mods ->
+    let redef = prlist_with_sep fnl pr_qualid mods in
+    fnl () ++ str "Redefined by:" ++ fnl () ++ redef
+  in
+  let l,t = split_ltac_fun tac.Tacenv.tac_body in
+  hv 2 (
+    hov 2 (str "Ltac" ++ spc() ++ pr_qualid qid ++
+           prlist pr_ltac_fun_arg l ++ spc () ++ str ":=")
+    ++ spc() ++ Pptactic.pr_glob_tactic (Global.env ()) t) ++ redefined
+
 let () =
   let open Prettyp in
-  let locate qid = try Some (Tacenv.locate_tactic qid) with Not_found -> None in
-  let locate_all = Tacenv.locate_extended_all_tactic in
-  let shortest_qualid = Tacenv.shortest_qualid_of_tactic in
-  let name kn = str "Ltac" ++ spc () ++ pr_path (Tacenv.path_of_tactic kn) in
-  let print kn =
-    let qid = qualid_of_path (Tacenv.path_of_tactic kn) in
-    Tacintern.print_ltac qid
-  in
+  let locate qid = try Some (qid, Tacenv.locate_tactic qid) with Not_found -> None in
+  let locate_all qid = List.map (fun kn -> (qid,kn)) (Tacenv.locate_extended_all_tactic qid) in
+  let shortest_qualid (qid,kn) = Tacenv.shortest_qualid_of_tactic kn in
+  let name (qid,kn) = str "Ltac" ++ spc () ++ pr_path (Tacenv.path_of_tactic kn) in
+  let print (qid,kn) =
+    let entries = Tacenv.ltac_entries () in
+    let tac = KNmap.find kn entries in
+    print_ltac_body qid tac in
   let about = name in
   register_locatable locatable_ltac {
     locate;
@@ -550,16 +601,27 @@ let () =
   }
 
 let print_located_tactic qid =
-  Feedback.msg_notice (Prettyp.print_located_other locatable_ltac qid)
+  Feedback.msg_notice (Prettyp.print_located_other (Global.env ()) locatable_ltac qid)
+
+let print_ltac id =
+ try
+  let kn = Tacenv.locate_tactic id in
+  let entries = Tacenv.ltac_entries () in
+  let tac = KNmap.find kn entries in
+  print_ltac_body id tac
+ with
+  Not_found ->
+   user_err
+    (pr_qualid id ++ spc() ++ str "is not a user defined tactic.")
 
 (** Grammar *)
 
 let () =
+  let open Procq.Entry in
   let entries = [
-    AnyEntry Pltac.tactic_expr;
-    AnyEntry Pltac.binder_tactic;
-    AnyEntry Pltac.simple_tactic;
-    AnyEntry Pltac.tactic_arg;
+    Any Pltac.ltac_expr;
+    Any Pltac.simple_tactic;
+    Any Pltac.tactic_value;
   ] in
   register_grammars_by_name "tactic" entries
 
@@ -615,6 +677,7 @@ let rec eval_sign : type a. a ty_sig -> a -> Geninterp.Val.t list -> Geninterp.i
 
 let eval : ty_ml -> Geninterp.Val.t list -> Geninterp.interp_sign -> unit Proofview.tactic = function
   | TyML (t,tac) -> eval_sign t tac
+let eval_of_ty_ml = eval
 
 let is_constr_entry = function
 | TUentry a -> Option.has_some @@ genarg_type_eq (ExtraArg a) Stdarg.wit_constr
@@ -635,7 +698,7 @@ let dummy_id = Id.of_string "_"
 let lift_constr_tac_to_ml_tac vars tac =
   let tac _ ist = Proofview.Goal.enter begin fun gl ->
     let env = Proofview.Goal.env gl in
-    let sigma = Tacmach.New.project gl in
+    let sigma = Tacmach.project gl in
     let map = function
     | Anonymous -> None
     | Name id ->
@@ -673,7 +736,7 @@ let tactic_extend plugin_name tacname ~level ?deprecation sign =
        node, the ML tactic retrieves its arguments in the [ist]
        environment instead.  This is the rôle of the
        [lift_constr_tac_to_ml_tac] function. *)
-    let body = Tacexpr.TacFun (vars, Tacexpr.TacML (CAst.make (ml, [])))in
+    let body = CAst.make (Tacexpr.TacFun (vars, CAst.make (Tacexpr.TacML (ml, [])))) in
     let id = Names.Id.of_string name in
     let obj () = Tacenv.register_ltac true false id body ?deprecation in
     let () = Tacenv.register_ml_tactic ml_tactic_name [|tac|] in
@@ -713,7 +776,7 @@ let ml_tactic_extend ~plugin ~name ~local ?deprecation sign tac =
   let args = List.init len (fun i -> Id.of_string (Printf.sprintf "arg%i" i)) in
   let vars = List.map (fun id -> Name id) args in
   let args = List.map (fun id -> Reference (Locus.ArgVar (CAst.make id))) args in
-  let body = Tacexpr.TacFun (vars, Tacexpr.TacML (CAst.make (ml, args))) in
+  let body = CAst.make (Tacexpr.TacFun (vars, CAst.make (Tacexpr.TacML (ml, args)))) in
   let id = Names.Id.of_string name in
   let obj () = Tacenv.register_ltac true local id body ?deprecation in
   let () = Tacenv.register_ml_tactic ml_tactic_name [|tac|] in
@@ -762,7 +825,7 @@ let in_tacval =
   let intern_fun _ e = Empty.abort e in
   let subst_fun s v = v in
   let () = Genintern.register_intern0 wit intern_fun in
-  let () = Genintern.register_subst0 wit subst_fun in
+  let () = Gensubst.register_subst0 wit subst_fun in
   (* No need to register a value tag for it via register_val0 since we will
      never access this genarg directly. *)
   let interp_fun ist tac =
@@ -780,9 +843,9 @@ let ml_val_tactic_extend ~plugin ~name ~local ?deprecation sign tac =
   let ml_tactic_name = { mltac_tactic = name; mltac_plugin = plugin } in
   let len = ml_sig_len sign in
   let vars = List.init len (fun i -> Id.of_string (Printf.sprintf "arg%i" i)) in
-  let body = TacGeneric (in_tacval { tacval_tac = ml_tactic_name;  tacval_var = vars }) in
+  let body = TacGeneric (None, in_tacval { tacval_tac = ml_tactic_name;  tacval_var = vars }) in
   let vars = List.map (fun id -> Name id) vars in
-  let body = Tacexpr.TacFun (vars, Tacexpr.TacArg (CAst.make body)) in
+  let body = CAst.make (Tacexpr.TacFun (vars, CAst.make (Tacexpr.TacArg body))) in
   let id = Names.Id.of_string name in
   let obj () = Tacenv.register_ltac true local id body ?deprecation in
   let () = assert (not @@ MLTacMap.mem ml_tactic_name !ml_table) in
@@ -803,15 +866,15 @@ type ('a, 'b) argument_intern =
 | ArgInternWit : ('a, 'b, 'c) Genarg.genarg_type -> ('a, 'b) argument_intern
 
 type 'b argument_subst =
-| ArgSubstFun : 'b Genintern.subst_fun -> 'b argument_subst
+| ArgSubstFun : 'b Gensubst.subst_fun -> 'b argument_subst
 | ArgSubstWit : ('a, 'b, 'c) Genarg.genarg_type -> 'b argument_subst
 
 type ('b, 'c) argument_interp =
 | ArgInterpRet : ('c, 'c) argument_interp
 | ArgInterpFun : ('b, Val.t) interp_fun -> ('b, 'c) argument_interp
 | ArgInterpWit : ('a, 'b, 'r) Genarg.genarg_type -> ('b, 'c) argument_interp
-| ArgInterpLegacy :
-  (Geninterp.interp_sign -> Goal.goal Evd.sigma -> 'b -> Evd.evar_map * 'c) -> ('b, 'c) argument_interp
+| ArgInterpSimple :
+  (Geninterp.interp_sign -> Environ.env -> Evd.evar_map -> 'b -> 'c) -> ('b, 'c) argument_interp
 
 type ('a, 'b, 'c) tactic_argument = {
   arg_parsing : 'a Vernacextend.argument_rule;
@@ -830,7 +893,7 @@ match arg.arg_intern with
     let ans = Genarg.out_gen (glbwit wit) (Tacintern.intern_genarg ist (Genarg.in_gen (rawwit wit) v)) in
     (ist, ans)
 
-let subst_fun (type a b c) (arg : (a, b, c) tactic_argument) : b Genintern.subst_fun =
+let subst_fun (type a b c) (arg : (a, b, c) tactic_argument) : b Gensubst.subst_fun =
 match arg.arg_subst with
 | ArgSubstFun f -> f
 | ArgSubstWit wit ->
@@ -844,17 +907,18 @@ match arg.arg_interp with
 | ArgInterpFun f -> f
 | ArgInterpWit wit ->
   (fun ist x -> Tacinterp.interp_genarg ist (Genarg.in_gen (glbwit wit) x))
-| ArgInterpLegacy f ->
-  (fun ist v -> Ftactic.enter (fun gl ->
-    let (sigma, v) = Tacmach.New.of_old (fun gl -> f ist gl v) gl in
-    let v = Geninterp.Val.inject tag v in
-    Proofview.tclTHEN (Proofview.Unsafe.tclEVARS sigma) (Ftactic.return v)
-  ))
+| ArgInterpSimple f ->
+  (fun ist v -> Ftactic.enter begin fun gl ->
+    let env = Proofview.Goal.env gl in
+    let sigma = Proofview.Goal.sigma gl in
+    let v = f ist env sigma v in
+    Ftactic.return (Geninterp.Val.inject tag v)
+  end)
 
-let argument_extend (type a b c) ~name (arg : (a, b, c) tactic_argument) =
+let argument_extend (type a b c) ~plugin ~name (arg : (a, b, c) tactic_argument) =
   let wit = Genarg.create_arg name in
   let () = Genintern.register_intern0 wit (intern_fun name arg) in
-  let () = Genintern.register_subst0 wit (subst_fun arg) in
+  let () = Gensubst.register_subst0 wit (subst_fun arg) in
   let tag = match arg.arg_tag with
   | None ->
     let () = register_val0 wit None in
@@ -866,17 +930,18 @@ let argument_extend (type a b c) ~name (arg : (a, b, c) tactic_argument) =
   let () = register_interp0 wit (interp_fun name arg tag) in
   let entry = match arg.arg_parsing with
   | Vernacextend.Arg_alias e ->
-    let () = Pcoq.register_grammar wit e in
+    let () = Procq.register_grammar wit e in
     e
   | Vernacextend.Arg_rules rules ->
-    let e = Pcoq.create_generic_entry Pcoq.utactic name (Genarg.rawwit wit) in
-    let () = Pcoq.grammar_extend e {pos=None; data=[(None, None, rules)]} in
+    let e = Procq.create_generic_entry2 name (Genarg.rawwit wit) in
+    let plugin_uid = (plugin, "argextend:"^name) in
+    let () = Egramml.grammar_extend ~plugin_uid e (Procq.Fresh (Gramlib.Gramext.First, [None, None, rules])) in
     e
   in
   let (rpr, gpr, tpr) = arg.arg_printer in
   let () = Pptactic.declare_extra_genarg_pprule wit rpr gpr tpr in
-  let () = create_ltac_quotation name
-    (fun (loc, v) -> Tacexpr.TacGeneric (Genarg.in_gen (Genarg.rawwit wit) v))
+  let () = create_ltac_quotation ~plugin name
+    (fun (loc, v) -> Tacexpr.TacGeneric (Some name,Genarg.in_gen (Genarg.rawwit wit) v))
     (entry, None)
   in
   (wit, entry)
