@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -19,7 +19,7 @@ open Tac2expr
 open Tac2entries.Pltac
 open Proofview.Notations
 
-let ltac2_plugin = "coq-core.plugins.ltac2"
+let ltac2_plugin = "rocq-runtime.plugins.ltac2"
 
 let constr_flags =
   let open Pretyping in
@@ -74,34 +74,34 @@ let format = repr_ext val_format
 let core_prefix path n = KerName.make path (Label.of_id (Id.of_string_soft n))
 
 let std_core n = core_prefix Tac2env.std_prefix n
-let coq_core n = core_prefix Tac2env.coq_prefix n
+let rocq_core n = core_prefix Tac2env.rocq_prefix n
 
 module Core =
 struct
 
-let t_unit = coq_core "unit"
+let t_unit = rocq_core "unit"
 let v_unit = Tac2ffi.of_unit ()
 
-let t_int = coq_core "int"
-let t_string = coq_core "string"
-let t_array = coq_core "array"
-let t_list = coq_core "list"
-let t_constr = coq_core "constr"
-let t_preterm = coq_core "preterm"
-let t_pattern = coq_core "pattern"
-let t_ident = coq_core "ident"
-let t_option = coq_core "option"
-let t_exn = coq_core "exn"
+let t_int = rocq_core "int"
+let t_string = rocq_core "string"
+let t_array = rocq_core "array"
+let t_list = rocq_core "list"
+let t_constr = rocq_core "constr"
+let t_preterm = rocq_core "preterm"
+let t_pattern = rocq_core "pattern"
+let t_ident = rocq_core "ident"
+let t_option = rocq_core "option"
+let t_exn = rocq_core "exn"
 let t_reference = std_core "reference"
 
-let c_nil = coq_core "[]"
-let c_cons = coq_core "::"
+let c_nil = rocq_core "[]"
+let c_cons = rocq_core "::"
 
-let c_none = coq_core "None"
-let c_some = coq_core "Some"
+let c_none = rocq_core "None"
+let c_some = rocq_core "Some"
 
-let c_true = coq_core "true"
-let c_false = coq_core "false"
+let c_true = rocq_core "true"
+let c_false = rocq_core "false"
 
 end
 
@@ -166,19 +166,19 @@ let projection = repr_ext val_projection
 (** Stdlib exceptions *)
 
 let err_notfocussed =
-  Tac2interp.LtacError (coq_core "Not_focussed", [||])
+  Tac2interp.LtacError (rocq_core "Not_focussed", [||])
 
 let err_outofbounds =
-  Tac2interp.LtacError (coq_core "Out_of_bounds", [||])
+  Tac2interp.LtacError (rocq_core "Out_of_bounds", [||])
 
 let err_notfound =
-  Tac2interp.LtacError (coq_core "Not_found", [||])
+  Tac2interp.LtacError (rocq_core "Not_found", [||])
 
 let err_matchfailure =
-  Tac2interp.LtacError (coq_core "Match_failure", [||])
+  Tac2interp.LtacError (rocq_core "Match_failure", [||])
 
 let err_division_by_zero =
-  Tac2interp.LtacError (coq_core "Division_by_zero", [||])
+  Tac2interp.LtacError (rocq_core "Division_by_zero", [||])
 
 (** Helper functions *)
 
@@ -737,7 +737,7 @@ let () =
   return (EConstr.Vars.closedn sigma n c)
 
 let () =
-  define "constr_occur_between" (int @-> int @-> constr @-> tac bool) @@ fun n m c ->
+  define "constr_noccur_between" (int @-> int @-> constr @-> tac bool) @@ fun n m c ->
   Proofview.tclEVARMAP >>= fun sigma ->
   return (EConstr.Vars.noccur_between sigma n m c)
 
@@ -867,6 +867,15 @@ let () =
   Proofview.tclEVARMAP >>= fun sigma ->
   return (Evarutil.has_undefined_evars sigma c)
 
+(** Uint63 *)
+
+let () = define "uint63_compare" (uint63 @-> uint63 @-> ret int) Uint63.compare
+
+let () = define "uint63_of_int" (int @-> ret uint63) Uint63.of_int
+
+let () = define "uint63_print" (uint63 @-> ret pp) @@ fun i ->
+  Pp.str (Uint63.to_string i)
+
 (** Extra equalities *)
 
 let () = define "evar_equal" (evar @-> evar @-> ret bool) Evar.equal
@@ -1123,6 +1132,15 @@ let () =
   else Tacticals.tclZEROMSG
     (str "Hypothesis " ++ quote (Id.print id) ++ str " not found") (* FIXME: Do something more sensible *)
 
+let () =
+  define "hyp_value" (ident @-> tac (option constr)) @@ fun id ->
+  pf_apply @@ fun env _ ->
+  match EConstr.lookup_named id env with
+  | d -> return (Context.Named.Declaration.get_value d)
+  | exception Not_found ->
+    Tacticals.tclZEROMSG
+    (str "Hypothesis " ++ quote (Id.print id) ++ str " not found") (* FIXME: Do something more sensible *)
+
 let () =
   define "hyps" (unit @-> tac valexpr) @@ fun _ ->
   pf_apply @@ fun env _ ->
@@ -1151,7 +1169,7 @@ let () =
 (** (unit -> constr) -> unit *)
 let () =
   define "refine" (closure @-> tac unit) @@ fun c ->
-  let c = thaw c >>= fun c -> Proofview.tclUNIT ((), Tac2ffi.to_constr c) in
+  let c = thaw c >>= fun c -> Proofview.tclUNIT ((), Tac2ffi.to_constr c, None) in
   Proofview.Goal.enter @@ fun gl ->
   Refine.generic_refine ~typecheck:true c gl
 
@@ -1592,20 +1610,33 @@ let to_lvar ist =
 
 let gtypref kn = GTypRef (Other kn, [])
 
-let intern_constr self ist c =
-  let (_, (c, _)) = Genintern.intern Stdarg.wit_constr ist c in
-  let v = match DAst.get c with
-    | GGenarg (GenArg (Glbwit tag, v)) ->
-      begin match genarg_type_eq tag wit_ltac2_var_quotation with
-      | Some Refl ->
-        begin match (fst v) with
-        | ConstrVar -> GlbTacexpr (GTacVar (snd v))
-        | _ -> GlbVal c
-        end
-      | None -> GlbVal c
+let of_glob_constr (c:Glob_term.glob_constr) =
+  match DAst.get c with
+  | GGenarg (GenArg (Glbwit tag, v)) ->
+    begin match genarg_type_eq tag wit_ltac2_var_quotation with
+    | Some Refl ->
+      begin match (fst v) with
+      | ConstrVar -> GlbTacexpr (GTacVar (snd v))
+      | _ -> GlbVal c
       end
-    | _ -> GlbVal c
-  in
+    | None -> GlbVal c
+    end
+  | _ -> GlbVal c
+
+let intern_constr ist c =
+  let {Genintern.ltacvars=lfun; genv=env; extra; intern_sign; strict_check} = ist in
+  let scope = Pretyping.WithoutTypeConstraint in
+  let ltacvars = {
+    Constrintern.ltac_vars = lfun;
+    ltac_bound = Id.Set.empty;
+    ltac_extra = extra;
+  } in
+  let c' = Constrintern.intern_core scope ~strict_check ~ltacvars env (Evd.from_env env) intern_sign c in
+  c'
+
+let intern_constr_tacexpr ist c =
+  let c = intern_constr ist c in
+  let v = of_glob_constr c in
   (v, gtypref t_constr)
 
 let interp_constr flags ist c =
@@ -1619,7 +1650,7 @@ let interp_constr flags ist c =
   end
 
 let () =
-  let intern = intern_constr in
+  let intern = intern_constr_tacexpr in
   let interp ist c = interp_constr constr_flags ist c in
   let print env sigma c = str "constr:(" ++ Printer.pr_lglob_constr_env env sigma c ++ str ")" in
   let raw_print env sigma c = str "constr:(" ++ Ppconstr.pr_constr_expr env sigma c ++ str ")" in
@@ -1634,7 +1665,7 @@ let () =
   define_ml_object Tac2quote.wit_constr obj
 
 let () =
-  let intern = intern_constr in
+  let intern = intern_constr_tacexpr in
   let interp ist c = interp_constr open_constr_no_classes_flags ist c in
   let print env sigma c = str "open_constr:(" ++ Printer.pr_lglob_constr_env env sigma c ++ str ")" in
   let raw_print env sigma c = str "open_constr:(" ++ Ppconstr.pr_constr_expr env sigma c ++ str ")" in
@@ -1652,7 +1683,7 @@ let () =
   let interp _ id = return (Tac2ffi.of_ident id) in
   let print _ _ id = str "ident:(" ++ Id.print id ++ str ")" in
   let obj = {
-    ml_intern = (fun _ _ id -> GlbVal id, gtypref t_ident);
+    ml_intern = (fun _ id -> GlbVal id, gtypref t_ident);
     ml_interp = interp;
     ml_subst = (fun _ id -> id);
     ml_print = print;
@@ -1661,7 +1692,7 @@ let () =
   define_ml_object Tac2quote.wit_ident obj
 
 let () =
-  let intern self {Genintern.ltacvars=lfun; genv=env; extra; intern_sign=_; strict_check} c =
+  let intern {Genintern.ltacvars=lfun; genv=env; extra; intern_sign=_; strict_check} c =
     let sigma = Evd.from_env env in
     let ltacvars = {
       Constrintern.ltac_vars = lfun;
@@ -1696,8 +1727,8 @@ let () =
   define_ml_object Tac2quote.wit_pattern obj
 
 let () =
-  let intern self ist c =
-    let (_, (c, _)) = Genintern.intern Stdarg.wit_constr ist c in
+  let intern ist c =
+    let c = intern_constr ist c in
     (GlbVal (Id.Set.empty,c), gtypref t_preterm)
   in
   let interp env (ids,c) =
@@ -1737,12 +1768,12 @@ let () =
   define_ml_object Tac2quote.wit_preterm obj
 
 let () =
-  let intern self ist ref = match ref.CAst.v with
+  let intern ist ref = match ref.CAst.v with
   | Tac2qexpr.QHypothesis id ->
     GlbVal (GlobRef.VarRef id), gtypref t_reference
   | Tac2qexpr.QReference qid ->
     let gr =
-      try Nametab.locate qid
+      try Smartlocate.locate_global_with_alias qid
       with Not_found as exn ->
         let _, info = Exninfo.capture exn in
         Nametab.error_global_not_found ~info qid
@@ -1867,8 +1898,7 @@ let () =
         | PatternVar -> str "pattern:"
       in
       str "$" ++ ppkind ++ Id.print id) in
-  let pr_top x = Util.Empty.abort x in
-  Genprint.register_print0 wit_ltac2_var_quotation pr_raw pr_glb pr_top
+  Genprint.register_noval_print0 wit_ltac2_var_quotation pr_raw pr_glb
 
 let warn_missing_notation_variable =
   CWarnings.create ~name:"ltac2-missing-notation-var" ~category:CWarnings.CoreCategories.ltac2
@@ -1910,18 +1940,6 @@ let () =
   in
   Genintern.register_ntn_subst0 wit_ltac2_constr subs
 
-let () =
-  let pr_raw _ = Genprint.PrinterBasic (fun _env _sigma -> assert false) in
-  let pr_glb (ids, e) =
-    let ids =
-      if List.is_empty ids then mt ()
-      else hov 0 (pr_sequence Id.print ids ++ str " |- ")
-    in
-    Genprint.PrinterBasic Pp.(fun _env _sigma -> ids ++ Tac2print.pr_glbexpr ~avoid:Id.Set.empty e)
-  in
-  let pr_top x = Util.Empty.abort x in
-  Genprint.register_print0 wit_ltac2in1 pr_raw pr_glb pr_top
-
 let () =
   let pr_raw e = Genprint.PrinterBasic (fun _env _sigma ->
       let avoid = Id.Set.empty in
@@ -1942,8 +1960,7 @@ let () =
     *)
     Genprint.PrinterBasic Pp.(fun _env _sigma -> ids ++ Tac2print.pr_glbexpr ~avoid:Id.Set.empty e)
   in
-  let pr_top e = Util.Empty.abort e in
-  Genprint.register_print0 wit_ltac2_constr pr_raw pr_glb pr_top
+  Genprint.register_noval_print0 wit_ltac2_constr pr_raw pr_glb
 
 (** Built-in notation scopes *)
 
@@ -1969,7 +1986,7 @@ let q_unit = CAst.make @@ CTacCst (AbsKn (Tuple 0))
 let add_generic_scope s entry arg =
   let parse = function
   | [] ->
-    let scope = Pcoq.Symbol.nterm entry in
+    let scope = Procq.Symbol.nterm entry in
     let act x = CAst.make @@ CTacExt (arg, x) in
     Tac2entries.ScopeRule (scope, act)
   | arg -> scope_fail s arg
@@ -1980,14 +1997,14 @@ open CAst
 
 let () = add_scope "keyword" begin function
 | [SexprStr {loc;v=s}] ->
-  let scope = Pcoq.Symbol.token (Tok.PKEYWORD s) in
+  let scope = Procq.Symbol.token (Tok.PKEYWORD s) in
   Tac2entries.ScopeRule (scope, (fun _ -> q_unit))
 | arg -> scope_fail "keyword" arg
 end
 
 let () = add_scope "terminal" begin function
 | [SexprStr {loc;v=s}] ->
-  let scope = Pcoq.Symbol.token (Pcoq.terminal s) in
+  let scope = Procq.Symbol.token (Procq.terminal s) in
   Tac2entries.ScopeRule (scope, (fun _ -> q_unit))
 | arg -> scope_fail "terminal" arg
 end
@@ -1995,13 +2012,13 @@ end
 let () = add_scope "list0" begin function
 | [tok] ->
   let Tac2entries.ScopeRule (scope, act) = Tac2entries.parse_scope tok in
-  let scope = Pcoq.Symbol.list0 scope in
+  let scope = Procq.Symbol.list0 scope in
   let act l = Tac2quote.of_list act l in
   Tac2entries.ScopeRule (scope, act)
 | [tok; SexprStr {v=str}] ->
   let Tac2entries.ScopeRule (scope, act) = Tac2entries.parse_scope tok in
-  let sep = Pcoq.Symbol.tokens [Pcoq.TPattern (Pcoq.terminal str)] in
-  let scope = Pcoq.Symbol.list0sep scope sep false in
+  let sep = Procq.Symbol.tokens [Procq.TPattern (Procq.terminal str)] in
+  let scope = Procq.Symbol.list0sep scope sep false in
   let act l = Tac2quote.of_list act l in
   Tac2entries.ScopeRule (scope, act)
 | arg -> scope_fail "list0" arg
@@ -2010,13 +2027,13 @@ end
 let () = add_scope "list1" begin function
 | [tok] ->
   let Tac2entries.ScopeRule (scope, act) = Tac2entries.parse_scope tok in
-  let scope = Pcoq.Symbol.list1 scope in
+  let scope = Procq.Symbol.list1 scope in
   let act l = Tac2quote.of_list act l in
   Tac2entries.ScopeRule (scope, act)
 | [tok; SexprStr {v=str}] ->
   let Tac2entries.ScopeRule (scope, act) = Tac2entries.parse_scope tok in
-  let sep = Pcoq.Symbol.tokens [Pcoq.TPattern (Pcoq.terminal str)] in
-  let scope = Pcoq.Symbol.list1sep scope sep false in
+  let sep = Procq.Symbol.tokens [Procq.TPattern (Procq.terminal str)] in
+  let scope = Procq.Symbol.list1sep scope sep false in
   let act l = Tac2quote.of_list act l in
   Tac2entries.ScopeRule (scope, act)
 | arg -> scope_fail "list1" arg
@@ -2025,7 +2042,7 @@ end
 let () = add_scope "opt" begin function
 | [tok] ->
   let Tac2entries.ScopeRule (scope, act) = Tac2entries.parse_scope tok in
-  let scope = Pcoq.Symbol.opt scope in
+  let scope = Procq.Symbol.opt scope in
   let act opt = match opt with
   | None ->
     CAst.make @@ CTacCst (AbsKn (Other Core.c_none))
@@ -2038,7 +2055,7 @@ end
 
 let () = add_scope "self" begin function
 | [] ->
-  let scope = Pcoq.Symbol.self in
+  let scope = Procq.Symbol.self in
   let act tac = tac in
   Tac2entries.ScopeRule (scope, act)
 | arg -> scope_fail "self" arg
@@ -2046,7 +2063,7 @@ end
 
 let () = add_scope "next" begin function
 | [] ->
-  let scope = Pcoq.Symbol.next in
+  let scope = Procq.Symbol.next in
   let act tac = tac in
   Tac2entries.ScopeRule (scope, act)
 | arg -> scope_fail "next" arg
@@ -2055,12 +2072,12 @@ end
 let () = add_scope "tactic" begin function
 | [] ->
   (* Default to level 5 parsing *)
-  let scope = Pcoq.Symbol.nterml ltac2_expr "5" in
+  let scope = Procq.Symbol.nterml ltac2_expr "5" in
   let act tac = tac in
   Tac2entries.ScopeRule (scope, act)
 | [SexprInt {loc;v=n}] as arg ->
   let () = if n < 0 || n > 6 then scope_fail "tactic" arg in
-  let scope = Pcoq.Symbol.nterml ltac2_expr (string_of_int n) in
+  let scope = Procq.Symbol.nterml ltac2_expr (string_of_int n) in
   let act tac = tac in
   Tac2entries.ScopeRule (scope, act)
 | arg -> scope_fail "tactic" arg
@@ -2081,7 +2098,7 @@ let () = add_scope "constr" (fun arg ->
         arg
     in
     let act e = Tac2quote.of_constr ~delimiters e in
-    Tac2entries.ScopeRule (Pcoq.Symbol.nterm Pcoq.Constr.constr, act)
+    Tac2entries.ScopeRule (Procq.Symbol.nterm Procq.Constr.constr, act)
   )
 
 let () = add_scope "open_constr" (fun arg ->
@@ -2091,7 +2108,7 @@ let () = add_scope "open_constr" (fun arg ->
         arg
     in
     let act e = Tac2quote.of_open_constr ~delimiters e in
-    Tac2entries.ScopeRule (Pcoq.Symbol.nterm Pcoq.Constr.constr, act)
+    Tac2entries.ScopeRule (Procq.Symbol.nterm Procq.Constr.constr, act)
   )
 
 let () = add_scope "preterm" (fun arg ->
@@ -2101,12 +2118,12 @@ let () = add_scope "preterm" (fun arg ->
         arg
     in
     let act e = Tac2quote.of_preterm ~delimiters e in
-    Tac2entries.ScopeRule (Pcoq.Symbol.nterm Pcoq.Constr.constr, act)
+    Tac2entries.ScopeRule (Procq.Symbol.nterm Procq.Constr.constr, act)
   )
 
 let add_expr_scope name entry f =
   add_scope name begin function
-  | [] -> Tac2entries.ScopeRule (Pcoq.Symbol.nterm entry, f)
+  | [] -> Tac2entries.ScopeRule (Procq.Symbol.nterm entry, f)
   | arg -> scope_fail name arg
   end
 
@@ -2131,13 +2148,13 @@ let () = add_expr_scope "pose" q_pose Tac2quote.of_pose
 let () = add_expr_scope "assert" q_assert Tac2quote.of_assertion
 let () = add_expr_scope "constr_matching" q_constr_matching Tac2quote.of_constr_matching
 let () = add_expr_scope "goal_matching" q_goal_matching Tac2quote.of_goal_matching
-let () = add_expr_scope "format" Pcoq.Prim.lstring Tac2quote.of_format
+let () = add_expr_scope "format" Procq.Prim.lstring Tac2quote.of_format
 
-let () = add_generic_scope "pattern" Pcoq.Constr.constr Tac2quote.wit_pattern
+let () = add_generic_scope "pattern" Procq.Constr.constr Tac2quote.wit_pattern
 
 (** seq scope, a bit hairy *)
 
-open Pcoq
+open Procq
 
 type _ converter =
 | CvNil : (Loc.t -> raw_tacexpr) converter
@@ -2153,17 +2170,17 @@ type seqrule =
 
 let rec make_seq_rule = function
 | [] ->
-  Seqrule (Pcoq.Rule.stop, CvNil)
+  Seqrule (Procq.Rule.stop, CvNil)
 | tok :: rem ->
   let Tac2entries.ScopeRule (scope, f) = Tac2entries.parse_scope tok in
   let scope =
-    match Pcoq.generalize_symbol scope with
+    match Procq.generalize_symbol scope with
     | None ->
       CErrors.user_err (str "Recursive symbols (self / next) are not allowed in local rules")
     | Some scope -> scope
   in
   let Seqrule (r, c) = make_seq_rule rem in
-  let r = Pcoq.Rule.next_norec r scope in
+  let r = Procq.Rule.next_norec r scope in
   let f = match tok with
   | SexprStr _ -> None (* Leave out mere strings *)
   | _ -> Some f
@@ -2173,7 +2190,7 @@ let rec make_seq_rule = function
 let () = add_scope "seq" begin fun toks ->
   let scope =
     let Seqrule (r, c) = make_seq_rule (List.rev toks) in
-    Pcoq.(Symbol.rules [Rules.make r (apply c [])])
+    Procq.(Symbol.rules [Rules.make r (apply c [])])
   in
   Tac2entries.ScopeRule (scope, (fun e -> e))
 end
