@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -17,7 +17,7 @@ open Tac2expr
 val register_ltac : ?deprecation:Deprecation.t -> ?local:bool -> ?mut:bool -> rec_flag ->
   (Names.lname * raw_tacexpr) list -> unit
 
-val register_type : ?local:bool -> rec_flag ->
+val register_type : ?local:bool -> ?abstract:bool -> rec_flag ->
   (qualid * redef_flag * raw_quant_typedef) list -> unit
 
 val register_primitive : ?deprecation:Deprecation.t -> ?local:bool ->
@@ -37,7 +37,7 @@ val perform_eval : pstate:Declare.Proof.t option -> raw_tacexpr -> unit
 (** {5 Notations} *)
 
 type scope_rule =
-| ScopeRule : (raw_tacexpr, _, 'a) Pcoq.Symbol.t * ('a -> raw_tacexpr) -> scope_rule
+| ScopeRule : (raw_tacexpr, _, 'a) Procq.Symbol.t * ('a -> raw_tacexpr) -> scope_rule
 
 type scope_interpretation = sexpr list -> scope_rule
 
@@ -78,32 +78,32 @@ val call_par : pstate:Declare.Proof.t -> with_end_tac:bool -> raw_tacexpr
 
 module Pltac :
 sig
-val ltac2_expr : raw_tacexpr Pcoq.Entry.t
-val tac2expr_in_env : (Id.t CAst.t list * raw_tacexpr) Pcoq.Entry.t
+val ltac2_expr : raw_tacexpr Procq.Entry.t
+val tac2expr_in_env : (Id.t CAst.t list * raw_tacexpr) Procq.Entry.t
 
 (** Quoted entries. To be used for complex notations. *)
 
 open Tac2qexpr
 
-val q_ident : Id.t CAst.t or_anti Pcoq.Entry.t
-val q_bindings : bindings Pcoq.Entry.t
-val q_with_bindings : bindings Pcoq.Entry.t
-val q_intropattern : intro_pattern Pcoq.Entry.t
-val q_intropatterns : intro_pattern list CAst.t Pcoq.Entry.t
-val q_destruction_arg : destruction_arg Pcoq.Entry.t
-val q_induction_clause : induction_clause Pcoq.Entry.t
-val q_conversion : conversion Pcoq.Entry.t
-val q_orient : orientation Pcoq.Entry.t
-val q_rewriting : rewriting Pcoq.Entry.t
-val q_clause : clause Pcoq.Entry.t
-val q_dispatch : dispatch Pcoq.Entry.t
-val q_occurrences : occurrences Pcoq.Entry.t
-val q_reference : reference or_anti Pcoq.Entry.t
-val q_strategy_flag : strategy_flag Pcoq.Entry.t
-val q_constr_matching : constr_matching Pcoq.Entry.t
-val q_goal_matching : goal_matching Pcoq.Entry.t
-val q_hintdb : hintdb Pcoq.Entry.t
-val q_move_location : move_location Pcoq.Entry.t
-val q_pose : pose Pcoq.Entry.t
-val q_assert : assertion Pcoq.Entry.t
+val q_ident : Id.t CAst.t or_anti Procq.Entry.t
+val q_bindings : bindings Procq.Entry.t
+val q_with_bindings : bindings Procq.Entry.t
+val q_intropattern : intro_pattern Procq.Entry.t
+val q_intropatterns : intro_pattern list CAst.t Procq.Entry.t
+val q_destruction_arg : destruction_arg Procq.Entry.t
+val q_induction_clause : induction_clause Procq.Entry.t
+val q_conversion : conversion Procq.Entry.t
+val q_orient : orientation Procq.Entry.t
+val q_rewriting : rewriting Procq.Entry.t
+val q_clause : clause Procq.Entry.t
+val q_dispatch : dispatch Procq.Entry.t
+val q_occurrences : occurrences Procq.Entry.t
+val q_reference : reference or_anti Procq.Entry.t
+val q_strategy_flag : strategy_flag Procq.Entry.t
+val q_constr_matching : constr_matching Procq.Entry.t
+val q_goal_matching : goal_matching Procq.Entry.t
+val q_hintdb : hintdb Procq.Entry.t
+val q_move_location : move_location Procq.Entry.t
+val q_pose : pose Procq.Entry.t
+val q_assert : assertion Procq.Entry.t
 end
