@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -13,7 +13,7 @@ open Names
 open Libnames
 open Nametab
 open Tac2expr
-open Tac2ffi
+open Tac2val
 
 (** Ltac2 global environment *)
 
@@ -29,6 +29,13 @@ type global_data = {
 val define_global : ltac_constant -> global_data -> unit
 val interp_global : ltac_constant -> global_data
 
+type compile_info = {
+  source : string;
+}
+
+val set_compiled_global : ltac_constant -> compile_info -> valexpr -> unit
+val get_compiled_global : ltac_constant -> (compile_info * valexpr) option
+
 val globals : unit -> global_data KNmap.t
 
 (** {5 Toplevel definition of types} *)
@@ -52,9 +59,12 @@ type constructor_data = {
       constructor is a member of an open type. *)
 }
 
-val define_constructor : ltac_constructor -> constructor_data -> unit
+val define_constructor : ?warn:UserWarn.t -> ltac_constructor -> constructor_data -> unit
 val interp_constructor : ltac_constructor -> constructor_data
 
+val find_all_constructors_in_type : type_constant -> constructor_data KNmap.t
+(** Useful for printing info about currently defined constructors of open types. *)
+
 (** {5 Toplevel definition of projections} *)
 
 type projection_data = {
@@ -83,12 +93,27 @@ type alias_data = {
 val define_alias : ?deprecation:Deprecation.t -> ltac_constant -> raw_tacexpr -> unit
 val interp_alias : ltac_constant -> alias_data
 
+(** {5 Toplevel definition of notations} *)
+
+(* no deprecation info: deprecation warning is emitted by the parser *)
+type notation_data =
+  | UntypedNota of raw_tacexpr
+  | TypedNota of {
+      nota_prms : int;
+      nota_argtys : int glb_typexpr Id.Map.t;
+      nota_ty : int glb_typexpr;
+      nota_body : glb_tacexpr;
+    }
+
+val define_notation : ltac_notation -> notation_data -> unit
+val interp_notation : ltac_notation -> notation_data
+
 (** {5 Name management} *)
 
 val push_ltac : visibility -> full_path -> tacref -> unit
 val locate_ltac : qualid -> tacref
 val locate_extended_all_ltac : qualid -> tacref list
-val shortest_qualid_of_ltac : tacref -> qualid
+val shortest_qualid_of_ltac : Id.Set.t -> tacref -> qualid
 val path_of_ltac : tacref -> full_path
 
 val push_constructor : visibility -> full_path -> ltac_constructor -> unit
@@ -97,10 +122,14 @@ val locate_extended_all_constructor : qualid -> ltac_constructor list
 val shortest_qualid_of_constructor : ltac_constructor -> qualid
 val path_of_constructor : ltac_constructor -> full_path
 
+(** Emit warning if any for the constructor. *)
+val constructor_user_warn : ?loc:Loc.t -> ltac_constructor -> unit
+
 val push_type : visibility -> full_path -> type_constant -> unit
 val locate_type : qualid -> type_constant
 val locate_extended_all_type : qualid -> type_constant list
 val shortest_qualid_of_type : ?loc:Loc.t -> type_constant -> qualid
+val path_of_type : type_constant -> full_path
 
 val push_projection : visibility -> full_path -> ltac_projection -> unit
 val locate_projection : qualid -> ltac_projection
@@ -112,8 +141,8 @@ val shortest_qualid_of_projection : ltac_projection -> qualid
 (** This state is not part of the summary, contrarily to the ones above. It is
     intended to be used from ML plugins to register ML-side functions. *)
 
-val define_primitive : ml_tactic_name -> closure -> unit
-val interp_primitive : ml_tactic_name -> closure
+val define_primitive : ml_tactic_name -> valexpr -> unit
+val interp_primitive : ml_tactic_name -> valexpr
 
 (** {5 ML primitive types} *)
 
@@ -128,10 +157,11 @@ type environment = {
 }
 
 type ('a, 'b) ml_object = {
-  ml_intern : 'r. (raw_tacexpr, glb_tacexpr, 'r) intern_fun -> ('a, 'b or_glb_tacexpr, 'r) intern_fun;
+  ml_intern : 'r. ('a, 'b or_glb_tacexpr, 'r) intern_fun;
   ml_subst : Mod_subst.substitution -> 'b -> 'b;
   ml_interp : environment -> 'b -> valexpr Proofview.tactic;
   ml_print : Environ.env -> Evd.evar_map -> 'b -> Pp.t;
+  ml_raw_print : Environ.env -> Evd.evar_map -> 'a -> Pp.t;
 }
 
 val define_ml_object : ('a, 'b) Tac2dyn.Arg.tag -> ('a, 'b) ml_object -> unit
@@ -139,7 +169,10 @@ val interp_ml_object : ('a, 'b) Tac2dyn.Arg.tag -> ('a, 'b) ml_object
 
 (** {5 Absolute paths} *)
 
-val coq_prefix : ModPath.t
+val rocq_prefix : ModPath.t
+(** Path where primitive datatypes are defined in Ltac2 plugin. *)
+
+val coq_prefix : ModPath.t [@@ocaml.deprecated "(9.0) Use rocq_prefix"]
 (** Path where primitive datatypes are defined in Ltac2 plugin. *)
 
 val std_prefix : ModPath.t
@@ -150,18 +183,17 @@ val ltac1_prefix : ModPath.t
 
 (** {5 Generic arguments} *)
 
-val wit_ltac2 : (Id.t CAst.t list * raw_tacexpr, Id.t list * glb_tacexpr, Util.Empty.t) genarg_type
-(** Ltac2 quotations in Ltac1 code *)
-
-val wit_ltac2_val : (Util.Empty.t, unit, Util.Empty.t) genarg_type
-(** Embedding Ltac2 closures of type [Ltac1.t -> Ltac1.t] inside Ltac1. There is
-    no relevant data because arguments are passed by conventional names. *)
-
 val wit_ltac2_constr : (raw_tacexpr, Id.Set.t * glb_tacexpr, Util.Empty.t) genarg_type
 (** Ltac2 quotations in Gallina terms *)
 
-val wit_ltac2_quotation : (Id.t Loc.located, Id.t, Util.Empty.t) genarg_type
-(** Ltac2 quotations for variables "$x" in Gallina terms *)
+type var_quotation_kind =
+  | ConstrVar
+  | PretermVar
+  | PatternVar
+
+val wit_ltac2_var_quotation : (lident option * lident, var_quotation_kind * Id.t, Util.Empty.t) genarg_type
+(** Ltac2 quotations for variables "$x" or "$kind:foo" in Gallina terms.
+    NB: "$x" means "$constr:x" *)
 
 (** {5 Helper functions} *)
 
