@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -11,6 +11,7 @@
 open Names
 open Genredexpr
 open Tac2expr
+open Tac2val
 open Tac2ffi
 open Tac2types
 open Tac2extffi
@@ -121,9 +122,34 @@ and to_or_and_intro_pattern v = match Value.to_block v with
 and to_intro_patterns il =
   Value.to_list to_intro_pattern il
 
-let intro_pattern = make_to_repr to_intro_pattern
+let rec of_intro_pattern = function
+  | IntroForthcoming b -> of_block (0, [|of_bool b|])
+  | IntroNaming pat -> of_block (1, [|of_intro_pattern_naming pat|])
+  | IntroAction act -> of_block (2, [|of_intro_pattern_action act|])
 
-let intro_patterns = make_to_repr to_intro_patterns
+and of_intro_pattern_naming = function
+  | IntroIdentifier id -> of_block (0, [|of_ident id|])
+  | IntroFresh id -> of_block (1, [|of_ident id|])
+  | IntroAnonymous -> of_int 0
+
+and of_intro_pattern_action = function
+  | IntroWildcard -> of_int 0
+  | IntroOrAndPattern op -> of_block (0, [|of_or_and_intro_pattern op|])
+  | IntroInjection inj -> of_block (1, [|of_list of_intro_pattern inj|])
+  | IntroApplyOn (c, ipat) ->
+    let c = repr_of (fun1 unit constr) c in
+    of_block (2, [|c; of_intro_pattern ipat|])
+  | IntroRewrite b -> of_block (3, [|of_bool b|])
+
+and of_or_and_intro_pattern = function
+  | IntroOrPattern ill -> of_block (0, [|of_list of_intro_patterns ill|])
+  | IntroAndPattern il -> of_block (1, [|of_intro_patterns il|])
+
+and of_intro_patterns il = of_list of_intro_pattern il
+
+let intro_pattern = make_repr of_intro_pattern to_intro_pattern
+
+let intro_patterns = make_repr of_intro_patterns to_intro_patterns
 
 let to_destruction_arg v = match Value.to_block v with
 | (0, [| c |]) ->
@@ -221,7 +247,7 @@ let generalize_arg = make_to_repr to_generalize_arg
 open Tac2externals
 
 let define s =
-  define { mltac_plugin = "coq-core.plugins.ltac2"; mltac_tactic = s }
+  define { mltac_plugin = "rocq-runtime.plugins.ltac2"; mltac_tactic = s }
 
 (** Tactics from Tacexpr *)
 
@@ -547,17 +573,17 @@ let () =
 
 let () =
   define "tac_trivial"
-    (debug @-> list (thunk constr) @-> option (list ident) @-> tac unit)
+    (debug @-> list reference @-> option (list ident) @-> tac unit)
     Tac2tactics.trivial
 
 let () =
   define "tac_eauto"
-    (debug @-> option int @-> list (thunk constr) @-> option (list ident) @-> tac unit)
+    (debug @-> option int @-> list reference @-> option (list ident) @-> tac unit)
     Tac2tactics.eauto
 
 let () =
   define "tac_auto"
-    (debug @-> option int @-> list (thunk constr) @-> option (list ident) @-> tac unit)
+    (debug @-> option int @-> list reference @-> option (list ident) @-> tac unit)
     Tac2tactics.auto
 
 let () =
@@ -592,3 +618,13 @@ let () =
   define "evarconv_unify"
     (transparent_state @-> constr @-> constr @-> tac unit)
     Tac2tactics.evarconv_unify
+
+let () =
+  define "congruence"
+  (option int @-> option (list constr) @-> tac unit)
+  Tac2tactics.congruence
+
+let () =
+  define "simple_congruence"
+  (option int @-> option (list constr) @-> tac unit)
+  Tac2tactics.simple_congruence
