@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -102,7 +102,7 @@ let subgoals_tys sigma (relctx, concl) =
  *    by eqid *)
 
 let get_eq_type env sigma =
-  Evd.fresh_global env sigma Coqlib.(lib_ref "core.eq.type")
+  Evd.fresh_global env sigma Rocqlib.(lib_ref "core.eq.type")
 
 type elim_what =
 | EConstr of
@@ -149,7 +149,7 @@ let mkTpat env sigma0 (sigma, t) = (* takes a term, refreshes it and makes a T p
 
 let redex_of_pattern env p = match redex_of_pattern p with
 | None -> CErrors.anomaly (Pp.str "pattern without redex.")
-| Some (sigma, e) -> Evarutil.nf_evar sigma e, Evd.evar_universe_context sigma
+| Some (sigma, e) -> Evarutil.nf_evar sigma e, Evd.ustate sigma
 
 let unif_redex env sigma0 nsigma p t = (* t is a hint for the redex of p *)
   let t, evs, ucst = abs_evars env sigma0 (nsigma, fire_subst nsigma t) in
@@ -162,7 +162,7 @@ let unif_redex env sigma0 nsigma p t = (* t is a hint for the redex of p *)
     with e when CErrors.noncritical e -> p
 
 let get_nth_arg n args =
-  List.find_map (fun (i, x, _) -> if Int.equal i n then Some x else None) args
+  List.find_map_exn (fun (i, x, _) -> if Int.equal i n then Some x else None) args
 
 let find_eliminator env sigma ~concl ~is_case ?elim oc c_gen =
   match elim with
@@ -201,16 +201,18 @@ let find_eliminator env sigma ~concl ~is_case ?elim oc c_gen =
     let sigma, c_ty = Typing.type_of env sigma c in
     let ((kn, i),_ as indu), unfolded_c_ty =
       Tacred.reduce_to_quantified_ind env sigma c_ty in
-    let sort = Retyping.get_sort_family_of env sigma concl in
+    let sort = Retyping.get_sort_of env sigma concl in
     let sigma, elim, elimty =
       if not is_case then
-        let sigma, elim = Evd.fresh_global env sigma (Indrec.lookup_eliminator env (kn,i) sort) in
+        let sigma, elim = Evd.fresh_global env sigma
+            (Indrec.lookup_eliminator env (kn,i) (EConstr.ESorts.family sigma sort))
+        in
         let elimty = Retyping.get_type_of env sigma elim in
         sigma, elim, elimty
       else
-        let indu = (fst indu, EConstr.EInstance.kind sigma (snd indu)) in
-        let (sigma, ind, indty) = Indrec.build_case_analysis_scheme env sigma indu true sort in
-        (sigma, EConstr.of_constr ind, EConstr.of_constr indty)
+        let sigma, case = Indrec.build_case_analysis_scheme env sigma indu true sort in
+        let (ind, indty) = Indrec.eval_case_analysis case in
+        (sigma, ind, indty)
     in
     let pred_id,n_elim_args,is_rec,elim_is_dep,n_pred_args,ctx_concl =
       analyze_eliminator elimty env sigma in
@@ -229,13 +231,13 @@ let find_eliminator env sigma ~concl ~is_case ?elim oc c_gen =
           tys
         in
         let drop_params x =
-          snd @@ EConstr.decompose_prod_n_assum sigma
+          snd @@ EConstr.decompose_prod_n_decls sigma
             mind.Declarations.mind_nparams (EConstr.of_constr x) in
         Array.map drop_params renamed_tys
       else
         subgoals_tys sigma ctx_concl
     in
-    let rctx = fst (EConstr.decompose_prod_assum sigma unfolded_c_ty) in
+    let rctx = fst (EConstr.decompose_prod_decls sigma unfolded_c_ty) in
     let n_c_args = Context.Rel.length rctx in
     let c, c_ty, t_args, sigma = saturate env sigma c ~ty:c_ty n_c_args in
     let elim, elimty, elim_args, sigma =
@@ -307,7 +309,7 @@ let check_pattern_instantiated env sigma patterns =
   let ev = List.fold_left Evar.Set.union Evar.Set.empty patterns_ev in
   let ty_ev = Evar.Set.fold (fun i e ->
         let ex = i in
-        let i_ty = Evd.evar_concl (Evd.find sigma ex) in
+        let i_ty = Evd.evar_concl (Evd.find_undefined sigma ex) in
         Evar.Set.union e (evars_of_term i_ty))
     ev Evar.Set.empty in
   let inter = Evar.Set.inter ev ty_ev in
@@ -358,7 +360,7 @@ let generate_pred env sigma0 ~concl patterns predty eqid is_rec deps elim_args n
         str"the defined ones matched")
     else match_all concl sigma postponed in
   let concl, sigma = match_all concl sigma patterns in
-  let pred_rctx, _ = EConstr.decompose_prod_assum sigma (fire_subst sigma predty) in
+  let pred_rctx, _ = EConstr.decompose_prod_decls sigma (fire_subst sigma predty) in
   let sigma, concl, gen_eq_tac, clr = match eqid with
   | Some (IPatId _) when not is_rec ->
       let k = List.length deps in
@@ -367,14 +369,14 @@ let generate_pred env sigma0 ~concl patterns predty eqid is_rec deps elim_args n
       let sigma, eq = get_eq_type env sigma in
       let sigma, gen_eq_tac, eq_ty =
         let refl = EConstr.mkApp (eq, [|t; c; c|]) in
-        let new_concl = EConstr.mkArrow refl Sorts.Relevant (EConstr.Vars.lift 1 concl0) in
+        let new_concl = EConstr.mkArrow refl EConstr.ERelevance.relevant (EConstr.Vars.lift 1 concl0) in
         let new_concl = fire_subst sigma new_concl in
         let sigma, erefl = mkRefl env sigma t c in
         let erefl = fire_subst sigma erefl in
         let erefl_ty = Retyping.get_type_of env sigma erefl in
         let eq_ty = Retyping.get_type_of env sigma erefl_ty in
-        let ucst = Evd.evar_universe_context sigma in
-        let evds = Evar.Map.bind (Evd.find sigma) @@
+        let ucst = Evd.ustate sigma in
+        let evds = Evar.Map.bind (Evd.find_undefined sigma) @@
           Evd.evars_of_term sigma new_concl in
         let gen_eq_tac =
           let open Proofview.Notations in
@@ -395,7 +397,7 @@ let generate_pred env sigma0 ~concl patterns predty eqid is_rec deps elim_args n
       in
       let rel = k + if c_is_head_p then 1 else 0 in
       let sigma, src = mkProt env sigma eq_ty EConstr.(mkApp (eq,[|t; c; mkRel rel|])) in
-      let concl = EConstr.mkArrow src Sorts.Relevant (EConstr.Vars.lift 1 concl) in
+      let concl = EConstr.mkArrow src EConstr.ERelevance.relevant (EConstr.Vars.lift 1 concl) in
       let clr = if deps <> [] then clr else [] in
       sigma, concl, gen_eq_tac, clr
   | _ -> sigma, concl, Tacticals.tclIDTAC, clr in
@@ -499,7 +501,7 @@ let ssrelim ?(is_case=false) deps what ?elim eqid elim_intro_tac =
   let elim = sigma, elim in
   let seed =
     Array.map (fun ty ->
-    let ctx,_ = EConstr.decompose_prod_assum sigma ty in
+    let ctx,_ = EConstr.decompose_prod_decls sigma ty in
     CList.rev_map Context.Rel.Declaration.get_name ctx) seed in
 
   let elim_tac =
@@ -526,9 +528,9 @@ let revtoptac n0 =
   let concl = Proofview.Goal.concl gl in
   let env = Proofview.Goal.env gl in
   let n = nb_prod sigma concl - n0 in
-  let dc, cl = EConstr.decompose_prod_n_assum sigma n concl in
+  let dc, cl = EConstr.decompose_prod_n_decls sigma n concl in
   let ty = EConstr.it_mkProd_or_LetIn cl (List.rev dc) in
-  let dc' = dc @ [Context.Rel.Declaration.LocalAssum(make_annot (Name rev_id) Sorts.Relevant, ty)] in
+  let dc' = dc @ [Context.Rel.Declaration.LocalAssum(make_annot (Name rev_id) EConstr.ERelevance.relevant, ty)] in
   Refine.refine ~typecheck:true begin fun sigma ->
     let f = EConstr.it_mkLambda_or_LetIn (mkEtaApp (EConstr.mkRel (n + 1)) (-n) 1) dc' in
     let sigma, ev = Evarutil.new_evar env sigma ty in
@@ -537,7 +539,7 @@ let revtoptac n0 =
   end
 
 let nothing_to_inject =
-   CWarnings.create ~name:"spurious-ssr-injection" ~category:"ssr"
+   CWarnings.create ~name:"spurious-ssr-injection" ~category:CWarnings.CoreCategories.ssr
      (fun (sigma, env, ty) ->
          Pp.(str "SSReflect: cannot obtain new equations out of" ++ fnl() ++
              str"  " ++ Printer.pr_econstr_env env sigma ty ++ fnl() ++
@@ -571,7 +573,7 @@ let injectl2rtac sigma c = match EConstr.kind sigma c with
 let is_injection_case env sigma c =
   let sigma, cty = Typing.type_of env sigma c in
   let (mind,_) = Tacred.eval_to_quantified_ind env sigma cty in
-  Coqlib.check_ind_ref "core.eq.type" mind
+  Rocqlib.check_ref "core.eq.type" (GlobRef.IndRef mind)
 
 let perform_injection c =
   let open Proofview.Notations in
@@ -587,7 +589,7 @@ let perform_injection c =
   let cl = Proofview.Goal.concl gl in
   let n = List.length dc in
   let c_eq = mkEtaApp c n 2 in
-  let cl1 = EConstr.mkLambda EConstr.(make_annot Anonymous Sorts.Relevant, mkArrow eqt Sorts.Relevant cl, mkApp (mkRel 1, [|c_eq|])) in
+  let cl1 = EConstr.mkLambda EConstr.(make_annot Anonymous ERelevance.relevant, mkArrow eqt ERelevance.relevant cl, mkApp (mkRel 1, [|c_eq|])) in
   let id = injecteq_id in
   let id_with_ebind = (EConstr.mkVar id, NoBindings) in
   let injtac = Tacticals.tclTHEN (introid id) (injectidl2rtac id id_with_ebind) in
