@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -15,7 +15,6 @@ open Util
 open Names
 open Constr
 open Context
-open Vars
 open Locus
 open Printer
 open Termops
@@ -29,12 +28,15 @@ open Ssrprinters
 open Ssrcommon
 open Proofview.Notations
 
+module ERelevance = EConstr.ERelevance
+
 let ssroldreworder = Summary.ref ~name:"SSR:oldreworder" false
 let () =
   Goptions.(declare_bool_option
-    { optkey   = ["SsrOldRewriteGoalsOrder"];
+    { optstage = Summary.Stage.Interp;
+      optkey   = ["SsrOldRewriteGoalsOrder"];
       optread  = (fun _ -> !ssroldreworder);
-      optdepr  = false;
+      optdepr  = None;
       optwrite = (fun b -> ssroldreworder := b) })
 
 (** The "simpl" tactic *)
@@ -85,7 +87,7 @@ let interp_congrarg_at env sigma ist ~concl n rf ty m =
       let rt = mkRApp congrn (args1 @  mkRApp rf (mkRHoles i) :: args2) in
       debug_ssr (fun () -> Pp.(str"rt=" ++ Printer.pr_glob_constr_env env sigma rt));
       Some (interp_refine env sigma ist ~concl rt)
-    with _ -> loop (i + 1) in
+    with e when CErrors.noncritical e -> loop (i + 1) in
   loop 0
 
 let pattern_id = mk_internal_id "pattern value"
@@ -98,7 +100,7 @@ let congrtac ((n, t), ty) ist =
   debug_ssr (fun () -> (Pp.str"===congr==="));
   debug_ssr (fun () -> Pp.(str"concl=" ++ Printer.pr_econstr_env env sigma concl));
   let nsigma, _ as it = interp_term env sigma ist t in
-  let sigma = Evd.merge_universe_context sigma (Evd.evar_universe_context nsigma) in
+  let sigma = Evd.merge_universe_context sigma (Evd.ustate nsigma) in
   let f, _, _ucst = abs_evars env sigma it in
   let ist' = {ist with lfun =
     Id.Map.add pattern_id (Tacinterp.Value.of_constr f) Id.Map.empty } in
@@ -151,22 +153,22 @@ let newssrcongrtac arg ist =
   (* utils *)
   let fs sigma t = Reductionops.nf_evar sigma t in
   let ssr_congr lr = EConstr.mkApp (arr, lr) in
-  let sigma, eq = Evd.fresh_global env sigma Coqlib.(lib_ref "core.eq.type") in
+  let sigma, eq = Evd.fresh_global env sigma Rocqlib.(lib_ref "core.eq.type") in
   (* here the two cases: simple equality or arrow *)
   let equality, _, eq_args, sigma' = saturate env sigma eq 3 in
   Proofview.Unsafe.tclEVARS sigma <*>
   tclMATCH_GOAL env sigma' equality
   (fun sigma' ->
-    let x = List.find_map (fun (n, x, _) -> if n = 0 then Some x else None) eq_args in
+    let x = List.find_map_exn (fun (n, x, _) -> if n = 0 then Some x else None) eq_args in
     let ty = fs sigma' x in
-    congrtac (arg, Detyping.detype Detyping.Now false Id.Set.empty env sigma ty) ist)
+    congrtac (arg, Detyping.detype Detyping.Now env sigma ty) ist)
   (fun () ->
     try
     let sigma, t_lhs = Evarutil.new_Type sigma in
     let sigma, t_rhs = Evarutil.new_Type sigma in
     let sigma, lhs = mk_evar env sigma t_lhs in
     let sigma, rhs = mk_evar env sigma t_rhs in
-    let arrow = EConstr.mkArrow lhs Sorts.Relevant (EConstr.Vars.lift 1 rhs) in
+    let arrow = EConstr.mkArrow lhs ERelevance.relevant (EConstr.Vars.lift 1 rhs) in
     tclMATCH_GOAL env sigma arrow
     (fun sigma ->
       let lr = [|fs sigma lhs;fs sigma rhs|] in
@@ -175,8 +177,8 @@ let newssrcongrtac arg ist =
                   ; Tactics.apply a
                   ; congrtac (arg, mkRType) ist ])
     (fun _ -> Tacticals.tclZEROMSG Pp.(str"Conclusion is not an equality nor an arrow"))
-    with e -> Proofview.tclZERO e (* FIXME *)
-    )
+    with e when CErrors.noncritical e -> Proofview.tclZERO e
+  )
   end
 
 (** 7. Rewriting tactics (rewrite, unlock) *)
@@ -240,11 +242,11 @@ let simplintac occ rdx sim =
   end
 
 let rec get_evalref env sigma c = match EConstr.kind sigma c with
-  | Var id -> Tacred.EvalVarRef id
-  | Const (k,_) -> Tacred.EvalConstRef k
+  | Var id -> Evaluable.EvalVarRef id
+  | Const (k,_) -> Evaluable.EvalConstRef k
   | App (c', _) -> get_evalref env sigma c'
   | Cast (c', _, _) -> get_evalref env sigma c'
-  | Proj(c,_) -> Tacred.EvalConstRef(Projection.constant c)
+  | Proj(p,_,_) -> Evaluable.EvalProjectionRef (Names.Projection.repr p)
   | _ -> errorstrm Pp.(str "The term " ++ pr_econstr_pat (Global.env ()) sigma c ++ str " is not unfoldable")
 
 (* Strip a pattern generated by a prenex implicit to its constant. *)
@@ -255,9 +257,9 @@ let strip_unfold_term _ ((sigma, t) as p) kt = match EConstr.kind sigma t with
   | Proj _ -> p, true
   | _ -> p, false
 
-let same_proj sigma t1 t2 =
+let same_proj env sigma t1 t2 =
   match EConstr.kind sigma t1, EConstr.kind sigma t2 with
-  | Proj(c1,_), Proj(c2, _) -> Projection.CanOrd.equal c1 c2
+  | Proj(c1,_,_), Proj(c2, _, _) -> Environ.QProjection.equal env c1 c2
   | _ -> false
 
 let classify_pattern p = match p with
@@ -274,7 +276,7 @@ let unfoldintac occ rdx t (kt,_) =
   let body env t c =
     Tacred.unfoldn [AllOccurrences, get_evalref env sigma t] env sigma0 c in
   let easy = occ = None && rdx = None in
-  let red_flags = if easy then CClosure.betaiotazeta else CClosure.betaiota in
+  let red_flags = if easy then RedFlags.betaiotazeta else RedFlags.betaiota in
   let beta env = Reductionops.clos_norm_flags red_flags env sigma0 in
   let unfold, conclude = match classify_pattern rdx with
   | None ->
@@ -296,13 +298,13 @@ let unfoldintac occ rdx t (kt,_) =
           match EConstr.kind sigma0 c with
           | Const _ when EConstr.eq_constr sigma0 c t -> body env t t
           | App (f,a) when EConstr.eq_constr sigma0 f t -> EConstr.mkApp (body env f f,a)
-          | Proj _ when same_proj sigma0 c t -> body env t c
+          | Proj _ when same_proj env sigma0 c t -> body env t c
           | _ ->
             let c = Reductionops.whd_betaiotazeta env sigma0 c in
             match EConstr.kind sigma0 c with
             | Const _ when EConstr.eq_constr sigma0 c t -> body env t t
             | App (f,a) when EConstr.eq_constr sigma0 f t -> EConstr.mkApp (body env f f,a)
-            | Proj _ when same_proj sigma0 c t -> body env t c
+            | Proj _ when same_proj env sigma0 c t -> body env t c
             | Const f -> aux (body env c c)
             | App (f, a) -> aux (EConstr.mkApp (body env f f, a))
             | _ -> errorstrm Pp.(str "The term "++ pr_econstr_env env sigma orig_c ++
@@ -310,7 +312,7 @@ let unfoldintac occ rdx t (kt,_) =
           in aux c
       else
         try body env t (fs (unify_HO env sigma c t) t)
-        with _ -> errorstrm Pp.(str "The term " ++
+        with e when CErrors.noncritical e -> errorstrm Pp.(str "The term " ++
           pr_econstr_env env sigma c ++spc()++ str "does not unify with " ++ pr_econstr_pat env sigma t)),
     ignore in
   let concl =
@@ -340,8 +342,9 @@ let foldtac occ rdx ft =
        try
          let sigma = unify_HO env sigma c t in
          Reductionops.nf_evar sigma t
-    with _ -> errorstrm Pp.(str "fold pattern " ++ pr_econstr_pat env sigma t ++ spc ()
-      ++ str "does not match redex " ++ pr_econstr_pat env sigma c)),
+       with e when CErrors.noncritical e ->
+         errorstrm Pp.(str "fold pattern " ++ pr_econstr_pat env sigma t ++ spc ()
+                       ++ str "does not match redex " ++ pr_econstr_pat env sigma c)),
     ignore in
   let concl = eval_pattern env0 sigma0 concl0 rdx occ fold in
   let () = conclude () in
@@ -352,16 +355,16 @@ let converse_dir = function L2R -> R2L | R2L -> L2R
 
 let rw_progress rhs lhs ise = not (EConstr.eq_constr ise lhs (Evarutil.nf_evar ise rhs))
 
-(* Coq has a more general form of "equation" (any type with a single *)
+(* Rocq has a more general form of "equation" (any type with a single *)
 (* constructor with no arguments with_rect_r elimination lemmas).    *)
 (* However there is no clear way of determining the LHS and RHS of   *)
 (* such a generic Leibniz equation -- short of inspecting the type   *)
 (* of the elimination lemmas.                                        *)
 
-let rec strip_prod_assum c = match Constr.kind c with
-  | Prod (_, _, c') -> strip_prod_assum c'
-  | LetIn (_, v, _, c') -> strip_prod_assum (subst1 v c)
-  | Cast (c', _, _) -> strip_prod_assum c'
+let rec strip_prod_expand_let sigma c = match EConstr.kind sigma c with
+  | Prod (_, _, c') -> strip_prod_expand_let sigma c'
+  | LetIn (_, v, _, c') -> strip_prod_expand_let sigma (EConstr.Vars.subst1 v c)
+  | Cast (c', _, _) -> strip_prod_expand_let sigma c'
   | _ -> c
 
 let rule_id = mk_internal_id "rewrite rule"
@@ -372,7 +375,7 @@ let id_map_redex _ sigma ~before:_ ~after = sigma, after
 
 (* Invariants expected from the arguments:
     ⊢ rdx : rdx_ty
-    pattern_id : rdx_ty ⊢ pred : Type@{s}
+    pattern_id : rdx_ty | #1 : rdx_ty ⊢ pred : Type@{s}
     ⊢ new_rdx : rdx_ty
     ⊢ c : c_ty
     ⊢ c_ty ≡ EQN rdx_ty rdx new_rdx
@@ -383,11 +386,8 @@ let pirrel_rewrite ?(under=false) ?(map_redex=id_map_redex) pred rdx rdx_ty new_
   Proofview.Goal.enter begin fun gl ->
 (*   ppdebug(lazy(str"sigma@pirrel_rewrite=" ++ pr_evar_map None sigma)); *)
   let env = pf_env gl in
-  let beta = Reductionops.clos_norm_flags CClosure.beta env sigma in
+  let beta = Reductionops.clos_norm_flags RedFlags.beta env sigma in
   let sigma, new_rdx = map_redex env sigma ~before:rdx ~after:new_rdx in
-  let sigma, p = (* The resulting goal *)
-    Evarutil.new_evar env sigma (beta (EConstr.Vars.subst1 new_rdx pred)) in
-  let pred = EConstr.mkNamedLambda sigma (make_annot pattern_id Sorts.Relevant) rdx_ty pred in
   let sigma, elim =
     let sort = Tacticals.elimination_sort_of_goal gl in
     match Equality.eq_elimination_ref (dir = L2R) sort with
@@ -401,22 +401,34 @@ let pirrel_rewrite ?(under=false) ?(map_redex=id_map_redex) pred rdx rdx_ty new_
       let mp,l = KerName.repr (Constant.canonical elim) in
       let l' = Label.of_id (Nameops.add_suffix (Label.to_id l) "_r")  in
       let c1' = Global.constant_of_delta_kn (Constant.canonical (Constant.make2 mp l')) in
-      sigma, EConstr.of_constr (mkConst c1')
+      Evd.fresh_global env sigma (ConstRef c1')
   in
+  (* The resulting goal *)
+  let sigma, p = Evarutil.new_evar env sigma (beta (EConstr.Vars.subst1 new_rdx pred)) in
   (* We check the proof is well typed. We assume that the type of [elim] is of
      the form [forall (A : Type) (x : A) (P : A -> Type@{s}), T] s.t. the only
      universes to unify are by checking the [A] and [P] arguments. *)
-  let sigma =
+  let sigma, p, pred =
     try
       let open EConstr in
       let elimT = Retyping.get_type_of env sigma elim in
       let (idA, tA, elimT) = destProd sigma elimT in
       let (_, _, elimT) = destProd sigma elimT in
       let (idP, tP, _) = destProd sigma elimT in
-      let sigma = Typing.check env sigma rdx_ty tA in
+      let sigma = Typing.check_actual_type env sigma (Retyping.get_judgment_of env sigma rdx_ty) tA in
       let tP = mkLetIn (idA, rdx_ty, tA, mkLetIn (anonR, mkProp, mkType Univ.Universe.type1, tP)) in
-      let sigma = Typing.check env sigma pred tP in
-      sigma
+      (* Do not fully retype pred, we already know that the domain is well-typed.
+         The way this is written makes it easier to profile which part of
+         typing is takes time. *)
+      let sigma, pred =
+        let id = make_annot (Name pattern_id) ERelevance.relevant in
+        let penv = EConstr.push_rel (LocalAssum (id, rdx_ty)) env in
+        let pred = Vars.subst_var sigma pattern_id pred in
+        let sigma, predty = Typing.type_of penv sigma pred in
+        sigma, { Environ.uj_val = mkLambda (id, rdx_ty, pred); uj_type = mkProd (id, rdx_ty, predty) }
+      in
+      let sigma = Typing.check_actual_type env sigma pred tP in
+      sigma, p, pred.uj_val
     with
     | Pretype_errors.PretypeError (env, sigma, te) -> raise (PRtype_error (Some (env, sigma, te)))
     | e when CErrors.noncritical e -> raise (PRtype_error None)
@@ -424,37 +436,37 @@ let pirrel_rewrite ?(under=false) ?(map_redex=id_map_redex) pred rdx rdx_ty new_
   let proof = EConstr.mkApp (elim, [| rdx_ty; new_rdx; pred; p; rdx; c |]) in
   debug_ssr (fun () -> Pp.(str"pirrel_rewrite: proof term: " ++ pr_econstr_env env sigma proof));
   Proofview.tclORELSE (refine_with
-    ~first_goes_last:(not !ssroldreworder || under) ~with_evars:under (sigma, proof))
-  (fun e ->
-    (* we generate a msg like: "Unable to find an instance for the variable" *)
-    let hd_ty, miss = match EConstr.kind sigma c with
-    | App (hd, args) ->
-        let hd_ty = Retyping.get_type_of env sigma hd in
-        let names = let rec aux env t = function 0 -> [] | n ->
-          let t = Reductionops.whd_all env sigma t in
-          let open EConstr in
-          match kind_of_type sigma t with
-          | ProdType (name, ty, t) ->
-              name.binder_name ::
-                aux (EConstr.push_rel (Context.Rel.Declaration.LocalAssum (name,ty)) env) t (n-1)
-          | _ ->
-              (* In the case the head is an HO constant it may accept more arguments *)
-              CList.init n (fun _ -> Names.Name.Anonymous) in aux env hd_ty (Array.length args) in
-        hd_ty, Util.List.map_filter (fun (t, name) ->
-          let evs = Evar.Set.elements (Evarutil.undefined_evars_of_term sigma t) in
-          let open_evs = List.filter (fun k ->
-            Sorts.InProp <> Retyping.get_sort_family_of
-              env sigma (Evd.evar_concl (Evd.find sigma k)))
-            evs in
-          if open_evs <> [] then Some name else None)
-          (List.combine (Array.to_list args) names)
-    | _ -> anomaly "rewrite rule not an application" in
-      tclZEROMSG Pp.(Himsg.explain_refiner_error env sigma (Logic.UnresolvedBindings miss)++
-      (Pp.fnl()++str"Rule's type:" ++ spc() ++ pr_econstr_env env sigma hd_ty)))
+                         ~first_goes_last:(not !ssroldreworder || under) ~with_evars:under (sigma, proof))
+    (fun (e, i) ->
+       (* we generate a msg like: "Unable to find an instance for the variable" *)
+       match EConstr.kind sigma c with
+       | App (hd, args) ->
+         let hd_ty = Retyping.get_type_of env sigma hd in
+         let names = let rec aux env t = function 0 -> [] | n ->
+             let t = Reductionops.whd_all env sigma t in
+             let open EConstr in
+             match kind_of_type sigma t with
+             | ProdType (name, ty, t) ->
+               name.binder_name ::
+               aux (EConstr.push_rel (Context.Rel.Declaration.LocalAssum (name,ty)) env) t (n-1)
+             | _ ->
+               (* In the case the head is an HO constant it may accept more arguments *)
+               CList.init n (fun _ -> Names.Name.Anonymous) in aux env hd_ty (Array.length args) in
+         let miss = Util.List.map_filter (fun (t, name) ->
+             let evs = Evar.Set.elements (Evarutil.undefined_evars_of_term sigma t) in
+             let open_evs = List.filter (fun k ->
+                 Sorts.InProp <> Retyping.get_sort_family_of
+                   env sigma (Evd.evar_concl (Evd.find_undefined sigma k)))
+                 evs in
+             if open_evs <> [] then Some name else None)
+             (List.combine (Array.to_list args) names) in
+         tclZEROMSG Pp.(Himsg.explain_refiner_error env sigma (Logic.UnresolvedBindings miss)++
+                        (Pp.fnl()++str"Rule's type:" ++ spc() ++ pr_econstr_env env sigma hd_ty))
+       | _ -> Proofview.tclZERO ~info:i e)
   end
 
 let pf_merge_uc_of s sigma =
-  Evd.merge_universe_context sigma (Evd.evar_universe_context s)
+  Evd.merge_universe_context sigma (Evd.ustate s)
 
 let rwcltac ?under ?map_redex cl rdx dir (sigma, r) =
   let open Proofview.Notations in
@@ -463,35 +475,35 @@ let rwcltac ?under ?map_redex cl rdx dir (sigma, r) =
   let sigma0 = Proofview.Goal.sigma gl in
   let sigma = resolve_typeclasses ~where:r ~fail:false env sigma in
   let r_n, evs, ucst = abs_evars env sigma0 (sigma, r) in
+  let sigma0 = Evd.set_universe_context sigma0 ucst in
   let n = List.length evs in
   let r_n' = abs_cterm env sigma0 n r_n in
   let r' = EConstr.Vars.subst_var sigma pattern_id r_n' in
-  let sigma0 = Evd.set_universe_context sigma0 ucst in
   let sigma, rdxt = Typing.type_of env sigma rdx in
   let () = debug_ssr (fun () -> Pp.(str"r@rwcltac=" ++ pr_econstr_env env sigma r)) in
   let cvtac, rwtac, sigma0 =
     if EConstr.Vars.closed0 sigma0 r' then
-      let c_eq = Coqlib.(lib_ref "core.eq.type") in
+      let c_eq = Rocqlib.(lib_ref "core.eq.type") in
       let sigma, c_ty = Typing.type_of env sigma r in
       let () = debug_ssr (fun () -> Pp.(str"c_ty@rwcltac=" ++ pr_econstr_env env sigma c_ty)) in
       let open EConstr in
       match kind_of_type sigma (Reductionops.whd_all env sigma c_ty) with
-      | AtomicType(e, a) when Ssrcommon.is_ind_ref sigma e c_eq ->
+      | AtomicType(e, a) when Ssrcommon.is_ind_ref env sigma e c_eq ->
           let new_rdx = if dir = L2R then a.(2) else a.(1) in
           pirrel_rewrite ?under ?map_redex cl rdx a.(0) new_rdx dir (sigma, r) c_ty, Tacticals.tclIDTAC, sigma0
       | _ ->
-          let cl' = EConstr.mkApp (EConstr.mkNamedLambda sigma (make_annot pattern_id Sorts.Relevant) rdxt cl, [|rdx|]) in
+          let cl' = EConstr.mkApp (EConstr.mkNamedLambda sigma (make_annot pattern_id ERelevance.relevant) rdxt cl, [|rdx|]) in
           let sigma, _ = Typing.type_of env sigma cl' in
           let sigma0 = pf_merge_uc_of sigma sigma0 in
           convert_concl ~check:true cl', rewritetac ?under dir r', sigma0
     else
-      let dc, r2 = EConstr.decompose_lam_n_assum sigma0 n r' in
+      let dc, r2 = EConstr.decompose_lambda_n_assum sigma0 n r' in
       let r3, _, r3t  =
-        try EConstr.destCast sigma0 r2 with _ ->
+        try EConstr.destCast sigma0 r2 with e when CErrors.noncritical e ->
         errorstrm Pp.(str "no cast from " ++ pr_econstr_pat env sigma0 r
                     ++ str " to " ++ pr_econstr_env env sigma0 r2) in
-      let cl' = EConstr.mkNamedProd sigma (make_annot rule_id Sorts.Relevant) (EConstr.it_mkProd_or_LetIn r3t dc) (EConstr.Vars.lift 1 cl) in
-      let cl'' = EConstr.mkNamedProd sigma (make_annot pattern_id Sorts.Relevant) rdxt cl' in
+      let cl' = EConstr.mkNamedProd sigma (make_annot rule_id ERelevance.relevant) (EConstr.it_mkProd_or_LetIn r3t dc) (EConstr.Vars.lift 1 cl) in
+      let cl'' = EConstr.mkNamedProd sigma (make_annot pattern_id ERelevance.relevant) rdxt cl' in
       let itacs = [introid pattern_id; introid rule_id] in
       let cltac = Tactics.clear [pattern_id; rule_id] in
       let rwtacs = [
@@ -515,7 +527,7 @@ let rwcltac ?under ?map_redex cl rdx dir (sigma, r) =
       then Tacticals.tclZEROMSG Pp.(str "Rewriting impacts evars" ++ error)
       else Tacticals.tclZEROMSG Pp.(str "Dependent type error in rewrite of "
         ++ pr_econstr_env env sigma0
-          (EConstr.mkNamedLambda sigma (make_annot pattern_id Sorts.Relevant) rdxt cl)
+          (EConstr.mkNamedLambda sigma (make_annot pattern_id ERelevance.relevant) rdxt cl)
         ++ error)
     | (e, info) -> Proofview.tclZERO ~info e
     end
@@ -524,20 +536,15 @@ let rwcltac ?under ?map_redex cl rdx dir (sigma, r) =
   Tacticals.tclTHEN cvtac' rwtac
   end
 
-[@@@ocaml.warning "-3"]
-let lz_coq_prod =
-  let prod = lazy (Coqlib.build_prod ()) in fun () -> Lazy.force prod
-
 let lz_setoid_relation =
-  let sdir = ["Classes"; "RelationClasses"] in
   let last_srel = ref None in
   fun env -> match !last_srel with
   | Some (env', srel) when env' == env -> srel
   | _ ->
     let srel =
        try Some (UnivGen.constr_of_monomorphic_global (Global.env ()) @@
-                 Coqlib.find_reference "Class_setoid" ("Coq"::sdir) "RewriteRelation" [@ocaml.warning "-3"])
-       with _ -> None in
+                 Rocqlib.lib_ref "rewrite.prop.RewriteRelation")
+       with e when CErrors.noncritical e -> None in
     last_srel := Some (env, srel); srel
 
 let ssr_is_setoid env =
@@ -555,33 +562,36 @@ let closed0_check env sigma cl p =
 let dir_org = function L2R -> 1 | R2L -> 2
 
 let rwprocess_rule env dir rule =
-  let coq_prod = lz_coq_prod () in
   let is_setoid = ssr_is_setoid env in
+  let prod_type = Rocqlib.lib_ref "core.prod.type" in
+  let prod_intro = Rocqlib.lib_ref "core.prod.intro" in
+  let prod_proj1 = Rocqlib.lib_ref "core.prod.proj1" in
+  let prod_proj2 = Rocqlib.lib_ref "core.prod.proj2" in
   let r_sigma, rules =
     let rec loop d sigma r t0 rs red =
       let t =
         if red = 1 then Tacred.hnf_constr env sigma t0
-        else Reductionops.clos_whd_flags CClosure.betaiotazeta env sigma t0 in
+        else Reductionops.clos_whd_flags RedFlags.betaiotazeta env sigma t0 in
       debug_ssr (fun () -> Pp.(str"rewrule="++pr_econstr_pat env sigma t));
       match EConstr.kind sigma t with
       | Prod (_, xt, at) ->
         let sigma = Evd.create_evar_defs sigma in
         let (sigma, x) = Evarutil.new_evar env sigma xt in
         loop d sigma EConstr.(mkApp (r, [|x|])) (EConstr.Vars.subst1 x at) rs 0
-      | App (pr, a) when is_ind_ref sigma pr coq_prod.Coqlib.typ ->
-        let r0 = Reductionops.clos_whd_flags CClosure.all env sigma r in
+      | App (pr, a) when is_ind_ref env sigma pr prod_type ->
+        let r0 = Reductionops.clos_whd_flags RedFlags.all env sigma r in
         let sigma, pL, pR = match EConstr.kind sigma r0 with
-        | App (c, ra) when is_construct_ref sigma c coq_prod.Coqlib.intro ->
+        | App (c, ra) when is_construct_ref env sigma c prod_intro ->
           (sigma, ra.(2), ra.(3))
         | _ ->
           let ra = Array.append a [|r|] in
-          let sigma, pi1 = Evd.fresh_global env sigma coq_prod.Coqlib.proj1 in
-          let sigma, pi2 = Evd.fresh_global env sigma coq_prod.Coqlib.proj2 in
+          let sigma, pi1 = Evd.fresh_global env sigma prod_proj1 in
+          let sigma, pi2 = Evd.fresh_global env sigma prod_proj2 in
           let pL = EConstr.mkApp (pi1, ra) in
           let pR = EConstr.mkApp (pi2, ra) in
           (sigma, pL, pR)
         in
-        if EConstr.isRefX sigma Coqlib.(lib_ref "core.True.type") a.(0) then
+        if EConstr.is_lib_ref env sigma "core.True.type" a.(0) then
          loop (converse_dir d) sigma pR a.(1) rs 0
         else
          let sigma, rs2 = loop d sigma pR a.(1) rs 0 in
@@ -589,9 +599,9 @@ let rwprocess_rule env dir rule =
       | App (r_eq, a) when Hipattern.match_with_equality_type env sigma t != None ->
         let (ind, u) = EConstr.destInd sigma r_eq and rhs = Array.last a in
         let np = Inductiveops.inductive_nparamdecls env ind in
-        let indu = (ind, EConstr.EInstance.kind sigma u) in
+        let indu = (ind, u) in
         let ind_ct = Inductiveops.type_of_constructors env indu in
-        let lhs0 = last_arg sigma (EConstr.of_constr (strip_prod_assum ind_ct.(0))) in
+        let lhs0 = last_arg sigma (strip_prod_expand_let sigma ind_ct.(0)) in
         let rdesc = match EConstr.kind sigma lhs0 with
         | Rel i ->
           let lhs = a.(np - i) in
@@ -644,8 +654,8 @@ let rwrxtac ?under ?map_redex occ rdx_pat dir rule =
         try
           let ise = unify_HO env (Evd.create_evar_defs r_sigma) lhs rdx in
           if not (rw_progress rhs rdx ise) then raise NoMatch else
-          d, (ise, Evd.evar_universe_context ise, Reductionops.nf_evar ise r)
-        with _ -> rwtac rs in
+          d, (ise, Evd.ustate ise, Reductionops.nf_evar ise r)
+        with e when CErrors.noncritical e -> rwtac rs in
      rwtac rules in
   let env0 = env in
   let concl0 = Proofview.Goal.concl gl in
@@ -712,10 +722,11 @@ let rwargtac ?under ?map_redex ist ((dir, mult), (((oclr, occ), grx), (kind, gt)
   let fail = ref false in
   let interp_rpattern env sigma gc =
     try interp_rpattern env sigma gc
-    with _ when snd mult = May -> fail := true; { pat_sigma = sigma; pat_pat = T EConstr.mkProp } in
+    with e when CErrors.noncritical e && snd mult = May ->
+      fail := true; { pat_sigma = sigma; pat_pat = T EConstr.mkProp } in
   let interp env sigma gc =
     try interp_term env sigma ist gc
-    with _ when snd mult = May -> fail := true; (sigma, EConstr.mkProp) in
+    with e when CErrors.noncritical e && snd mult = May -> fail := true; (sigma, EConstr.mkProp) in
   let rwtac =
     Proofview.Goal.enter begin fun gl ->
     let env = Proofview.Goal.env gl in
@@ -724,10 +735,10 @@ let rwargtac ?under ?map_redex ist ((dir, mult), (((oclr, occ), grx), (kind, gt)
     (* Evarmaps below are extensions of sigma, so setting the universe context is correct *)
     let sigma = match rx with
     | None -> sigma
-    | Some { pat_sigma = s } -> Evd.set_universe_context sigma (Evd.evar_universe_context s)
+    | Some { pat_sigma = s } -> Evd.set_universe_context sigma (Evd.ustate s)
     in
     let t = interp env sigma gt in
-    let sigma = Evd.set_universe_context sigma  (Evd.evar_universe_context (fst t)) in
+    let sigma = Evd.set_universe_context sigma  (Evd.ustate (fst t)) in
     Proofview.Unsafe.tclEVARS sigma <*>
     (match kind with
     | RWred sim -> simplintac occ rx sim
@@ -760,7 +771,7 @@ let unfoldtac occ ko t kt =
   let concl = Evarutil.nf_evar sigma concl in
   let cl, c = fill_occ_term env sigma concl occ (fst (strip_unfold_term env t kt)) in
   let cl' = EConstr.Vars.subst1 (Tacred.unfoldn [OnlyOccurrences [1], get_evalref env sigma c] env sigma c) cl in
-  let f = if ko = None then CClosure.betaiotazeta else CClosure.betaiota in
+  let f = if ko = None then RedFlags.betaiotazeta else RedFlags.betaiota in
   convert_concl ~check:true (Reductionops.clos_norm_flags f env sigma cl')
   end
 
