@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -26,7 +26,7 @@ module AdaptorDb = struct
 
   module AdaptorKind = struct
     type t = kind
-    let compare = pervasives_compare
+    let compare = Stdlib.compare
   end
   module AdaptorMap = Map.Make(AdaptorKind)
 
@@ -85,6 +85,7 @@ type vstate = {
 include Ssrcommon.MakeState(struct
   type state = vstate option
   let init = None
+  let name = "ssrview"
 end)
 
 let vsINIT ~view ~subject_name ~to_clear =
@@ -157,18 +158,19 @@ let is_tac_in_term ?extra_scope { annotation; body; glob_env; interp_env } =
         extra = ist.ast_extra;
         intern_sign = ist.ast_intern_sign;
         genv;
+        strict_check = true;
       }
     in
     (* We open extra_scope *)
     let body =
       match extra_scope with
       | None -> body
-      | Some s -> CAst.make (Constrexpr.CDelimiters(s,body))
+      | Some s -> CAst.make Constrexpr.(CDelimiters(DelimUnboundedScope,s,body))
     in
     (* We unravel notations *)
     let g = intern_constr_expr ist sigma body in
     match DAst.get g with
-    | Glob_term.GHole (_,_, Some x)
+    | Glob_term.GGenarg x
       when Genarg.has_type x (Genarg.glbwit Tacarg.wit_tactic)
         -> tclUNIT (`Tac (Genarg.out_gen (Genarg.glbwit Tacarg.wit_tactic) x))
     | _ -> tclUNIT (`Term (annotation, interp_env, g))
@@ -184,7 +186,7 @@ let tclINJ_CONSTR_IST ist p =
 
 let mkGHole =
   DAst.make
-    (Glob_term.GHole(Evar_kinds.InternalHole, Namegen.IntroAnonymous, None))
+    (Glob_term.GHole (GInternalHole))
 let rec mkGHoles n = if n > 0 then mkGHole :: mkGHoles (n - 1) else []
 let mkGApp f args =
   if args = [] then f
@@ -201,8 +203,7 @@ let interp_glob ist glob = Goal.enter_one ~__LOC__ begin fun goal ->
     Ssrprinters.debug_ssr (fun () ->
       Pp.(str"interp-out: " ++ Printer.pr_econstr_env env sigma term));
     tclUNIT (env,sigma,term)
-  with e ->
-    (* XXX this is another catch all! *)
+  with e when CErrors.noncritical e ->
     let e, info = Exninfo.capture e in
     Ssrprinters.debug_ssr (fun () ->
     Pp.(str"interp-err: " ++ Printer.pr_glob_constr_env env sigma glob));
@@ -237,14 +238,14 @@ let guess_max_implicits ist glob =
   Proofview.tclORELSE
     (interp_glob ist (mkGApp glob (mkGHoles 6)) >>= fun (env,sigma,term) ->
      let term_ty = Retyping.get_type_of env sigma term in
-     let ctx, _ = Reductionops.splay_prod env sigma term_ty in
+     let ctx, _ = Reductionops.whd_decompose_prod env sigma term_ty in
      tclUNIT (List.length ctx + 6))
   (fun _ -> tclUNIT 5)
 
 let pad_to_inductive ist glob = Goal.enter_one ~__LOC__ begin fun goal ->
   interp_glob ist glob >>= fun (env, sigma, term as ot) ->
   let term_ty = Retyping.get_type_of env sigma term in
-  let ctx, i = Reductionops.splay_prod env sigma term_ty in
+  let ctx, i = Reductionops.whd_decompose_prod env sigma term_ty in
   let rel_ctx =
     List.map (fun (a,b) -> Context.Rel.Declaration.LocalAssum(a,b)) ctx in
   if not (Ssrcommon.isAppInd (EConstr.push_rel_context rel_ctx env) sigma i)
@@ -307,20 +308,21 @@ Goal.enter_one ~__LOC__ begin fun g ->
   let sigma = Typeclasses.resolve_typeclasses ~fail:false ~filter env sigma in
   let _, sigma = Evd.pop_shelf sigma in
   let p = Reductionops.nf_evar sigma p in
-  let get_body = function Evd.Evar_defined x -> x | _ -> assert false in
+  let get_body : type a. a Evd.evar_body -> EConstr.t = function Evd.Evar_defined x -> x | Evd.Evar_empty -> assert false in
   let evars_of_econstr sigma t =
     Evarutil.undefined_evars_of_term sigma t in
   let rigid_of s =
     List.fold_left (fun l k ->
       if Evd.is_defined sigma k then
-        let bo = get_body Evd.(evar_body (find sigma k)) in
+        let EvarInfo evi = Evd.find sigma k in
+        let bo = get_body (Evd.evar_body evi) in
           k :: l @ Evar.Set.elements (evars_of_econstr sigma bo)
       else l
     ) [] s in
   let env0 = Proofview.Goal.env s0 in
   let sigma0 = Proofview.Goal.sigma s0 in
   let und0 = (* Unassigned evars in the initial goal *)
-    let g0info = Evd.find sigma0 (Proofview.Goal.goal s0) in
+    let EvarInfo g0info = Evd.find sigma0 (Proofview.Goal.goal s0) in
     let g0 = Evd.evars_of_filtered_evar_info sigma0 g0info in
     List.filter (fun k -> Evar.Set.mem k g0)
       (List.map fst (Evar.Map.bindings (Evd.undefined_map sigma0))) in
@@ -335,7 +337,7 @@ Goal.enter_one ~__LOC__ begin fun g ->
 end
 
 let pose_proof subject_name p =
-  Tactics.generalize [p] <*>
+  Generalize.generalize [p] <*>
   begin match subject_name with
   | id :: _ -> Ssrcommon.tclRENAME_HD_PROD (Name.Name id)
   | _ -> tclUNIT() end
