@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -17,7 +17,6 @@ open Environ
 open Util
 open Libobject
 
-module NamedDecl = Context.Named.Declaration
 (*i*)
 
 let name_table =
@@ -26,16 +25,16 @@ let name_table =
 
 type req =
   | ReqLocal
-  | ReqGlobal of GlobRef.t * Name.t list
+  | ReqGlobal
 
-let load_rename_args _ (_, (_, (r, names))) =
+let load_rename_args _ (_, (r, names)) =
   name_table := GlobRef.Map.add r names !name_table
 
 let cache_rename_args o = load_rename_args 1 o
 
 let classify_rename_args = function
   | ReqLocal, _ -> Dispose
-  | ReqGlobal _, _ as o -> Substitute o
+  | ReqGlobal, _ -> Substitute
 
 let subst_rename_args (subst, (_, (r, names as orig))) =
   ReqLocal,
@@ -43,29 +42,33 @@ let subst_rename_args (subst, (_, (r, names as orig))) =
   if r==r' then orig else (r', names)
 
 let discharge_rename_args = function
-  | _, (ReqGlobal (c, names), _ as req) when not (isVarRef c && Lib.is_in_section c) ->
+  | ReqGlobal, (c, names) as req when not (isVarRef c && Lib.is_in_section c) ->
      (try
-       let vars = Lib.variable_section_segment_of_reference c in
-       let var_names = List.map (NamedDecl.get_id %> Name.mk_name) vars in
-       let names' = var_names @ names in
-       Some (ReqGlobal (c, names), (c, names'))
+       let var_names = Array.map_to_list (fun c -> Name (destVar c)) (Lib.section_instance c) in
+       let names = var_names @ names in
+       Some (ReqGlobal, (c, names))
      with Not_found -> Some req)
   | _ -> None
 
-let rebuild_rename_args x = x
-
 let inRenameArgs = declare_object { (default_object "RENAME-ARGUMENTS" ) with
   load_function = load_rename_args;
   cache_function = cache_rename_args;
   classify_function = classify_rename_args;
   subst_function = subst_rename_args;
   discharge_function = discharge_rename_args;
-  rebuild_function = rebuild_rename_args;
 }
 
 let rename_arguments local r names =
-  let req = if local then ReqLocal else ReqGlobal (r, names) in
-  Lib.add_anonymous_leaf (inRenameArgs (req, (r, names)))
+  let () = match r with
+    | GlobRef.VarRef id ->
+      CErrors.user_err
+        Pp.(str "Arguments of section variables such as "
+            ++ Id.print id
+            ++ str" may not be renamed.")
+    | _ -> ()
+  in
+  let req = if local then ReqLocal else ReqGlobal in
+  Lib.add_leaf (inRenameArgs (req, (r, names)))
 
 let arguments_names r = GlobRef.Map.find r !name_table
 
@@ -88,18 +91,6 @@ let rename_type ty ref =
   try rename_type_aux ty (arguments_names ref)
   with Not_found -> ty
 
-let rename_type_of_constant env c =
-  let ty = Typeops.type_of_constant_in env c in
-  rename_type ty (GlobRef.ConstRef (fst c))
-
-let rename_type_of_inductive env ind =
-  let ty = Inductiveops.type_of_inductive env ind in
-  rename_type ty (GlobRef.IndRef (fst ind))
-
-let rename_type_of_constructor env cstruct =
-  let ty = Inductiveops.type_of_constructor env cstruct in
-  rename_type ty (GlobRef.ConstructRef (fst cstruct))
-
 let rename_typing env c =
   let j = Typeops.infer env c in
   let j' =
