@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -45,14 +45,15 @@ open Esubst
 type cbv_value =
   | VAL of int * constr
   | STACK of int * cbv_value * cbv_stack
-  | LAMBDA of int * (Name.t Context.binder_annot * types) list * constr * cbv_value subs
-  | PROD of Name.t Context.binder_annot * types * types * cbv_value subs
-  | LETIN of Name.t Context.binder_annot * cbv_value * types * constr * cbv_value subs
+  | LAMBDA of int * (Name.t Constr.binder_annot * types) list * constr * cbv_value subs
+  | PROD of Name.t Constr.binder_annot * types * types * cbv_value subs
+  | LETIN of Name.t Constr.binder_annot * cbv_value * types * constr * cbv_value subs
   | FIX of fixpoint * cbv_value subs * cbv_value array
   | COFIX of cofixpoint * cbv_value subs * cbv_value array
   | CONSTRUCT of constructor UVars.puniverses * cbv_value array
   | PRIMITIVE of CPrimitives.t * pconstant * cbv_value array
   | ARRAY of UVars.Instance.t * cbv_value Parray.t * cbv_value
+  | SYMBOL of { cst: Constant.t UVars.puniverses; unfoldfix: bool; rules: Declarations.rewrite_rule list; stk: cbv_stack }
 
 (* type of terms with a hole. This hole can appear only under App or Case.
  *   TOP means the term is considered without context
@@ -100,6 +101,13 @@ let rec shift_value n = function
       PRIMITIVE(op,c,Array.map (shift_value n) args)
   | ARRAY (u,t,ty) ->
       ARRAY(u, Parray.map (shift_value n) t, shift_value n ty)
+  | SYMBOL s -> SYMBOL { s with stk = shift_stack n s.stk }
+
+and shift_stack n = function (* Slow *)
+  | TOP -> TOP
+  | APP (args, stk) -> APP (List.map (shift_value n) args, shift_stack n stk)
+  | CASE (u,pms,c,b,iv,i,s,stk) -> CASE (u,pms,c,b,iv,i,subs_shft(n,s), shift_stack n stk)
+  | PROJ (p, r, stk) -> PROJ (p, r, shift_stack n stk)
 
 let shift_value n v =
   if Int.equal n 0 then v else shift_value n v
@@ -165,7 +173,7 @@ end)
 
 type cbv_infos = {
   env : Environ.env;
-  tab : (cbv_value, Empty.t) Declarations.constant_def KeyTable.t;
+  tab : (cbv_value, Empty.t, bool * Declarations.rewrite_rule list) Declarations.constant_def KeyTable.t;
   reds : RedFlags.reds;
   sigma : Evd.evar_map;
   strong : bool;
@@ -190,8 +198,22 @@ let strip_appl head stack =
     | COFIX (cofix,env,app) -> (COFIX(cofix,env,[||]), stack_vect_app app stack)
     | CONSTRUCT (c,app) -> (CONSTRUCT(c,[||]), stack_vect_app app stack)
     | PRIMITIVE(op,c,app) -> (PRIMITIVE(op,c,[||]), stack_vect_app app stack)
-    | LETIN _ | VAL _ | STACK _ | PROD _ | LAMBDA _ | ARRAY _ -> (head, stack)
+    | LETIN _ | VAL _ | STACK _ | PROD _ | LAMBDA _ | ARRAY _ | SYMBOL _ -> (head, stack)
 
+let destack head stack =
+  match head with
+  | FIX (fix,env,app) -> (FIX(fix,env,[||]), stack_vect_app app stack)
+  | COFIX (cofix,env,app) -> (COFIX(cofix,env,[||]), stack_vect_app app stack)
+  | CONSTRUCT (c,app) -> (CONSTRUCT(c,[||]), stack_vect_app app stack)
+  | PRIMITIVE(op,c,app) -> (PRIMITIVE(op,c,[||]), stack_vect_app app stack)
+  | STACK (k, v, stk) -> (shift_value k v, stack_concat (shift_stack k stk) stack)
+  | SYMBOL ({ stk } as s) -> (SYMBOL { s with stk=TOP }, stack_concat stk stack)
+  | LETIN _ | VAL _ | PROD _ | LAMBDA _ | ARRAY _ -> (head, stack)
+
+let rec fixp_reducible_symb_stk = function
+  | TOP -> true
+  | APP (_, stk) -> fixp_reducible_symb_stk stk
+  | CASE _ | PROJ _ -> false
 
 (* Tests if fixpoint reduction is possible. *)
 let fixp_reducible flgs ((reci,i),_) stk =
@@ -203,6 +225,8 @@ let fixp_reducible flgs ((reci,i),_) stk =
         | v :: appl ->
           if Int.equal n 0 then match v with
           | CONSTRUCT _ -> true
+          | SYMBOL { unfoldfix=true; stk; _ } ->
+              fixp_reducible_symb_stk stk
           | _ -> false
           else check (n - 1) appl
         in
@@ -251,6 +275,14 @@ module VNativeEntries =
         | _ -> raise Primred.NativeDestKO)
       | _ -> raise Primred.NativeDestKO
 
+    let get_string () e =
+      match e with
+      | VAL(_, cf) ->
+        (match kind cf with
+        | String s -> s
+        | _ -> raise Primred.NativeDestKO)
+      | _ -> raise Primred.NativeDestKO
+
     let get_parray () e =
       match e with
       | ARRAY(_u,t,_ty) -> t
@@ -260,6 +292,8 @@ module VNativeEntries =
 
     let mkFloat env f = VAL(0, mkFloat f)
 
+    let mkString env s = VAL(0, mkString s)
+
     let mkBool env b =
       let (ct,cf) = get_bool_constructors env in
       CONSTRUCT(UVars.in_punivs (if b then ct else cf), [||])
@@ -405,6 +439,8 @@ and reify_value = function (* reduction under binders *)
   | ARRAY (u,t,ty) ->
     let t, def = Parray.to_array t in
       mkArray(u, Array.map reify_value t, reify_value def, reify_value ty)
+  | SYMBOL { cst; stk; _ } ->
+      reify_stack (mkConstU cst) stk
 
 and apply_env env t =
   match kind t with
@@ -418,6 +454,22 @@ and apply_env env t =
   | _ ->
     map_with_binders subs_lift apply_env env t
 
+let apply_env_context e ctx =
+  let open Context.Rel.Declaration in
+  let rec subst_context ctx = match ctx with
+  | [] -> e, []
+  | LocalAssum (na, ty) :: ctx ->
+    let e, ctx = subst_context ctx in
+    let ty = apply_env e ty in
+    subs_lift e, LocalAssum (na, ty) :: ctx
+  | LocalDef (na, ty, bdy) :: ctx ->
+    let e, ctx = subst_context ctx in
+    let ty = apply_env e ty in
+    let bdy = apply_env e bdy in
+    subs_lift e, LocalDef (na, ty, bdy) :: ctx
+  in
+  snd @@ subst_context ctx
+
 let rec strip_app = function
   | APP (args,st) -> APP (args,strip_app st)
   | s -> TOP
@@ -455,6 +507,8 @@ let cbv_subst_of_rel_context_instance_list mkclos sign args env =
  * argument is [].  Because we must put all the applied terms in the
  * stack. *)
 
+exception PatternFailure
+
 let rec norm_head info env t stack =
   (* no reduction under binders *)
   match kind t with
@@ -473,8 +527,7 @@ let rec norm_head info env t stack =
 
   | Proj (p, r, c) ->
     let p' =
-      if red_set info.reds (fCONST (Projection.constant p))
-        && red_set info.reds fBETA
+      if red_set info.reds (fPROJ (Projection.repr p))
       then Projection.unfold p
       else p
     in
@@ -537,7 +590,7 @@ let rec norm_head info env t stack =
     (ARRAY (u,t,ty), stack)
 
   (* neutral cases *)
-  | (Sort _ | Meta _ | Ind _ | Int _ | Float _) -> (VAL(0, t), stack)
+  | (Sort _ | Meta _ | Ind _ | Int _ | Float _ | String _) -> (VAL(0, t), stack)
   | Prod (na,t,u) -> (PROD(na,t,u,env), stack)
 
 and norm_head_ref k info env stack normt t =
@@ -552,6 +605,13 @@ and norm_head_ref k info env stack normt t =
           | RelKey _ | VarKey _ -> assert false
         in
         (PRIMITIVE(op,c,[||]),stack)
+      | Declarations.Symbol (unfoldfix, rules) ->
+        assert (k = 0);
+        let cst = match normt with
+          | ConstKey c -> c
+          | RelKey _ | VarKey _ -> assert false
+        in
+        (SYMBOL { cst; unfoldfix; rules; stk=TOP }, stack)
       | Declarations.OpaqueDef _ | Declarations.Undef _ ->
          debug_cbv (fun () -> Pp.(str "Not unfolding " ++ debug_pr_key normt));
          (VAL(0,make_constr_ref k normt t),stack)
@@ -654,7 +714,15 @@ and cbv_stack_value info env = function
         (* partial application *)
               (assert (stk = TOP);
                PRIMITIVE(op,c,Array.of_list appl))
-      end
+        end
+    | SYMBOL ({ cst; rules; stk } as s ), stk' ->
+        let stk = stack_concat stk stk' in
+        begin try
+          let rhs, stack = cbv_apply_rules info env (snd cst) rules stk in
+          cbv_stack_value info env (destack rhs stack)
+        with PatternFailure ->
+          SYMBOL { s with stk }
+        end
 
     (* definitely a value *)
     | (head,stk) -> mkSTACK(head, stk)
@@ -683,10 +751,159 @@ and cbv_value_cache info ref =
         Declarations.Def v
       with
       | Environ.NotEvaluableConst (Environ.IsPrimitive (_u,op)) -> Declarations.Primitive op
+      | Environ.NotEvaluableConst (Environ.HasRules (u, b, r)) -> Declarations.Symbol (b, r)
       | Not_found | Environ.NotEvaluableConst _ -> Declarations.Undef None
     in
     KeyTable.add info.tab ref v; v
 
+
+and it_mkLambda_or_LetIn info ctx t =
+  let open Context.Rel.Declaration in
+  match List.rev ctx with
+  | [] -> t
+  | LocalAssum (n, ty) :: ctx ->
+      let assums, ctx = List.map_until (function LocalAssum (n, ty) -> Some (n, ty) | LocalDef _ -> None) ctx in
+      let assums = (n, ty) :: assums in
+      LAMBDA (List.length assums, assums, Term.it_mkLambda_or_LetIn (reify_value t) (List.rev ctx), subs_id 0)
+  | LocalDef _ :: _ ->
+      cbv_stack_term info TOP (subs_id 0) (Term.it_mkLambda_or_LetIn (reify_value t) ctx)
+
+and cbv_match_arg_pattern info env ctx psubst p t =
+  let open Declarations in
+  let t' = it_mkLambda_or_LetIn info ctx t in
+  match p with
+  | EHole i -> Partial_subst.add_term i t' psubst
+  | EHoleIgnored -> psubst
+  | ERigid (ph, es) ->
+      let t, stk = destack t TOP in
+      let psubst = cbv_match_rigid_arg_pattern info env ctx psubst ph t in
+      let psubst, stk = cbv_apply_rule info env ctx psubst es stk in
+      match stk with
+      | TOP -> psubst
+      | APP _| CASE _ | PROJ _ -> raise PatternFailure
+
+and cbv_match_arg_pattern_lift info env ctx n psubst p t =
+  let env = subs_liftn n env in
+  cbv_match_arg_pattern info env ctx psubst p
+    (cbv_stack_term info TOP env t)
+
+and match_sort ps s subst =
+  match Sorts.pattern_match ps s subst with
+  | Some subst -> subst
+  | None -> raise PatternFailure
+
+and match_instance pu u psubst =
+  match UVars.Instance.pattern_match pu u psubst with
+  | Some subst -> subst
+  | None -> raise PatternFailure
+
+
+and cbv_match_rigid_arg_pattern info env ctx psubst p t =
+  let open Declarations in
+  match [@ocaml.warning "-4"] p, t with
+  | PHInd (ind, pu), VAL(0, t') ->
+    begin match kind t' with Ind (ind', u) when Ind.CanOrd.equal ind ind' -> match_instance pu u psubst | _ -> raise PatternFailure end
+  | PHConstr (constr, pu), CONSTRUCT ((constr', u), [||]) ->
+    if Construct.CanOrd.equal constr constr' then match_instance pu u psubst else raise PatternFailure
+  | PHRel i, VAL(k, t') ->
+    begin match kind t' with Rel n when Int.equal i (k + n) -> psubst | _ -> raise PatternFailure end
+  | PHSort ps, VAL(0, t') ->
+    begin match kind t' with Sort s -> match_sort ps s psubst | _ -> raise PatternFailure end
+  | PHSymbol (c, pu), SYMBOL { cst = c', u; _ } ->
+    if Constant.CanOrd.equal c c' then match_instance pu u psubst else raise PatternFailure
+  | PHInt i, VAL(0, t') ->
+    begin match kind t' with Int i' when Uint63.equal i i' -> psubst | _ -> raise PatternFailure end
+  | PHFloat f, VAL(0, t') ->
+    begin match kind t' with Float f' when Float64.equal f f' -> psubst | _ -> raise PatternFailure end
+  | PHString s, VAL(0, t') ->
+    begin match kind t' with String s' when Pstring.equal s s' -> psubst | _ -> raise PatternFailure end
+  | PHLambda (ptys, pbod), LAMBDA (nlam, ntys, body, env) ->
+    let np = Array.length ptys in
+    if np > nlam then raise PatternFailure;
+    let ntys, body =
+      if np = nlam then ntys, body
+      else (* np < nlam *)
+        let ntys, tys' = List.chop np ntys in
+        ntys, Term.compose_lam (List.rev tys') body
+    in
+    let ctx' = List.rev_map (fun (n, ty) -> Context.Rel.Declaration.LocalAssum (n, ty)) ntys in
+    let ctx' = apply_env_context env ctx' in
+    let tys = Array.of_list ntys in
+    let contexts_upto = Array.init np (fun i -> List.lastn i ctx' @ ctx) in
+    let psubst = Array.fold_left3_i (fun i psubst ctx pty (_, ty) -> cbv_match_arg_pattern_lift info env ctx i psubst pty ty) psubst contexts_upto ptys tys in
+    let psubst = cbv_match_arg_pattern_lift info env (ctx' @ ctx) np psubst pbod body in
+    psubst
+  | PHProd (ptys, pbod), PROD (na, ty, body, env) ->
+    let ntys, _ = Term.decompose_prod body in
+    let np = Array.length ptys in
+    let nprod = 1 + List.length ntys in
+    if np > nprod then raise PatternFailure;
+    let ntys, body = Term.decompose_prod_n (np-1) body in
+    let ctx' = List.map (fun (n, ty) -> Context.Rel.Declaration.LocalAssum (n, ty)) (ntys @ [na, ty]) in
+    let ctx' = apply_env_context env ctx' in
+    let tys = Array.of_list ((na, ty) :: List.rev ntys) in
+    let na = Array.length tys in
+    let contexts_upto = Array.init na (fun i -> List.lastn i ctx' @ ctx) in
+    let psubst = Array.fold_left3_i (fun i psubst ctx pty (_, ty) -> cbv_match_arg_pattern_lift info env ctx i psubst pty ty) psubst contexts_upto ptys tys in
+    let psubst = cbv_match_arg_pattern_lift info env (ctx' @ ctx) na psubst pbod body in
+    psubst
+  | (PHInd _ | PHConstr _ | PHRel _ | PHInt _ | PHFloat _ | PHString _ | PHSort _ | PHSymbol _ | PHLambda _ | PHProd _), _ -> raise PatternFailure
+
+
+and cbv_apply_rule info env ctx psubst es stk =
+  match [@ocaml.warning "-4"] es, stk with
+  | [], _ -> psubst, stk
+  | Declarations.PEApp pargs :: e, APP (args, s) ->
+      let args = Array.of_list args in
+      let np = Array.length pargs in
+      let na = Array.length args in
+      if np == na then
+        let psubst = Array.fold_left2 (cbv_match_arg_pattern info env ctx) psubst pargs args in
+        cbv_apply_rule info env ctx psubst e s
+      else if np < na then (* more real arguments *)
+        let usedargs, remargs = Array.chop np args in
+        let psubst = Array.fold_left2 (cbv_match_arg_pattern info env ctx) psubst pargs usedargs in
+        cbv_apply_rule info env ctx psubst e (APP (Array.to_list remargs, s))
+      else (* more pattern arguments *)
+        let usedpargs, rempargs = Array.chop na pargs in
+        let psubst = Array.fold_left2 (cbv_match_arg_pattern info env ctx) psubst usedpargs args in
+        cbv_apply_rule info env ctx psubst (PEApp rempargs :: e) s
+  | Declarations.PECase (pind, pu, pret, pbrs) :: e, CASE (u, pms, (p, _), brs, iv, ci, env, s) ->
+      if not @@ Ind.CanOrd.equal pind ci.ci_ind then raise PatternFailure;
+      let specif = Inductive.lookup_mind_specif info.env ci.ci_ind in
+      let ntys_ret = Inductive.expand_arity specif (ci.ci_ind, u) pms (fst p) in
+      let ntys_ret = apply_env_context env ntys_ret in
+      let ntys_brs = Inductive.expand_branch_contexts specif u pms brs in
+      let psubst = match_instance pu u psubst in
+      let brs = Array.map2 (fun ctx' br -> List.length ctx', ctx' @ ctx, (snd br)) ntys_brs brs in
+      let psubst = cbv_match_arg_pattern_lift info env (ntys_ret @ ctx) (List.length ntys_ret) psubst pret (snd p) in
+      let psubst = Array.fold_left2 (fun psubst pat (n, ctx, br) -> cbv_match_arg_pattern_lift info env (apply_env_context env ctx) n psubst pat br) psubst pbrs brs in
+      cbv_apply_rule info env ctx psubst e s
+  | Declarations.PEProj proj :: e, PROJ (proj', r, s) ->
+      if not @@ Projection.(Repr.CanOrd.equal proj (repr proj')) then raise PatternFailure;
+      cbv_apply_rule info env ctx psubst e s
+  | _, _ -> raise PatternFailure
+
+
+and cbv_apply_rules info env u r stk =
+  match r with
+  | [] -> raise PatternFailure
+  | { lhs_pat = (pu, elims); nvars; rhs } :: rs ->
+    try
+      let psubst = Partial_subst.make nvars in
+      let psubst = match_instance pu u psubst in
+      let psubst, stk = cbv_apply_rule info env [] psubst elims stk in
+      let subst, qsubst, usubst = Partial_subst.to_arrays psubst in
+      let subst = Array.fold_right subs_cons subst env in
+      let usubst = UVars.Instance.of_array (qsubst, usubst) in
+      let rhsu = Vars.subst_instance_constr usubst rhs in
+      let rhs' = cbv_stack_term info TOP subst rhsu in
+      rhs', stk
+    with PatternFailure -> cbv_apply_rules info env u rs stk
+
+
+
+
 (* When we are sure t will never produce a redex with its stack, we
  * normalize (even under binders) the applied terms and we build the
  * final term
@@ -773,6 +990,7 @@ and cbv_norm_value info = function
     let t, def = Parray.to_array t in
     let def = cbv_norm_value info def in
       mkArray(u, Array.map (cbv_norm_value info) t, def, ty)
+  | SYMBOL { cst; stk; _ } -> apply_stack info (mkConstU cst) stk
 
 (* with profiling *)
 let cbv_norm infos constr =
