@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -108,7 +108,7 @@ let get_polymorphic_positions env sigma f =
     let mib,oib = Inductive.lookup_mind_specif env ind in
       (match mib.mind_template with
       | None -> assert false
-      | Some templ -> templ.template_param_levels)
+      | Some templ -> templ.template_param_arguments)
   | _ -> assert false
 
 let refresh_universes ?(status=univ_rigid) ?(onlyalg=false) ?(refreshset=false)
@@ -116,8 +116,12 @@ let refresh_universes ?(status=univ_rigid) ?(onlyalg=false) ?(refreshset=false)
   let evdref = ref evd in
   (* direction: true for fresh universes lower than the existing ones *)
   let refresh_sort status ~direction s =
-    let s = ESorts.kind !evdref s in
-    let sigma, s' = new_sort_variable status !evdref in
+    let sigma, l = new_univ_level_variable status !evdref in
+    let s' = match ESorts.kind sigma s with
+      | QSort (q, _) -> Sorts.qsort q (Univ.Universe.make l)
+      | _ -> Sorts.sort_of_univ @@ Univ.Universe.make l
+    in
+    let s' = ESorts.make s' in
     evdref := sigma;
     let evd =
       if direction then set_leq_sort env !evdref s' s
@@ -128,14 +132,15 @@ let refresh_universes ?(status=univ_rigid) ?(onlyalg=false) ?(refreshset=false)
     match EConstr.kind !evdref t with
     | Sort s ->
       begin match ESorts.kind !evdref s with
-      | Type u ->
+      | Type u | QSort (_, u) ->
          (* TODO: check if max(l,u) is not ok as well *)
         (match Univ.Universe.level u with
         | None -> refresh_sort status ~direction s
         | Some l ->
            (match Evd.universe_rigidity !evdref l with
             | UnivRigid ->
-               if not onlyalg then refresh_sort status ~direction s
+              if not onlyalg && (not (Univ.Level.is_set l) || (refreshset && not direction))
+              then refresh_sort status ~direction s
                else t
             | UnivFlexible alg ->
                (if alg then
@@ -145,7 +150,7 @@ let refresh_universes ?(status=univ_rigid) ?(onlyalg=false) ?(refreshset=false)
        (* Cannot make a universe "lower" than "Set",
           only refreshing when we want higher universes. *)
        refresh_sort status ~direction s
-      | _ -> t
+      | Prop | SProp | Set -> t
       end
     | Prod (na,u,v) ->
        let v' = refresh ~onlyalg status ~direction v in
@@ -155,7 +160,7 @@ let refresh_universes ?(status=univ_rigid) ?(onlyalg=false) ?(refreshset=false)
   (* Refresh the types of evars under template polymorphic references *)
   let rec refresh_term_evars ~onevars ~top t =
     match EConstr.kind !evdref t with
-    | App (f, args) when Termops.is_template_polymorphic_ind env !evdref f ->
+    | App (f, args) when Termops.is_template_polymorphic_ref env !evdref f ->
       let pos = get_polymorphic_positions env !evdref f in
         refresh_polymorphic_positions args pos; t
     | App (f, args) when top && isEvar !evdref f ->
@@ -164,7 +169,7 @@ let refresh_universes ?(status=univ_rigid) ?(onlyalg=false) ?(refreshset=false)
        if f' == f && args' == args then t
        else mkApp (f', args')
     | Evar (ev, a) when onevars ->
-      let evi = Evd.find !evdref ev in
+      let evi = Evd.find_undefined !evdref ev in
       let ty = Evd.evar_concl evi in
       let ty' = refresh ~onlyalg univ_flexible ~direction:true ty in
       if ty == ty' then t
@@ -177,11 +182,11 @@ let refresh_universes ?(status=univ_rigid) ?(onlyalg=false) ?(refreshset=false)
     | _ -> EConstr.map !evdref (refresh_term_evars ~onevars ~top:false) t
   and refresh_polymorphic_positions args pos =
     let rec aux i = function
-      | Some l :: ls ->
+      | true :: ls ->
         if i < Array.length args then
           ignore(refresh_term_evars ~onevars:true ~top:false args.(i));
         aux (succ i) ls
-      | None :: ls ->
+      | false :: ls ->
         if i < Array.length args then
           ignore(refresh_term_evars ~onevars:false ~top:false args.(i));
         aux (succ i) ls
@@ -198,7 +203,7 @@ let refresh_universes ?(status=univ_rigid) ?(onlyalg=false) ?(refreshset=false)
     else refresh_term_evars ~onevars:false ~top:true t
   in !evdref, t'
 
-let get_type_of_refresh  ?(polyprop=true) ?(lax=false) env evars t =
+let get_type_of_refresh ?(lax=false) env evars t =
   let tty = Retyping.get_type_of env evars t in
   let evars', tty = refresh_universes ~onlyalg:true
     ~status:(Evd.UnivFlexible false) (Some false) env evars tty in
@@ -206,36 +211,34 @@ let get_type_of_refresh  ?(polyprop=true) ?(lax=false) env evars t =
 
 let add_conv_oriented_pb ?(tail=true) (pbty,env,t1,t2) evd =
   match pbty with
-  | Some true -> add_conv_pb ~tail (Reduction.CUMUL,env,t1,t2) evd
-  | Some false -> add_conv_pb ~tail (Reduction.CUMUL,env,t2,t1) evd
-  | None -> add_conv_pb ~tail (Reduction.CONV,env,t1,t2) evd
+  | Some true -> add_conv_pb ~tail (Conversion.CUMUL,env,t1,t2) evd
+  | Some false -> add_conv_pb ~tail (Conversion.CUMUL,env,t2,t1) evd
+  | None -> add_conv_pb ~tail (Conversion.CONV,env,t1,t2) evd
 
 (* We retype applications to ensure the universe constraints are collected *)
-
 exception IllTypedInstance of env * evar_map * EConstr.types * EConstr.types
 exception IllTypedInstanceFun of env * evar_map * EConstr.constr * EConstr.types
 
+let checked_appvect, checked_appvect_hook = Hook.make ()
+
 let recheck_applications unify flags env evdref t =
   let rec aux env t =
+    (* the order matters: if the sub-applications are incorrect, checked_appvect may fail badly *)
+    iter_with_full_binders env !evdref (fun d env -> push_rel d env) aux env t;
     match EConstr.kind !evdref t with
     | App (f, args) ->
-       let () = aux env f in
-       let fty = Retyping.get_type_of env !evdref f in
-       let argsty = Array.map (fun x -> aux env x; Retyping.get_type_of env !evdref x) args in
-       let rec aux i ty =
-         if i < Array.length argsty then
-         match EConstr.kind !evdref (whd_all env !evdref ty) with
-         | Prod (na, dom, codom) ->
-            (match unify flags TypeUnification env !evdref Reduction.CUMUL argsty.(i) dom with
-             | Success evd -> evdref := evd;
-                             aux (succ i) (subst1 args.(i) codom)
-             | UnifFailure (evd, reason) -> raise (IllTypedInstance (env, evd, argsty.(i), dom)))
-         | _ -> raise (IllTypedInstanceFun (env, !evdref, f, ty))
-       else ()
-     in aux 0 fty
-    | _ ->
-       iter_with_full_binders env !evdref (fun d env -> push_rel d env) aux env t
-  in aux env t
+      let evd, _ = Hook.get checked_appvect env !evdref f args in
+      evdref := evd
+    | _ -> ()
+  in
+  try aux env t
+  with PretypeError (env,sigma,e) ->
+  match e with
+  | CantApplyBadTypeExplained (((_,expected,argty),_,_),_) ->
+    raise (IllTypedInstance (env,sigma,argty, expected))
+  | TypingError (CantApplyNonFunctional (fj,_)) ->
+    raise (IllTypedInstanceFun (env,sigma,fj.uj_val,fj.uj_type))
+  | _ -> assert false
 
 
 (*------------------------------------*
@@ -272,7 +275,7 @@ let restrict_applied_evar evd (evk,argsv) filter candidates =
   let newargsv = match filter with
   | None -> (* optim *) argsv
   | Some filter ->
-      let evi = Evd.find evd evk in
+      let EvarInfo evi = Evd.find evd evk in
       let subfilter = Filter.compose (evar_filter evi) filter in
       Filter.filter_slist subfilter argsv in
   evd,(newevk,newargsv)
@@ -284,7 +287,7 @@ let restrict_evar evd evk filter candidates =
 (* Restrict an evar in the current evar_map *)
 let restrict_instance evd evk filter argsv =
   match filter with None -> argsv | Some filter ->
-  let evi = Evd.find evd evk in
+  let EvarInfo evi = Evd.find evd evk in
   Filter.filter_slist (Filter.compose (evar_filter evi) filter) argsv
 
 open Context.Rel.Declaration
@@ -305,7 +308,7 @@ let noccur_evar env evd evk c =
        (match decl with
         | LocalAssum _ -> ()
         | LocalDef (_,b,_) -> cache := Int.Set.add (i-k) !cache; occur_rec false acc (lift i (EConstr.of_constr b)))
-  | Proj (p,c) -> occur_rec true acc c
+  | Proj (p,_,c) -> occur_rec true acc c
   | _ -> iter_with_full_binders env evd (fun rd (k,env) -> (succ k, push_rel rd env))
     (occur_rec check_types) acc c
   in
@@ -616,7 +619,7 @@ let distinct_actual_deps env evd aliases l t =
 
 open Context.Named.Declaration
 let remove_instance_local_defs evd evk args =
-  let evi = Evd.find evd evk in
+  let EvarInfo evi = Evd.find evd evk in
   let rec aux sign args = match sign, args with
   | [], [] -> []
   | LocalAssum _ :: sign, c :: args -> c :: aux sign args
@@ -728,7 +731,7 @@ let make_constructor_subst sigma sign args =
   | [], None -> accu
   | LocalAssum ({ binder_name = id }, _) :: decls, Some (Some a, args) ->
     let accu = fold decls args accu in
-    let a', args = decompose_app_vect sigma a in
+    let a', args = decompose_app sigma a in
     begin match EConstr.kind sigma a' with
     | Construct (cstr, _) ->
       let l = try Constrmap.find cstr accu with Not_found -> [] in
@@ -816,7 +819,8 @@ let define_evar_from_virtual_equation define_fun env evd src t_in_env ty_t_in_si
   let (evd, evk) = new_pure_evar sign evd ty_t_in_sign ~filter ~src in
   let t_in_env = whd_evar evd t_in_env in
   let evd = define_fun env evd None (evk, inst_in_env) t_in_env in
-  let inst_in_sign = evar_identity_subst (Evd.find evd evk) in
+  let EvarInfo evi = Evd.find evd evk in
+  let inst_in_sign = evar_identity_subst evi in
   let evar_in_sign = mkEvar (evk, inst_in_sign) in
   (evd,whd_evar evd evar_in_sign)
 
@@ -892,8 +896,7 @@ let restrict_upon_filter evd evk p args =
   let len = Array.length args in
   Filter.restrict_upon oldfullfilter len (fun i -> p (Array.unsafe_get args i))
 
-let check_evar_instance unify flags env evd evk1 body =
-  let evi = Evd.find evd evk1 in
+let check_evar_instance_evi unify flags env evd evi body =
   let evenv = evar_env env evi in
   (* FIXME: The body might be ill-typed when this is called from w_merge *)
   (* This happens in practice, cf MathClasses build failure on 2013-3-15 *)
@@ -902,10 +905,14 @@ let check_evar_instance unify flags env evd evk1 body =
     with Retyping.RetypeError _ ->
       let loc, _ = Evd.evar_source evi in user_err ?loc (Pp.(str "Ill-typed evar instance"))
   in
-  match unify flags TypeUnification evenv evd Reduction.CUMUL ty (Evd.evar_concl evi) with
+  match unify flags TypeUnification evenv evd Conversion.CUMUL ty (Evd.evar_concl evi) with
   | Success evd -> evd
   | UnifFailure _ -> raise (IllTypedInstance (evenv,evd,ty, Evd.evar_concl evi))
 
+let check_evar_instance unify flags env evd evk body =
+  let evi = try Evd.find_undefined evd evk with Not_found -> assert false in
+  check_evar_instance_evi unify flags env evd evi body
+
 (***************)
 (* Unification *)
 
@@ -958,7 +965,7 @@ let find_projectable_constructor env evd cstr k args cstr_subst =
 
 type evar_projection =
 | ProjectVar
-| ProjectEvar of EConstr.existential * evar_info * Id.t * evar_projection
+| ProjectEvar of EConstr.existential * undefined evar_info * Id.t * evar_projection
 
 exception NotUnique
 exception NotUniqueInType of (Id.t * evar_projection) list
@@ -984,7 +991,7 @@ let rec find_projectable_vars aliases sigma y subst =
     else if Evd.is_defined sigma (fst c) then subst2 (* already solved *)
     else
       let (evk,argsv as t) = c in
-      let evi = Evd.find sigma evk in
+      let evi = Evd.find_undefined sigma evk in
       let subst = make_projectable_subst aliases sigma (evar_filtered_context evi) argsv in
       let l = find_projectable_vars aliases sigma y subst in
       match l with
@@ -1267,7 +1274,7 @@ let postpone_non_unique_projection env evd pbty (evk,argsv as ev) sols rhs =
 
 let filter_compatible_candidates unify flags env evd evi args rhs c =
   let c' = instantiate_evar_array evd evi c args in
-  match unify flags TermUnification env evd Reduction.CONV rhs c' with
+  match unify flags TermUnification env evd Conversion.CONV rhs c' with
   | Success evd -> Inl (c,evd)
   | UnifFailure _ -> Inr c'
 
@@ -1278,8 +1285,8 @@ let filter_compatible_candidates unify flags env evd evi args rhs c =
 exception DoesNotPreserveCandidateRestriction
 
 let restrict_candidates unify flags env evd filter1 (evk1,argsv1) (evk2,argsv2) =
-  let evi1 = Evd.find evd evk1 in
-  let evi2 = Evd.find evd evk2 in
+  let evi1 = Evd.find_undefined evd evk1 in
+  let evi2 = Evd.find_undefined evd evk2 in
   match Evd.evar_candidates evi1, Evd.evar_candidates evi2 with
   | _, None -> filter_candidates evd evk1 filter1 NoUpdate
   | None, Some _ -> raise DoesNotPreserveCandidateRestriction
@@ -1315,7 +1322,7 @@ exception CannotProject of evar_map * EConstr.existential
 *)
 
 let rec is_constrainable_in top env evd k (evk,(fv_rels,fv_ids) as g) t =
-  let f,args = decompose_app_vect evd t in
+  let f,args = decompose_app evd t in
   match EConstr.kind evd f with
   | Construct ((ind,_),u) ->
     let n = Inductiveops.inductive_nparams env ind in
@@ -1363,7 +1370,7 @@ let project_evar_on_evar force unify flags env evd aliases k2 pbty (evk1,argsv1
        from [env]. In particular its filter must be trivial. *)
     Int.equal (SList.length argsv2) (Range.length (Environ.named_context_val env).env_named_idx) &&
     SList.Skip.for_all (fun arg -> noccur_evar env evd evk2 arg && closed0 evd arg) argsv1 &&
-    Option.is_empty (Evd.evar_candidates (Evd.find evd evk2))
+    let evi2 = Evd.find_undefined evd evk2 in Option.is_empty (Evd.evar_candidates evi2)
   then
     evd, EConstr.mkEvar ev1
   else
@@ -1395,19 +1402,21 @@ let project_evar_on_evar force unify flags env evd aliases k2 pbty (evk1,argsv1
 
 let update_evar_info ev1 ev2 evd =
   (* We update the source of obligation evars during evar-evar unifications. *)
-  let loc, evs1 = evar_source (Evd.find evd ev1) in
+  let EvarInfo evi1 = Evd.find evd ev1 in
+  let loc, evs1 = evar_source evi1 in
   Evd.update_source evd ev2 (loc, evs1)
 
 let solve_evar_evar_l2r force f unify flags env evd aliases pbty ev1 (evk2,_ as ev2) =
   try
     let evd,body = project_evar_on_evar force unify flags env evd aliases 0 pbty ev1 ev2 in
-    let evd' = Evd.define_with_evar evk2 body evd in
-    let evd' =
+    let evd =
       if is_obligation_evar evd evk2 then
-        update_evar_info evk2 (fst (destEvar evd' body)) evd'
-      else evd'
+        update_evar_info evk2 (fst (destEvar evd body)) evd
+      else evd
     in
-    check_evar_instance unify flags env evd' evk2 body
+    let evi = Evd.find_undefined evd evk2 in
+    let evd' = Evd.define_with_evar evk2 body evd in
+    check_evar_instance_evi unify flags env evd' evi body
   with EvarSolvedOnTheFly (evd,c) ->
     f env evd pbty ev2 c
 
@@ -1446,7 +1455,7 @@ let solve_evar_evar_aux force f unify flags env evd pbty (evk1,args1 as ev1) (ev
 (** Precondition: evk1 is not frozen *)
 let solve_evar_evar ?(force=false) f unify flags env evd pbty (evk1,args1 as ev1) (evk2,args2 as ev2) =
   let pbty = if force then None else pbty in
-  let evi = Evd.find evd evk1 in
+  let evi = Evd.find_undefined evd evk1 in
   let downcast evk t evd = downcast evk t evd in
   let evd =
     try
@@ -1454,11 +1463,9 @@ let solve_evar_evar ?(force=false) f unify flags env evd pbty (evk1,args1 as ev1
          The body of ?X and ?Y just has to be of type Π Δ. Type k for some k <= i, j. *)
       let evienv = Evd.evar_env env evi in
       let ctx1, i = Reductionops.dest_arity evienv evd (Evd.evar_concl evi) in
-      let evi2 = Evd.find evd evk2 in
+      let evi2 = Evd.find_undefined evd evk2 in
       let evi2env = Evd.evar_env env evi2 in
       let ctx2, j = Reductionops.dest_arity evi2env evd (Evd.evar_concl evi2) in
-      let i = ESorts.kind evd i in
-      let j = ESorts.kind evd j in
         if i == j || Evd.check_eq evd i j
         then (* Shortcut, i = j *)
           evd
@@ -1498,7 +1505,7 @@ let solve_refl ?(can_drop=false) unify flags env evd pbty evk argsv1 argsv2 =
   let args = List.map2 (fun a1 a2 -> (a1, a2)) argsv1e argsv2e in
   let untypedfilter =
     restrict_upon_filter evd evk
-      (fun (a1,a2) -> unify flags TermUnification env evd Reduction.CONV a1 a2) args in
+      (fun (a1,a2) -> unify flags TermUnification env evd Conversion.CONV a1 a2) args in
   let candidates = filter_candidates evd evk untypedfilter NoUpdate in
   let filter = closure_of_filter ~can_drop:false evd evk untypedfilter in
   let evd',ev1 = restrict_applied_evar evd (evk, argsv1) filter candidates in
@@ -1526,7 +1533,7 @@ exception NoCandidates
 exception IncompatibleCandidates of EConstr.t
 
 let solve_candidates unify flags env evd (evk,argsv) rhs =
-  let evi = Evd.find evd evk in
+  let evi = Evd.find_undefined evd evk in
   match Evd.evar_candidates evi with
   | None -> raise NoCandidates
   | Some l ->
@@ -1553,19 +1560,16 @@ let solve_candidates unify flags env evd (evk,argsv) rhs =
       | l, [] -> evd
 
 let occur_evar_upto_types sigma n c =
-  let c = EConstr.Unsafe.to_constr c in
   let seen = ref Evar.Set.empty in
-  (* FIXME: Is that supposed to be evar-insensitive? *)
-  let rec occur_rec c = match Constr.kind c with
+  let rec occur_rec c = match EConstr.kind sigma c with
     | Evar (sp,_) when Evar.equal sp n -> raise Occur
     | Evar (sp,args as e) ->
        if Evar.Set.mem sp !seen then
          SList.Skip.iter occur_rec args
        else (
          seen := Evar.Set.add sp !seen;
-         Option.iter occur_rec (existential_opt_value0 sigma e);
-         occur_rec (Evd.existential_type0 sigma e))
-    | _ -> Constr.iter occur_rec c
+         occur_rec (Evd.existential_type sigma e))
+    | _ -> EConstr.iter sigma occur_rec c
   in
   try occur_rec c; false with Occur -> true
 
@@ -1608,7 +1612,7 @@ exception MetaOccurInBodyInternal
 
 let rec invert_definition unify flags choose imitate_defs
                           env evd pbty (evk,argsv as ev) rhs =
-  let evi = Evd.find evd evk in
+  let EvarInfo evi = Evd.find evd evk in
   let sign = evar_filtered_context evi in
   let rhs = whd_beta env evd rhs (* heuristic *) in
   let fast =
@@ -1729,9 +1733,12 @@ let rec invert_definition unify flags choose imitate_defs
           let evd =
              (* Try now to invert args in terms of args' *)
             try
+              if not @@ is_evar_allowed flags evk' then
+                raise (CannotProject (evd, ev''));
               let evd,body = project_evar_on_evar false unify flags env' evd aliases 0 None ev'' ev' in
+              let evi = Evd.find_undefined evd evk' in
               let evd = Evd.define evk' body evd in
-                check_evar_instance unify flags env' evd evk' body
+                check_evar_instance_evi unify flags env' evd evi body
             with
             | EvarSolvedOnTheFly _ -> assert false (* ev has no candidates *)
             | CannotProject (evd,ev'') ->
@@ -1739,10 +1746,17 @@ let rec invert_definition unify flags choose imitate_defs
               add_conv_oriented_pb (None,env',mkEvar ev'',mkEvar ev') evd in
           evdref := evd;
           evar'')
+    | App (f, args) when EConstr.isLambda !evdref f ->
+        let p = !progress in
+        progress := true;
+        (try
+          map_constr_with_full_binders env' !evdref (fun d (env,k) -> push_rel d env, k+1)
+                                        imitate envk t
+        with _ -> progress := p; imitate envk (whd_beta env' !evdref t))
     | _ ->
         progress := true;
         match
-          let c,args = decompose_app_vect !evdref t in
+          let c,args = decompose_app !evdref t in
           match EConstr.kind !evdref c with
           | Construct (cstr,u) when noccur_between !evdref 1 k t ->
             (* This is common case when inferring the return clause of match *)
@@ -1791,7 +1805,7 @@ let rec invert_definition unify flags choose imitate_defs
 
 and evar_define unify flags ?(choose=false) ?(imitate_defs=true) env evd pbty (evk,argsv as ev) rhs =
   match EConstr.kind evd rhs with
-  | Evar (evk2,argsv2 as ev2) ->
+  | Evar (evk2,argsv2 as ev2) when is_evar_allowed flags evk2 ->
       if Evar.equal evk evk2 then
         solve_refl ~can_drop:choose
           (test_success unify) flags env evd pbty evk argsv argsv2
@@ -1855,12 +1869,8 @@ and evar_define unify flags ?(choose=false) ?(imitate_defs=true) env evd pbty (e
  * ass.
  *)
 
-let status_changed evd lev (pbty,_,t1,t2) =
-  (try Evar.Set.mem (head_evar evd t1) lev with NoHeadEvar -> false) ||
-  (try Evar.Set.mem (head_evar evd t2) lev with NoHeadEvar -> false)
-
 let reconsider_unif_constraints unify flags evd =
-  let (evd,pbs) = extract_changed_conv_pbs evd (status_changed evd) in
+  let (evd,pbs) = extract_changed_conv_pbs evd in
   List.fold_left
     (fun p (pbty,env,t1,t2 as x) ->
        match p with
