@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -12,7 +12,7 @@ open CErrors
 open Util
 open Names
 open Term
-open Constr
+open EConstr
 open Vars
 open Context
 open Declarations
@@ -25,10 +25,11 @@ open Context.Rel.Declaration
    Inductive, but they expect an env *)
 
 let type_of_inductive env (ind,u) =
+ let u = EConstr.Unsafe.to_instance u in
  let (mib,_ as specif) = Inductive.lookup_mind_specif env ind in
  Typeops.check_hyps_inclusion env (GlobRef.IndRef ind) mib.mind_hyps;
  let t = Inductive.type_of_inductive (specif,u) in
- Arguments_renaming.rename_type t (IndRef ind)
+ EConstr.of_constr @@ Arguments_renaming.rename_type t (IndRef ind)
 
 let e_type_of_inductive env sigma (ind,u) =
  let (mib,_ as specif) = Inductive.lookup_mind_specif env ind in
@@ -38,11 +39,12 @@ let e_type_of_inductive env sigma (ind,u) =
 
 (* Return type as quoted by the user *)
 let type_of_constructor env (cstr,u) =
+ let u = EConstr.Unsafe.to_instance u in
  let (mib,_ as specif) =
    Inductive.lookup_mind_specif env (inductive_of_constructor cstr) in
  Typeops.check_hyps_inclusion env (GlobRef.ConstructRef cstr) mib.mind_hyps;
  let t = Inductive.type_of_constructor (cstr,u) specif in
- Arguments_renaming.rename_type t (ConstructRef cstr)
+ EConstr.of_constr @@ Arguments_renaming.rename_type t (ConstructRef cstr)
 
 let e_type_of_constructor env sigma (cstr,u) =
  let (mib,_ as specif) =
@@ -53,16 +55,18 @@ let e_type_of_constructor env sigma (cstr,u) =
 
 (* Return constructor types in user form *)
 let type_of_constructors env (ind,u as indu) =
- let specif = Inductive.lookup_mind_specif env ind in
-  Inductive.type_of_constructors indu specif
+  let indu = on_snd EConstr.Unsafe.to_instance indu in
+  let specif = Inductive.lookup_mind_specif env ind in
+  Array.map EConstr.of_constr (Inductive.type_of_constructors indu specif)
 
 (* Return constructor types in normal form *)
 let arities_of_constructors env (ind,u as indu) =
- let specif = Inductive.lookup_mind_specif env ind in
-  Inductive.arities_of_constructors indu specif
+  let indu = on_snd EConstr.Unsafe.to_instance indu in
+  let specif = Inductive.lookup_mind_specif env ind in
+  Array.map EConstr.of_constr (Inductive.arities_of_constructors indu specif)
 
 (* [inductive_family] = [inductive_instance] applied to global parameters *)
-type inductive_family = pinductive * constr list
+type inductive_family = inductive puniverses * constr list
 
 let make_ind_family (mis, params) = (mis,params)
 let dest_ind_family (mis,params) : inductive_family = (mis,params)
@@ -74,8 +78,12 @@ let lift_inductive_family n = liftn_inductive_family n 1
 
 let substnl_ind_family l n = map_ind_family (substnl l n)
 
+let relevance_of_inductive env ind =
+  let ind = on_snd EConstr.Unsafe.to_instance ind in
+  ERelevance.make @@ Inductive.relevance_of_inductive env ind
+
 let relevance_of_inductive_family env (ind,_ : inductive_family) =
-  Inductive.relevance_of_inductive env ind
+  relevance_of_inductive env ind
 
 type inductive_type = IndType of inductive_family * EConstr.constr list
 
@@ -85,8 +93,7 @@ let make_ind_type (indf, realargs) = IndType (indf,realargs)
 let dest_ind_type (IndType (indf,realargs)) = (indf,realargs)
 
 let map_inductive_type f (IndType (indf, realargs)) =
-  let f' c = EConstr.Unsafe.to_constr (f (EConstr.of_constr c)) in
-  IndType (map_ind_family f' indf, List.map f realargs)
+  IndType (map_ind_family f indf, List.map f realargs)
 
 let liftn_inductive_type n d = map_inductive_type (EConstr.Vars.liftn n d)
 let lift_inductive_type n = liftn_inductive_type n 1
@@ -97,30 +104,38 @@ let relevance_of_inductive_type env (IndType (indf, _)) =
   relevance_of_inductive_family env indf
 
 let mkAppliedInd (IndType ((ind,params), realargs)) =
-  let open EConstr in
-  let ind = on_snd EInstance.make ind in
-  applist (mkIndU ind, (List.map EConstr.of_constr params)@realargs)
+  applist (mkIndU ind, params @ realargs)
+
+let dest_recarg p = match Rtree.Kind.kind p with
+| Rtree.Kind.Node (ra, _) -> ra
+| Rtree.Kind.Var _ -> assert false
+
+let dest_subterms p = match Rtree.Kind.kind p with
+| Rtree.Kind.Node (ra, cstrs) ->
+  let () = assert (match ra with Norec -> false | _ -> true) in
+  cstrs
+| Rtree.Kind.Var _ -> assert false
 
 (* Does not consider imbricated or mutually recursive types *)
 let mis_is_recursive_subset listind rarg =
   let one_is_rec rvec =
-    List.exists
+    Array.exists
       (fun ra ->
         match dest_recarg ra with
-          | Mrec (_,i) -> Int.List.mem i listind
-          | _ -> false) rvec
+          | Mrec (RecArgInd ind) -> List.exists (Names.Ind.CanOrd.equal ind) listind
+          | Mrec (RecArgPrim _) | Norec -> false) rvec
   in
   Array.exists one_is_rec (dest_subterms rarg)
 
-let mis_is_recursive (ind,mib,mip) =
-  mis_is_recursive_subset (List.interval 0 (mib.mind_ntypes - 1))
-    mip.mind_recargs
+let mis_is_recursive ((ind,_),mib,mip) =
+  mis_is_recursive_subset (List.init mib.mind_ntypes (fun i -> (ind,i)))
+    (Rtree.Kind.make mip.mind_recargs)
 
 let mis_nf_constructor_type ((_,j),u) (mib,mip) =
   let nconstr = Array.length mip.mind_consnames in
   if j > nconstr then user_err Pp.(str "Not enough constructors in the type.");
   let (ctx, cty) = mip.mind_nf_lc.(j - 1) in
-  subst_instance_constr u (Term.it_mkProd_or_LetIn cty ctx)
+  subst_instance_constr u (EConstr.it_mkProd_or_LetIn (EConstr.of_constr cty) (EConstr.of_rel_context ctx))
 
 (* Number of constructors *)
 
@@ -203,12 +218,13 @@ let inductive_nalldecls env ind =
 (* Others *)
 
 let inductive_paramdecls env (ind,u) =
+  let u = EConstr.Unsafe.to_instance u in
   let (mib,mip) = Inductive.lookup_mind_specif env ind in
-    Inductive.inductive_paramdecls (mib,u)
+  EConstr.of_rel_context @@ Inductive.inductive_paramdecls (mib,u)
 
 let inductive_alldecls env (ind,u) =
   let (mib,mip) = Inductive.lookup_mind_specif env ind in
-    Vars.subst_instance_context u mip.mind_arity_ctxt
+  Vars.subst_instance_context u (EConstr.of_rel_context mip.mind_arity_ctxt)
 
 let inductive_alltags env ind =
   let (mib,mip) = Inductive.lookup_mind_specif env ind in
@@ -243,34 +259,65 @@ let quality_leq q q' =
       -> true
     | (QProp|QType), _ -> false
     end
+  | QVar _, QConstant QType -> true
   | (QVar _|QConstant _), _ -> false
 
 type squash = SquashToSet | SquashToQuality of Sorts.Quality.t
 
 let is_squashed sigma ((_,mip),u) =
-  match mip.mind_arity with
-  | TemplateArity _ -> None (* template is never squashed *)
-  | RegularArity a ->
-    match mip.mind_squashed with
-    | None -> None
-    | Some squash ->
-      let indq = EConstr.ESorts.quality sigma
-          (EConstr.ESorts.make @@ UVars.subst_instance_sort u a.mind_sort)
-      in
-      match squash with
-      | AlwaysSquashed -> begin match a.mind_sort with
-          | Sorts.Set -> Some SquashToSet
-          | _ -> Some (SquashToQuality indq)
-        end
-      | SometimesSquashed squash ->
-        (* impredicative set squashes are always AlwaysSquashed,
-           so here if inds=Set it is a sort poly squash (see "foo6" in test sort_poly.v) *)
-        if Sorts.Quality.Set.for_all (fun q ->
-            let q = UVars.subst_instance_quality u q in
-            let q = UState.nf_quality (Evd.evar_universe_context sigma) q in
-            quality_leq q indq) squash
-        then None
-        else Some (SquashToQuality indq)
+  let s = match mip.mind_arity with
+    | RegularArity a -> a.mind_sort
+    | TemplateArity a -> a.template_level
+  in
+  match mip.mind_squashed with
+  | None -> None
+  | Some squash ->
+    let u = EConstr.Unsafe.to_instance u in
+    let indq = EConstr.ESorts.quality sigma
+        (EConstr.ESorts.make @@ UVars.subst_instance_sort u s)
+    in
+    match squash with
+    | AlwaysSquashed -> begin match s with
+        | Sorts.Set -> Some SquashToSet
+        | _ -> Some (SquashToQuality indq)
+      end
+    | SometimesSquashed squash ->
+      (* impredicative set squashes are always AlwaysSquashed,
+         so here if inds=Set it is a sort poly squash (see "foo6" in test sort_poly.v) *)
+      if Sorts.Quality.Set.for_all (fun q ->
+          let q = UVars.subst_instance_quality u q in
+          let q = UState.nf_quality (Evd.ustate sigma) q in
+          quality_leq q indq) squash
+      then None
+      else Some (SquashToQuality indq)
+
+let squash_elim_sort env sigma squash rtnsort = match squash with
+| SquashToSet ->
+  (* Squashed inductive in Set, only happens with impredicative Set *)
+  begin match ESorts.kind sigma rtnsort with
+  | Set | SProp | Prop -> sigma
+  | QSort _ | Type _ ->
+    Evd.set_eq_sort env sigma rtnsort ESorts.set
+  end
+| SquashToQuality (QConstant QProp) ->
+  (* Squashed inductive in Prop, return sort must be Prop or SProp *)
+  begin match ESorts.kind sigma rtnsort with
+  | SProp | Prop -> sigma
+  | QSort _ | Type _ | Set ->
+    Evd.set_eq_sort env sigma rtnsort ESorts.prop
+  end
+| SquashToQuality (QConstant QSProp) ->
+  (* Squashed inductive in SProp, return sort must be SProp. *)
+  begin match ESorts.kind sigma rtnsort with
+  | SProp -> sigma
+  | Type _ | Set | Prop | QSort _ ->
+    Evd.set_eq_sort env sigma rtnsort ESorts.sprop
+  end
+| SquashToQuality (QConstant QType) ->
+  (* Sort poly squash to type *)
+  Evd.set_leq_sort env sigma ESorts.set rtnsort
+| SquashToQuality (QVar q) ->
+  Evd.set_leq_sort env sigma (ESorts.make (Sorts.qsort q Univ.Universe.type0)) rtnsort
 
 let is_allowed_elimination sigma ((mib,_),_ as specifu) s =
   let open Sorts in
@@ -288,9 +335,34 @@ let is_allowed_elimination sigma ((mib,_),_ as specifu) s =
       end
     | Some (SquashToQuality indq) -> quality_leq (EConstr.ESorts.quality sigma s) indq
 
+let make_allowed_elimination env sigma ((mib,_),_ as specifu) s =
+  let open Sorts in
+  match mib.mind_record with
+  | PrimRecord _ -> Some sigma
+  | NotRecord | FakeRecord ->
+    match is_squashed sigma specifu with
+    | None -> Some sigma
+    | Some SquashToSet ->
+      begin match EConstr.ESorts.kind sigma s with
+      | SProp|Prop|Set -> Some sigma
+      | QSort _ | Type _ ->
+        try Some (Evd.set_leq_sort env sigma s ESorts.set)
+        with UGraph.UniverseInconsistency _ -> None
+      end
+    | Some (SquashToQuality indq) ->
+      let sq = EConstr.ESorts.quality sigma s in
+      if quality_leq sq indq then Some sigma
+      else
+        let mk q = ESorts.make @@ Sorts.make q Univ.Universe.type0 in
+        try Some (Evd.set_leq_sort env sigma (mk sq) (mk indq))
+        with UGraph.UniverseInconsistency _ -> None
+
+(* XXX questionable for sort poly inductives *)
 let elim_sort (_,mip) =
   if Option.is_empty mip.mind_squashed then Sorts.InType
-  else Inductive.inductive_sort_family mip
+  else match mip.mind_arity with
+    | TemplateArity x -> Sorts.family x.template_level
+    | RegularArity s -> Sorts.family s.mind_sort
 
 let top_allowed_sort env (kn,i as ind) =
   let specif = Inductive.lookup_mind_specif env ind in
@@ -319,14 +391,8 @@ let has_dependent_elim (mib,mip) =
 (* Annotation for cases *)
 let make_case_info env ind style =
   let (mib,mip) = Inductive.lookup_mind_specif env ind in
-  let ind_tags =
-    Context.Rel.to_tags (List.firstn mip.mind_nrealdecls mip.mind_arity_ctxt) in
-  let cstr_tags =
-    Array.map2 (fun (d, _) n ->
-      Context.Rel.to_tags (List.firstn n d))
-      mip.mind_nf_lc mip.mind_consnrealdecls in
-  let print_info = { ind_tags; cstr_tags; style } in
-  { ci_ind     = ind;
+  let print_info = { Constr.style } in
+  { Constr.ci_ind     = ind;
     ci_npar    = mib.mind_nparams;
     ci_cstr_ndecls = mip.mind_consnrealdecls;
     ci_cstr_nargs = mip.mind_consnrealargs;
@@ -335,10 +401,10 @@ let make_case_info env ind style =
 (*s Useful functions *)
 
 type constructor_summary = {
-  cs_cstr : pconstructor;
+  cs_cstr : constructor puniverses;
   cs_params : constr list;
   cs_nargs : int;
-  cs_args : Constr.rel_context;
+  cs_args : EConstr.rel_context;
   cs_concl_realargs : constr array
 }
 
@@ -355,26 +421,27 @@ let instantiate_params t params sign =
   let nnonrecpar = Context.Rel.nhyps sign - List.length params in
   (* Adjust the signature if recursively non-uniform parameters are not here *)
   let _,sign = Context.Rel.chop_nhyps nnonrecpar sign in
-  let _,t = decompose_prod_n_decls (Context.Rel.length sign) t in
+  let _,t = Term.decompose_prod_n_decls (Context.Rel.length sign) t in
   let subst = subst_of_rel_context_instance_list sign params in
-  substl subst t
+  substl subst (EConstr.of_constr t)
 
 let instantiate_constructor_params (_,u as cstru) (mib,_ as mind_specif) params =
   let typi = mis_nf_constructor_type cstru mind_specif in
-  let ctx = Vars.subst_instance_context u mib.mind_params_ctxt in
-  instantiate_params typi params ctx
+  let ctx = Vars.subst_instance_context u (EConstr.of_rel_context mib.mind_params_ctxt) in
+  instantiate_params (EConstr.Unsafe.to_constr typi) params ctx
 
 let get_constructor ((ind,u),mib,mip,params) j =
   assert (j <= Array.length mip.mind_consnames);
   let typi = instantiate_constructor_params ((ind,j),u) (mib,mip) params in
-  let (args,ccl) = decompose_prod_decls typi in
-  let (_,allargs) = decompose_app_list ccl in
+  let typi = EConstr.Unsafe.to_constr typi in
+  let (args,ccl) = Term.decompose_prod_decls typi in
+  let (_,allargs) = Constr.decompose_app_list ccl in
   let vargs = List.skipn (List.length params) allargs in
   { cs_cstr = (ith_constructor_of_inductive ind j,u);
     cs_params = params;
     cs_nargs = Context.Rel.length args;
-    cs_args = args;
-    cs_concl_realargs = Array.of_list vargs }
+    cs_args = EConstr.of_rel_context args;
+    cs_concl_realargs = Array.map_of_list EConstr.of_constr vargs }
 
 let get_constructors env (ind,params) =
   let (mib,mip) = Inductive.lookup_mind_specif env (fst ind) in
@@ -383,13 +450,13 @@ let get_constructors env (ind,params) =
 
 let get_projections = Environ.get_projections
 
-let make_case_invert env (IndType (((ind,u),params),indices)) ~case_relevance:r ci =
+let make_case_invert env sigma (IndType (((ind,u),params),indices)) ~case_relevance:r ci =
+  let r = ERelevance.kind sigma r in
   if Typeops.should_invert_case env r ci
-  then CaseInvert {indices=Array.of_list indices}
-  else NoInvert
+  then Constr.CaseInvert {indices=Array.of_list indices}
+  else Constr.NoInvert
 
 let make_project env sigma ind pred c branches ps =
-  let open EConstr in
   assert(Array.length branches == 1);
   let na, ty, t = destLambda sigma pred in
   let mib, mip as specif = Inductive.lookup_mind_specif env ind in
@@ -407,10 +474,10 @@ let make_project env sigma ind pred c branches ps =
   let mkProj i c =
     let p, r = ps.(i) in
     let r = UVars.subst_instance_relevance u r in
-    mkProj (Projection.make p true, r, c)
+    mkProj (Projection.make p true, ERelevance.make r, c)
   in
   let proj = match EConstr.destRel sigma br with
-    | exception DestKO -> None
+    | exception Constr.DestKO -> None
     | i ->
       begin match List.skipn (i-1) ctx with
       | exception Failure _ -> None
@@ -442,20 +509,18 @@ let make_project env sigma ind pred c branches ps =
   mkLetIn (na, c, ty, it_mkLambda_or_LetIn (Vars.liftn 1 (mip.mind_consnrealdecls.(0) + 1) br) ctx)
 
 let simple_make_case_or_project env sigma ci pred invert c branches =
-  let open EConstr in
-  let ind = ci.ci_ind in
+  let ind = ci.Constr.ci_ind in
   let projs = get_projections env ind in
   match projs with
   | None -> mkCase (EConstr.contract_case env sigma (ci, pred, invert, c, branches))
   | Some ps -> make_project env sigma ind (fst pred) c branches ps
 
 let make_case_or_project env sigma indt ci pred c branches =
-  let open EConstr in
   let IndType (((ind,_),_),_) = indt in
   let projs = get_projections env ind in
   match projs with
   | None ->
-     let invert = make_case_invert env indt ~case_relevance:(snd pred) ci in
+     let invert = make_case_invert env sigma indt ~case_relevance:(snd pred) ci in
      mkCase (EConstr.contract_case env sigma (ci, pred, invert, c, branches))
   | Some ps -> make_project env sigma ind (fst pred) c branches ps
 
@@ -475,6 +540,7 @@ let get_arity env ((ind,u),params) =
     (* Dynamically detect if called with an instance of recursively
        uniform parameter only or also of recursively non-uniform
        parameters *)
+    let u = EConstr.Unsafe.to_instance u in
     let nparams = List.length params in
     if Int.equal nparams mib.mind_nparams then
       Inductive.inductive_paramdecls (mib,u)
@@ -482,8 +548,10 @@ let get_arity env ((ind,u),params) =
       assert (Int.equal nparams mib.mind_nparams_rec);
       snd (Inductive.inductive_nonrec_rec_paramdecls (mib,u))
     end in
+  let parsign = EConstr.of_rel_context parsign in
   let arproperlength = List.length mip.mind_arity_ctxt - List.length parsign in
   let arsign,_ = List.chop arproperlength mip.mind_arity_ctxt in
+  let arsign = EConstr.of_rel_context arsign in
   let subst = subst_of_rel_context_instance_list parsign params in
   let arsign = Vars.subst_instance_context u arsign in
   substl_rel_context subst arsign
@@ -506,20 +574,18 @@ let build_dependent_inductive env ((ind, params) as indf) =
 
 let make_arity_signature env sigma dep (ind, _ as indf) =
   let arsign = get_arity env indf in
-  let r = Inductive.relevance_of_inductive env ind in
+  let r = relevance_of_inductive env ind in
   let anon = make_annot Anonymous r in
-  let arsign = List.map (fun d -> Termops.map_rel_decl EConstr.of_constr d) arsign in
   if dep then
     (* We need names everywhere *)
     Namegen.name_context env sigma
-      ((LocalAssum (anon,EConstr.of_constr (build_dependent_inductive env indf)))::arsign)
+      ((LocalAssum (anon, build_dependent_inductive env indf)) :: arsign)
       (* Costly: would be better to name once for all at definition time *)
   else
     (* No need to enforce names *)
     arsign
 
 let make_arity env sigma dep indf s =
-  let open EConstr in
   it_mkProd_or_LetIn (mkSort s) (make_arity_signature env sigma dep indf)
 
 (**************************************************)
@@ -529,20 +595,22 @@ let make_arity env sigma dep indf s =
     The term built is expecting to be substituted first by
     a substitution of the form [params, x : ind params] *)
 let compute_projections env (kn, i as ind) =
-  let open Term in
   let mib = Environ.lookup_mind kn env in
   let u = UVars.make_abstract_instance (Declareops.inductive_polymorphic_context mib) in
+  let u = EInstance.make u in
   let x = match mib.mind_record with
   | NotRecord | FakeRecord ->
     anomaly Pp.(str "Trying to build primitive projections for a non-primitive record")
   | PrimRecord info ->
     let id, _, _, _ = info.(i) in
-    make_annot (Name id) mib.mind_packets.(i).mind_relevance
+    make_annot (Name id) (ERelevance.make mib.mind_packets.(i).mind_relevance)
   in
   let pkt = mib.mind_packets.(i) in
   let { mind_nparams = nparamargs; mind_params_ctxt = params } = mib in
+  let params = EConstr.of_rel_context params in
   let ctx, _ = pkt.mind_nf_lc.(0) in
   let ctx, paramslet = List.chop pkt.mind_consnrealdecls.(0) ctx in
+  let ctx = EConstr.of_rel_context ctx in
   (* We build a substitution smashing the lets in the record parameters so
      that typechecking projections requires just a substitution and not
      matching with a parameter context. *)
@@ -621,10 +689,9 @@ let find_rectype env sigma c =
     | Ind (ind,u) ->
         let (mib,mip) = Inductive.lookup_mind_specif env ind in
         if mib.mind_nparams > List.length l then raise Not_found;
-        let l = List.map EConstr.Unsafe.to_constr l in
         let (par,rargs) = List.chop mib.mind_nparams l in
-        let indu = (ind, EInstance.kind sigma u) in
-        IndType((indu, par),List.map EConstr.of_constr rargs)
+        let indu = (ind, u) in
+        IndType ((indu, par), rargs)
     | _ -> raise Not_found
 
 let find_inductive env sigma c =
@@ -633,7 +700,6 @@ let find_inductive env sigma c =
   match EConstr.kind sigma t with
     | Ind ind
         when (fst (Inductive.lookup_mind_specif env (fst ind))).mind_finite <> CoFinite ->
-        let l = List.map EConstr.Unsafe.to_constr l in
         (ind, l)
     | _ -> raise Not_found
 
@@ -643,7 +709,6 @@ let find_coinductive env sigma c =
   match EConstr.kind sigma t with
     | Ind ind
         when (fst (Inductive.lookup_mind_specif env (fst ind))).mind_finite == CoFinite ->
-        let l = List.map EConstr.Unsafe.to_constr l in
         (ind, l)
     | _ -> raise Not_found
 
@@ -651,10 +716,10 @@ let find_coinductive env sigma c =
 (* Type of Case predicates *)
 let arity_of_case_predicate env (ind,params) dep k =
   let arsign = get_arity env (ind,params) in
-  let r = Inductive.relevance_of_inductive env ind in
+  let r = relevance_of_inductive env ind in
   let mind = build_dependent_inductive env (ind,params) in
   let concl = if dep then mkArrow mind r (mkSort k) else mkSort k in
-  Term.it_mkProd_or_LetIn concl arsign
+  it_mkProd_or_LetIn concl arsign
 
 (***********************************************)
 (* Inferring the sort of parameters of a polymorphic inductive type
@@ -670,11 +735,12 @@ let univ_level_mem l s = match s with
    conclusion, and the other ones by fresh universes. *)
 let rec instantiate_universes env evdref scl is = function
   | (LocalDef _ as d)::sign, exp ->
-      d :: instantiate_universes env evdref scl is (sign, exp)
-  | d::sign, None::exp ->
-      d :: instantiate_universes env evdref scl is (sign, exp)
-  | (LocalAssum (na,ty))::sign, Some l::exp ->
-      let ctx,_ = Reduction.dest_arity env ty in
+    EConstr.of_rel_decl d :: instantiate_universes env evdref scl is (sign, exp)
+  | d::sign, false::exp ->
+    EConstr.of_rel_decl d :: instantiate_universes env evdref scl is (sign, exp)
+  | (LocalAssum (na,ty))::sign, true::exp ->
+      let ctx,s = Reduction.dest_arity env ty in
+      let l = match s with Type u -> Option.get @@ Univ.Universe.level u | _ -> assert false in
       let u = Univ.Universe.make l in
       let s =
         (* Does the sort of parameter [u] appear in (or equal)
@@ -687,14 +753,15 @@ let rec instantiate_universes env evdref scl is = function
           let evm = Evd.set_leq_sort env evm s (EConstr.ESorts.make (Sorts.sort_of_univ u)) in
             evdref := evm; s
       in
-      let s = EConstr.ESorts.kind !evdref s in
-      (LocalAssum (na,mkArity(ctx,s))) :: instantiate_universes env evdref scl is (sign, exp)
-  | sign, [] -> sign (* Uniform parameters are exhausted *)
+      let ctx = EConstr.of_rel_context ctx in
+      let na = EConstr.of_binder_annot na in
+      (LocalAssum (na, mkArity (ctx, s))) :: instantiate_universes env evdref scl is (sign, exp)
+  | sign, [] -> EConstr.of_rel_context sign (* Uniform parameters are exhausted *)
   | [], _ -> assert false
 
 let type_of_inductive_knowing_conclusion env sigma ((mib,mip),u) conclty =
   match mip.mind_arity with
-  | RegularArity s -> sigma, EConstr.of_constr (subst_instance_constr u s.mind_user_arity)
+  | RegularArity s -> sigma, subst_instance_constr u (EConstr.of_constr s.mind_user_arity)
   | TemplateArity ar ->
     let templ = match mib.mind_template with
     | None -> assert false
@@ -705,16 +772,15 @@ let type_of_inductive_knowing_conclusion env sigma ((mib,mip),u) conclty =
     let evdref = ref sigma in
     let ctx =
       instantiate_universes
-        env evdref scl ar.template_level (ctx,templ.template_param_levels) in
-    let scl = EConstr.ESorts.kind !evdref scl in
-      !evdref, EConstr.of_constr (mkArity (List.rev ctx,scl))
+        env evdref scl ar.template_level (ctx,templ.template_param_arguments) in
+    !evdref, mkArity (List.rev ctx, scl)
 
 let type_of_projection_constant env (p,u) =
   let _, pty = lookup_projection p env in
-  Vars.subst_instance_constr u pty
+  EConstr.Vars.subst_instance_constr u (EConstr.of_constr pty)
 
 let type_of_projection_knowing_arg env sigma p c ty =
-  let c = EConstr.Unsafe.to_constr c in
+  let open EConstr.Vars in
   let IndType(pars,realargs) =
     try find_rectype env sigma ty
     with Not_found ->
@@ -730,15 +796,16 @@ let type_of_projection_knowing_arg env sigma p c ty =
    syntactic conditions *)
 
 let control_only_guard env sigma c =
+  let evars = Evd.evar_handler sigma in
   let c = Evarutil.nf_evar sigma c in
   let check_fix_cofix e c =
     (* [c] has already been normalized upfront *)
     let c = EConstr.Unsafe.to_constr c in
-    match kind c with
+    match Constr.kind c with
     | CoFix (_,(_,_,_) as cofix) ->
-      Inductive.check_cofix e cofix
+      Inductive.check_cofix ~evars e cofix
     | Fix fix ->
-      Inductive.check_fix e fix
+      Inductive.check_fix ~evars e fix
     | _ -> ()
   in
   let rec iter env c =
@@ -747,4 +814,6 @@ let control_only_guard env sigma c =
   in
   try iter env c
   with Type_errors.TypeError (env, e) ->
-    raise (Pretype_errors.PretypeError (env, sigma, TypingError (Type_errors.map_ptype_error EConstr.of_constr e)))
+    raise (Pretype_errors.PretypeError
+             (env, sigma,
+              TypingError (Pretype_errors.of_type_error e)))
