@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -27,13 +27,15 @@ type key =
   | KRel
   | KInt
   | KFloat
+  | KString
+  | KArray
 
 module KeyOrdered = struct
   type t = key
 
   let hash gr =
     match gr with
-    | KGlob gr -> 9 + GlobRef.Ordered.hash gr
+    | KGlob gr -> 9 + GlobRef.UserOrd.hash gr
     | KLam -> 0
     | KLet -> 1
     | KProd -> 2
@@ -44,17 +46,19 @@ module KeyOrdered = struct
     | KRel -> 7
     | KInt -> 8
     | KFloat -> 9
+    | KString -> 10
+    | KArray -> 11
 
   let compare gr1 gr2 =
     match gr1, gr2 with
-    | KGlob gr1, KGlob gr2 -> GlobRef.Ordered.compare gr1 gr2
+    | KGlob gr1, KGlob gr2 -> GlobRef.UserOrd.compare gr1 gr2
     | _, KGlob _ -> -1
     | KGlob _, _ -> 1
     | k, k' -> Int.compare (hash k) (hash k')
 
   let equal k1 k2 =
     match k1, k2 with
-    | KGlob gr1, KGlob gr2 -> GlobRef.Ordered.equal gr1 gr2
+    | KGlob gr1, KGlob gr2 -> GlobRef.UserOrd.equal gr1 gr2
     | _, KGlob _ -> false
     | KGlob _, _ -> false
     | k, k' -> k == k'
@@ -79,9 +83,11 @@ let equiv_keys k k' =
     try Keyset.mem k' (Keymap.find k !keys)
     with Not_found -> false
 
+let mkKGlob env gr = KGlob (Environ.QGlobRef.canonize env gr)
+
 (** Registration of keys as an object *)
 
-let load_keys _ (_,(ref,ref')) =
+let load_keys _ (ref,ref') =
   add_keys ref ref'
 
 let cache_keys o =
@@ -89,7 +95,7 @@ let cache_keys o =
 
 let subst_key subst k =
   match k with
-  | KGlob gr -> KGlob (subst_global_reference subst gr)
+  | KGlob gr -> mkKGlob (Global.env ()) (subst_global_reference subst gr)
   | _ -> k
 
 let subst_keys (subst,(k,k')) =
@@ -99,7 +105,7 @@ let discharge_key = function
   | KGlob (GlobRef.VarRef _ as g) when Lib.is_in_section g -> None
   | x -> Some x
 
-let discharge_keys (_,(k,k')) =
+let discharge_keys (k,k') =
   match discharge_key k, discharge_key k' with
   | Some x, Some y -> Some (x, y)
   | _ -> None
@@ -113,18 +119,18 @@ let inKeys : key_obj -> obj =
     ~discharge:discharge_keys
 
 let declare_equiv_keys ref ref' =
-  Lib.add_anonymous_leaf (inKeys (ref,ref'))
+  Lib.add_leaf (inKeys (ref,ref'))
 
-let constr_key kind c =
+let constr_key env kind c =
   try
     let rec aux k =
       match kind k with
-      | Const (c, _) -> KGlob (GlobRef.ConstRef c)
-      | Ind (i, u) -> KGlob (GlobRef.IndRef i)
-      | Construct (c,u) -> KGlob (GlobRef.ConstructRef c)
-      | Var id -> KGlob (GlobRef.VarRef id)
+      | Const (c, _) -> mkKGlob env (GlobRef.ConstRef c)
+      | Ind (i, u) -> mkKGlob env (GlobRef.IndRef i)
+      | Construct (c,u) -> mkKGlob env (GlobRef.ConstructRef c)
+      | Var id -> mkKGlob env (GlobRef.VarRef id)
       | App (f, _) -> aux f
-      | Proj (p, _) -> KGlob (GlobRef.ConstRef (Projection.constant p))
+      | Proj (p, _, _) -> mkKGlob env (GlobRef.ConstRef (Projection.constant p))
       | Cast (p, _, _) -> aux p
       | Lambda _ -> KLam
       | Prod _ -> KProd
@@ -138,6 +144,8 @@ let constr_key kind c =
       | LetIn _ -> KLet
       | Int _ -> KInt
       | Float _ -> KFloat
+      | String _ -> KString
+      | Array _ -> KArray
     in Some (aux c)
   with Not_found -> None
 
@@ -155,6 +163,8 @@ let pr_key pr_global = function
   | KRel -> str"Rel"
   | KInt -> str"Int"
   | KFloat -> str"Float"
+  | KString -> str"String"
+  | KArray -> str"Array"
 
 let pr_keyset pr_global v =
   prlist_with_sep spc (pr_key pr_global) (Keyset.elements v)
