@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -54,10 +54,11 @@ let cache_reduction_effect (con,funkey) =
 let subst_reduction_effect (subst,(con,funkey)) =
   (subst_constant subst con,funkey)
 
-let inReductionEffect : Constant.t * string -> obj =
-  declare_object @@ global_object_nodischarge "REDUCTION-EFFECT"
+let inReductionEffect : Libobject.locality * (Constant.t * string) -> obj =
+  declare_object @@ object_with_locality "REDUCTION-EFFECT"
     ~cache:cache_reduction_effect
     ~subst:(Some subst_reduction_effect)
+    ~discharge:(fun x -> x)
 
 let declare_reduction_effect funkey f =
   if String.Map.mem funkey !effect_table then
@@ -65,8 +66,8 @@ let declare_reduction_effect funkey f =
   effect_table := String.Map.add funkey f !effect_table
 
 (** A function to set the value of the print function *)
-let set_reduction_effect x funkey =
-  Lib.add_leaf (inReductionEffect (x,funkey))
+let set_reduction_effect local x funkey =
+  Lib.add_leaf (inReductionEffect (local,(x,funkey)))
 
 
 (** Machinery to custom the behavior of the reduction *)
@@ -87,6 +88,8 @@ module ReductionBehaviour = struct
     | UnfoldWhen x -> UnfoldWhen (more_args_when k x)
     | UnfoldWhenNoMatch x -> UnfoldWhenNoMatch (more_args_when k x)
 
+  type table = Cpred.t * t Cmap.t
+
 (* We need to have a fast way to know the set of all constants that
   have the NeverUnfold flag.  Therefore, the table has a distinct subpart
   that is this set. *)
@@ -95,8 +98,9 @@ module ReductionBehaviour = struct
 
   let load _ (_,(r, b)) =
     table := (match b with
-                | NeverUnfold -> Cpred.add r (fst !table), Cmap.remove r (snd !table)
-                | _ -> Cpred.remove r (fst !table), Cmap.add r b (snd !table))
+                | None -> Cpred.remove r (fst !table), Cmap.remove r (snd !table)
+                | Some NeverUnfold -> Cpred.add r (fst !table), Cmap.remove r (snd !table)
+                | Some b -> Cpred.remove r (fst !table), Cmap.add r b (snd !table))
 
   let cache o = load 1 o
 
@@ -113,7 +117,7 @@ module ReductionBehaviour = struct
         if Lib.is_in_section gr then
           let vars = Lib.section_instance gr in
           let extra = Array.length vars in
-          more_args extra b
+          Option.map (more_args extra) b
         else b
       in
       Some (false, (gr, b))
@@ -131,18 +135,16 @@ module ReductionBehaviour = struct
   let set ~local r b =
     Lib.add_leaf (inRedBehaviour (local, (r, b)))
 
-  let get r =
-    if Cpred.mem r (fst !table) then
+  let get_from_db table r =
+    if Cpred.mem r (fst table) then
       Some NeverUnfold
     else
-      Cmap.find_opt r (snd !table)
+      Cmap.find_opt r (snd table)
 
-  let all_never_unfold () = fst !table
-
-  let print ref =
+  let print_from_db table ref =
     let open Pp in
     let pr_global c = Nametab.pr_global_env Id.Set.empty (ConstRef c) in
-    match get ref with
+    match get_from_db table ref with
     | None -> mt ()
     | Some b ->
        let pp_nomatch = spc () ++ str "but avoid exposing match constructs" in
@@ -173,6 +175,18 @@ module ReductionBehaviour = struct
        in
        hov 2 (str "The reduction tactics " ++ pp_behavior b)
 
+  module Db = struct
+    type t = table
+    let get () = !table
+    let empty = (Cpred.empty, Cmap.empty)
+    let print = print_from_db
+    let all_never_unfold table = fst table
+  end
+
+  let get r = get_from_db (Db.get ()) r
+
+  let print c = print_from_db (Db.get ()) c
+
 end
 
 (** The type of (machine) stacks (= lambda-bar-calculus' contexts) *)
@@ -182,13 +196,14 @@ sig
   type app_node
   val pr_app_node : (EConstr.t -> Pp.t) -> app_node -> Pp.t
 
-  type case_stk =
-    case_info * EInstance.t * EConstr.t array * EConstr.t pcase_return * EConstr.t  pcase_invert * EConstr.t pcase_branch array
+  type case_stk = case_info * EInstance.t * EConstr.t array * EConstr.case_return * EConstr.t pcase_invert * EConstr.case_branch array
+
+  val mkCaseStk : case_info * EInstance.t * EConstr.t array * EConstr.case_return * EConstr.t pcase_invert * EConstr.case_branch array -> case_stk
 
   type member =
   | App of app_node
   | Case of case_stk
-  | Proj of Projection.t * Sorts.relevance
+  | Proj of Projection.t * ERelevance.t
   | Fix of fixpoint * t
   | Primitive of CPrimitives.t * (Constant.t * EInstance.t) * t * CPrimitives.args_red
 
@@ -215,7 +230,7 @@ sig
   val zip : evar_map -> constr * t -> constr
   val check_native_args : CPrimitives.t -> t -> bool
   val get_next_primitive_args : CPrimitives.args_red -> t -> CPrimitives.args_red * (t * EConstr.t * t) option
-  val expand_case : env -> evar_map -> case_stk -> case_info * EInstance.t * constr array * ((rel_context * constr) * Sorts.relevance) * (rel_context * constr) array
+  val expand_case : env -> evar_map -> case_stk -> case_info * EInstance.t * constr array * ((rel_context * constr) * ERelevance.t) * (rel_context * constr) array
 end =
 struct
   open EConstr
@@ -236,12 +251,14 @@ struct
 
 
   type case_stk =
-    case_info * EInstance.t * EConstr.t array * EConstr.t pcase_return * EConstr.t  pcase_invert * EConstr.t pcase_branch array
+    case_info * EInstance.t * EConstr.t array * EConstr.case_return * EConstr.t pcase_invert * EConstr.case_branch array
+
+  let mkCaseStk x = x
 
   type member =
   | App of app_node
   | Case of case_stk
-  | Proj of Projection.t * Sorts.relevance
+  | Proj of Projection.t * ERelevance.t
   | Fix of fixpoint * t
   | Primitive of CPrimitives.t * (Constant.t * EInstance.t) * t * CPrimitives.args_red
 
@@ -258,7 +275,7 @@ struct
          prvect_with_sep (pr_bar) (fun (_, c) -> pr_c c) br
        ++ str ")"
     | Proj (p,_)  ->
-      str "ZProj(" ++ Constant.debug_print (Projection.constant p) ++ str ")"
+      str "ZProj(" ++ Projection.debug_print p ++ str ")"
     | Fix (f,args) ->
        str "ZFix(" ++ Constr.debug_print_fix pr_c f
        ++ pr_comma () ++ pr pr_c args ++ str ")"
@@ -451,9 +468,11 @@ let pr_state env sigma (tm,sk) =
 (*** Reduction Functions Operators ***)
 (*************************************)
 
-let safe_meta_value sigma ev =
-  try Some (Evd.meta_value sigma ev)
-  with Not_found -> None
+type meta_handler = { meta_value : metavariable -> EConstr.t option }
+
+let safe_meta_value metas ev = match metas with
+| None -> None
+| Some f -> f.meta_value ev
 
 (*************************************)
 (*** Reduction using bindingss ***)
@@ -537,6 +556,11 @@ struct
     | Float f -> f
     | _ -> raise Primred.NativeDestKO
 
+  let get_string evd e =
+    match EConstr.kind evd e with
+    | String s -> s
+    | _ -> raise Primred.NativeDestKO
+
   let get_parray evd e =
     match EConstr.kind evd e with
     | Array(_u,t,def,_ty) -> Parray.of_array t def
@@ -548,6 +572,9 @@ struct
   let mkFloat env f =
     mkFloat f
 
+  let mkString env s =
+    mkString s
+
   let mkBool env b =
     let (ct,cf) = get_bool_constructors env in
     mkConstruct (if b then ct else cf)
@@ -680,7 +707,119 @@ let apply_branch env sigma (ind, i) args (ci, u, pms, iv, r, lf) =
     let ctx = expand_branch env sigma u pms (ind, i) br in
     applist (it_mkLambda_or_LetIn (snd br) ctx, args)
 
-let whd_state_gen flags env sigma =
+
+exception PatternFailure
+
+let match_einstance sigma pu u psubst =
+  match UVars.Instance.pattern_match pu (EInstance.kind sigma u) psubst with
+  | Some psubst -> psubst
+  | None -> raise PatternFailure
+
+let match_sort ps s psubst =
+  match Sorts.pattern_match ps s psubst with
+  | Some psubst -> psubst
+  | None -> raise PatternFailure
+
+let rec match_arg_pattern whrec env sigma ctx psubst p t =
+  let open Declarations in
+  let t' = EConstr.it_mkLambda_or_LetIn t ctx in
+  match p with
+  | EHole i -> Partial_subst.add_term i t' psubst
+  | EHoleIgnored -> psubst
+  | ERigid (ph, es) ->
+      let t, stk = whrec (t, Stack.empty) in
+      let psubst = match_rigid_arg_pattern whrec env sigma ctx psubst ph t in
+      let psubst, stk = apply_rule whrec env sigma ctx psubst es stk in
+      if Stack.is_empty stk then psubst else raise PatternFailure
+
+and match_rigid_arg_pattern whrec env sigma ctx psubst p t =
+  match [@ocaml.warning "-4"] p, EConstr.kind sigma t with
+  | PHInd (ind, pu), Ind (ind', u) ->
+    if Ind.CanOrd.equal ind ind' then match_einstance sigma pu u psubst else raise PatternFailure
+  | PHConstr (constr, pu), Construct (constr', u) ->
+    if Construct.CanOrd.equal constr constr' then match_einstance sigma pu u psubst else raise PatternFailure
+  | PHRel i, Rel n when i = n -> psubst
+  | PHSort ps, Sort s -> match_sort ps (ESorts.kind sigma s) psubst
+  | PHSymbol (c, pu), Const (c', u) ->
+    if Constant.CanOrd.equal c c' then match_einstance sigma pu u psubst else raise PatternFailure
+  | PHInt i, Int i' ->
+    if Uint63.equal i i' then psubst else raise PatternFailure
+  | PHFloat f, Float f' ->
+    if Float64.equal f f' then psubst else raise PatternFailure
+  | PHString s, String s' ->
+    if Pstring.equal s s' then psubst else raise PatternFailure
+  | PHLambda (ptys, pbod), _ ->
+    let ntys, _ = EConstr.decompose_lambda sigma t in
+    let na = List.length ntys and np = Array.length ptys in
+    if np > na then raise PatternFailure;
+    let ntys, body = EConstr.decompose_lambda_n sigma np t in
+    let ctx' = List.map (fun (n, ty) -> Context.Rel.Declaration.LocalAssum (n, ty)) ntys in
+    let tys = Array.of_list @@ List.rev_map snd ntys in
+    let na = Array.length tys in
+    let contexts_upto = Array.init na (fun i -> List.skipn (na - i) ctx' @ ctx) in
+    let psubst = Array.fold_left3 (fun psubst ctx -> match_arg_pattern whrec env sigma ctx psubst) psubst contexts_upto ptys tys in
+    let psubst = match_arg_pattern whrec env sigma (ctx' @ ctx) psubst pbod body in
+    psubst
+  | PHProd (ptys, pbod), _ ->
+    let ntys, _ = EConstr.decompose_prod sigma t in
+    let na = List.length ntys and np = Array.length ptys in
+    if np > na then raise PatternFailure;
+    let ntys, body = EConstr.decompose_prod_n sigma np t in
+    let ctx' = List.map (fun (n, ty) -> Context.Rel.Declaration.LocalAssum (n, ty)) ntys in
+    let tys = Array.of_list @@ List.rev_map snd ntys in
+    let na = Array.length tys in
+    let contexts_upto = Array.init na (fun i -> List.skipn (na - i) ctx' @ ctx) in
+    let psubst = Array.fold_left3 (fun psubst ctx -> match_arg_pattern whrec env sigma ctx psubst) psubst contexts_upto ptys tys in
+    let psubst = match_arg_pattern whrec env sigma (ctx' @ ctx) psubst pbod body in
+    psubst
+  | (PHInd _ | PHConstr _ | PHRel _ | PHSort _ | PHSymbol _ | PHInt _ | PHFloat _ | PHString _), _ -> raise PatternFailure
+
+and extract_n_stack args n s =
+  if n = 0 then List.rev args, s else
+  match Stack.decomp s with
+  | Some (arg, rest) -> extract_n_stack (arg :: args) (n-1) rest
+  | None -> raise PatternFailure
+
+and apply_rule whrec env sigma ctx psubst es stk =
+  match [@ocaml.warning "-4"] es, stk with
+  | [], _ -> psubst, stk
+  | Declarations.PEApp pargs :: e, s ->
+      let np = Array.length pargs in
+      let pargs = Array.to_list pargs in
+      let args, s = extract_n_stack [] np s in
+      let psubst = List.fold_left2 (match_arg_pattern whrec env sigma ctx) psubst pargs args in
+      apply_rule whrec env sigma ctx psubst e s
+  | Declarations.PECase (pind, pu, pret, pbrs) :: e, Stack.Case (ci, u, pms, p, iv, brs) :: s ->
+      if not @@ Ind.CanOrd.equal pind ci.ci_ind then raise PatternFailure;
+      let dummy = mkProp in
+      let psubst = match_einstance sigma pu u psubst in
+      let (_, _, _, ((ntys_ret, ret), _), _, _, brs) = EConstr.annotate_case env sigma (ci, u, pms, p, NoInvert, dummy, brs) in
+      let psubst = match_arg_pattern whrec env sigma (ntys_ret @ ctx) psubst pret ret in
+      let psubst = Array.fold_left2 (fun psubst pat (ctx', br) -> match_arg_pattern whrec env sigma (ctx' @ ctx) psubst pat br) psubst pbrs brs in
+      apply_rule whrec env sigma ctx psubst e s
+  | Declarations.PEProj proj :: e, Stack.Proj (proj', r) :: s ->
+      if not @@ Projection.(Repr.CanOrd.equal proj (repr proj')) then raise PatternFailure;
+      apply_rule whrec env sigma ctx psubst e s
+  | _, _ -> raise PatternFailure
+
+
+let rec apply_rules whrec env sigma u r stk =
+  let open Declarations in
+  match r with
+  | [] -> raise PatternFailure
+  | { lhs_pat = (pu, elims); nvars; rhs } :: rs ->
+    try
+      let psubst = Partial_subst.make nvars in
+      let psubst = match_einstance sigma pu u psubst in
+      let psubst, stk = apply_rule whrec env sigma [] psubst elims stk in
+      let subst, qsubst, usubst = Partial_subst.to_arrays psubst in
+      let usubst = UVars.Instance.of_array (qsubst, usubst) in
+      let rhsu = subst_instance_constr (EConstr.EInstance.make usubst) (EConstr.of_constr rhs) in
+      let rhs' = substl (Array.to_list subst) rhsu in
+      (rhs', stk)
+    with PatternFailure -> apply_rules whrec env sigma u rs stk
+
+let whd_state_gen flags ?metas env sigma =
   let open Context.Named.Declaration in
   let rec whrec (x, stack) : state =
     let () =
@@ -709,7 +848,7 @@ let whd_state_gen flags env sigma =
       | _ -> fold ())
     | Evar ev -> fold ()
     | Meta ev ->
-      (match safe_meta_value sigma ev with
+      (match safe_meta_value metas ev with
       | Some body -> whrec (body, stack)
       | None -> fold ())
     | Const (c,u as const) ->
@@ -729,6 +868,19 @@ let whd_state_gen flags env sigma =
           (* Should not fail thanks to [check_native_args] *)
           let (before,a,after) = Option.get o in
           whrec (a,Stack.Primitive(p,const,before,kargs)::after)
+       | exception NotEvaluableConst (HasRules (u', b, r)) ->
+          begin try
+            let rhs, stack = apply_rules whrec env sigma u r stack in
+            whrec (rhs, stack)
+          with PatternFailure ->
+            if not b then fold () else
+            match Stack.strip_app stack with
+            | args, (Stack.Fix (f,s')::s'') when RedFlags.red_set flags RedFlags.fFIX ->
+              let x' = Stack.zip sigma (x, args) in
+              let out_sk = s' @ (Stack.append_app [|x'|] s'') in
+              whrec (reduce_and_refold_fix env sigma f out_sk)
+            | _ -> fold ()
+          end
        | exception NotEvaluableConst _ -> fold ()
       else fold ()
     | Proj (p, r, c) when RedFlags.red_projection flags p ->
@@ -782,7 +934,7 @@ let whd_state_gen flags env sigma =
         |_ -> fold ()
       else fold ()
 
-    | Int _ | Float _ | Array _ ->
+    | Int _ | Float _ | String _ | Array _ ->
       begin match Stack.strip_app stack with
        | (_, Stack.Primitive(p,(_, u as kn),rargs,kargs)::s) ->
          let more_to_reduce = List.exists (fun k -> CPrimitives.Kwhnf = k) kargs in
@@ -813,7 +965,7 @@ let whd_state_gen flags env sigma =
   whrec
 
 (** reduction machine without global env and refold machinery *)
-let local_whd_state_gen flags env sigma =
+let local_whd_state_gen flags ?metas env sigma =
   let rec whrec (x, stack) =
     let c0 = EConstr.kind sigma x in
     let s = (EConstr.of_kind c0, stack) in
@@ -841,7 +993,7 @@ let local_whd_state_gen flags env sigma =
 
     | Evar ev -> s
     | Meta ev ->
-      (match safe_meta_value sigma ev with
+      (match safe_meta_value metas ev with
         Some c -> whrec (c,stack)
       | None -> s)
 
@@ -871,64 +1023,75 @@ let local_whd_state_gen flags env sigma =
       else s
 
     | Rel _ | Var _ | Sort _ | Prod _ | LetIn _ | Const _  | Ind _ | Proj _
-      | Int _ | Float _ | Array _ -> s
+      | Int _ | Float _ | String _ | Array _ -> s
 
   in
   whrec
 
-let stack_red_of_state_red f =
-  let f env sigma x = EConstr.decompose_app_list sigma (Stack.zip sigma (f env sigma (x, Stack.empty))) in
+let stack_red_of_state_red f ?metas =
+  let f env sigma x = EConstr.decompose_app_list sigma (Stack.zip sigma (f ?metas env sigma (x, Stack.empty))) in
   f
 
-let red_of_state_red f env sigma x =
-  Stack.zip sigma (f env sigma (x,Stack.empty))
+let red_of_state_red ?metas ~delta f env sigma x =
+  let rec is_whnf c = match Constr.kind c with
+    | Const _ | Var _ -> not delta
+    | Construct _ | Ind _ | Int _ | Float _ | String _
+    | Sort _ | Prod _ -> true
+    | App (h,_) -> is_whnf h
+    | _ -> false
+  in
+  (* preserve physical equality if possible
+     not sure if anything relies on reduction unfolding head evars
+     for now use Unsafe.to_constr to keep such unfolds *)
+  if is_whnf (EConstr.Unsafe.to_constr x) then x
+  else Stack.zip sigma (f ?metas env sigma (x,Stack.empty))
 
 (* 0. No Reduction Functions *)
 
 let whd_nored_state = local_whd_state_gen RedFlags.nored
 let whd_nored_stack = stack_red_of_state_red whd_nored_state
-let whd_nored = red_of_state_red whd_nored_state
+let whd_nored ?metas = red_of_state_red ?metas ~delta:false whd_nored_state
 
 (* 1. Beta Reduction Functions *)
 
 let whd_beta_state = local_whd_state_gen RedFlags.beta
 let whd_beta_stack = stack_red_of_state_red whd_beta_state
-let whd_beta = red_of_state_red whd_beta_state
+let whd_beta = red_of_state_red ~delta:false whd_beta_state
 
 let whd_betalet_state = local_whd_state_gen RedFlags.betazeta
 let whd_betalet_stack = stack_red_of_state_red whd_betalet_state
-let whd_betalet = red_of_state_red whd_betalet_state
+let whd_betalet = red_of_state_red ~delta:false whd_betalet_state
 
 (* 2. Delta Reduction Functions *)
 
 let whd_const_state c e = whd_state_gen RedFlags.(mkflags [fCONST c]) e
-let whd_const c = red_of_state_red (whd_const_state c)
+let whd_const c = red_of_state_red ~delta:true (fun ?metas -> whd_const_state c)
 
-let whd_delta_state e = whd_state_gen RedFlags.delta e
+let whd_delta_state = whd_state_gen RedFlags.delta
 let whd_delta_stack = stack_red_of_state_red whd_delta_state
-let whd_delta = red_of_state_red whd_delta_state
+let whd_delta = red_of_state_red ~delta:true whd_delta_state
 
 let whd_betadeltazeta_state = whd_state_gen RedFlags.betadeltazeta
 let whd_betadeltazeta_stack = stack_red_of_state_red whd_betadeltazeta_state
-let whd_betadeltazeta = red_of_state_red whd_betadeltazeta_state
+let whd_betadeltazeta = red_of_state_red ~delta:true whd_betadeltazeta_state
 
 (* 3. Iota reduction Functions *)
 
 let whd_betaiota_state = local_whd_state_gen RedFlags.betaiota
 let whd_betaiota_stack = stack_red_of_state_red whd_betaiota_state
-let whd_betaiota = red_of_state_red whd_betaiota_state
+let whd_betaiota ?metas = red_of_state_red ?metas ~delta:false whd_betaiota_state
 
 let whd_betaiotazeta_state = local_whd_state_gen RedFlags.betaiotazeta
 let whd_betaiotazeta_stack = stack_red_of_state_red whd_betaiotazeta_state
-let whd_betaiotazeta = red_of_state_red whd_betaiotazeta_state
+let whd_betaiotazeta ?metas = red_of_state_red ?metas ~delta:false whd_betaiotazeta_state
 
 let whd_all_state = whd_state_gen RedFlags.all
 let whd_all_stack = stack_red_of_state_red whd_all_state
-let whd_all = red_of_state_red whd_all_state
+let whd_all ?metas = red_of_state_red ?metas ~delta:true whd_all_state
 
 let whd_allnolet_state = whd_state_gen RedFlags.allnolet
 let whd_allnolet_stack = stack_red_of_state_red whd_allnolet_state
-let whd_allnolet = red_of_state_red whd_allnolet_state
+let whd_allnolet = red_of_state_red ~delta:true whd_allnolet_state
 
 let whd_stack_gen reds = stack_red_of_state_red (whd_state_gen reds)
 
@@ -953,12 +1116,8 @@ let shrink_eta sigma c =
           if noccurn sigma 1 u then pop u else x
         | _ -> x
       else x
-    | Meta ev ->
-      (match safe_meta_value sigma ev with
-        Some c -> whrec c
-      | None -> x)
-    | App _ | Case _ | Fix _ | Construct _ | CoFix _ | Evar _ | Rel _ | Var _ | Sort _ | Prod _
-    | LetIn _ | Const _  | Ind _ | Proj _ | Int _ | Float _ | Array _ -> x
+    | Meta _ | App _ | Case _ | Fix _ | Construct _ | CoFix _ | Evar _ | Rel _ | Var _ | Sort _ | Prod _
+    | LetIn _ | Const _  | Ind _ | Proj _ | Int _ | Float _ | String _ | Array _ -> x
   in
   whrec c
 
@@ -966,7 +1125,7 @@ let shrink_eta sigma c =
 
 let whd_zeta_state = local_whd_state_gen RedFlags.zeta
 let whd_zeta_stack = stack_red_of_state_red whd_zeta_state
-let whd_zeta = red_of_state_red whd_zeta_state
+let whd_zeta = red_of_state_red ~delta:false whd_zeta_state
 
 (****************************************************************************)
 (*                   Reduction Functions                                    *)
@@ -1008,7 +1167,7 @@ let nf_all env sigma =
 (********************************************************************)
 
 let is_transparent e k =
-  match Conv_oracle.get_strategy (Environ.oracle e) k with
+  match Conv_oracle.get_strategy (Environ.oracle e) (Evaluable.to_kevaluable k) with
   | Conv_oracle.Opaque -> false
   | _ -> true
 
@@ -1045,30 +1204,27 @@ struct
 open Conversion
 
 let check_eq univs u u' =
-  if not (Evd.check_eq univs u u') then raise NotConvertible
+  if Evd.check_eq univs u u' then Result.Ok univs else Result.Error None
 
 let check_leq univs u u' =
-  if not (Evd.check_leq univs u u') then raise NotConvertible
+  if Evd.check_leq univs u u' then Result.Ok univs else Result.Error None
 
-let check_sort_cmp_universes pb s0 s1 univs =
+let checked_sort_cmp_universes _env pb s0 s1 univs =
   let s0 = ESorts.make s0 in
   let s1 = ESorts.make s1 in
   match pb with
   | CUMUL -> check_leq univs s0 s1
   | CONV -> check_eq univs s0 s1
 
-let checked_sort_cmp_universes _env pb s0 s1 univs =
-  check_sort_cmp_universes pb s0 s1 univs; univs
-
 let check_convert_instances ~flex:_ u u' univs =
   let csts = UVars.enforce_eq_instances u u' (Sorts.QConstraints.empty,Constraints.empty) in
-  if Evd.check_quconstraints univs csts then univs else raise NotConvertible
+  if Evd.check_quconstraints univs csts then Result.Ok univs else Result.Error None
 
 (* general conversion and inference functions *)
 let check_inductive_instances cv_pb variance u1 u2 univs =
   let csts = get_cumulativity_constraints cv_pb variance u1 u2 in
-  if (Evd.check_quconstraints univs csts) then univs
-  else raise NotConvertible
+  if (Evd.check_quconstraints univs csts) then Result.Ok univs
+  else Result.Error None
 
 let checked_universes =
   { compare_sorts = checked_sort_cmp_universes;
@@ -1090,9 +1246,12 @@ let is_fconv ?(reds=TransparentState.full) pb env sigma t1 t2 =
     let evars = Evd.evar_handler sigma in
     try
       let env = Environ.set_universes (Evd.universes sigma) env in
-      let _ = Conversion.generic_conv ~l2r:false pb ~evars reds env (sigma, CheckUnivs.checked_universes) t1 t2 in
-      true
-    with Conversion.NotConvertible -> false
+      begin match Conversion.generic_conv ~l2r:false pb ~evars reds env (sigma, CheckUnivs.checked_universes) t1 t2 with
+      | Result.Ok (_ : Evd.evar_map) -> true
+      | Result.Error None -> false
+      | Result.Error (Some e) -> Empty.abort e
+      end
+    with
     | e ->
       let e = Exninfo.capture e in
       report_anomaly e
@@ -1106,20 +1265,27 @@ let check_conv ?(pb=Conversion.CUMUL) ?(ts=TransparentState.full) env sigma x y
 
 let sigma_compare_sorts env pb s0 s1 sigma =
   match pb with
-  | Conversion.CONV -> Evd.set_eq_sort env sigma (ESorts.make s0) (ESorts.make s1)
-  | Conversion.CUMUL -> Evd.set_leq_sort env sigma (ESorts.make s0) (ESorts.make s1)
+  | Conversion.CONV ->
+    begin
+      try Result.Ok (Evd.set_eq_sort env sigma (ESorts.make s0) (ESorts.make s1))
+      with UGraph.UniverseInconsistency err -> Result.Error (Some err)
+    end
+  | Conversion.CUMUL ->
+    begin
+      try Result.Ok (Evd.set_leq_sort env sigma (ESorts.make s0) (ESorts.make s1))
+      with UGraph.UniverseInconsistency err -> Result.Error (Some err)
+    end
 
 let sigma_compare_instances ~flex i0 i1 sigma =
-  try Evd.set_eq_instances ~flex sigma i0 i1
-  with Evd.UniversesDiffer
-     | UGraph.UniverseInconsistency _ ->
-        raise Conversion.NotConvertible
+  match Evd.set_eq_instances ~flex sigma i0 i1 with
+  | sigma -> Result.Ok sigma
+  | exception Evd.UniversesDiffer -> Result.Error None
+  | exception UGraph.UniverseInconsistency err -> Result.Error (Some err)
 
 let sigma_check_inductive_instances cv_pb variance u1 u2 sigma =
   match Evarutil.compare_cumulative_instances cv_pb variance u1 u2 sigma with
-  | Inl sigma -> sigma
-  | Inr _ ->
-    raise Conversion.NotConvertible
+  | Inl sigma -> Result.Ok sigma
+  | Inr err -> Result.Error (Some err)
 
 let sigma_univ_state =
   let open Conversion in
@@ -1130,13 +1296,14 @@ let sigma_univ_state =
 let univproblem_compare_sorts env pb s0 s1 uset =
   let open UnivProblem in
   match pb with
-  | Conversion.CONV -> UnivProblem.Set.add (UEq (s0, s1)) uset
-  | Conversion.CUMUL -> UnivProblem.Set.add (ULe (s0, s1)) uset
+  | Conversion.CONV -> Result.Ok (UnivProblem.Set.add (UEq (s0, s1)) uset)
+  | Conversion.CUMUL -> Result.Ok (UnivProblem.Set.add (ULe (s0, s1)) uset)
 
 let univproblem_compare_instances ~flex i0 i1 uset =
-  UnivProblem.enforce_eq_instances_univs flex i0 i1 uset
+  Result.Ok (UnivProblem.enforce_eq_instances_univs flex i0 i1 uset)
 
-let univproblem_check_inductive_instances = UnivProblem.compare_cumulative_instances
+let univproblem_check_inductive_instances cv_pb variance u1 u2 sigma =
+  Result.Ok (UnivProblem.compare_cumulative_instances cv_pb variance u1 u2 sigma)
 
 let univproblem_univ_state =
   let open Conversion in
@@ -1144,6 +1311,11 @@ let univproblem_univ_state =
     compare_instances = univproblem_compare_instances;
     compare_cumul_instances = univproblem_check_inductive_instances; }
 
+type genconv = {
+  genconv : 'a 'err. conv_pb -> l2r:bool -> Evd.evar_map -> TransparentState.t ->
+    Environ.env -> ('a, 'err) Conversion.generic_conversion_function
+}
+
 let infer_conv_gen conv_fun ?(catch_incon=true) ?(pb=Conversion.CUMUL)
     ?(ts=TransparentState.full) env sigma x y =
   try
@@ -1165,19 +1337,35 @@ let infer_conv_gen conv_fun ?(catch_incon=true) ?(pb=Conversion.CUMUL)
         let x = EConstr.Unsafe.to_constr x in
         let y = EConstr.Unsafe.to_constr y in
         let env = Environ.set_universes (Evd.universes sigma) env in
-        let sigma' =
-          conv_fun pb ~l2r:false sigma ts
-            env (sigma, sigma_univ_state) x y in
-        Some sigma'
+        (* First try conversion with postponed universe problems as a kind of FO
+           approximation. This may result in unsatisfiable constraints even if
+           some unfoldings of the arguments could have been unified, but this
+           should be exceedingly rare. *)
+        let ans = match conv_fun.genconv pb ~l2r:false sigma ts env (UnivProblem.Set.empty, univproblem_univ_state) x y with
+        | Result.Ok cstr -> Some cstr
+        | Result.Error None ->
+          None (* no universe unification can make these terms convertible *)
+        | Result.Error (Some e) -> Empty.abort e
+        in
+        match ans with
+        | None -> None
+        | Some cstr ->
+          match Evd.add_universe_constraints sigma cstr with
+          | sigma -> Some sigma
+          | exception UGraph.UniverseInconsistency _ | exception Evd.UniversesDiffer ->
+            (* Retry with local universe checking, which may imply constant unfolding *)
+            match conv_fun.genconv pb ~l2r:false sigma ts env (sigma, sigma_univ_state) x y with
+            | Result.Ok sigma -> Some sigma
+            | Result.Error None -> None
+            | Result.Error (Some e) -> raise (UGraph.UniverseInconsistency e)
   with
-  | Conversion.NotConvertible -> None
   | UGraph.UniverseInconsistency _ when catch_incon -> None
   | e ->
     let e = Exninfo.capture e in
     report_anomaly e
 
-let infer_conv = infer_conv_gen (fun pb ~l2r sigma ->
-      Conversion.generic_conv pb ~l2r ~evars:(Evd.evar_handler sigma))
+let infer_conv = infer_conv_gen { genconv = fun pb ~l2r sigma ->
+      Conversion.generic_conv pb ~l2r ~evars:(Evd.evar_handler sigma) }
 
 let infer_conv_ustate ?(catch_incon=true) ?(pb=Conversion.CUMUL)
     ?(ts=TransparentState.full) env sigma x y =
@@ -1194,12 +1382,14 @@ let infer_conv_ustate ?(catch_incon=true) ?(pb=Conversion.CUMUL)
         let x = EConstr.Unsafe.to_constr x in
         let y = EConstr.Unsafe.to_constr y in
         let env = Environ.set_universes (Evd.universes sigma) env in
-        let cstr =
+        match
           Conversion.generic_conv pb ~l2r:false ~evars:(Evd.evar_handler sigma) ts
-            env (UnivProblem.Set.empty, univproblem_univ_state) x y in
-        Some cstr
+            env (UnivProblem.Set.empty, univproblem_univ_state) x y
+        with
+        | Result.Ok cstr -> Some cstr
+        | Result.Error None -> None
+        | Result.Error (Some e) -> raise (UGraph.UniverseInconsistency e)
   with
-  | Conversion.NotConvertible -> None
   | UGraph.UniverseInconsistency _ when catch_incon -> None
   | e ->
     let e = Exninfo.capture e in
@@ -1209,15 +1399,15 @@ let evars_of_evar_map sigma =
   { Genlambda.evars_val = Evd.evar_handler sigma }
 
 let vm_infer_conv ?(pb=Conversion.CUMUL) env sigma t1 t2 =
-  infer_conv_gen (fun pb ~l2r sigma ts ->
-      Vconv.vm_conv_gen pb (evars_of_evar_map sigma))
+  infer_conv_gen { genconv = fun pb ~l2r sigma ts ->
+      Vconv.vm_conv_gen pb (evars_of_evar_map sigma) }
     ~catch_incon:true ~pb env sigma t1 t2
 
 let native_conv_generic pb sigma t =
   Nativeconv.native_conv_gen pb (evars_of_evar_map sigma) t
 
 let native_infer_conv ?(pb=Conversion.CUMUL) env sigma t1 t2 =
-  infer_conv_gen (fun pb ~l2r sigma ts -> native_conv_generic pb sigma)
+  infer_conv_gen { genconv = fun pb ~l2r sigma ts -> native_conv_generic pb sigma }
     ~catch_incon:true ~pb env sigma t1 t2
 
 let check_hyps_inclusion env sigma x hyps =
@@ -1229,46 +1419,6 @@ let check_hyps_inclusion env sigma x hyps =
 (*             Special-Purpose Reduction                            *)
 (********************************************************************)
 
-(* [instance] is used for [res_pf]; the call to [local_strong whd_betaiota]
-   has (unfortunately) different subtle side effects:
-
-   - ** Order of subgoals **
-     If the lemma is a case analysis with parameters, it will move the
-     parameters as first subgoals (e.g. "case H" applied on
-     "H:D->A/\B|-C" will present the subgoal |-D first while w/o
-     betaiota the subgoal |-D would have come last).
-
-   - ** Betaiota-contraction in statement **
-     If the lemma has a parameter which is a function and this
-     function is applied in the lemma, then the _strong_ betaiota will
-     contract the application of the function to its argument (e.g.
-     "apply (H (fun x => x))" in "H:forall f, f 0 = 0 |- 0=0" will
-     result in applying the lemma 0=0 in which "(fun x => x) 0" has
-     been contracted). A goal to rewrite may then fail or succeed
-     differently.
-
-   - ** Naming of hypotheses **
-     If a lemma is a function of the form "fun H:(forall a:A, P a)
-     => .. F H .." where the expected type of H is "forall b:A, P b",
-     then, without reduction, the application of the lemma will
-     generate a subgoal "forall a:A, P a" (and intro will use name
-     "a"), while with reduction, it will generate a subgoal "forall
-     b:A, P b" (and intro will use name "b").
-
-   - ** First-order pattern-matching **
-     If a lemma has the type "(fun x => p) t" then rewriting t may fail
-     if the type of the lemma is first beta-reduced (this typically happens
-     when rewriting a single variable and the type of the lemma is obtained
-     by meta_instance (with empty map) which itself calls instance with this
-     empty map).
- *)
-
-let instance env sigma c =
-  (* if s = [] then c else *)
-  (* No need to compute contexts under binders as whd_betaiota is local *)
-  let rec strongrec t = EConstr.map sigma strongrec (whd_betaiota env sigma t) in
-  strongrec c
-
 (* pseudo-reduction rule:
  * [hnf_prod_app env s (Prod(_,B)) N --> B[N]
  * with an HNF on the first argument to produce a product.
@@ -1459,7 +1609,7 @@ let is_sort env sigma t =
 (* reduction to head-normal-form allowing delta/zeta only in argument
    of case/fix (heuristic used by evar_conv) *)
 
-let whd_betaiota_deltazeta_for_iota_state ts env sigma s =
+let whd_betaiota_deltazeta_for_iota_state ts ?metas env sigma s =
   let all' = RedFlags.red_add_transparent RedFlags.all ts in
   (* Unset the sharing flag to get a call-by-name reduction. This matters for
      the shape of the generated term. *)
@@ -1478,7 +1628,20 @@ let whd_betaiota_deltazeta_for_iota_state ts env sigma s =
     | _ -> None
   in
   let rec whrec s =
-    let (t, stack as s) = whd_state_gen RedFlags.betaiota env sigma s in
+    let (t, stack as s) = whd_state_gen RedFlags.betaiota ?metas env sigma s in
+    let rewrite_step =
+      match kind sigma t with
+      | Const (cst, u) when Environ.is_symbol env cst ->
+        let r = Cmap_env.get cst env.symb_pats in
+        begin match apply_rules whrec env sigma u r stack with
+        | r -> Some r
+        | exception PatternFailure -> None
+        end
+      | _ -> None
+    in
+    match rewrite_step with
+    | Some r -> whrec r
+    | None ->
     match Stack.strip_app stack with
       |args, (Stack.Case _ :: _ as stack') ->
         begin match whd_opt (t, args) with
@@ -1515,30 +1678,23 @@ let is_arity env sigma c =
     | Sort _ -> true
     | _ -> false
 
-(*************************************)
-(* Metas *)
-
-let meta_instance env sigma b =
-  let fm = b.freemetas in
-  if Metaset.is_empty fm then b.rebus
-  else
-    instance env sigma b.rebus
-
 module Infer = struct
 
 open Conversion
 
 let infer_eq (univs, cstrs as cuniv) u u' =
-  if UGraph.check_eq_sort univs u u' then cuniv
-  else
+  if UGraph.check_eq_sort univs u u' then Result.Ok cuniv
+  else try
     let cstrs' = UnivSubst.enforce_eq_sort u u' Constraints.empty in
-    UGraph.merge_constraints cstrs' univs, Constraints.union cstrs cstrs'
+    Result.Ok (UGraph.merge_constraints cstrs' univs, Constraints.union cstrs cstrs')
+  with UGraph.UniverseInconsistency err -> Result.Error (Some err)
 
 let infer_leq (univs, cstrs as cuniv) u u' =
-  if UGraph.check_leq_sort univs u u' then cuniv
-  else
-    let cstrs', univs = UnivSubst.enforce_leq_alg_sort u u' univs in
-    univs, Univ.Constraints.union cstrs cstrs'
+  if UGraph.check_leq_sort univs u u' then Result.Ok cuniv
+  else match UnivSubst.enforce_leq_alg_sort u u' univs with
+  | cstrs', univs ->
+    Result.Ok (univs, Univ.Constraints.union cstrs cstrs')
+  | exception UGraph.UniverseInconsistency err -> Result.Error (Some err)
 
 let infer_cmp_universes _env pb s0 s1 univs =
   match pb with
@@ -1547,19 +1703,24 @@ let infer_cmp_universes _env pb s0 s1 univs =
 
 let infer_convert_instances ~flex u u' (univs,cstrs as cuniv) =
   if flex then
-    if UGraph.check_eq_instances univs u u' then cuniv
-    else raise NotConvertible
+    if UGraph.check_eq_instances univs u u' then Result.Ok cuniv
+    else Result.Error None
   else
     let qcstrs, cstrs' = UVars.enforce_eq_instances u u' Sorts.QUConstraints.empty in
-    let () = if not (Sorts.QConstraints.trivial qcstrs) then raise NotConvertible in
-    (univs, Constraints.union cstrs cstrs')
+    if Sorts.QConstraints.trivial qcstrs then
+      Result.Ok (univs, Constraints.union cstrs cstrs')
+    else
+      Result.Error None
 
 let infer_inductive_instances cv_pb variance u1 u2 (univs,csts) =
   let qcsts, csts' = get_cumulativity_constraints cv_pb variance u1 u2 in
-    let () = if not (Sorts.QConstraints.trivial qcsts) then raise NotConvertible in
-  (UGraph.merge_constraints csts' univs, Univ.Constraints.union csts csts')
+  if Sorts.QConstraints.trivial qcsts then
+    match UGraph.merge_constraints csts' univs with
+    | univs -> Result.Ok (univs, Univ.Constraints.union csts csts')
+    | exception (UGraph.UniverseInconsistency err) -> Result.Error (Some err)
+  else Result.Error None
 
-let inferred_universes : (UGraph.t * Univ.Constraints.t) universe_compare =
+let inferred_universes =
   { compare_sorts = infer_cmp_universes;
     compare_instances = infer_convert_instances;
     compare_cumul_instances = infer_inductive_instances; }
