@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -8,9 +8,6 @@
 (*         *     (see LICENSE file for the text of the license)         *)
 (************************************************************************)
 
-module CVars = Vars
-
-open Pp
 open CErrors
 open Util
 open Term
@@ -22,40 +19,57 @@ open Vars
 open Reductionops
 open Inductive
 open Inductiveops
-open Typeops
 open Arguments_renaming
 open Pretype_errors
 open Context.Rel.Declaration
 
 module GR = Names.GlobRef
 
-let meta_type env evd mv =
-  let ty =
-    try Evd.meta_ftype evd mv
-    with Not_found -> anomaly (str "unknown meta ?" ++ str (Nameops.string_of_meta mv) ++ str ".") in
-  meta_instance env evd ty
-
-let make_param_univs env sigma indu spec jl =
-  Array.to_list @@ Array.mapi (fun i j ~expected ->
-      match ESorts.kind sigma @@ Reductionops.sort_of_arity env sigma j.uj_type with
-      | Sorts.SProp | exception Reduction.NotArity ->
-        let indty = EConstr.of_constr @@
-          Inductive.type_of_inductive (spec, Unsafe.to_instance @@ snd indu)
+let fresh_template_context env0 sigma ind (mib, mip as spec) args =
+  let templ = match mib.Declarations.mind_template with
+  | None -> assert false
+  | Some t -> Array.of_list t.template_param_arguments
+  in
+  let ctx = List.rev (EConstr.of_rel_context mib.Declarations.mind_params_ctxt) in
+  let rec freshen i env sigma accu sorts = function
+  | [] -> sigma, List.rev sorts
+  | LocalAssum (na, t) as decl :: ctx ->
+    let sigma, decl, s =
+      if templ.(i) then
+        let decls, s0 = Reductionops.dest_arity env sigma t in
+        let sigma, s =
+          if i < Array.length args then match Reductionops.sort_of_arity env sigma args.(i).uj_type with
+          | s -> sigma, s
+          | exception Reduction.NotArity -> Evd.new_sort_variable Evd.univ_flexible sigma
+          else sigma, s0
         in
-        error_cant_apply_bad_type env sigma
-          (i+1, mkType (Univ.Universe.make expected), j.uj_type)
-          (make_judge (mkIndU indu) indty)
-          jl
-      | Sorts.Prop -> TemplateProp
-      | Sorts.Set -> TemplateUniv Univ.Universe.type0
-      | Sorts.Type u | Sorts.QSort (_, u) -> TemplateUniv u)
-    jl
+        let t = EConstr.it_mkProd_or_LetIn (mkSort s) decls in
+        let s ~expected = match ESorts.kind sigma s with
+        | Sorts.SProp ->
+          let indty = EConstr.of_constr @@
+            Inductive.type_of_inductive (spec, UVars.Instance.empty)
+          in
+          error_cant_apply_bad_type env0 sigma
+            (i+1, mkType (Univ.Universe.make expected), args.(i).uj_type)
+            (make_judge (mkIndU (ind, EInstance.empty)) indty)
+            args
+        | Sorts.Prop -> TemplateProp
+        | Sorts.Set -> TemplateUniv Univ.Universe.type0
+        | Sorts.Type u | Sorts.QSort (_, u) -> TemplateUniv u
+        in
+        sigma, LocalAssum (na, t), s
+      else
+        sigma, decl, (fun ~expected -> assert false)
+    in
+    freshen (i + 1) (push_rel decl env) sigma (decl :: accu) (s :: sorts) ctx
+  | LocalDef (na, b, t) as decl :: ctx ->
+    freshen i (push_rel decl env) sigma (decl :: accu) sorts ctx
+  in
+  freshen 0 env0 sigma [] [] ctx
 
-let inductive_type_knowing_parameters env sigma (ind,u as indu) jl =
-  let u = Unsafe.to_instance u in
-  let mspec = lookup_mind_specif env ind in
-  let paramstyp = make_param_univs env sigma indu mspec jl in
-  Inductive.type_of_inductive_knowing_parameters (mspec,u) paramstyp
+let get_template_parameters env sigma ind args =
+  let spec = Inductive.lookup_mind_specif env ind in
+  fresh_template_context env sigma ind spec args
 
 let type_judgment env sigma j =
   match EConstr.kind sigma (whd_all env sigma j.uj_type) with
@@ -72,7 +86,7 @@ let assumption_of_judgment env sigma j =
   with Type_errors.TypeError _ | PretypeError _ ->
     error_assumption env sigma j
 
-let judge_of_apply env sigma funj argjv =
+let judge_of_apply_core env sigma funj argjv =
   let rec apply_rec sigma n subs typ = function
   | [] ->
     let typ = Vars.esubst Vars.lift_substituend subs typ in
@@ -95,7 +109,10 @@ let judge_of_apply env sigma funj argjv =
         let (_, c1, c2) = destProd sigma t in
         sigma, c1, subs, c2
       | _ ->
-        error_cant_apply_not_functional env sigma funj argjv
+        let seen, rest = Array.chop (n-1) argjv in
+        error_cant_apply_not_functional env sigma
+          (make_judge (mkApp (funj.uj_val, Array.map j_val seen)) typ)
+          rest
     in
     match Evarconv.unify_leq_delay env sigma hj.uj_type c1 with
     | sigma ->
@@ -105,22 +122,55 @@ let judge_of_apply env sigma funj argjv =
   in
   apply_rec sigma 1 (Esubst.subs_id 0) funj.uj_type (Array.to_list argjv)
 
+let judge_of_applied ~check env sigma funj argjv =
+  let sigma =
+    if check then
+      let (sigma, _) = judge_of_apply_core env sigma funj argjv in
+      sigma
+    else sigma
+  in
+  let typ = hnf_prod_appvect env sigma (j_type funj) (Array.map j_val argjv) in
+  sigma, { uj_val = (mkApp (j_val funj, Array.map j_val argjv)); uj_type = typ }
+
+let judge_of_applied_inductive_knowing_parameters ~check env sigma (ind, u) argjv =
+  let (mib,_ as specif) = Inductive.lookup_mind_specif env ind in
+  let () = if check then Reductionops.check_hyps_inclusion env sigma (GR.IndRef ind) mib.mind_hyps in
+  let sigma, paramstyp = fresh_template_context env sigma ind specif argjv in
+  let u0 = EInstance.kind sigma u in
+  let ty, csts = Inductive.type_of_inductive_knowing_parameters (specif, u0) paramstyp in
+  let sigma = Evd.add_constraints sigma csts in
+  let funj = { uj_val = mkIndU (ind, u); uj_type = EConstr.of_constr (rename_type ty (GR.IndRef ind)) } in
+  judge_of_applied ~check env sigma funj argjv
+
+let judge_of_applied_constructor_knowing_parameters ~check env sigma ((ind, _ as cstr), u) argjv =
+  let (mib,_ as specif) = Inductive.lookup_mind_specif env ind in
+  let () = if check then Reductionops.check_hyps_inclusion env sigma (GR.IndRef ind) mib.mind_hyps in
+  let sigma, paramstyp = fresh_template_context env sigma ind specif argjv in
+  let u0 = EInstance.kind sigma u in
+  let ty, csts = Inductive.type_of_constructor_knowing_parameters (cstr, u0) specif paramstyp in
+  let sigma = Evd.add_constraints sigma csts in
+  let funj = { uj_val = mkConstructU (cstr, u); uj_type = (EConstr.of_constr (rename_type ty (GR.ConstructRef cstr))) } in
+  judge_of_applied ~check env sigma funj argjv
+
+let judge_of_apply env sigma fj args =
+  match EConstr.kind sigma fj.uj_val with
+  | Ind (ind, u) when EInstance.is_empty u && Environ.template_polymorphic_ind ind env ->
+    judge_of_applied_inductive_knowing_parameters ~check:true env sigma (ind, u) args
+  | Construct (cstr, u) when EInstance.is_empty u && Environ.template_polymorphic_ind (fst cstr) env ->
+    judge_of_applied_constructor_knowing_parameters ~check:true env sigma (cstr, u) args
+  | _ ->
+    (* No template polymorphism *)
+    judge_of_apply_core env sigma fj args
+
 let checked_appvect env sigma f args =
   let mk c = Retyping.get_judgment_of env sigma c in
   let sigma, j = judge_of_apply env sigma (mk f) (Array.map mk args) in
   sigma, j.uj_val
 
+let () = Hook.set Evarsolve.checked_appvect_hook checked_appvect
+
 let checked_applist env sigma f args = checked_appvect env sigma f (Array.of_list args)
 
-let judge_of_applied_inductive_knowing_parameters_nocheck env sigma funj ind argjv =
-  let ar = inductive_type_knowing_parameters env sigma ind argjv in
-  let typ = hnf_prod_appvect env sigma (EConstr.of_constr ar) (Array.map j_val argjv) in
-  sigma, { uj_val = (mkApp (j_val funj, Array.map j_val argjv)); uj_type = typ }
-
-let judge_of_applied_inductive_knowing_parameters env sigma funj ind argjv =
-  let (sigma, j) = judge_of_apply env sigma funj argjv in
-  judge_of_applied_inductive_knowing_parameters_nocheck env sigma funj ind argjv
-
 let check_branch_types env sigma (ind,u) cj (lfj,explft) =
   if not (Int.equal (Array.length lfj) (Array.length explft)) then
     error_number_branches env sigma cj (Array.length explft);
@@ -131,41 +181,36 @@ let check_branch_types env sigma (ind,u) cj (lfj,explft) =
         error_ill_formed_branch env sigma cj.uj_val ((ind,i+1),u) lfj.uj_type explft)
     sigma lfj explft
 
-let max_sort l =
-  if List.mem_f Sorts.family_equal InType l then InType else
-  if List.mem_f Sorts.family_equal InSet l then InSet else InProp
-
 let is_correct_arity env sigma c pj ind specif params =
   let arsign = make_arity_signature env sigma true (make_ind_family (ind,params)) in
-  let allowed_sorts = sorts_below (elim_sort specif) in
-  let error () = Pretype_errors.error_elim_arity env sigma ind c None in
+  let error rtnsort = Pretype_errors.error_elim_arity env sigma ind c rtnsort in
   let rec srec env sigma pt ar =
     let pt' = whd_all env sigma pt in
     match EConstr.kind sigma pt', ar with
     | Prod (na1,a1,t), (LocalAssum (_,a1'))::ar' ->
       begin match Evarconv.unify_leq_delay env sigma a1 a1' with
-        | exception Evarconv.UnableToUnify _ -> error ()
+        | exception Evarconv.UnableToUnify _ -> error None
         | sigma ->
           srec (push_rel (LocalAssum (na1,a1)) env) sigma t ar'
       end
     | Sort s, [] ->
-        let sigma = match ESorts.kind sigma s with
-        | QSort (_, u) ->
-          (* Arbitrarily set the return sort to Type *)
-          Evd.set_eq_sort env sigma s (ESorts.make (Sorts.sort_of_univ u))
-        | Set | Type _ | Prop | SProp -> sigma
+      begin match is_squashed sigma (specif, snd ind) with
+      | None -> sigma, s
+      | Some squash ->
+        let sigma =
+          try squash_elim_sort env sigma squash s
+          with UGraph.UniverseInconsistency _ -> error (Some s)
         in
-        if not (List.mem_f Sorts.family_equal (ESorts.family sigma s) allowed_sorts)
-        then error ()
-        else sigma, s
+        sigma, s
+      end
     | Evar (ev,_), [] ->
-        let sigma, s = Evd.fresh_sort_in_family sigma (max_sort allowed_sorts) in
+        let sigma, s = Evd.fresh_sort_in_family sigma (elim_sort specif) in
         let sigma = Evd.define ev (mkSort s) sigma in
         sigma, s
     | _, (LocalDef _ as d)::ar' ->
         srec (push_rel d env) sigma (lift 1 pt') ar'
     | _ ->
-        error ()
+        error None
   in
   srec env sigma pj.uj_type (List.rev arsign)
 
@@ -184,16 +229,17 @@ let type_case_branches env sigma (ind,largs) specif pj c =
   let nparams = inductive_params specif in
   let (params,realargs) = List.chop nparams largs in
   let p = pj.uj_val in
-  let params = List.map EConstr.Unsafe.to_constr params in
   let sigma, ps = is_correct_arity env sigma c pj ind specif params in
+  let ind = on_snd EConstr.Unsafe.to_instance ind in
+  let params = List.map EConstr.Unsafe.to_constr params in
   let lc = build_branches_type ind specif params (EConstr.to_constr ~abort_on_undefined_evars:false sigma p) in
   let lc = Array.map EConstr.of_constr lc in
   let n = (snd specif).Declarations.mind_nrealdecls in
   let ty = whd_betaiota env sigma (lambda_applist_decls sigma (n+1) p (realargs@[c])) in
-  sigma, (lc, ty, ESorts.relevance_of_sort sigma ps)
+  sigma, (lc, ty, ESorts.relevance_of_sort ps)
 
 let unify_relevance sigma r1 r2 =
-  match Evarutil.nf_relevance sigma r1, Evarutil.nf_relevance sigma r2 with
+  match ERelevance.kind sigma r1, ERelevance.kind sigma r2 with
   | Relevant, Relevant | Irrelevant, Irrelevant -> Some sigma
   | Relevant, Irrelevant | Irrelevant, Relevant -> None
   | Irrelevant, RelevanceVar q | RelevanceVar q, Irrelevant ->
@@ -211,12 +257,14 @@ let unify_relevance sigma r1 r2 =
     in
     Some sigma
   | RelevanceVar q1, RelevanceVar q2 ->
-    let sigma =
-      Evd.add_quconstraints sigma
-        (Sorts.QConstraints.singleton (QVar q1, Equal, QVar q2),
-         Univ.Constraints.empty)
-    in
-    Some sigma
+    if Sorts.QVar.equal q1 q2 then Some sigma
+    else
+      let sigma =
+        Evd.add_quconstraints sigma
+          (Sorts.QConstraints.singleton (QVar q1, Equal, QVar q2),
+           Univ.Constraints.empty)
+      in
+      Some sigma
 
 let judge_of_case env sigma case ci (pj,rp) iv cj lfj =
   let ((ind, u), spec) =
@@ -224,20 +272,20 @@ let judge_of_case env sigma case ci (pj,rp) iv cj lfj =
     with Not_found -> error_case_not_inductive env sigma cj in
   let specif = lookup_mind_specif env ind in
   let () = if Inductive.is_private specif then Type_errors.error_case_on_private_ind env ind in
-  let indspec = ((ind, EInstance.kind sigma u), spec) in
+  let indspec = ((ind, u), spec) in
   let sigma, (bty,rslty,rci) = type_case_branches env sigma indspec specif pj cj.uj_val in
   (* should we have evar map aware should_invert_case? *)
   let sigma, rp =
-    if Sorts.relevance_equal rp rci then sigma, rp
-    else match unify_relevance sigma rp rci with
+    match unify_relevance sigma rp rci with
     | None ->
       raise_type_error (env,sigma,Type_errors.BadCaseRelevance (rp, mkCase case))
     | Some sigma -> sigma, rci
   in
-  let () = check_case_info env (fst indspec) ci in
-  let sigma = check_branch_types env sigma (fst indspec) cj (lfj,bty) in
+  let ind = on_snd (EInstance.kind sigma) (fst indspec) in
+  let () = check_case_info env ind ci in
+  let sigma = check_branch_types env sigma ind cj (lfj,bty) in
   let () = if (match iv with | NoInvert -> false | CaseInvert _ -> true)
-              != should_invert_case env rp ci
+              != Typeops.should_invert_case env (ERelevance.kind sigma rp) ci
     then Type_errors.error_bad_invert env
   in
   sigma, { uj_val  = mkCase case;
@@ -264,10 +312,10 @@ let check_allowed_sort env sigma ind c p =
     | Sort s -> s
     | _ -> error_elim_arity env sigma ind c None
   in
-  if Inductiveops.is_allowed_elimination sigma (specif,(snd ind)) sort then
-    ESorts.relevance_of_sort sigma sort
-  else
-    error_elim_arity env sigma ind c (Some (pj, sort))
+  match Inductiveops.make_allowed_elimination env sigma (specif, (snd ind)) sort with
+  | Some sigma -> sigma, ESorts.relevance_of_sort sort
+  | None ->
+    error_elim_arity env sigma ind c (Some sort)
 
 let check_actual_type env sigma cj t =
   try Evarconv.unify_leq_delay env sigma cj.uj_type t
@@ -282,12 +330,14 @@ let judge_of_cast env sigma cj k tj =
 let check_fix env sigma pfix =
   let inj c = EConstr.to_constr ~abort_on_undefined_evars:false sigma c in
   let (idx, (ids, cs, ts)) = pfix in
-  check_fix env (idx, (ids, Array.map inj cs, Array.map inj ts))
+  let ids = Array.map EConstr.Unsafe.to_binder_annot ids in
+  check_fix ~evars:(Evd.evar_handler sigma) env (idx, (ids, Array.map inj cs, Array.map inj ts))
 
 let check_cofix env sigma pcofix =
   let inj c = EConstr.to_constr sigma c in
   let (idx, (ids, cs, ts)) = pcofix in
-  check_cofix env (idx, (ids, Array.map inj cs, Array.map inj ts))
+  let ids = Array.map EConstr.Unsafe.to_binder_annot ids in
+  check_cofix ~evars:(Evd.evar_handler sigma) env (idx, (ids, Array.map inj cs, Array.map inj ts))
 
 (* The typing machine with universes and existential variables. *)
 
@@ -316,14 +366,17 @@ let judge_of_sort s =
   in
   { uj_val = EConstr.mkSort (ESorts.make s); uj_type = EConstr.mkType u }
 
+let type_of_relative env n =
+  EConstr.of_constr (Typeops.type_of_relative env n)
+
 let judge_of_relative env v =
-  Environ.on_judgment EConstr.of_constr (judge_of_relative env v)
+  { uj_val = mkRel v; uj_type = type_of_relative env v }
 
 let type_of_variable env id =
-  EConstr.of_constr (type_of_variable env id)
+  EConstr.of_constr (Typeops.type_of_variable env id)
 
 let judge_of_variable env id =
-  Environ.on_judgment EConstr.of_constr (judge_of_variable env id)
+  { uj_val = mkVar id; uj_type = type_of_variable env id }
 
 let judge_of_projection env sigma p cj =
   let pr, pty = lookup_projection p env in
@@ -331,28 +384,27 @@ let judge_of_projection env sigma p cj =
     try find_mrectype env sigma cj.uj_type
     with Not_found -> error_case_not_inductive env sigma cj
   in
-  let u = EInstance.kind sigma u in
-  let pr = UVars.subst_instance_relevance u pr in
-  let ty = EConstr.of_constr (CVars.subst_instance_constr u pty) in
+  let pr = Vars.subst_instance_relevance u (ERelevance.make pr) in
+  let ty = Vars.subst_instance_constr u (EConstr.of_constr pty) in
   let ty = substl (cj.uj_val :: List.rev args) ty in
   {uj_val = EConstr.mkProj (p,pr,cj.uj_val);
    uj_type = ty}
 
 let judge_of_abstraction env sigma name var j =
-  let r = ESorts.relevance_of_sort sigma var.utj_type in
+  let r = ESorts.relevance_of_sort var.utj_type in
   { uj_val = mkLambda (make_annot name r, var.utj_val, j.uj_val);
     uj_type = mkProd (make_annot name r, var.utj_val, j.uj_type) }
 
 let judge_of_product env sigma name t1 t2 =
   let s1 = ESorts.kind sigma t1.utj_type in
   let s2 = ESorts.kind sigma t2.utj_type in
-  let r = Sorts.relevance_of_sort s1 in
-  let s = ESorts.make (sort_of_product env s1 s2) in
+  let r = ERelevance.make @@ Sorts.relevance_of_sort s1 in
+  let s = ESorts.make (Typeops.sort_of_product env s1 s2) in
   { uj_val = mkProd (make_annot name r, t1.utj_val, t2.utj_val);
     uj_type = mkSort s }
 
 let judge_of_letin env sigma name defj typj j =
-  let r = ESorts.relevance_of_sort sigma typj.utj_type in
+  let r = ESorts.relevance_of_sort typj.utj_type in
   { uj_val = mkLetIn (make_annot name r, defj.uj_val, typj.utj_val, j.uj_val) ;
     uj_type = subst1 defj.uj_val j.uj_type }
 
@@ -383,11 +435,20 @@ let type_of_constructor env sigma ((ind,_ as ctor),u) =
   let sigma = Evd.add_constraints sigma csts in
   sigma, (EConstr.of_constr (rename_type ty (GR.ConstructRef ctor)))
 
+let type_of_int env = EConstr.of_constr (Typeops.type_of_int env)
+
 let judge_of_int env v =
-  Environ.on_judgment EConstr.of_constr (judge_of_int env v)
+  { uj_val = mkInt v; uj_type = type_of_int env }
+
+let type_of_float env = EConstr.of_constr (Typeops.type_of_float env)
 
 let judge_of_float env v =
-  Environ.on_judgment EConstr.of_constr (judge_of_float env v)
+  { uj_val = mkFloat v; uj_type = type_of_float env }
+
+let type_of_string env = EConstr.of_constr (Typeops.type_of_string env)
+
+let judge_of_string env v =
+  { uj_val = mkString v; uj_type = type_of_string env }
 
 let judge_of_array env sigma u tj defj tyj =
   let ulev = match UVars.Instance.to_array u with
@@ -400,21 +461,28 @@ let judge_of_array env sigma u tj defj tyj =
   let check_one sigma j = check_actual_type env sigma j tyj.utj_val in
   let sigma = check_one sigma defj in
   let sigma = Array.fold_left check_one sigma tj in
-  let arr = EConstr.of_constr @@ type_of_array env u in
+  let arr = EConstr.of_constr @@ Typeops.type_of_array env u in
   let j = make_judge (mkArray(EInstance.make u, Array.map j_val tj, defj.uj_val, tyj.utj_val))
       (mkApp (arr, [|tyj.utj_val|]))
   in
   sigma, j
 
-let bad_relevance_msg = CWarnings.create_msg Typeops.bad_relevance_warning ()
+type ('constr,'types,'r) bad_relevance =
+| BadRelevanceBinder of 'r * ('constr,'types,'r) Context.Rel.Declaration.pt
+| BadRelevanceCase of 'r * 'constr
+
+let bad_relevance_warning =
+  CWarnings.create_warning ~name:"bad-relevance" ~default:CWarnings.AsError ()
+
+let bad_relevance_msg = CWarnings.create_msg bad_relevance_warning ()
 (* no need for default printer, pretyping is always linked with himsg in practice *)
 
 let warn_bad_relevance_binder ?loc env sigma rlv bnd =
-  match CWarnings.warning_status Typeops.bad_relevance_warning with
+  match CWarnings.warning_status bad_relevance_warning with
 | CWarnings.Disabled | CWarnings.Enabled ->
-  CWarnings.warn bad_relevance_msg ?loc (env,sigma,Typeops.BadRelevanceBinder (rlv, bnd))
+  CWarnings.warn bad_relevance_msg ?loc (env,sigma,BadRelevanceBinder (rlv, bnd))
 | CWarnings.AsError ->
-  CErrors.anomaly (CErrors.print (PretypeError (env, sigma, TypingError (Type_errors.BadBinderRelevance (rlv, bnd)))))
+  Loc.raise ?loc (PretypeError (env, sigma, TypingError (Type_errors.BadBinderRelevance (rlv, bnd))))
 
 type relevance_preunify =
   | Trivial
@@ -422,7 +490,7 @@ type relevance_preunify =
   | DummySort of ESorts.t
 
 let check_binder_relevance env sigma s decl =
-  let preunify = match ESorts.kind sigma s, Evarutil.nf_relevance sigma (get_relevance decl) with
+  let preunify = match ESorts.kind sigma s, ERelevance.kind sigma (get_relevance decl) with
     | (Prop | Set | Type _), Relevant -> Trivial
     | (Prop | Set | Type _), Irrelevant -> Impossible
     | SProp, Irrelevant -> Trivial
@@ -446,7 +514,7 @@ let check_binder_relevance env sigma s decl =
   | Some sigma -> sigma, decl
   | None ->
     (* TODO always anomaly *)
-    let rs = ESorts.relevance_of_sort sigma s in
+    let rs = ESorts.relevance_of_sort s in
     let () =
       if not (UGraph.type_in_type (Evd.universes sigma))
       then warn_bad_relevance_binder env sigma rs decl
@@ -458,8 +526,7 @@ let check_binder_relevance env sigma s decl =
 let rec execute env sigma cstr =
   let cstr = whd_evar sigma cstr in
   match EConstr.kind sigma cstr with
-    | Meta n ->
-        sigma, { uj_val = cstr; uj_type = meta_type env sigma n }
+    | Meta n -> assert false (* Typing should always be performed on meta-free terms *)
 
     | Evar ev ->
         let ty = EConstr.existential_type sigma ev in
@@ -531,15 +598,9 @@ let rec execute env sigma cstr =
       sigma, judge_of_projection env sigma p cj
 
     | App (f,args) ->
+        let sigma, fj = execute env sigma f in
         let sigma, jl = execute_array env sigma args in
-        (match EConstr.kind sigma f with
-            | Ind (ind, u) when EInstance.is_empty u && Environ.template_polymorphic_ind ind env ->
-               let sigma, fj = execute env sigma f in
-               judge_of_applied_inductive_knowing_parameters env sigma fj (ind, u) jl
-            | _ ->
-               (* No template polymorphism *)
-               let sigma, fj = execute env sigma f in
-               judge_of_apply env sigma fj jl)
+        judge_of_apply env sigma fj jl
 
     | Lambda (name,c1,c2) ->
         let sigma, j = execute env sigma c1 in
@@ -580,6 +641,9 @@ let rec execute env sigma cstr =
     | Float f ->
         sigma, judge_of_float env f
 
+    | String s ->
+        sigma, judge_of_string env s
+
     | Array(u,t,def,ty) ->
       let sigma, tyj = execute env sigma ty in
       let sigma, tyj = type_judgment env sigma tyj in
@@ -704,7 +768,7 @@ let rec recheck_against env sigma good c =
     match kind sigma good, kind sigma c with
     (* No subterms *)
     | _, (Meta _ | Rel _ | Var _ | Const _ | Ind _ | Construct _
-         | Sort _ | Int _ | Float _) ->
+         | Sort _ | Int _ | Float _ | String _) ->
       default ()
 
     (* Evar (todo deal with Evar differently??? execute recurses on its type)
@@ -770,7 +834,9 @@ let rec recheck_against env sigma good c =
         let sigma, jl = execute_array env sigma args in
         (match EConstr.kind sigma f with
          | Ind (ind, u) when EInstance.is_empty u && Environ.template_polymorphic_ind ind env ->
-           maybe_changed (judge_of_applied_inductive_knowing_parameters env sigma fj (ind, u) jl)
+           maybe_changed (judge_of_applied_inductive_knowing_parameters ~check:true env sigma (ind, u) jl)
+         | Construct (cstr, u) when EInstance.is_empty u && Environ.template_polymorphic_ind (fst cstr) env ->
+           maybe_changed (judge_of_applied_constructor_knowing_parameters ~check:true env sigma (cstr, u) jl)
          | _ ->
            (* No template polymorphism *)
            maybe_changed (judge_of_apply env sigma fj jl))
@@ -785,11 +851,16 @@ let rec recheck_against env sigma good c =
         if unchanged changedargs && bodyonly changedf
         then assume_unchanged_type sigma
         else
+          (* FIXME: the template poly cases are generating useless constraints *)
           (match EConstr.kind sigma f with
            | Ind (ind, u) when EInstance.is_empty u && Environ.template_polymorphic_ind ind env ->
              let sigma, _ = judge_of_apply_against env sigma changedf fj jl in
              let jl = Array.map snd jl in
-             maybe_changed (judge_of_applied_inductive_knowing_parameters_nocheck env sigma fj (ind, u) jl)
+             maybe_changed (judge_of_applied_inductive_knowing_parameters ~check:false env sigma (ind, u) jl)
+           | Construct (cstr, u) when EInstance.is_empty u && Environ.template_polymorphic_ind (fst cstr) env ->
+             let sigma, _ = judge_of_apply_against env sigma changedf fj jl in
+             let jl = Array.map snd jl in
+             maybe_changed (judge_of_applied_constructor_knowing_parameters ~check:false env sigma (cstr, u) jl)
            | _ ->
              (* No template polymorphism *)
              maybe_changed (judge_of_apply_against env sigma changedf fj jl))
