@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -68,10 +68,11 @@ struct
     | DCoFix  of int
     | DInt    of Uint63.t
     | DFloat  of Float64.t
+    | DString of Pstring.t
     | DArray
 
   let compare_ci ci1 ci2 =
-    let c = Ind.CanOrd.compare ci1.ci_ind ci2.ci_ind in
+    let c = Label.compare (MutInd.label @@ fst ci1.ci_ind) (MutInd.label @@ fst ci2.ci_ind) in
     if c = 0 then
       let c = Int.compare ci1.ci_npar ci2.ci_npar in
       if c = 0 then
@@ -87,7 +88,7 @@ struct
   | DRel, _ -> -1 | _, DRel -> 1
   | DSort, DSort -> 0
   | DSort, _ -> -1 | _, DSort -> 1
-  | DRef gr1, DRef gr2 -> GlobRef.CanOrd.compare gr1 gr2
+  | DRef gr1, DRef gr2 -> GlobRef.UserOrd.compare gr1 gr2
   | DRef _, _ -> -1 | _, DRef _ -> 1
 
   | DProd, DProd -> 0
@@ -124,6 +125,10 @@ struct
 
   | DFloat _, _ -> -1 | _, DFloat _ -> 1
 
+  | DString s1, DString s2 -> Pstring.compare s1 s2
+
+  | DString _, _ -> -1 | _, DString _ -> 1
+
   | DArray, DArray -> 1
 
 end
@@ -143,7 +148,7 @@ sig
 
   (** [add c i dn] adds the binding [(c,i)] to [dn]. [c] can be a
      closed term or a pattern (with untyped Evars). No Metas accepted *)
-  val add : constr -> ident -> t -> t
+  val add : Environ.env -> constr -> ident -> t -> t
 
   (*
    * High-level primitives describing specific search problems
@@ -151,7 +156,7 @@ sig
 
   (** [search_pattern dn c] returns all terms/patterns in dn
      matching/matched by c *)
-  val search_pattern : t -> constr -> ident list
+  val search_pattern : Environ.env -> t -> constr -> ident list
 
   (** [find_all dn] returns all idents contained in dn *)
   val find_all : t -> ident list
@@ -174,18 +179,18 @@ struct
   open DTerm
   open TDnet
 
-  let pat_of_constr c : (unit DTerm.t * Constr.t list) option =
+  let pat_of_constr env c : (unit DTerm.t * Constr.t list) option =
     let open GlobRef in
     let rec pat_of_constr c = match Constr.kind c with
     | Rel _          -> Some (DRel, [])
     | Sort _         -> Some (DSort, [])
     | Var i          -> Some (DRef (VarRef i), [])
-    | Const (c,u)    -> Some (DRef (ConstRef c), [])
-    | Ind (i,u)      -> Some (DRef (IndRef i), [])
-    | Construct (c,u)-> Some (DRef (ConstructRef c), [])
+    | Const (c,u)    -> Some (DRef (ConstRef (Environ.QConstant.canonize env c)), [])
+    | Ind (i,u)      -> Some (DRef (IndRef (Environ.QInd.canonize env i)), [])
+    | Construct (c,u)-> Some (DRef (ConstructRef (Environ.QConstruct.canonize env c)), [])
     | Meta _         -> assert false
     | Evar (i,_)     -> None
-    | Case (ci,u1,pms1,c1,_iv,c2,ca)     ->
+    | Case (ci,u1,pms1,(c1,_),_iv,c2,ca)     ->
       let f_ctx (_, p) = p in
       Some (DCase(ci), [f_ctx c1; c2] @ Array.map_to_list f_ctx ca)
     | Fix ((ia,i),(_,ta,ca)) ->
@@ -201,9 +206,13 @@ struct
       let a = ca.(len - 1) in
       let ca = Array.sub ca 0 (len - 1) in
       Some (DApp, [mkApp (f, ca); a])
-    | Proj (p,c) -> pat_of_constr (mkApp (mkConst (Projection.constant p), [|c|]))
+    | Proj (p,_,c) ->
+      (* UnsafeMonomorphic is fine because the term will only be used
+         by pat_of_constr which ignores universes *)
+      pat_of_constr (mkApp (UnsafeMonomorphic.mkConst (Projection.constant p), [|c|]))
     | Int i -> Some (DInt i, [])
     | Float f -> Some (DFloat f, [])
+    | String s -> Some (DString s, [])
     | Array (_u,t,def,ty) ->
       Some (DArray, Array.to_list t @ [def ; ty])
     in
@@ -215,12 +224,12 @@ struct
 
   let empty = TDnet.empty
 
-  let add (c:constr) (id:Ident.t) (dn:t) =
+  let add env (c:constr) (id:Ident.t) (dn:t) =
     (* We used to consider the types of the product as well, but since the dnet
        is only computing an approximation rectified by [filtering] we do not
        anymore. *)
-    let (ctx, c) = Term.decompose_prod_assum c in
-    let c = TDnet.pattern pat_of_constr c in
+    let (ctx, c) = Term.decompose_prod_decls c in
+    let c = TDnet.pattern (fun c -> pat_of_constr env c) c in
     TDnet.add dn c id
 
 (* App(c,[t1,...tn]) -> ([c,t1,...,tn-1],tn)
@@ -236,56 +245,56 @@ let split_app sigma c = match EConstr.kind sigma c with
 
 exception CannotFilter
 
-let filtering env sigma cv_pb c1 c2 =
+let filtering env sigma ctx cv_pb c1 c2 =
   let open EConstr in
   let open Vars in
   let evm = ref Evar.Map.empty in
   let define cv_pb e1 ev c1 =
     try let (e2,c2) = Evar.Map.find ev !evm in
     let shift = e1 - e2 in
-    if Termops.constr_cmp sigma cv_pb c1 (lift shift c2) then () else raise CannotFilter
+    if Termops.constr_cmp env sigma cv_pb c1 (lift shift c2) then () else raise CannotFilter
     with Not_found ->
       evm := Evar.Map.add ev (e1,c1) !evm
   in
-  let rec aux env cv_pb c1 c2 =
+  let rec aux ctx cv_pb c1 c2 =
     match EConstr.kind sigma c1, EConstr.kind sigma c2 with
       | App _, App _ ->
         let ((p1,l1),(p2,l2)) = (split_app sigma c1),(split_app sigma c2) in
-        let () = aux env cv_pb l1 l2 in
+        let () = aux ctx cv_pb l1 l2 in
         begin match p1, p2 with
         | [], [] -> ()
         | (h1 :: p1), (h2 :: p2) ->
-          aux env cv_pb (applist (h1, p1)) (applist (h2, p2))
+          aux ctx cv_pb (applist (h1, p1)) (applist (h2, p2))
         | _ -> assert false
         end
       | Prod (n,t1,c1), Prod (_,t2,c2) ->
-          aux env cv_pb t1 t2;
-          aux (env + 1) cv_pb c1 c2
-      | _, Evar (ev,_) -> define cv_pb env ev c1
-      | Evar (ev,_), _ -> define cv_pb env ev c2
+          aux ctx cv_pb t1 t2;
+          aux (ctx + 1) cv_pb c1 c2
+      | _, Evar (ev,_) -> define cv_pb ctx ev c1
+      | Evar (ev,_), _ -> define cv_pb ctx ev c2
       | _ ->
-          if Termops.compare_constr_univ sigma
-          (fun pb c1 c2 -> aux env pb c1 c2; true) cv_pb c1 c2 then ()
+          if Termops.compare_constr_univ env sigma
+          (fun pb c1 c2 -> aux ctx pb c1 c2; true) cv_pb c1 c2 then ()
           else raise CannotFilter
           (* TODO: le reste des binders *)
   in
-  try let () = aux env cv_pb c1 c2 in true with CannotFilter -> false
+  try let () = aux ctx cv_pb c1 c2 in true with CannotFilter -> false
 
 let align_prod_letin sigma c a =
-  let (lc,_) = EConstr.decompose_prod_assum sigma c in
-  let (l,a) = EConstr.decompose_prod_assum sigma a in
+  let (lc,_) = EConstr.decompose_prod_decls sigma c in
+  let (l,a) = EConstr.decompose_prod_decls sigma a in
   let lc = List.length lc in
   let n = List.length l in
   if n < lc then invalid_arg "align_prod_letin";
   let l1 = CList.firstn lc l in
   n - lc, EConstr.it_mkProd_or_LetIn a l1
 
-  let decomp pat = match pat_of_constr pat with
+  let decomp env pat = match pat_of_constr env pat with
   | None -> Dn.Everything
   | Some (lbl, args) -> Dn.Label (lbl, args)
 
-  let search_pattern dn cpat =
-    let _dctx, dpat = Term.decompose_prod_assum cpat in
+  let search_pattern env dn cpat =
+    let _dctx, dpat = Term.decompose_prod_decls cpat in
     let whole_c = EConstr.of_constr cpat in
     List.fold_left
       (fun acc id ->
@@ -293,9 +302,9 @@ let align_prod_letin sigma c a =
          let (ctx,wc) =
            try align_prod_letin Evd.empty whole_c c_id (* FIXME *)
            with Invalid_argument _ -> 0, c_id in
-        if filtering ctx Evd.empty Reduction.CUMUL whole_c wc then id :: acc
+        if filtering env Evd.empty ctx Conversion.CUMUL whole_c wc then id :: acc
         else acc
-      ) (TDnet.lookup dn decomp dpat) []
+      ) (TDnet.lookup dn (fun c -> decomp env c) dpat) []
 
   let find_all dn = TDnet.lookup dn (fun () -> Everything) ()
 
@@ -330,9 +339,9 @@ let find_rewrites bas =
   let sort r1 r2 = Int.compare (KNmap.find r2.rew_id db.rdb_order) (KNmap.find r1.rew_id db.rdb_order) in
   List.sort sort (HintDN.find_all db.rdb_hintdn)
 
-let find_matches bas pat =
+let find_matches env bas pat =
   let base = find_base bas in
-  let res = HintDN.search_pattern base.rdb_hintdn pat in
+  let res = HintDN.search_pattern env base.rdb_hintdn pat in
   let sort r1 r2 = Int.compare (KNmap.find r2.rew_id base.rdb_order) (KNmap.find r1.rew_id base.rdb_order) in
   List.sort sort res
 
@@ -364,6 +373,7 @@ let one_base where conds tac_main bas =
   Proofview.Goal.enter begin fun gl ->
     let sigma = Proofview.Goal.sigma gl in
     let subst, ctx' = UnivGen.fresh_universe_context_set_instance h.rew_ctx in
+    let subst = Sorts.QVar.Map.empty, subst in
     let c' = Vars.subst_univs_level_constr subst h.rew_lemma in
     let sigma = Evd.merge_context_set Evd.univ_flexible sigma ctx' in
     Proofview.tclTHEN (Proofview.Unsafe.tclEVARS sigma) (rewrite h.rew_l2r c' tc)
@@ -373,11 +383,11 @@ let one_base where conds tac_main bas =
   let eval h =
     let tac = match h.rew_tac with
     | None -> Proofview.tclUNIT ()
-    | Some (Genarg.GenArg (Genarg.Glbwit wit, tac)) ->
+    | Some tac ->
       let ist = { Geninterp.lfun = Id.Map.empty
                 ; poly
                 ; extra = Geninterp.TacStore.empty } in
-      Ftactic.run (Geninterp.interp wit ist tac) (fun _ -> Proofview.tclUNIT ())
+      Ftactic.run (Geninterp.generic_interp ist tac) (fun _ -> Proofview.tclUNIT ())
     in
     Tacticals.tclREPEAT_MAIN (Tacticals.tclTHENFIRST (try_rewrite h tac) tac_main)
   in
@@ -462,7 +472,7 @@ let auto_multi_rewrite_with ?(conds=Naive) tac_main lbas cl =
 let cache_hintrewrite (rbase,lrl) =
   let base = try raw_find_base rbase with Not_found -> empty_rewrite_db in
   let fold accu r = {
-    rdb_hintdn = HintDN.add r.rew_pat r accu.rdb_hintdn;
+    rdb_hintdn = HintDN.add (Global.env ()) r.rew_pat r accu.rdb_hintdn;
     rdb_order = KNmap.add r.rew_id accu.rdb_maxuid accu.rdb_order;
     rdb_maxuid = accu.rdb_maxuid + 1;
   } in
@@ -475,7 +485,7 @@ let subst_hintrewrite (subst,(rbase,list as node)) =
     let cst' = subst_mps subst hint.rew_lemma in
     let typ' = subst_mps subst hint.rew_type in
     let pat' = subst_mps subst hint.rew_pat in
-    let t' = Option.Smart.map (Genintern.generic_substitute subst) hint.rew_tac in
+    let t' = Option.Smart.map (Gensubst.generic_substitute subst) hint.rew_tac in
       if hint.rew_id == id' && hint.rew_lemma == cst' && hint.rew_type == typ' &&
          hint.rew_tac == t' && hint.rew_pat == pat' then hint else
         { hint with
@@ -487,17 +497,12 @@ let subst_hintrewrite (subst,(rbase,list as node)) =
       (rbase,list')
 
 (* Declaration of the Hint Rewrite library object *)
-let inGlobalHintRewrite : string * rew_rule list -> Libobject.obj =
+let inHintRewrite : Libobject.locality * (string * rew_rule list) -> Libobject.obj =
   let open Libobject in
-  declare_object @@ superglobal_object_nodischarge "HINT_REWRITE_GLOBAL"
-    ~cache:cache_hintrewrite
-    ~subst:(Some subst_hintrewrite)
-
-let inExportHintRewrite : string * rew_rule list -> Libobject.obj =
-  let open Libobject in
-  declare_object @@ global_object_nodischarge ~cat:Hints.hint_cat "HINT_REWRITE_EXPORT"
+  declare_object @@ object_with_locality "HINT_REWRITE_GLOBAL"
     ~cache:cache_hintrewrite
     ~subst:(Some subst_hintrewrite)
+    ~discharge:(fun _ -> assert false)
 
 type hypinfo = {
   hyp_ty : EConstr.types;
@@ -509,8 +514,8 @@ let decompose_applied_relation env sigma c ctype left2right =
     (* FIXME: this is nonsense, we generate evars and then we drop the
        corresponding evarmap. This sometimes works because [Term_dnet] performs
        evar surgery via [Termops.filtering]. *)
-    let sigma, ty = Clenv.make_evar_clause env sigma ty in
-    let (_, args) = Termops.decompose_app_vect sigma ty.Clenv.cl_concl in
+    let sigma, ty = EClause.make_evar_clause env sigma ty in
+    let (_, args) = EConstr.decompose_app sigma ty.EClause.cl_concl in
     let len = Array.length args in
     if 2 <= len then
       let c1 = args.(len - 2) in
@@ -521,7 +526,7 @@ let decompose_applied_relation env sigma c ctype left2right =
     match find_rel ctype with
     | Some c -> Some { hyp_pat = c; hyp_ty = ctype }
     | None ->
-        let ctx,t' = Reductionops.splay_prod_assum env sigma ctype in (* Search for underlying eq *)
+        let ctx,t' = Reductionops.whd_decompose_prod_decls env sigma ctype in (* Search for underlying eq *)
         let ctype = EConstr.it_mkProd_or_LetIn t' ctx in
         match find_rel ctype with
         | Some c -> Some { hyp_pat = c; hyp_ty = ctype }
@@ -536,30 +541,12 @@ let find_applied_relation ?loc env sigma c left2right =
                     (str"The type" ++ spc () ++ Printer.pr_econstr_env env sigma ctype ++
                        spc () ++ str"of this term does not end with an applied relation.")
 
-let warn_deprecated_hint_rewrite_without_locality =
-  CWarnings.create ~name:"deprecated-hint-rewrite-without-locality" ~category:"deprecated"
-    ~default:CWarnings.AsError
-    (fun () -> strbrk "The default value for rewriting hint locality is \
-    currently \"global\" outside sections, but is scheduled to change to \
-    \"export\" in the next release (Coq 8.18). In Coq 8.17, not providing \
-    an explicit locality outside sections triggers a fatal warning, to \
-    ensure that hint localities are made explicit before the upcoming \
-    change in the default value. It is recommended to use \"export\" \
-    whenever possible. Use the attributes \
-    #[local], #[global] and #[export] depending on your choice. For example: \
-    \"#[export] Hint Rewrite foo : bar.\" This is supported since Coq 8.14.")
-
-let default_hint_rewrite_locality () =
-  if Global.sections_are_opened () then Hints.Local
-  else
-    let () = warn_deprecated_hint_rewrite_without_locality () in
-    Hints.SuperGlobal
-
 (* To add rewriting rules to a base *)
 let add_rew_rules ~locality base lrul =
+  let () = Locality.check_locality_nodischarge locality in
   let env = Global.env () in
   let sigma = Evd.from_env env in
-  let ist = Genintern.empty_glob_sign (Global.env ()) in
+  let ist = Genintern.empty_glob_sign ~strict:true (Global.env ()) in
   let intern tac = snd (Genintern.generic_intern ist tac) in
   let map {CAst.loc;v=((c,ctx),b,t)} =
     let sigma = Evd.merge_context_set Evd.univ_rigid sigma ctx in
@@ -571,20 +558,4 @@ let add_rew_rules ~locality base lrul =
       rew_tac = Option.map intern t }
   in
   let lrul = List.map map lrul in
-  let open Hints in
-  match locality with
-  | Local -> cache_hintrewrite (base,lrul)
-  | SuperGlobal ->
-    let () =
-      if Global.sections_are_opened () then
-      CErrors.user_err Pp.(str
-        "This command does not support the global attribute in sections.");
-    in
-    Lib.add_leaf (inGlobalHintRewrite (base,lrul))
-  | Export ->
-    let () =
-      if Global.sections_are_opened () then
-        CErrors.user_err Pp.(str
-          "This command does not support the export attribute in sections.");
-    in
-    Lib.add_leaf (inExportHintRewrite (base,lrul))
+  Lib.add_leaf (inHintRewrite (locality,(base,lrul)))
