@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -21,7 +21,7 @@
      existing non-dependent schemes eq_rect and eq_rect_r. There is in
      particular a problem with the dependent rewriting schemes in
      hypotheses for which the inductive types cannot be in last
-     position of the scheme as it is the general rule in Coq. This has
+     position of the scheme as it is the general rule in Rocq. This has
      an effect on the order of generated goals (side-conditions of the
      lemma after or before the main goal). The non-dependent case can be
      fixed but to the price of a lost of uniformity wrt side-conditions
@@ -99,8 +99,8 @@ let my_it_mkLambda_or_LetIn_name env s c =
   let mkLambda_or_LetIn_name d b = mkLambda_or_LetIn (name_assumption env d) b in
   List.fold_left (fun c d -> mkLambda_or_LetIn_name d c) c s
 
-let get_coq_eq env ctx =
-  let eq = Globnames.destIndRef (Coqlib.lib_ref "core.eq.type") in
+let get_rocq_eq env ctx =
+  let eq = Globnames.destIndRef (Rocqlib.lib_ref "core.eq.type") in
   let eq, ctx = with_context_set ctx
       (UnivGen.fresh_inductive_instance env eq) in
   mkIndU eq, mkConstructUi (eq,1), ctx
@@ -200,7 +200,7 @@ let build_sym_scheme env _handle ind =
     get_sym_eq_data env indu in
   let cstr n =
     mkApp (mkConstructUi(indu,1),Context.Rel.instance mkRel n mib.mind_params_ctxt) in
-  let inds = inductive_sort_family mip in
+  let inds = Indrec.pseudo_sort_family_for_elim ind mip in
   let varH,_ = fresh env (default_id_of_sort inds) Id.Set.empty in
   let applied_ind = build_dependent_inductive indu specif in
   let indr = UVars.subst_instance_relevance u mip.mind_relevance in
@@ -231,7 +231,7 @@ let build_sym_scheme env _handle ind =
   c, UState.of_context_set ctx
 
 let sym_scheme_kind =
-  declare_individual_scheme_object "_sym_internal"
+  declare_individual_scheme_object "sym_internal"
   build_sym_scheme
 
 (**********************************************************************)
@@ -260,10 +260,10 @@ let build_sym_involutive_scheme env handle ind =
   let (ind,u as indu), ctx = UnivGen.fresh_inductive_instance env ind in
   let (mib,mip as specif),nrealargs,realsign,paramsctxt,paramsctxt1 =
     get_sym_eq_data env indu in
-  let eq,eqrefl,ctx = get_coq_eq env ctx in
+  let eq,eqrefl,ctx = get_rocq_eq env ctx in
   let sym, ctx = const_of_scheme sym_scheme_kind env handle ind ctx in
   let cstr n = mkApp (mkConstructUi (indu,1),Context.Rel.instance mkRel n paramsctxt) in
-  let inds = inductive_sort_family mip in
+  let inds = Indrec.pseudo_sort_family_for_elim ind mip in
   let indr = UVars.subst_instance_relevance u mip.mind_relevance in
   let varH,_ = fresh env (default_id_of_sort inds) Id.Set.empty in
   let applied_ind = build_dependent_inductive indu specif in
@@ -304,7 +304,7 @@ let build_sym_involutive_scheme env handle ind =
   in (c, UState.of_context_set ctx)
 
 let sym_involutive_scheme_kind =
-  declare_individual_scheme_object "_sym_involutive"
+  declare_individual_scheme_object "sym_involutive"
   ~deps:(fun _ ind -> [SchemeIndividualDep (ind, sym_scheme_kind)])
   build_sym_involutive_scheme
 
@@ -374,12 +374,12 @@ let build_l2r_rew_scheme dep env handle ind kind =
     get_sym_eq_data env indu in
   let sym, ctx = const_of_scheme sym_scheme_kind env handle ind ctx in
   let sym_involutive, ctx = const_of_scheme sym_involutive_scheme_kind env handle ind ctx in
-  let eq,eqrefl,ctx = get_coq_eq env ctx in
+  let eq,eqrefl,ctx = get_rocq_eq env ctx in
   let cstr n p =
     mkApp (mkConstructUi(indu,1),
       Array.concat [Context.Rel.instance mkRel n paramsctxt1;
                     rel_vect p nrealargs]) in
-  let inds = inductive_sort_family mip in
+  let inds = Indrec.pseudo_sort_family_for_elim ind mip in
   let indr = UVars.subst_instance_relevance u mip.mind_relevance in
   let varH,avoid = fresh env (default_id_of_sort inds) Id.Set.empty in
   let varHC,avoid = fresh env (Id.of_string "HC") avoid in
@@ -497,7 +497,7 @@ let build_l2r_forward_rew_scheme dep env ind kind =
     mkApp (mkConstructUi(indu,1),
       Array.concat [Context.Rel.instance mkRel n paramsctxt1;
                     rel_vect p nrealargs]) in
-  let inds = inductive_sort_family mip in
+  let inds = Indrec.pseudo_sort_family_for_elim ind mip in
   let indr = UVars.subst_instance_relevance u mip.mind_relevance in
   let varH,avoid = fresh env (default_id_of_sort inds) Id.Set.empty in
   let varHC,avoid = fresh env (Id.of_string "HC") avoid in
@@ -575,7 +575,7 @@ let build_l2r_forward_rew_scheme dep env ind kind =
 (* Note that the dependent elimination here is not a dependency       *)
 (* in the conclusion of the scheme but a dependency in the premise of *)
 (* the scheme. This is unfortunately incompatible with the standard   *)
-(* pattern for schemes in Coq which expects that the eliminated       *)
+(* pattern for schemes in Rocq which expects that the eliminated      *)
 (* object is the last premise of the scheme. We then have no choice   *)
 (* than following the more liberal pattern of having the eliminated   *)
 (* object coming before the premises.                                 *)
@@ -592,8 +592,8 @@ let build_r2l_forward_rew_scheme dep env ind kind =
   let cstr n =
     mkApp (mkConstructUi(indu,1),Context.Rel.instance mkRel n mib.mind_params_ctxt) in
   let constrargs_cstr = constrargs@[cstr 0] in
-  let inds = inductive_sort_family mip in
-  let indr = Sorts.relevance_of_sort_family inds in
+  let inds = Indrec.pseudo_sort_family_for_elim ind mip in
+  let indr = Inductive.relevance_of_ind_body mip u in
   let varH,avoid = fresh env (default_id_of_sort inds) Id.Set.empty in
   let varHC,avoid = fresh env (Id.of_string "HC") avoid in
   let varP,_ = fresh env (Id.of_string "P") avoid in
@@ -635,7 +635,7 @@ let build_r2l_forward_rew_scheme dep env ind kind =
 (**********************************************************************)
 (* This function "repairs" the non-dependent r2l forward rewriting    *)
 (* scheme by making it comply with the standard pattern of schemes    *)
-(* in Coq. Otherwise said, it turns a scheme of type                  *)
+(* in Rocq. Otherwise said, it turns a scheme of type                 *)
 (*                                                                    *)
 (*  forall q1..pm a1..an, I q1..qm a1..an ->                          *)
 (*  forall (P: forall a1..an, kind),                                  *)
@@ -690,9 +690,11 @@ let fix_r2l_forward_rew_scheme env (c, ctx') =
 let build_r2l_rew_scheme dep env ind k =
   let sigma = Evd.from_env env in
   let (sigma, indu) = Evd.fresh_inductive_instance env sigma ind in
+  let indu = on_snd EConstr.EInstance.make indu in
+  let sigma, k = Evd.fresh_sort_in_family ~rigid:UnivRigid sigma k in
   let (sigma, c) = build_case_analysis_scheme env sigma indu dep k in
   let (c, _) = Indrec.eval_case_analysis c in
-  EConstr.Unsafe.to_constr c, Evd.evar_universe_context sigma
+  EConstr.Unsafe.to_constr c, Evd.ustate sigma
 
 (**********************************************************************)
 (* Register the rewriting schemes                                     *)
@@ -705,7 +707,7 @@ let build_r2l_rew_scheme dep env ind k =
 (* with H:I p1..pn a1..an in Gamma                                    *)
 (**********************************************************************)
 let rew_l2r_dep_scheme_kind =
-  declare_individual_scheme_object "_rew_r_dep"
+  declare_individual_scheme_object "rew_r_dep"
   ~deps:(fun _ ind -> [
     SchemeIndividualDep (ind, sym_scheme_kind);
     SchemeIndividualDep (ind, sym_involutive_scheme_kind);
@@ -719,7 +721,7 @@ let rew_l2r_dep_scheme_kind =
 (* or   H:I b1..bn a1..an in Gamma (symmetric case)                   *)
 (**********************************************************************)
 let rew_r2l_dep_scheme_kind =
-  declare_individual_scheme_object "_rew_dep"
+  declare_individual_scheme_object "rew_dep"
   (fun env _ ind -> build_r2l_rew_scheme true env ind InType)
 
 (**********************************************************************)
@@ -729,7 +731,7 @@ let rew_r2l_dep_scheme_kind =
 (* or   H:I b1..bn a1..an in Gamma (symmetric case)                   *)
 (**********************************************************************)
 let rew_r2l_forward_dep_scheme_kind =
-  declare_individual_scheme_object "_rew_fwd_dep"
+  declare_individual_scheme_object "rew_fwd_dep"
   (fun env _ ind -> build_r2l_forward_rew_scheme true env ind InType)
 
 (**********************************************************************)
@@ -739,7 +741,7 @@ let rew_r2l_forward_dep_scheme_kind =
 (* with H:I p1..pn a1..an in Gamma                                    *)
 (**********************************************************************)
 let rew_l2r_forward_dep_scheme_kind =
-  declare_individual_scheme_object "_rew_fwd_r_dep"
+  declare_individual_scheme_object "rew_fwd_r_dep"
   (fun env _ ind -> build_l2r_forward_rew_scheme true env ind InType)
 
 (**********************************************************************)
@@ -749,10 +751,10 @@ let rew_l2r_forward_dep_scheme_kind =
 (* we adopt r2l_forward_rew (this one introduces a blocked beta-      *)
 (* expansion but since the guard condition supports commutative cuts  *)
 (* this is not a problem; we need though a fix to adjust it to the    *)
-(* standard form of schemes in Coq)                                   *)
+(* standard form of schemes in Rocq)                                  *)
 (**********************************************************************)
 let rew_l2r_scheme_kind =
-  declare_individual_scheme_object "_rew_r"
+  declare_individual_scheme_object "rew_r"
   (fun env _ ind -> fix_r2l_forward_rew_scheme env
      (build_r2l_forward_rew_scheme false env ind InType))
 
@@ -763,7 +765,7 @@ let rew_l2r_scheme_kind =
 (* introducing commutative cuts, we adopt it                          *)
 (**********************************************************************)
 let rew_r2l_scheme_kind =
-  declare_individual_scheme_object "_rew"
+  declare_individual_scheme_object "rew"
   (fun env _ ind -> build_r2l_rew_scheme false env ind InType)
 
 (* End of rewriting schemes *)
@@ -848,7 +850,7 @@ let build_congr env (eq,refl,ctx) ind =
             mkApp (mkVar varf, [|lift (mip.mind_nrealargs+3) b|])|])|])))))))
   in c, UState.of_context_set ctx
 
-let congr_scheme_kind = declare_individual_scheme_object "_congr"
+let congr_scheme_kind = declare_individual_scheme_object "congr"
   (fun env _ ind ->
      (* May fail if equality is not defined *)
-   build_congr env (get_coq_eq env UnivGen.empty_sort_context) ind)
+   build_congr env (get_rocq_eq env UnivGen.empty_sort_context) ind)
