@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -11,10 +11,8 @@
 open Util
 open Names
 open Constr
-open Context
 open CErrors
 open Evd
-open Evarutil
 open Evarsolve
 open Locus
 open Context.Named.Declaration
@@ -27,17 +25,11 @@ module NamedDecl = Context.Named.Declaration
 (* Instantiation of existential variables *)
 (******************************************)
 
-let depends_on_evar sigma evk _ (pbty,_,t1,t2) =
-  try Evar.equal (head_evar sigma t1) evk
-  with NoHeadEvar ->
-  try Evar.equal (head_evar sigma t2) evk
-  with NoHeadEvar -> false
-
 let define_and_solve_constraints evk c env evd =
   if Termops.occur_evar evd evk c then
     Pretype_errors.error_occur_check env evd evk c;
   let evd = define evk c evd in
-  let (evd,pbs) = extract_changed_conv_pbs evd (depends_on_evar evd evk) in
+  let (evd,pbs) = extract_changed_conv_pbs_from evd (Some (Evar.Set.singleton evk)) in
   match
     List.fold_left
       (fun p (pbty,env,t1,t2) -> match p with
@@ -65,6 +57,9 @@ let w_refine evk rawc env sigma =
       expand_evars = true;
       program_mode = false;
       polymorphic = false;
+      undeclared_evars_patvars = false;
+      patvars_abstract = false;
+      unconstrained_sorts = false;
     } in
     let expected_type = Pretyping.OfType (Evd.evar_concl evi) in
     try Pretyping.understand_uconstr ~flags ~expected_type env sigma rawc
@@ -147,18 +142,3 @@ let let_evar name typ =
     Tacticals.tclTHEN (Proofview.Unsafe.tclEVARS sigma)
     (Tactics.pose_tac (Name.Name id) evar)
   end
-
-let hget_evar n =
-  let open EConstr in
-  Proofview.Goal.enter begin fun gl ->
-  let sigma = Tacmach.project gl in
-  let concl = Proofview.Goal.concl gl in
-  let evl = evar_list sigma concl in
-  if List.length evl < n then
-    user_err Pp.(str "Not enough uninstantiated existential variables.");
-  if n <= 0 then user_err Pp.(str "Incorrect existential variable index.");
-  let ev = List.nth evl (n-1) in
-  let ev_type = EConstr.existential_type sigma ev in
-  let r = Retyping.relevance_of_type (Proofview.Goal.env gl) sigma ev_type in
-  Tactics.change_concl (mkLetIn (make_annot Name.Anonymous r,mkEvar ev,ev_type,concl))
-  end
