@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -31,6 +31,8 @@ val is_quantified_hypothesis : Id.t -> Proofview.Goal.t -> bool
 
 (** {6 Primitive tactics. } *)
 
+exception NotConvertible
+
 val introduction    : Id.t -> unit Proofview.tactic
 val convert_concl   : cast:bool -> check:bool -> types -> cast_kind -> unit Proofview.tactic
 val convert_hyp     : check:bool -> reorder:bool -> named_declaration -> unit Proofview.tactic
@@ -52,6 +54,7 @@ val intro                : unit Proofview.tactic
 val introf               : unit Proofview.tactic
 val intro_move        : Id.t option -> Id.t Logic.move_location -> unit Proofview.tactic
 val intro_move_avoid  : Id.t option -> Id.Set.t -> Id.t Logic.move_location -> unit Proofview.tactic
+val intros_move       : (Id.t * Id.t Logic.move_location) list -> unit Proofview.tactic
 
   (** [intro_avoiding idl] acts as intro but prevents the new Id.t
      to belong to [idl] *)
@@ -59,13 +62,13 @@ val intro_avoiding       : Id.Set.t -> unit Proofview.tactic
 
 val intro_replacing      : Id.t -> unit Proofview.tactic
 val intro_using          : Id.t -> unit Proofview.tactic
-[@@ocaml.deprecated "Prefer [intro_using_then] to avoid renaming issues."]
+[@@ocaml.deprecated "(8.13) Prefer [intro_using_then] to avoid renaming issues."]
 val intro_using_then     : Id.t -> (Id.t -> unit Proofview.tactic) -> unit Proofview.tactic
 val intro_mustbe_force   : Id.t -> unit Proofview.tactic
 val intros_mustbe_force  : Id.t list -> unit Proofview.tactic
 val intro_then           : (Id.t -> unit Proofview.tactic) -> unit Proofview.tactic
 val intros_using         : Id.t list -> unit Proofview.tactic
-[@@ocaml.deprecated "Prefer [intros_using_then] to avoid renaming issues."]
+[@@ocaml.deprecated "(8.13) Prefer [intros_using_then] to avoid renaming issues."]
 val intros_using_then    : Id.t list -> (Id.t list -> unit Proofview.tactic) -> unit Proofview.tactic
 val intros_replacing     : Id.t list -> unit Proofview.tactic
 val intros_possibly_replacing : Id.t list -> unit Proofview.tactic
@@ -113,7 +116,7 @@ val force_destruction_arg : evars_flag -> env -> evar_map ->
     evar_map * constr with_bindings destruction_arg
 
 val finish_evar_resolution : ?flags:Pretyping.inference_flags ->
-  env -> evar_map -> (evar_map * constr) -> evar_map * constr
+  env -> evar_map -> (evar_map option * constr) -> evar_map * constr
 
 (** Tell if a used hypothesis should be cleared by default or not *)
 
@@ -171,11 +174,11 @@ val normalise_in_hyp  : hyp_location -> unit Proofview.tactic
 val normalise_option  : goal_location -> unit Proofview.tactic
 val normalise_vm_in_concl : unit Proofview.tactic
 val unfold_in_concl   :
-  (occurrences * Tacred.evaluable_global_reference) list -> unit Proofview.tactic
+  (occurrences * Evaluable.t) list -> unit Proofview.tactic
 val unfold_in_hyp     :
-  (occurrences * Tacred.evaluable_global_reference) list -> hyp_location -> unit Proofview.tactic
+  (occurrences * Evaluable.t) list -> hyp_location -> unit Proofview.tactic
 val unfold_option     :
-  (occurrences * Tacred.evaluable_global_reference) list -> goal_location -> unit Proofview.tactic
+  (occurrences * Evaluable.t) list -> goal_location -> unit Proofview.tactic
 val change            :
   check:bool -> constr_pattern option -> change_arg -> clause -> unit Proofview.tactic
 val pattern_option    :
@@ -196,12 +199,9 @@ val specialize    : constr with_bindings -> intro_pattern option -> unit Proofvi
 val move_hyp      : Id.t -> Id.t Logic.move_location -> unit Proofview.tactic
 val rename_hyp    : (Id.t * Id.t) list -> unit Proofview.tactic
 
-val revert        : Id.t list -> unit Proofview.tactic
-
 (** {6 Resolution tactics. } *)
 
 val apply_type : typecheck:bool -> constr -> constr list -> unit Proofview.tactic
-val bring_hyps : named_context -> unit Proofview.tactic
 
 val apply                 : constr -> unit Proofview.tactic
 val eapply : ?with_classes:bool -> constr -> unit Proofview.tactic
@@ -210,7 +210,7 @@ val apply_with_bindings_gen :
   ?with_classes:bool -> advanced_flag -> evars_flag -> (clear_flag * constr with_bindings CAst.t) list -> unit Proofview.tactic
 
 val apply_with_delayed_bindings_gen :
-  advanced_flag -> evars_flag -> (clear_flag * delayed_open_constr_with_bindings CAst.t) list -> unit Proofview.tactic
+  advanced_flag -> evars_flag -> (clear_flag * constr with_bindings Proofview.tactic CAst.t) list -> unit Proofview.tactic
 
 val apply_with_bindings   : constr with_bindings -> unit Proofview.tactic
 val eapply_with_bindings : ?with_classes:bool -> constr with_bindings -> unit Proofview.tactic
@@ -224,59 +224,13 @@ val apply_in :
 
 val apply_delayed_in :
   advanced_flag -> evars_flag -> Id.t ->
-    (clear_flag * delayed_open_constr_with_bindings CAst.t) list ->
+    (clear_flag * constr with_bindings Proofview.tactic CAst.t) list ->
     intro_pattern option -> unit Proofview.tactic -> unit Proofview.tactic
 
 (** {6 Elimination tactics. } *)
 
-(*
-   The general form of an induction principle is the following:
-
-   forall prm1 prm2 ... prmp,                          (induction parameters)
-   forall Q1...,(Qi:Ti_1 -> Ti_2 ->...-> Ti_ni),...Qq, (predicates)
-   branch1, branch2, ... , branchr,                    (branches of the principle)
-   forall (x1:Ti_1) (x2:Ti_2) ... (xni:Ti_ni),         (induction arguments)
-   (HI: I prm1..prmp x1...xni)                         (optional main induction arg)
-   -> (Qi x1...xni HI        (f prm1...prmp x1...xni)).(conclusion)
-                   ^^        ^^^^^^^^^^^^^^^^^^^^^^^^
-               optional                optional
-               even if HI      argument added if principle
-             present above   generated by functional induction
-             [indarg]          [farg]
-
-  HI is not present when the induction principle does not come directly from an
-  inductive type (like when it is generated by functional induction for
-  example). HI is present otherwise BUT may not appear in the conclusion
-  (dependent principle). HI and (f...) cannot be both present.
-
-  Principles taken from functional induction have the final (f...).
-*)
-
-(** [rel_contexts] and [rel_declaration] actually contain triples, and
-   lists are actually in reverse order to fit [compose_prod]. *)
-type elim_scheme = {
-  elimt: types;
-  indref: GlobRef.t option;
-  params: rel_context;      (** (prm1,tprm1);(prm2,tprm2)...(prmp,tprmp) *)
-  nparams: int;               (** number of parameters *)
-  predicates: rel_context;  (** (Qq, (Tq_1 -> Tq_2 ->...-> Tq_nq)), (Q1,...) *)
-  npredicates: int;           (** Number of predicates *)
-  branches: rel_context;    (** branchr,...,branch1 *)
-  nbranches: int;             (** Number of branches *)
-  args: rel_context;        (** (xni, Ti_ni) ... (x1, Ti_1) *)
-  nargs: int;                 (** number of arguments *)
-  indarg: rel_declaration option;  (** Some (H,I prm1..prmp x1...xni)
-                                                 if HI is in premisses, None otherwise *)
-  concl: types;               (** Qi x1...xni HI (f...), HI and (f...)
-                                  are optional and mutually exclusive *)
-  indarg_in_concl: bool;      (** true if HI appears at the end of conclusion *)
-  farg_in_concl: bool;        (** true if (f...) appears at the end of conclusion *)
-}
-
-val compute_elim_sig : evar_map -> types -> elim_scheme
-
 val general_elim_clause : evars_flag -> unify_flags -> Id.t option ->
-  (EConstr.t * EConstr.types) -> EConstr.t -> unit Proofview.tactic
+  ((metavariable list * Unification.Meta.t) * EConstr.t * EConstr.types) -> Constant.t -> unit Proofview.tactic
 
 val default_elim  : evars_flag -> clear_flag -> constr with_bindings ->
   unit Proofview.tactic
@@ -284,31 +238,14 @@ val simplest_elim : constr -> unit Proofview.tactic
 val elim :
   evars_flag -> clear_flag -> constr with_bindings -> constr with_bindings option -> unit Proofview.tactic
 
-val induction : evars_flag -> clear_flag -> constr -> or_and_intro_pattern option ->
-  constr with_bindings option -> unit Proofview.tactic
-
 (** {6 Case analysis tactics. } *)
 
 val general_case_analysis : evars_flag -> clear_flag -> constr with_bindings ->  unit Proofview.tactic
 val simplest_case         : constr -> unit Proofview.tactic
 
-val destruct : evars_flag -> clear_flag -> constr -> or_and_intro_pattern option ->
-  constr with_bindings option -> unit Proofview.tactic
-
-(** {6 Generic case analysis / induction tactics. } *)
-
-(** Implements user-level "destruct" and "induction" *)
-
-val induction_destruct : rec_flag -> evars_flag ->
-  (delayed_open_constr_with_bindings destruction_arg
-   * (intro_pattern_naming option * or_and_intro_pattern option)
-   * clause option) list *
-  constr with_bindings option -> unit Proofview.tactic
-
 (** {6 Eliminations giving the type instead of the proof. } *)
 
-val case_type         : types -> unit Proofview.tactic
-val elim_type         : types -> unit Proofview.tactic
+val exfalso : unit Proofview.tactic
 
 (** {6 Constructor tactics. } *)
 
@@ -356,10 +293,6 @@ val assert_after_replacing : Id.t -> types -> unit Proofview.tactic
 val assert_before : Name.t -> types -> unit Proofview.tactic
 val assert_after  : Name.t -> types -> unit Proofview.tactic
 
-val assert_as     : (* true = before *) bool ->
-  (* optionally tell if a specialization of some hyp: *) Id.t option ->
-  intro_pattern option -> constr -> unit Proofview.tactic
-
 (** Implements the tactics assert, enough and pose proof; note that "by"
     applies on the first goal for both assert and enough *)
 
@@ -389,16 +322,7 @@ val letin_tac : (bool * intro_pattern_naming) option ->
 (** Common entry point for user-level "set", "pose" and "remember" *)
 
 val letin_pat_tac : evars_flag -> (bool * intro_pattern_naming) option ->
-  Name.t -> (evar_map * constr) -> clause -> unit Proofview.tactic
-
-(** {6 Generalize tactics. } *)
-
-val generalize      : constr list -> unit Proofview.tactic
-val generalize_gen  : (constr Locus.with_occurrences * Name.t) list -> unit Proofview.tactic
-
-val new_generalize_gen  : ((occurrences * constr) * Name.t) list -> unit Proofview.tactic
-
-val generalize_dep  : ?with_let:bool (** Don't lose let bindings *) -> constr  -> unit Proofview.tactic
+  Name.t -> (evar_map option * constr) -> clause -> unit Proofview.tactic
 
 (** {6 Other tactics. } *)
 
@@ -408,8 +332,8 @@ val generalize_dep  : ?with_let:bool (** Don't lose let bindings *) -> constr  -
 val constr_eq : strict:bool -> constr -> constr -> unit Proofview.tactic
 
 val unify           : ?state:TransparentState.t -> constr -> constr -> unit Proofview.tactic
+val evarconv_unify : ?state:TransparentState.t -> ?with_ho:bool -> constr -> constr -> unit Proofview.tactic
 
-val abstract_generalize : ?generalize_vars:bool -> ?force_dep:bool -> Id.t -> unit Proofview.tactic
 val specialize_eqs : Id.t -> unit Proofview.tactic
 
 val general_rewrite_clause :
@@ -419,7 +343,7 @@ val subst_one :
   (bool -> Id.t -> Id.t * constr * bool -> unit Proofview.tactic) Hook.t
 
 val declare_intro_decomp_eq :
-  ((int -> unit Proofview.tactic) -> Coqlib.coq_eq_data * types *
+  ((int -> unit Proofview.tactic) -> Rocqlib.rocq_eq_data * types *
    (types * constr * constr) ->
    constr * types -> unit Proofview.tactic) -> unit
 
@@ -451,3 +375,24 @@ val refine : typecheck:bool -> (evar_map -> evar_map * constr) -> unit Proofview
 
 val reduce_after_refine : unit Proofview.tactic
 (** The reducing tactic called after {!refine}. *)
+
+(** {6 Internals, do not use} *)
+
+module Internal :
+sig
+
+val explicit_intro_names : 'a intro_pattern_expr CAst.t list -> Id.Set.t
+
+val check_name_unicity : env -> Id.t list -> Id.t list -> 'a intro_pattern_expr CAst.t list -> unit
+
+val clear_gen : (env -> evar_map -> Id.t -> Evarutil.clear_dependency_error ->
+  GlobRef.t option -> evar_map * named_context_val * types) ->
+  Id.t list -> unit Proofview.tactic
+
+val clear_wildcards : lident list -> unit Proofview.tactic
+
+val dest_intro_patterns : evars_flag -> Id.Set.t -> lident list ->
+  Id.t Logic.move_location -> intro_patterns ->
+  (Id.t list -> lident list -> unit Proofview.tactic) -> unit Proofview.tactic
+
+end
