@@ -2,7 +2,7 @@
 
 . ../template/init.sh
 
-coq_makefile -f _CoqProject -o Makefile
+rocq makefile -f _CoqProject -o Makefile
 cat Makefile.conf
 make
 make html mlihtml
@@ -21,14 +21,18 @@ make install-doc DSTROOT="$PWD/tmp"
 sort -u > desired <<EOT
 .
 ./test
-./test/test_plugin.cmi
-./test/test_plugin.cmx
-./test/test_plugin.cmxa
-./test/test_plugin.cmxs
+./test/.coq-native
+./test/.coq-native/Ntest_test.cmi
+./test/.coq-native/Ntest_test.cmx
+./test/.coq-native/Ntest_test.cmxs
 ./test/test.glob
 ./test/test.v
 ./test/test.vo
 ./test/sub
+./test/sub/.coq-native
+./test/sub/.coq-native/Ntest_sub_testsub.cmi
+./test/sub/.coq-native/Ntest_sub_testsub.cmx
+./test/sub/.coq-native/Ntest_sub_testsub.cmxs
 ./test/sub/testsub.glob
 ./test/sub/testsub.v
 ./test/sub/testsub.vo
@@ -56,4 +60,16 @@ sort -u > desired <<EOT
 ./test/html/coqdoc.css
 ./test/html/test.test.html
 EOT
-exec diff -u desired actual
+(coqc -config | grep -q "NATIVE_COMPILER_DEFAULT=yes") || sed -i.bak '/\.coq-native/d' desired
+diff -u desired actual
+
+(cd "$(find tmp -name coq-test-suite)" && find .) | sort > actual
+sort > desired <<EOT
+.
+./META
+./test_plugin.cmi
+./test_plugin.cmx
+./test_plugin.cmxa
+./test_plugin.cmxs
+EOT
+diff -u desired actual
