@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -23,7 +23,8 @@ type input_buffer = {
   mutable str : Bytes.t; (* buffer of already read characters *)
   mutable len : int;    (* number of chars in the buffer *)
   mutable bols : int list; (* offsets in str of beginning of lines *)
-  mutable tokens : Pcoq.Parsable.t; (* stream of tokens *)
+  mutable stream : (unit,char) Gramlib.Stream.t; (* stream of chars *)
+  mutable tokens : Procq.Parsable.t; (* stream of tokens *)
   mutable start : int } (* stream count of the first char of the buffer *)
 
 (* Double the size of the buffer. *)
@@ -33,10 +34,23 @@ let resize_buffer ibuf = let open Bytes in
   blit ibuf.str 0 nstr 0 (length ibuf.str);
   ibuf.str <- nstr
 
+let peek_to_newline ibuf =
+  (* peek to see a newline following the latest command
+     cf #19355 *)
+  let rec aux n =
+    let l = Gramlib.Stream.npeek () n ibuf.stream in
+    if List.length l < n then ()
+    else match CList.last l with
+      | '\n' -> ()
+      | ' ' | '\t' -> aux (n+1)
+      | _ -> (* nonblank character: the latest command is not the last on this line *) ()
+  in
+  aux 1
+
 (* Delete all irrelevant lines of the input buffer. Keep the last line
    in the buffer (useful when there are several commands on the same line). *)
-
 let resynch_buffer ibuf =
+  let () = peek_to_newline ibuf in
   match ibuf.bols with
     | ll::_ ->
         let new_len = ibuf.len - ll in
@@ -54,7 +68,7 @@ let emacs_prompt_endstring   () = if !print_emacs then "</prompt>" else ""
 
 (* Read a char in an input channel, displaying a prompt at every
    beginning of line. *)
-let prompt_char doc ic ibuf count =
+let prompt_char doc ic ibuf () =
   let bol = match ibuf.bols with
     | ll::_ -> Int.equal ibuf.len ll
     | [] -> Int.equal ibuf.len 0
@@ -70,15 +84,6 @@ let prompt_char doc ic ibuf count =
   with End_of_file ->
     None
 
-(* Reinitialize the char stream (after a Drop) *)
-
-let reset_input_buffer doc ic ibuf =
-  ibuf.str <- Bytes.empty;
-  ibuf.len <- 0;
-  ibuf.bols <- [];
-  ibuf.tokens <- Pcoq.Parsable.make (Stream.from (prompt_char doc ic ibuf));
-  ibuf.start <- 0
-
 (* Functions to print underlined locations from an input buffer. *)
 module TopErr = struct
 
@@ -99,7 +104,11 @@ let get_bols_of_loc ibuf (bp,ep) =
         let nafter = if ll < ep then add_line (ll,ba) after else after in
         lines_rec ll nafter fl
   in
-  let (fl,ll) = lines_rec ibuf.len ([],None) ibuf.bols in
+  let bols = match ibuf.bols with
+    | [] -> [ibuf.len+1] (* no newline at the end of the command, pretend there was one *)
+    | _ :: _ as bols -> bols
+  in
+  let (fl,ll) = lines_rec ibuf.len ([],None) bols in
   (fl,Option.get ll)
 
 let dotted_location (b,e) =
@@ -149,7 +158,7 @@ let print_highlight_location ib loc =
   highlight_lines
 
 let valid_buffer_loc ib loc =
-  let (b,e) = Loc.unloc loc in b-ib.start >= 0 && e-ib.start < ib.len && b<=e
+  let (b,e) = Loc.unloc loc in b-ib.start >= 0 && e-ib.start <= ib.len && b<=e
 
 (* Toplevel error explanation. *)
 let error_info_for_buffer ?loc buf =
@@ -171,11 +180,11 @@ let error_info_for_buffer ?loc buf =
       | Loc.InFile _ -> Topfmt.pr_phase ~loc ()
 
 (* Actual printing routine *)
-let print_error_for_buffer ?loc lvl msg buf =
+let print_error_for_buffer ?loc ?qf lvl msg buf =
   let pre_hdr = error_info_for_buffer ?loc buf in
   if !print_emacs
   then Topfmt.emacs_logger ?pre_hdr lvl msg
-  else Topfmt.std_logger   ?pre_hdr lvl msg
+  else Topfmt.std_logger   ?pre_hdr ?qf lvl msg
 
 (*
 let print_toplevel_parse_error (e, info) buf =
@@ -186,17 +195,17 @@ let print_toplevel_parse_error (e, info) buf =
 *)
 end
 
-(*s The Coq prompt is the name of the focused proof, if any, and "Coq"
+(*s The Rocq prompt is the name of the focused proof, if any, and "Rocq"
     otherwise. We trap all exceptions to prevent the error message printing
     from cycling. *)
 let make_prompt () =
   try
     (Names.Id.to_string (Vernacstate.Declare.get_current_proof_name ())) ^ " < "
   with Vernacstate.Declare.NoCurrentProof ->
-    "Coq < "
+    "Rocq < "
   [@@ocaml.warning "-3"]
 
-(* the coq prompt added to the default one when in emacs mode
+(* the rocq prompt added to the default one when in emacs mode
    The prompt contains the current state label [n] (for global
    backtracking) and the current proof state [p] (for proof
    backtracking) plus the list of open (nested) proofs (for proof
@@ -226,13 +235,25 @@ let top_buffer =
     ^ make_emacs_prompt doc
     ^ emacs_prompt_endstring()
   in
+  let stream = Gramlib.Stream.empty () in
   { prompt = pr;
     str = Bytes.empty;
     len = 0;
     bols = [];
-    tokens = Pcoq.Parsable.make (Stream.of_list []);
+    stream;
+    tokens = Procq.Parsable.make stream;
     start = 0 }
 
+(* Intialize or reinitialize the char stream *)
+let reset_input_buffer ~state =
+  let stream = Gramlib.Stream.from (prompt_char state.Vernac.State.doc stdin top_buffer) in
+  top_buffer.str <- Bytes.empty;
+  top_buffer.len <- 0;
+  top_buffer.bols <- [];
+  top_buffer.stream <- stream;
+  top_buffer.tokens <- Procq.Parsable.make stream;
+  top_buffer.start <- 0
+
 let set_prompt prompt =
   top_buffer.prompt
   <- (fun doc ->
@@ -242,12 +263,12 @@ let set_prompt prompt =
 
 (* Read the input stream until a dot is encountered *)
 let parse_to_dot =
-  let rec dot tok st = match Stream.next st with
+  let rec dot kwstate st = match Gramlib.LStream.next kwstate st with
     | Tok.KEYWORD ("."|"...") -> ()
     | Tok.EOI -> ()
-    | _ -> dot tok st
+    | _ -> dot kwstate st
   in
-  Pcoq.Entry.of_parser "Coqtoplevel.dot" dot
+  Procq.Entry.(of_parser "Coqtoplevel.dot" { parser_fun = dot })
 
 (* If an error occurred while parsing, we try to read the input until a dot
    token is encountered.
@@ -255,9 +276,9 @@ let parse_to_dot =
 
 let rec discard_to_dot () =
   try
-    Pcoq.Entry.parse parse_to_dot top_buffer.tokens
+    Procq.Entry.parse parse_to_dot top_buffer.tokens
   with
-    | CLexer.Error.E _ -> discard_to_dot ()
+    | CLexer.Error.E _ -> (* Lexer failed *) discard_to_dot ()
     | e when CErrors.noncritical e -> ()
 
 let read_sentence ~state input =
@@ -266,7 +287,22 @@ let read_sentence ~state input =
   try Stm.parse_sentence ~doc:state.doc state.sid ~entry:G_toplevel.vernac_toplevel input
   with reraise ->
     let reraise = Exninfo.capture reraise in
-    discard_to_dot ();
+    (* When typing Ctrl-C, two situations may arise:
+       - if a lexer/parsing arrived first, the rest of the ill-formed
+         sentence needs to be discarded, and, if Ctrl-C is found while
+         trying to discarding (in discard_to_dot), let it bypass the
+         reporting of the parsing error and report the Sys.Break
+         instead.
+       - if a Ctrl-C arrives after a valid start of sentence, do not
+         discard_to_dot since Ctrl-C is the last read character and
+         there is nothing left to discard. *)
+    (match fst reraise with
+     | Sys.Break -> Pp.pp_with !Topfmt.err_ft (Pp.fnl ())
+     | _ ->
+        try discard_to_dot ()
+        with Sys.Break ->
+          Pp.pp_with !Topfmt.err_ft (Pp.fnl ());
+          raise Sys.Break);
     (* The caller of read_sentence does the error printing now, this
        should be re-enabled once we rely on the feedback error
        printer again *)
@@ -298,16 +334,16 @@ let coqloop_feed (fb : Feedback.feedback) = let open Feedback in
   | FileLoaded (_,_) -> ()
   | Custom (_,_,_) -> ()
   (* Re-enable when we switch back to feedback-based error printing *)
-  | Message (Error,loc,msg) -> ()
+  | Message (Error,loc,_,msg) -> ()
   (* TopErr.print_error_for_buffer ?loc lvl msg top_buffer *)
-  | Message (Warning,loc,msg) ->
+  | Message (Warning,loc,qf,msg) ->
     let loc = extract_default_loc loc fb.doc_id fb.span_id in
-    TopErr.print_error_for_buffer ?loc Warning msg top_buffer
-  | Message (lvl,loc,msg) ->
-    TopErr.print_error_for_buffer ?loc lvl msg top_buffer
+    TopErr.print_error_for_buffer ?loc ~qf Warning msg top_buffer
+  | Message (lvl,loc,qf,msg) ->
+    TopErr.print_error_for_buffer ?loc ~qf lvl msg top_buffer
 
-(** Main coq loop : read vernacular expressions until Drop is entered.
-    Ctrl-C is handled internally as Sys.Break instead of aborting Coq.
+(** Main rocq loop : read vernacular expressions until Drop is entered.
+    Ctrl-C is handled internally as Sys.Break instead of aborting Rocq.
     Normally, the only exceptions that can come out of [do_vernac] and
     exit the loop are Drop and Quit. Any other exception there indicates
     an issue with [print_toplevel_error] above. *)
@@ -324,14 +360,12 @@ let loop_flush_all () =
 let pequal cmp1 cmp2 (a1,a2) (b1,b2) = cmp1 a1 b1 && cmp2 a2 b2
 let evleq e1 e2 = CList.equal Evar.equal e1 e2
 let cproof p1 p2 =
-  let Proof.{goals=a1;stack=a2;shelf=a3;given_up=a4} = Proof.data p1 in
-  let Proof.{goals=b1;stack=b2;shelf=b3;given_up=b4} = Proof.data p2 in
+  let Proof.{goals=a1;stack=a2;sigma=sigma1} = Proof.data p1 in
+  let Proof.{goals=b1;stack=b2;sigma=sigma2} = Proof.data p2 in
   evleq a1 b1 &&
   CList.equal (pequal evleq evleq) a2 b2 &&
-  CList.equal Evar.equal a3 b3 &&
-  CList.equal Evar.equal a4 b4
-
-let drop_last_doc = ref None
+  CList.equal Evar.equal (Evd.shelf sigma1) (Evd.shelf sigma2) &&
+  Evar.Set.equal (Evd.given_up sigma1) (Evd.given_up sigma2)
 
 (* todo: could add other Set/Unset commands, such as "Printing Universes" *)
 let print_anyway_opts = [
@@ -341,9 +375,25 @@ let print_anyway_opts = [
 let print_anyway c =
   let open Vernacexpr in
   match c.expr with
-  | VernacSetOption (_, opt, _) -> List.mem opt print_anyway_opts
+  | VernacSynterp (VernacSetOption (_, opt, _)) -> List.mem opt print_anyway_opts
   | _ -> false
 
+(* print the proof step, possibly with diffs highlighted, *)
+let print_and_diff oldp proof =
+  let output =
+    if Proof_diffs.show_diffs () then
+      try Printer.pr_open_subgoals ~diffs:oldp proof
+      with Pp_diff.Diff_Failure msg -> begin
+        (* todo: print the unparsable string (if we know it) *)
+        Feedback.msg_warning Pp.(str ("Diff failure: " ^ msg) ++ cut()
+            ++ str "Showing results without diff highlighting" );
+        Printer.pr_open_subgoals proof
+      end
+    else
+      Printer.pr_open_subgoals proof
+  in
+  Feedback.msg_notice output
+
 (* We try to behave better when goal printing raises an exception
    [usually Ctrl-C]
 
@@ -351,12 +401,12 @@ let print_anyway c =
    generic way, but that'll do for now *)
 let top_goal_print ~doc c oldp newp =
   try
-    let proof_changed = not (Option.equal cproof oldp newp) in
+    let proof_changed = not (Option.equal cproof oldp (Some newp)) in
     let print_goals = proof_changed && Vernacstate.Declare.there_are_pending_proofs () ||
                       print_anyway c in
     if not !Flags.quiet && print_goals then begin
       let dproof = Stm.get_prev_proof ~doc (Stm.get_current_state ~doc) in
-      Printer.print_and_diff dproof newp
+      print_and_diff dproof newp
     end
   with
   | exn ->
@@ -366,109 +416,99 @@ let top_goal_print ~doc c oldp newp =
     TopErr.print_error_for_buffer ?loc Feedback.Error msg top_buffer
   [@@ocaml.warning "-3"]
 
-let exit_on_error =
+let { Goptions.get = exit_on_error } =
   let open Goptions in
-  declare_bool_option_and_ref ~depr:false ~key:["Coqtop";"Exit";"On";"Error"]
+  declare_bool_option_and_ref
+    ~key:["Coqtop";"Exit";"On";"Error"]
     ~value:false
+    ()
 
-(* XXX: This is duplicated with Vernacentries.show_proof , at some
-   point we should consolidate the code *)
-let show_proof_diff_to_pp pstate =
-  let p = Option.get pstate in
-  let sigma, env = Proof.get_proof_context p in
-  let pprf = Proof.partial_proof p in
-  Pp.prlist_with_sep Pp.fnl (Printer.pr_econstr_env env sigma) pprf
-
-let show_proof_diff_cmd ~state removed =
+let show_proof_diff_cmd ~state diff_opt =
   let open Vernac.State in
-  try
-    let n_pp = show_proof_diff_to_pp state.proof in
-    if true (*Proof_diffs.show_diffs ()*) then
-      let doc = state.doc in
-      let oproof = Stm.get_prev_proof ~doc (Stm.get_current_state ~doc) in
-      try
-        let o_pp = show_proof_diff_to_pp oproof in
-        let tokenize_string = Proof_diffs.tokenize_string in
-        let show_removed = Some removed in
-        Pp_diff.diff_pp_combined ~tokenize_string ?show_removed o_pp n_pp
-      with
-      | Proof.NoSuchGoal _
-      | Option.IsNone -> n_pp
-      | Pp_diff.Diff_Failure msg -> begin
-          (* todo: print the unparsable string (if we know it) *)
-          Feedback.msg_warning Pp.(str ("Diff failure: " ^ msg) ++ cut()
-                                   ++ str "Showing results without diff highlighting" );
-          n_pp
-        end
-    else
-      n_pp
-  with
-  | Proof.NoSuchGoal _
-  | Option.IsNone ->
-    CErrors.user_err (str "No goals to show.")
+  match state.proof with
+  | None -> CErrors.user_err (str "No proofs to diff.")
+  | Some proof ->
+      let old = Stm.get_prev_proof ~doc:state.doc state.sid in
+      Proof_diffs.diff_proofs ~diff_opt ?old proof
 
+let ml_toplevel_state = ref None
+let ml_toplevel_include_ran = ref false
+
+(* Initialises the Ocaml toplevel before launching it, so that it can
+   find the "include" file in the *source* directory *)
+let init_ocaml_path () =
+  let env = Boot.Env.init () in
+  let corelib = Boot.Env.corelib env |> Boot.Path.to_string in
+  let add_subdir dl = Mltop.add_ml_dir (Filename.concat corelib dl) in
+  List.iter add_subdir ("dev" :: Coq_config.all_src_dirs)
+
+let init_and_run_ml_toploop () =
+  init_ocaml_path ();
+  Flags.with_option Flags.in_ml_toplevel (Mltop.ocaml_toploop ~init_file:"ml_toplevel/include") ()
+
+(* We return whether the execution should continue and a new state *)
 let process_toplevel_command ~state stm =
   let open Vernac.State in
   let open G_toplevel in
   match stm with
-  (* Usually handled in the caller *)
   | VernacDrop ->
-    state
+    if Mltop.is_ocaml_top() then begin
+      (* Save the last state for [go ()] *)
+      ml_toplevel_state := Some state;
+      (* Initialise and launch the OCaml toplevel *)
+      init_and_run_ml_toploop ();
+      (* Reinitialize the char stream *)
+      reset_input_buffer ~state;
+      (* [go ()] was potentially executed — get the new state *)
+      let state = Option.get !ml_toplevel_state in
+      true, state
+    end else begin
+      Feedback.msg_warning (str "There is no ML toplevel.");
+      true, state
+    end
 
   | VernacBackTo bid ->
     let bid = Stateid.of_int bid in
     let doc, res = Stm.edit_at ~doc:state.doc bid in
-    assert (res = `NewTip);
-    { state with doc; sid = bid }
+    assert (res = Stm.NewTip);
+    true, { state with doc; sid = bid }
 
   | VernacQuit ->
-    exit 0
+    false, state
 
   | VernacControl { CAst.loc; v=c } ->
     let nstate = Vernac.process_expr ~state (CAst.make ?loc c) in
-    top_goal_print ~doc:state.doc c state.proof nstate.proof;
-    nstate
+    let () = match nstate.proof with
+    | None -> ()
+    | Some proof -> top_goal_print ~doc:state.doc c state.proof proof
+    in
+    true, nstate
 
   | VernacShowGoal { gid; sid } ->
     let proof = Stm.get_proof ~doc:state.doc (Stateid.of_int sid) in
     let goal = Printer.pr_goal_emacs ~proof gid sid in
-    let evars =
-      match proof with
-      | None -> mt()
-      | Some p ->
-        let gl = (Evar.unsafe_of_int gid) in
-        let { Proof.sigma } = Proof.data p in
-        try Printer.print_dependent_evars (Some gl) sigma [ gl ]
-        with Not_found -> mt()
-    in
-    Feedback.msg_notice (v 0 (goal ++ evars));
-    state
+    let () = Feedback.msg_notice goal in
+    true, state
 
-  | VernacShowProofDiffs removed ->
+  | VernacShowProofDiffs diff_opt ->
     (* We print nothing if there are no goals left *)
     if not (Proof_diffs.color_enabled ()) then
       CErrors.user_err Pp.(str "Show Proof Diffs requires setting the \"-color\" command line argument to \"on\" or \"auto\".")
     else
-      let out = show_proof_diff_cmd ~state removed in
+      let out = show_proof_diff_cmd ~state diff_opt in
       Feedback.msg_notice out;
-    state
-
-(* We return a new state and true if we got a `Drop` command  *)
-let read_and_execute_base ~state =
-  let input = top_buffer.tokens in
-  match read_sentence ~state input with
-  | Some G_toplevel.VernacDrop ->
-    if Mltop.is_ocaml_top()
-    then (drop_last_doc := Some state; state, true)
-    else (Feedback.msg_warning (str "There is no ML toplevel."); state, false)
-  | Some stm ->
-    process_toplevel_command ~state stm, false
-  (* End of file *)
-  | None ->
-    top_stderr (fnl ()); exit 0
+    true, state
 
 let read_and_execute ~state =
-  try read_and_execute_base ~state
+  try
+    let input = top_buffer.tokens in
+    match read_sentence ~state input with
+    | Some stm ->
+      process_toplevel_command ~state stm
+    (* End of file *)
+    | None ->
+      top_stderr (fnl ());
+      false, state
   with
   (* Exception printing should be done by the feedback listener,
      however this is not yet ready so we rely on the exception for
@@ -489,43 +529,31 @@ let read_and_execute ~state =
     let msg = CErrors.iprint (e, info) in
     TopErr.print_error_for_buffer ?loc Feedback.Error msg top_buffer;
     if exit_on_error () then exit 1;
-    state, false
+    true, state
 
-(* This function will only return on [Drop], careful to keep it tail-recursive *)
-let rec vernac_loop ~state =
-  let open Vernac.State in
-  loop_flush_all ();
-  top_stderr (fnl());
-  if !print_emacs then top_stderr (str (top_buffer.prompt state.doc));
-  resynch_buffer top_buffer;
-  let state, drop = read_and_execute ~state in
-  if drop then state else vernac_loop ~state
+let loop ~state =
+  (* Initialize buffer *)
+  reset_input_buffer ~state;
+  Flags.without_option Flags.in_ml_toplevel (fun () ->
+    (* The main loop, as a tail-recursive function *)
+    let rec aux state =
+      loop_flush_all ();
+      top_stderr (fnl());
+      let open Vernac.State in
+      if !print_emacs then top_stderr (str (top_buffer.prompt state.doc));
+      resynch_buffer top_buffer;
+      let new_running, new_state = read_and_execute ~state:state in
+      if new_running then
+        (aux [@ocaml.tailcall]) new_state
+      else
+        new_state in
+    aux state
+  ) ()
 
-(* Default toplevel loop, machinery for drop is below *)
-
-let drop_args = ref None
-
-(* Initialises the Ocaml toplevel before launching it, so that it can
-   find the "include" file in the *source* directory *)
-let init_ocaml_path ~coqlib =
-  let add_subdir dl = Mltop.add_ml_dir (Filename.concat coqlib dl) in
-  List.iter add_subdir ("dev" :: Coq_config.all_src_dirs)
-
-let loop ~opts ~state =
-  drop_args := Some opts;
+let run ~opts ~state =
   let open Coqargs in
   print_emacs := opts.config.print_emacs;
   (* We initialize the console only if we run the toploop_run *)
   let tl_feed = Feedback.add_feeder coqloop_feed in
-  (* Initialize buffer *)
-  Sys.catch_break true;
-  reset_input_buffer state.Vernac.State.doc stdin top_buffer;
-  (* Call the main loop *)
-  let _ : Vernac.State.t = vernac_loop ~state in
-  (* Initialise and launch the Ocaml toplevel *)
-  let coqlib = Envars.coqlib () in
-  init_ocaml_path ~coqlib;
-  Mltop.ocaml_toploop();
-  (* We delete the feeder after the OCaml toploop has ended so users
-     of Drop can see the feedback. *)
+  let _ : Vernac.State.t = loop ~state in
   Feedback.del_feeder tl_feed
