@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -18,8 +18,6 @@ open Coercionops
 val try_add_new_coercion_with_target
   :  GlobRef.t
   -> local:bool
-  -> poly:bool
-  -> nonuniform:bool
   -> reversible:bool
   -> source:cl_typ
   -> target:cl_typ
@@ -27,8 +25,7 @@ val try_add_new_coercion_with_target
 
 (** [try_add_new_coercion ref s] declares [ref], assumed to be of type
    [(x1:T1)...(xn:Tn)src->tg], as a coercion from [src] to [tg] *)
-val try_add_new_coercion : GlobRef.t -> local:bool -> poly:bool ->
-  nonuniform:bool ->
+val try_add_new_coercion : GlobRef.t -> local:bool ->
   reversible:bool -> unit
 
 (** [try_add_new_coercion_subclass cst s] expects that [cst] denotes a
@@ -36,13 +33,12 @@ val try_add_new_coercion : GlobRef.t -> local:bool -> poly:bool ->
    an identity coercion from [cst] to [tg], named something like
    ["Id_cst_tg"] *)
 val try_add_new_coercion_subclass : cl_typ -> local:bool -> poly:bool ->
-  nonuniform:bool ->
   reversible:bool -> unit
 
 (** [try_add_new_coercion_with_source ref s src] declares [ref] as a coercion
    from [src] to [tg] where the target is inferred from the type of [ref] *)
 val try_add_new_coercion_with_source : GlobRef.t -> local:bool ->
-  poly:bool -> nonuniform:bool -> reversible:bool -> source:cl_typ -> unit
+  reversible:bool -> source:cl_typ -> unit
 
 (** [try_add_new_identity_coercion id s src tg] enriches the
    environment with a new definition of name [id] declared as an
@@ -52,14 +48,14 @@ val try_add_new_identity_coercion
   -> local:bool
   -> poly:bool -> source:cl_typ -> target:cl_typ -> unit
 
-val add_coercion_hook : poly:bool -> nonuniform:bool -> reversible:bool -> Declare.Hook.t
+val add_coercion_hook : reversible:bool -> Declare.Hook.t
 
 val add_subclass_hook : poly:bool -> reversible:bool -> Declare.Hook.t
 
 val class_of_global : GlobRef.t -> cl_typ
 
 (** Attribute to silence warning for coercions that don't satisfy
-   the uniform inheritance condition. *)
+   the uniform inheritance condition. (deprecated in 8.18) *)
 val nonuniform : bool option Attributes.attribute
 
 val change_reverse : GlobRef.t -> reversible:bool -> unit
