@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -13,6 +13,14 @@ open Constrexpr
 
 (** {6 Inductive and coinductive types} *)
 
+type flags = {
+  poly : bool;
+  cumulative : bool;
+  template : bool option;
+  finite : Declarations.recursivity_kind;
+  mode : Hints.hint_mode list option;
+}
+
 (** Entry points for the vernacular commands Inductive and CoInductive *)
 
 type uniform_inductive_flag =
@@ -20,15 +28,12 @@ type uniform_inductive_flag =
   | NonUniformParameters
 
 val do_mutual_inductive
-  :  template:bool option
-  -> cumul_univ_decl_expr option
-  -> (one_inductive_expr * decl_notation list) list
-  -> cumulative:bool
-  -> poly:bool
+  : flags:flags
   -> ?typing_flags:Declarations.typing_flags
+  -> cumul_univ_decl_expr option
+  -> (one_inductive_expr * notation_declaration list) list
   -> private_ind:bool
   -> uniform:uniform_inductive_flag
-  -> Declarations.recursivity_kind
   -> unit
 
 (** User-interface API *)
@@ -46,12 +51,14 @@ module Mind_decl : sig
 (** inductive_expr at the constr level *)
 type t = {
   mie : Entries.mutual_inductive_entry;
+  default_dep_elim : DeclareInd.default_dep_elim list;
   nuparams : int option;
   univ_binders : UnivNames.universe_binders;
   implicits : DeclareInd.one_inductive_impls list;
   uctx : Univ.ContextSet.t;
-  where_notations : Metasyntax.where_decl_notation list;
+  where_notations : Metasyntax.notation_interpretation_decl list;
   coercions : Libnames.qualid list;
+  indlocs : DeclareInd.indlocs;
 }
 
 end
@@ -59,36 +66,33 @@ end
 (** elaborates an inductive declaration (the first half of do_mutual_inductive) *)
 val interp_mutual_inductive
   :  env:Environ.env
-  -> template:bool option
-  -> cumul_univ_decl_expr option
-  -> (one_inductive_expr * decl_notation list) list
-  -> cumulative:bool
-  -> poly:bool
+  -> flags:flags
   -> ?typing_flags:Declarations.typing_flags
+  -> cumul_univ_decl_expr option
+  -> (one_inductive_expr * notation_declaration list) list
   -> private_ind:bool
   -> uniform:uniform_inductive_flag
-  -> Declarations.recursivity_kind
   -> Mind_decl.t
 
+type syntax_allows_template_poly = SyntaxAllowsTemplatePoly | SyntaxNoTemplatePoly
+
 (** the post-elaboration part of interp_mutual_inductive, mainly dealing with
     universe levels *)
 val interp_mutual_inductive_constr
-  : sigma:Evd.evar_map
-  -> template:bool option
+  :  sigma:Evd.evar_map
+  -> flags:flags
   -> udecl:UState.universe_decl
   -> variances:Entries.variance_entry
-  -> ctx_params:(EConstr.t, EConstr.t) Context.Rel.Declaration.pt list
+  -> ctx_params:EConstr.rel_context
   -> indnames:Names.Id.t list
+  -> arities_explicit:bool list
   -> arities:EConstr.t list
-  -> arityconcl:(bool * EConstr.ESorts.t) option list
-  -> constructors:(Names.Id.t list * Constr.constr list) list
+  -> template_syntax:syntax_allows_template_poly list
+  -> constructors:(Names.Id.t list * EConstr.constr list) list
   -> env_ar_params:Environ.env
   (** Environment with the inductives and parameters in the rel_context *)
-  -> cumulative:bool
-  -> poly:bool
   -> private_ind:bool
-  -> finite:Declarations.recursivity_kind
-  -> Entries.mutual_inductive_entry * UnivNames.universe_binders * Univ.ContextSet.t
+  -> DeclareInd.default_dep_elim list * Entries.mutual_inductive_entry * UnivNames.universe_binders * Univ.ContextSet.t
 
 (************************************************************************)
 (** Internal API, exported for Record                                   *)
@@ -96,11 +100,10 @@ val interp_mutual_inductive_constr
 
 val compute_template_inductive
   : user_template:bool option
-  -> env_ar_params:Environ.env
-  -> ctx_params:(Constr.constr, Constr.constr) Context.Rel.Declaration.pt list
+  -> ctx_params:Constr.rel_context
   -> univ_entry:UState.universes_entry
   -> Entries.one_inductive_entry
-  -> Sorts.t option
+  -> syntax_allows_template_poly
   -> Entries.inductive_universes_entry * Univ.ContextSet.t
 (** [compute_template_inductive] computes whether an inductive can be template
     polymorphic. *)
@@ -123,5 +126,25 @@ val variance_of_entry
 
 module Internal :
 sig
-  val compute_constructor_level : Environ.env -> Evd.evar_map -> EConstr.rel_context -> Sorts.t
+  (** Returns the modified arities (the result sort may be replaced by Prop).
+      Should be called with minimized universes. *)
+  val inductive_levels
+    : Environ.env
+    -> Evd.evar_map
+    -> poly:bool
+    -> indnames:Names.Id.t list
+    -> arities_explicit:bool list
+    (* whether the arities were explicit from the user (for auto Prop lowering) *)
+    -> EConstr.constr list
+    (* arities *)
+    -> EConstr.rel_context list list
+    (* constructors *)
+    -> Evd.evar_map * (DeclareInd.default_dep_elim list * EConstr.t list)
+
+  val error_differing_params
+    : kind:string
+    -> (Names.lident * Vernacexpr.inductive_params_expr)
+    -> (Names.lident * Vernacexpr.inductive_params_expr)
+    -> 'a
+
 end
