@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -49,7 +49,7 @@ let find_with_logical_path dirpath =
   List.filter (fun p -> Names.DirPath.equal p.path_logical dirpath) !load_paths
 
 let warn_file_found_multiple_times =
-  CWarnings.create ~name:"ambiguous-extra-dep" ~category:"filesystem"
+  CWarnings.create ~name:"ambiguous-extra-dep" ~category:CWarnings.CoreCategories.filesystem
     (fun (file,from,other,extra) ->
       Pp.(str "File " ++ str file ++ str " found twice in " ++
       Names.DirPath.print from ++ str":" ++ spc () ++ str other ++ str " (selected)," ++
@@ -88,17 +88,17 @@ let remove_load_path dir =
   load_paths := List.filter filter !load_paths
 
 let warn_overriding_logical_loadpath =
-  CWarnings.create ~name:"overriding-logical-loadpath" ~category:"loadpath"
-    (fun (phys_path, old_path, coq_path) ->
+  CWarnings.create ~name:"overriding-logical-loadpath" ~category:CWarnings.CoreCategories.filesystem
+    (fun (phys_path, old_path, rocq_path) ->
        Pp.(seq [str phys_path; strbrk " was previously bound to "
                ; DP.print old_path; strbrk "; it is remapped to "
-               ; DP.print coq_path]))
+               ; DP.print rocq_path]))
 
-let add_load_path root phys_path coq_path ~implicit =
+let add_load_path root phys_path rocq_path ~implicit =
   let phys_path = CUnix.canonical_path_name phys_path in
   let filter p = String.equal p.path_physical phys_path in
   let binding = {
-    path_logical = coq_path;
+    path_logical = rocq_path;
     path_physical = phys_path;
     path_implicit = implicit;
     path_root = root;
@@ -108,13 +108,13 @@ let add_load_path root phys_path coq_path ~implicit =
     load_paths := binding :: !load_paths
   | [{ path_logical = old_path; path_implicit = old_implicit }] ->
     let replace =
-      if DP.equal coq_path old_path then
+      if DP.equal rocq_path old_path then
         implicit <> old_implicit
       else
         let () =
           (* Do not warn when overriding the default "-I ." path *)
           if not (DP.equal old_path Libnames.default_root_prefix) then
-          warn_overriding_logical_loadpath (phys_path, old_path, coq_path)
+          warn_overriding_logical_loadpath (phys_path, old_path, rocq_path)
         in
         true in
     if replace then
@@ -176,49 +176,49 @@ let locate_file fname =
 (************************************************************************)
 (*s Locate absolute or partially qualified library names in the path *)
 
-type locate_error = LibUnmappedDir | LibNotFound
-type 'a locate_result = ('a, locate_error) result
+module Error = struct
+  type t = LibUnmappedDir | LibNotFound
 
-let warn_several_object_files =
-  CWarnings.create ~name:"several-object-files" ~category:"require"
-    Pp.(fun (vi, vo) ->
-        seq [ str "Loading"; spc (); str vi
-            ; strbrk " instead of "; str vo
-            ; strbrk " because it is more recent."
-            ])
+  let unmapped_dir qid =
+    let prefix, _ = Libnames.repr_qualid qid in
+    CErrors.user_err
+      Pp.(seq [ str "Cannot load "; Libnames.pr_qualid qid; str ":"; spc ()
+              ; str "no physical path bound to"; spc ()
+              ; Names.DirPath.print prefix; fnl ()
+              ])
 
+  let lib_not_found dir =
+    let vos = !Flags.load_vos_libraries in
+    let vos_msg = if vos then [Pp.str " (while searching for a .vos file)"] else [] in
+    CErrors.user_err
+      Pp.(seq ([ str "Cannot find library "; Names.DirPath.print dir; str" in loadpath"]@vos_msg))
+
+  let raise dp = function
+    | LibUnmappedDir ->
+      unmapped_dir (Libnames.qualid_of_dirpath dp)
+    | LibNotFound ->
+      lib_not_found dp
+end
+
+(* If [!Flags.load_vos_libraries]
+      and the .vos file exists
+      and this file is not empty
+   Then load this library
+   Else load the .vo file or raise error if both are missing *)
 let select_vo_file ~find base =
   let find ext =
     try
       let name = Names.Id.to_string base ^ ext in
       let lpath, file = find name in
-      Some (lpath, file)
-    with Not_found -> None in
-  (* If [!Flags.load_vos_libraries]
-        and the .vos file exists
-        and this file is not empty
-     Then load this library
-     Else load the most recent between the .vo file and the .vio file,
-          or if there is only of the two files, take this one,
-          or raise an error if both are missing. *)
-  let load_most_recent_of_vo_and_vio () =
-    match find ".vo", find ".vio" with
-    | None, None ->
-      Error LibNotFound
-    | Some res, None | None, Some res ->
-      Ok res
-    | Some (_, vo), Some (_, vi as resvi)
-      when Unix.((stat vo).st_mtime < (stat vi).st_mtime) ->
-      warn_several_object_files (vi, vo);
-      Ok resvi
-    | Some resvo, Some _ ->
-      Ok resvo
-    in
-  if !Flags.load_vos_libraries then begin
+      Ok (lpath, file)
+    with Not_found -> Error Error.LibNotFound in
+  if !Flags.load_vos_libraries
+  then begin
     match find ".vos" with
-    | Some (_, vos as resvos) when (Unix.stat vos).Unix.st_size > 0 -> Ok resvos
-    | _ -> load_most_recent_of_vo_and_vio()
-  end else load_most_recent_of_vo_and_vio()
+    | Ok (_, vos as resvos) when (Unix.stat vos).Unix.st_size > 0 -> Ok resvos
+    | _ -> find ".vo"
+  end
+  else find ".vo"
 
 let find_first loadpath base =
   match System.all_in_path loadpath base with
@@ -233,7 +233,7 @@ let find_unique fullqid loadpath base =
     CErrors.user_err Pp.(str "Required library " ++ Libnames.pr_qualid fullqid ++
       strbrk " matches several files in path (found " ++ pr_enum str (List.map snd l) ++ str ").")
 
-let locate_absolute_library dir : CUnix.physical_path locate_result =
+let locate_absolute_library dir : (CUnix.physical_path, Error.t) Result.t =
   (* Search in loadpath *)
   let pref, base = Libnames.split_dirpath dir in
   let loadpath = filter_path (fun dir -> DP.equal dir pref) in
@@ -245,7 +245,7 @@ let locate_absolute_library dir : CUnix.physical_path locate_result =
     | Error fail -> Error fail
 
 let locate_qualified_library ?root qid :
-  (DP.t * CUnix.physical_path) locate_result =
+  (DP.t * CUnix.physical_path, Error.t) Result.t =
   (* Search library in loadpath *)
   let dir, base = Libnames.repr_qualid qid in
   match expand_path ?root dir with
@@ -258,7 +258,7 @@ let locate_qualified_library ?root qid :
       | Error _ ->
          (* Looking otherwise in -R/-Q blocks of partial matches *)
         let rec aux = function
-          | [] -> Error LibUnmappedDir
+          | [] -> Error Error.LibUnmappedDir
           | block :: rest ->
             match select_vo_file ~find:(find_unique qid block) base with
             | Ok _ as x -> x
@@ -272,27 +272,26 @@ let locate_qualified_library ?root qid :
       Ok (library, file)
     | Error _ as e -> e
 
-let error_unmapped_dir qid =
-  let prefix, _ = Libnames.repr_qualid qid in
-  CErrors.user_err
-    Pp.(seq [ str "Cannot load "; Libnames.pr_qualid qid; str ":"; spc ()
-            ; str "no physical path bound to"; spc ()
-            ; DP.print prefix; fnl ()
-            ])
+let warn_deprecated_missing_stdlib =
+  CWarnings.create ~name:"deprecated-missing-stdlib"
+    ~category:Deprecation.Version.v9_0
+    (fun qid ->
+      Pp.(str "Loading Stdlib without prefix is deprecated." ++ spc ()
+          ++ str "Use \"From Stdlib Require " ++ Libnames.pr_qualid qid
+          ++ str "\"" ++ spc() ++ str "or the deprecated \"From Coq Require "
+          ++ Libnames.pr_qualid qid ++ str "\"" ++ spc ()
+          ++ str "for compatibility with older Coq versions."))
 
-let error_lib_not_found qid =
-  let vos = !Flags.load_vos_libraries in
-  let vos_msg = if vos then [Pp.str " (while searching for a .vos file)"] else [] in
-  CErrors.user_err
-    Pp.(seq ([ str "Cannot find library "; Libnames.pr_qualid qid; str" in loadpath"]@vos_msg))
-
-let try_locate_absolute_library dir =
-  match locate_absolute_library dir with
-  | Ok res -> res
-  | Error LibUnmappedDir ->
-    error_unmapped_dir (Libnames.qualid_of_dirpath dir)
-  | Error LibNotFound ->
-    error_lib_not_found (Libnames.qualid_of_dirpath dir)
+(* temporary handling deprecated loading of stdlib without root *)
+let locate_qualified_library ?root qid =
+  let root_stdlib =
+    Names.(Libnames.add_dirpath_suffix DirPath.empty (Id.of_string "Stdlib")) in
+  match root, locate_qualified_library ?root qid with
+  | Some _, r | None, (Ok _ as r) -> r
+  | None, (Error _ as e) ->
+     match locate_qualified_library ~root:root_stdlib qid with
+     | Error _ -> e
+     | Ok _ as o -> warn_deprecated_missing_stdlib ?loc:qid.loc qid; o
 
 (** { 5 Extending the load path } *)
 
@@ -304,21 +303,19 @@ type vo_path =
   ; implicit  : bool
   (** [implicit = true] avoids having to qualify with [coq_path]
       true for -R, false for -Q in command line *)
-  ; has_ml    : bool
-  (** If [has_ml] is true, the directory will also be added to the ml include path *)
   ; recursive : bool
   (** [recursive] will determine whether we explore sub-directories  *)
   }
 
 let warn_cannot_open_path =
-  CWarnings.create ~name:"cannot-open-path" ~category:"filesystem"
+  CWarnings.create ~name:"cannot-open-path" ~category:CWarnings.CoreCategories.filesystem
     (fun unix_path -> Pp.(str "Cannot open " ++ str unix_path))
 
 let warn_cannot_use_directory =
-  CWarnings.create ~name:"cannot-use-directory" ~category:"filesystem"
+  CWarnings.create ~name:"cannot-use-directory" ~category:CWarnings.CoreCategories.filesystem
     (fun d ->
        Pp.(str "Directory " ++ str d ++
-           strbrk " cannot be used as a Coq identifier (skipped)"))
+           strbrk " cannot be used as a Rocq identifier (skipped)"))
 
 let convert_string d =
   try Names.Id.of_string d
@@ -343,15 +340,6 @@ let add_vo_path lp =
       with Exit -> None
     in
     let dirs = List.map_filter convert_dirs dirs in
-    let () =
-      if lp.has_ml && not lp.recursive then
-        Mltop.add_ml_dir unix_path
-      else if lp.has_ml && lp.recursive then
-        (List.iter (fun (lp,_) -> Mltop.add_ml_dir lp) dirs;
-         Mltop.add_ml_dir unix_path)
-      else
-        ()
-    in
     let root = (unix_path,lp.coq_path) in
     let add (path, dir) = add_load_path root path ~implicit dir in
     (* deeper dirs registered first and thus be found last *)
