@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -35,19 +35,19 @@ let entry_buf = Buffer.create 64
 let pr_entry e =
   let () = Buffer.clear entry_buf in
   let ft = Format.formatter_of_buffer entry_buf in
-  let () = Pcoq.Entry.print ft e in
+  let () = Procq.Entry.print ft e in
   str (Buffer.contents entry_buf)
 
 let error_unknown_entry ?loc name =
   user_err ?loc Pp.(str "Unknown or unprintable grammar entry " ++ str name ++ str".")
 
 let pr_registered_grammar name =
-  let gram = Pcoq.find_grammars_by_name name in
+  let gram = Procq.find_grammars_by_name name in
   match gram with
   | [] -> error_unknown_entry name
   | entries ->
-    let pr_one (Pcoq.Entry.Any e) =
-      str "Entry " ++ str (Pcoq.Entry.name e) ++ str " is" ++ fnl () ++
+    let pr_one (Procq.Entry.Any e) =
+      str "Entry " ++ str (Procq.Entry.name e) ++ str " is" ++ fnl () ++
       pr_entry e
     in
     prlist pr_one entries
@@ -58,19 +58,19 @@ let pr_grammar_subset grammar =
       | entries ->
         str "Entry " ++ str name ++ str " is" ++ fnl() ++
         prlist_with_sep (fun () -> str "or" ++ fnl())
-          (fun (Pcoq.Entry.Any e) -> pr_entry e)
+          (fun (Procq.Entry.Any e) -> pr_entry e)
           entries)
       grammar
   in
   let pp = CString.Map.bindings pp in
   prlist_with_sep fnl (fun (_,pp) -> pp) pp
 
-let is_known = let open Pcoq.Entry in function
+let is_known = let open Procq.Entry in function
   | "constr" | "term" | "binder_constr" ->
-    Some [ Any Pcoq.Constr.constr;
-      Any Pcoq.Constr.lconstr;
-      Any Pcoq.Constr.binder_constr;
-      Any Pcoq.Constr.term;
+    Some [ Any Procq.Constr.constr;
+      Any Procq.Constr.lconstr;
+      Any Procq.Constr.binder_constr;
+      Any Procq.Constr.term;
     ]
   | "vernac" ->
     Some Pvernac.Vernac_.[
@@ -83,21 +83,21 @@ let is_known = let open Pcoq.Entry in function
         Any gallina_ext;
     ]
   | name ->
-    let gram = Pcoq.find_grammars_by_name name in
+    let gram = Procq.find_grammars_by_name name in
     match gram with
     | [] -> None
     | entries -> Some entries
 
 let full_grammar () =
   let open Pvernac.Vernac_ in
-  let open Pcoq.Entry in
+  let open Procq.Entry in
   let proof_modes = List.map (fun (_,e) -> Any e)
       (CString.Map.bindings (Pvernac.list_proof_modes()))
   in
   let entries = (Any vernac_control) :: (Any noedit_mode) :: proof_modes in
-  Pcoq.Entry.accumulate_in entries
+  Procq.Entry.accumulate_in entries
 
-let same_entry (Pcoq.Entry.Any e) (Pcoq.Entry.Any e') = (Obj.magic e) == (Obj.magic e')
+let same_entry (Procq.Entry.Any e) (Procq.Entry.Any e') = (Obj.magic e) == (Obj.magic e')
 
 let pr_grammar = function
   | [] ->
@@ -119,8 +119,8 @@ let pr_grammar = function
     in
     let other = String.Set.of_list other in
     let grammar = String.Map.filter (fun name _ -> String.Set.mem name other) grammar in
-    let grammar = List.fold_left (fun grammar (Pcoq.Entry.Any e as any) ->
-        String.Map.update (Pcoq.Entry.name e) (function
+    let grammar = List.fold_left (fun grammar (Procq.Entry.Any e as any) ->
+        String.Map.update (Procq.Entry.name e) (function
             | None -> Some [any]
             | Some vl as v ->
               if List.mem_f same_entry any vl
@@ -133,7 +133,7 @@ let pr_grammar = function
 let pr_custom_grammar name = pr_registered_grammar ("custom:"^name)
 
 let pr_keywords () =
-  Pp.prlist_with_sep Pp.fnl Pp.str (CString.Set.elements (CLexer.keywords (Pcoq.get_keyword_state())))
+  Pp.prlist_with_sep Pp.fnl Pp.str (CString.Set.elements (CLexer.keywords (Procq.get_keyword_state())))
 
 (** **************************************************************** **)
 (** Parse a format (every terminal starting with a letter or a single
@@ -718,7 +718,7 @@ let prod_entry_type = function
 
 let keyword_needed need s =
   (* Ensure that IDENT articulation terminal symbols are keywords *)
-  match Pcoq.terminal s with
+  match Procq.terminal s with
   | Tok.PIDENT (Some k) ->
     if need then
       Flags.if_verbose Feedback.msg_info (str "Identifier '" ++ str k ++ str "' now a keyword");
@@ -894,7 +894,7 @@ let check_and_extend_constr_grammar ntn rule =
       error_parsing_incompatible_level ntn ntn_for_grammar oldprec oldtyps prec typs;
     if oldparsing = None then raise Not_found
   with Not_found ->
-    Egramcoq.extend_constr_grammar rule
+    Egramrocq.extend_constr_grammar rule
 
 let warn_prefix_incompatible_level =
   CWarnings.create ~name:"notation-incompatible-prefix"
@@ -921,7 +921,7 @@ let check_prefix_incompatible_level ntn prec nottyps =
        let pref_nottyps = Notgram_ops.non_terminals_of_notation pref in
        let pref_nottyps = CList.firstn k pref_nottyps in
        let nottyps = CList.firstn k nottyps in
-       if not (level_eq prec pref_prec && List.for_all2 Extend.constr_entry_key_eq nottyps pref_nottyps) then
+       if not (level_eq prec pref_prec && List.for_all2 Extend.constr_entry_key_eq_ignore_binder_kind nottyps pref_nottyps) then
          warn_prefix_incompatible_level (pref, ntn, pref_prec, pref_nottyps, prec, nottyps);
      with Not_found | Failure _ -> ()
 
@@ -1011,7 +1011,7 @@ let () = CErrors.register_handler @@ function
   | _ -> None
 
 let check_custom_entry entry =
-  if not (Egramcoq.exists_custom_entry entry) then
+  if not (Egramrocq.exists_custom_entry entry) then
     raise @@ UnknownCustomEntry entry
 
 let check_entry_type = function
@@ -1076,7 +1076,7 @@ let check_useless_entry_types recvars mainvars etyps =
 type notation_main_data = {
   onlyparsing  : bool;
   onlyprinting : bool;
-  user_warns   : UserWarn.t option;
+  user_warns   : Globnames.extended_global_reference UserWarn.with_qf option;
   entry        : notation_entry;
   format       : unparsing Loc.located list option;
   itemscopes  : (Id.t * scope_name) list;
@@ -1470,7 +1470,7 @@ let check_locality_compatibility local custom i_typs =
     let subcustom = List.map_filter (function _,ETConstr (InCustomEntry s,_,_) -> Some s | _ -> None) i_typs in
     let allcustoms = match custom with InCustomEntry s -> s::subcustom | _ -> subcustom in
     List.iter (fun s ->
-        if Egramcoq.locality_of_custom_entry s then
+        if Egramrocq.locality_of_custom_entry s then
           user_err (strbrk "Notation has to be declared local as it depends on custom entry " ++ str s ++
                     strbrk " which is local."))
       (List.uniquize allcustoms)
@@ -1675,7 +1675,7 @@ let with_lib_stk_protection f x =
 
 let with_syntax_protection f x =
   with_lib_stk_protection
-    (Pcoq.with_grammar_rule_protection
+    (Procq.with_grammar_rule_protection
        (with_notation_protection f)) x
 
 (** **************************************************************** **)
@@ -1855,7 +1855,7 @@ let make_notation_interpretation ~local main_data notation_symbols ntn syntax_ru
     notobj_use = use;
     notobj_interp = (vars, ac);
     notobj_coercion = coe;
-    notobj_user_warns = main_data.user_warns;
+    notobj_user_warns = main_data.user_warns |> Option.map UserWarn.drop_qf;
     notobj_notation = df';
     notobj_specific_pp_rules = sy_pp_rules;
   }
@@ -1916,6 +1916,7 @@ let set_notation_for_interpretation env impls (ntn_decl, main_data, notation_sym
 let build_notation_syntax ~local ~infix user_warns ntn_decl =
   let { ntn_decl_string = {CAst.loc;v=df}; ntn_decl_modifiers = modifiers; ntn_decl_interp = c } = ntn_decl in
   (* Extract the modifiers not affecting the parsing rule *)
+  let user_warns = Option.map UserWarn.with_empty_qf user_warns in
   let (main_data,syntax_modifiers) = interp_non_syntax_modifiers ~reserved:false ~infix ~abbrev:false user_warns modifiers in
   (* Extract the modifiers not affecting the parsing rule *)
   let notation_symbols, is_prim_token = analyze_notation_tokens ~onlyprinting:main_data.onlyprinting ~infix main_data.entry df in
@@ -2098,8 +2099,8 @@ let warn_custom_entry =
           strbrk "Custom entry " ++ str s ++ strbrk " has been overridden.")
 
 let load_custom_entry _ (local,s) =
-  if Egramcoq.exists_custom_entry s then warn_custom_entry s
-  else Egramcoq.create_custom_entry ~local s
+  if Egramrocq.exists_custom_entry s then warn_custom_entry s
+  else Egramrocq.create_custom_entry ~local s
 
 let cache_custom_entry o = load_custom_entry 1 o
 
@@ -2117,7 +2118,7 @@ let inCustomEntry : locality_flag * string -> obj =
       classify_function = classify_custom_entry}
 
 let declare_custom_entry local s =
-  if Egramcoq.exists_custom_entry s then
+  if Egramrocq.exists_custom_entry s then
     user_err Pp.(str "Custom entry " ++ str s ++ str " already exists.")
   else
     Lib.add_leaf (inCustomEntry (local,s))
