@@ -1,5 +1,5 @@
 (************************************************************************)
-(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*         *      The Rocq Prover / The Rocq Development Team           *)
 (*  v      *         Copyright INRIA, CNRS and contributors             *)
 (* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
@@ -13,8 +13,13 @@
 
 val pr_set_entry_type : ('a -> Pp.t) -> 'a Extend.constr_entry_key_gen -> Pp.t
 
+val pr_syntax_modifier : Vernacexpr.syntax_modifier CAst.t -> Pp.t
+
 (** Prints a fixpoint body *)
-val pr_rec_definition : Vernacexpr.fixpoint_expr -> Pp.t
+val pr_rec_definition : Constrexpr.fixpoint_order_expr option * Vernacexpr.recursive_expr_gen -> Pp.t
+
+(** Prints a scheme *)
+val pr_onescheme : Names.lident option * Vernacexpr.scheme -> Pp.t
 
 (** Prints a vernac expression without dot *)
 val pr_vernac_expr : Vernacexpr.vernac_expr -> Pp.t
